#ifdef HAVE_CONFIG_H
#include <config_analysis.h>
#endif
#include <iostream>
#include <iomanip>
#include <sstream>
#include <vector>
#include <csignal>
#include <cmath>
#include <analysis/vmedaq.h>
/**
 * A test program for libVMEdaq
 * It reads VATAGP data
 *
 * We first define a class which derives from VATAGP and implements
 * some of the hooks that receive the data already digested.
 */

#if defined(HAS_ROOT)
#include <TROOT.h>
#include <TFile.h>
#include <TH1.h>
#include <TH2.h>

#endif
class MyData : public VATAGP
{
    private:
#if defined(HAS_ROOT)
        int   n_ped_record;
        TFile *ofile;
        TH1   *hSig;
        TH1   *hPed;
        TH1   *hNoise;
        TH2   *hHitMap;
        TH1   *hHit1D;
#endif
    public:
        MyData(bool up = false);
        ~MyData();

        void onNewFile(const std::string &fnam);
        void onNewRun();
        void onPedestals(const Module *);
        void onMap(const Module *);
        void onNewPoint(const PointList &);
        void onScan();
        void onNewData(const Module *, int evtcnt, const ChannelList &);
};

MyData::MyData(bool up)
        : VATAGP(up)
{
#if defined(HAS_ROOT)
    n_ped_record = 0;
    ofile = new TFile("tstlib.root", "RECREATE");
    hSig  = new TH1F("hSig", "Signal", 256, 0., 1024.);
    ofile->mkdir("pedestals");
    ofile->cd();
    hHitMap = 0;
#endif
}

MyData::~MyData()
{
#if defined(HAS_ROOT)
    std::cout << "Saving output file " << ofile->GetName() << std::endl;
    ofile->Write();
    ofile->Close();
    delete ofile;
#endif
}

void MyData::onNewFile(const std::string &fnam)
{
    std::cout << "\n++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n"
    << "+ new file: " << fnam << std::endl
    << "++++++++++++++++++++++++++++++++++++++++++++++++++++++++"
    << std::endl;
}

void MyData::onNewRun()
{
    std::cout << "\nNew Run" << std::endl;
}

void MyData::onPedestals(const Module *md)
{

#if defined(HAS_ROOT)
    std::ostringstream oname, otit;
    ofile->cd("pedestals");
    oname << "hPed_" << n_ped_record;
    hPed = new TH1F(oname.str().c_str(), "Pedestals", md->ntot(), 0., md->ntot());

    otit << "hNoise_" << n_ped_record;
    hNoise = new TH1F(otit.str().c_str(), "Noise", md->ntot(), 0., md->ntot());
    ofile->cd();
    
    std::vector<double>::const_iterator ip;
    const std::vector<double> &ped = md->pedestals();
    const std::vector<double> &noise = md->noise();

    int i;
    for (i = 0, ip = ped.begin();ip != ped.end();++ip, i++)
    {
        hPed->Fill(i + 0.5, *ip);
    }
    for (i = 0, ip = noise.begin();ip != noise.end();++ip, i++)
    {
        hNoise->Fill(i + 0.5, *ip);
    }
    n_ped_record++;
#endif
}

void MyData::onMap(const Module *md)
{
    std::cout << "\nPixelMap for module " << md->id() << std::endl;

#if defined(HAS_ROOT)
    if (!hHitMap)
    {
        hHitMap = new TH2F("hHitMap", "Hit map",
                           md->pad_map()->nx(), 0, md->pad_map()->nx(),
                           md->pad_map()->ny(), 0, md->pad_map()->ny()
                          );
        hHitMap->SetXTitle("X");
        hHitMap->SetYTitle("Y");
        hHit1D = new TH1F("hHitMap1D","Hit map",md->nchan(), -0.5, md->nchan()-0.5);
        
    }
#endif
}

void MyData::onNewPoint(const PointList &)
{
    std::cout << "\nNew ScanPoint" << std::endl;
}

void MyData::onScan()
{
    std::cout << "\nA Scan record " << std::endl;
}

void MyData::onNewData(const Module *md, int evtcnt, const ChannelList &data)
{
    if ( !(ndata() % 10000) )
        std::cout << '\r' << std::setw(8) << ndata() << std::flush;

    ChannelList::const_iterator ip;
    for (ip = data.begin();ip != data.end();++ip)
    {
        VATAChannel *ch = dynamic_cast<VATAChannel *>(*ip);
        if (ch->adj())
            continue;

#if defined(HAS_ROOT)
        double val = ch->signal() - md->pedestal(ch->channel());
        double ns  = md->noise(ch->channel());
        double sn  =  ns > 0.1 ? val/ns : 0.;
        if ( sn>6. )
        {
            hSig->Fill(val);
            if (hHitMap)
            {
                int ix, iy;
                md->pad_map()->pad(ch->channel(), ix, iy);
                if (ix != 0xff && iy != 0xff)
                    hHitMap->Fill( ix + 0.5, iy + 0.5 );
                
                hHit1D->Fill(ch->channel());
            }
        }
#endif
    }
}

int main(int argc, char **argv)
{
    VMEdaq vmedaq;
    MyData obs;

    int i;
    std::vector<std::string> files;
    for (i = 1;i < argc;i++)
        files.push_back(argv[i]);

    if (files.empty())
    {
        std::cout << "I need 1 or more input files" << std::endl;
        return 1;
    }

    vmedaq.open(files);
    vmedaq.add( &obs );

    vmedaq.loop();

    vmedaq.close();
    std::cout << "\n\nBye !!!" << std::endl;
    return 0;
}
