/* -*- mode: c++ -*- */
#ifndef __vmedaq_h__
#define __vmedaq_h__
#include <string>
#include <map>
#include <queue>
#include <memory>
#include <zlib.h>

#include <analysis/Module.h>
#include <analysis/DataFormat.h>
#include <analysis/PadMap.h>
#include <analysis/ScanPoint.h>

class VMEdaqObserver;

/** \addtogroup ANALYSIS
 * API for data file analysis.
 * 
 * This API defines the interface and the basic utilitites to read vmedaq
 * output files. There is a VMEdaq class that reads the files and mantains
 * a list of VMEdaqObservers that subscribe to receive some information 
 * while VMEdaq goes through the data blocks in the file. The observer is 
 * the responsible of decoding the data.
 *
 */

/** \ingroup ANALYSIS
 * 
 * VMEdaq is the class in charge of opening the file and find the
 * different data blocks.
 */
class VMEdaq
{
    private:
        /// file
        gzFile _in;

        /// list of observers subscribed to receive some information
        typedef std::map<VMEdaqObserver *, unsigned long> ObserverList;
        ObserverList observers;

        // opens the next file in the queue, if any.
        // returns 0 on sucess and !=1 otherwise
        int open_next_file();

        // list of files to process
        std::queue<std::string> _files;

        /// memory allocator to temporary buffers
        static std::allocator<unsigned int> alloc;

        /// last event number seen
        int _last_event;
    public:
        /// Constructor
        VMEdaq();

        /// Destructor
        virtual ~VMEdaq();

        /** Opens a new file. 
         *  It accepts as input either a single file name or a list of file names.
         *  It will just ignore non-existing files
         */
        template <class T> void open(const T &fl)
        {
            typename T::const_iterator ip;
            for (ip = fl.begin(); ip != fl.end(); ++ip)
            {
                _files.push( *ip);
            }
        }

        /** Opens a single file
         */
        void open(const std::string &s)
        {
            _files.push(s);
        }

        /// closes the file
        void close();

        /// rewinds the file
        void rewind();

        /// read one record from the file
        unsigned int read();

        /// Loop on all the records of the file
        void loop();

        /// Resets the state and closes all the files
        void reset();

        /**
         *  Registers a new observer
         * 
         *  \param type   the data type the observer wants to receive
         *  \param o      the observer
         */
        void add(unsigned long type, VMEdaqObserver *o);
        
        /**
         * Registers a new obsrever.
         * This observer registers to any type of data
         * 
         * \param o the observer
         */
        void add(VMEdaqObserver *o)
        {
            add(0xffffffff, o);
        }

        /// Unregisters a registered observer
        void remove(VMEdaqObserver *o)
        {
            observers.erase(o);
        }

        /// Returns the current event number
        int event_number() const
        {
            return _last_event;
        }
        /**
         *  methods that can be overloaded
         * 
         **/
        /// Called at the beginning of each event
        virtual void NewEvent(int evt_num)
        {
        }
        
        /// Called at the end of each event
        virtual void EndOfEvent(int evt_num)
        {
            
        }
};

/** \ingroup ANALYSIS
 * 
 * \brief class Channel
 * 
 * This class encapsulates the information of a channel as obtained
 * from a DataRecord. The basic information that it stores are the 
 * channel number and the signal.
 */
class Channel
{
    private:
        int _channel;
        double _signal;

        void cpy(const Channel &c)
        {
            _channel = c._channel;
            _signal = c._signal;
        }
    public:
        /// Constructor
        Channel(int ch = 0, double val = 0.) :
            _channel(ch), _signal(val)
        {
        }
        
        /// copy constructor
        Channel(const Channel &c)
        {
            cpy(c);
        }
        
        /// destructor
        virtual ~Channel()
        {
        }
        
        /// Asignment operator
        Channel const &operator=(const Channel &c)
        {
            if (&c != this)
                cpy(c);
            return *this;
        }

        /// Returns the channel number
        int channel() const
        {
            return _channel;
        }
        
        /// Sets the channel number
        void channel(int x)
        {
            _channel=x;
        }
        
        /// Returns the channel signal
        double signal() const
        {
            return _signal;
        }
        
        /// Sets the channel signal
        void signal(double x)
        {
            _signal=x;
        }
        
        /// Returns a copy of itself
        virtual Channel *Clone();
};

/** \ingroup ANALYSIS
 *  \brief class ChannelList
 * 
 * This is a list of Channels.
 * 
 */
class ChannelList : public std::vector<Channel *>
{
    private:
        void clear_pointers()
        {
            ChannelList::iterator ip;
            for (ip = begin(); ip != end(); ++ip)
                delete *ip;
        }
        void cpy(const ChannelList &L);
    public:
        ChannelList() {}
        ChannelList( const ChannelList &L);
        ChannelList &operator=(const ChannelList &L);
        ~ChannelList()
        {
            clear_pointers();
        }
        void clear()
        {
            clear_pointers();
            std::vector<Channel *>::clear();
        }
};

/** \ingroup ANALYSIS
 * 
 * This class represents an event buffer with a collection
 * of Channels. 
 * 
 * It derives from Channel and its channel id 
 * corresponds to the "local" event number. For a given event, 
 * the ChannelList passed to onNewData will be a collection of
 * ChannelLists each of which representing a data buffer.
 */
class DataBuffer : public Channel
{
    private:
        ChannelList _channels;
        void cpy(const DataBuffer &b);
    public:
        DataBuffer(int cntr=0) :
            Channel(cntr, 0)
        {
        }
        DataBuffer(const DataBuffer &b);
        DataBuffer &operator=(const DataBuffer &b);
        ChannelList &list()
        {
            return _channels;
        }
        int gate_counter() const
        {
            return channel();
        }
        void gate_counter(int x)
        {
            channel(x);
        }
        Channel *Clone();
};

/** \ingroup ANALISYS
 * 
 * This class is the responsible for reading and decoding VMEDAQ
 * data. 
 * 
 * It provides a number of hooks to decode the data. It offers
 * two levels of complexity. A first layer that receives the raw data
 * and decodes and a second level, which receives the data digested by
 * the first level.
 */
class VMEdaqObserver
{
    public:
        typedef std::map<int, Module *> ModuleList;
        typedef std::pair<int, int> Point;
        typedef std::vector<Point> PointList;
        typedef std::vector<int> Idlist;
        typedef std::map<int, Idlist> DataTypeList;

    protected:
        /// This is a generic event counter. In the case of a scan it is
        /// reset at the beginning of any new scan point
        unsigned int _cnt;

        /// Global counter. Just counts the number of data records
        unsigned int _ndata;

        /// ScanPoint counter
        int _npoint;

        /// Run number
        unsigned int _run_number;

        /// List of module IDs
        ModuleList _idlist;

        /// List of variables in a scan
        ScanPoint _scan_point;

        /// Number of events per scan point
        int _nevt_per_point;

        /// starting time of th erun as read from StartOfRun
        unsigned int _run_starts;

        /// time of end of run as read from EndOfRun
        unsigned int _run_stops;

        /// last event number read
        int _last_event;

        /// the data type we want to receive
        unsigned long _type;
        
        /// List of data types
        DataTypeList _data_types;
        
        /** Sets the data type. This is done by VMEdaq at
         *  registration.
         */
        void data_type(unsigned long t) { _type=t; }
        friend class VMEdaq;
    public:
        VMEdaqObserver() :
            _cnt(0), _ndata(0), _npoint(-1), _run_number(0), _nevt_per_point(0),
            _run_starts(0), _run_stops(0), _last_event(-1)
        {
        }

        virtual ~VMEdaqObserver();

        /// Returns the data type
        unsigned long data_type() const { return _type; }
        
        
        /// Sets the value of the last event
        void last_event(int x)
        {
            _last_event = x;
        }
        
        /// Returns the value of the last event
        int last_event() const
        {
            return _last_event;
        }
        
        /** 
         * Returns the time at which the run started.
         * The value is a unix time_t suitable for functions as ctime
         */
        unsigned int run_starts() const
        {
            return _run_starts;
        }
        
        /**
         * Returns the time at which the run (or file) stops.
         * See run_starts for a description of the return value
         */
        unsigned int run_stops() const
        {
            return _run_stops;
        }
        
        /**
         *  Returns a generic event counter. In the case of a scan it is
         *  reset at the beginning of any new scan point
         */
        unsigned int cnt() const
        {
            return _cnt;
        }
        
        /**
         * Returns the number of data blocks received so far
         */
        unsigned int ndata() const
        {
            return _ndata;
        }
        
        /**
         * Returns the scan point number
         */
        unsigned int npoint() const
        {
            return _npoint;
        }
        
        /**
         * Returns the actual run number
         */
        unsigned int run_number() const
        {
            return _run_number;
        }
        
        /**
         * Number of events per point in a scan
         */
        int nevt_per_point() const
        {
            return _nevt_per_point;
        }

        /**
         * Contains the scan definition as read from the ScanRecord
         */
        const ScanPoint &scan() const { return _scan_point; }
        
        /**
         * Returns a pointer to the Module with the given id
         */
        const Module *module(int id) const
        {
            ModuleList::const_iterator ip = _idlist.find(id);
            return ip == _idlist.end() ? 0 : ip->second;
        }

        /**
         * Returns a pointer to the Module with the given id
         */
        Module *module(int id)
        {
            ModuleList::iterator ip = _idlist.find(id);
            return ip == _idlist.end() ? 0 : ip->second;
        }

        ModuleList::const_iterator first_module() const
        {
            return _idlist.begin();
        }
        ModuleList::const_iterator last_module() const
        {
            return _idlist.end();
        }
        int number_of_modules() const
        {
            return _idlist.size();
        }

        int get_module_data_type(int id) const;
        
        /** Very simple tests on the validity of the data. It just checks that
         *  the internal data structure is OK
         */
        bool data_valid(unsigned int *);

        /**
         * The Raw data level.
         * All the methods here receive the raw data and are responsible
         * of decoding and passing the digested information to the
         * second level.
         *
         */
        // Called when a StartOfRun record is found
        virtual void StartOfRunRecord(unsigned int *data);

        // Called when an EndOfRunRecord is found
        virtual void EndOfRunRecord(unsigned int *);

        // Called when a DataRecord is found
        // This one should be implemented by each
        virtual void DataRecord(unsigned int *) = 0;

        // Called when a PedestalRecord is found
        virtual void PedestalRecord(unsigned int *);

        // Called when a new point in the scan starts
        virtual void NewPointRecord(unsigned int *);

        // Called when a MapRecord is found
        virtual void MapRecord(unsigned int *);

        // Called when a ScanRecord is found
        virtual void ScanRecord(unsigned int *);

        // Called at the end of each record with the full raw data
        virtual void EndOfRecord(unsigned int *)
        {
        }

        // Called when a ConfigRecord is found
        virtual void ConfigRecord(unsigned int *)
        {
            
        }
        
        /**
         * The user friendly level. The methods below receive the data
         * already decoded.
         */
        // Called when a new file is opened
        virtual void onNewFile(const std::string&)
        {
        }

        // After a StartOfRun record
        virtual void onNewRun()
        {
        }

        // After an end of run
        virtual void onEndOfRun()
        {
        }

        // After a DataRecord
        virtual void onNewData(const Module *, int evtcnt, const ChannelList &)
        {
        }

        // After a PedestalRecord
        virtual void onPedestals(const Module *)
        {
        }

        /** After a new Point in the scan.
         *  PointList is a vector of Point, and a Point is a 
         *  pair of 2 integers, the first is the variable type and the
         *  second the variable name
         */
        virtual void onNewPoint(const PointList &)
        {
        }

        // At the end of a scan Point
        virtual void onEndPoint()
        {
        }

        // After a MapRecord
        virtual void onMap(const Module *)
        {
        }

        // After a ConfigRecord
        virtual void onConfig(const Module *)
        {
        }
        
        // After a ScanRecord
        virtual void onScan()
        {
        }

        // Called at the end of an event
        virtual void onEndOfEvent(int evt, const Module *)
        {
        }
};

/** \ingroup ANALYSIS
 * This class encapsulates the configuration of a VATAmodule.
 */
class VATAConfig : public ModuleConfig
{
    private:
        int _thrs;
        int _chip_type;
        std::vector<bool> _mask;
        std::vector<bool> _test;
        std::vector<int>  _trim;
        
    public:
        VATAConfig(int ntot);
        ~VATAConfig() {}
        
        int  thrs() const { return _thrs; }
        bool mask(int ch) const { return _mask[ch]; }
        bool test(int ch) const { return _test[ch]; }
        int  trim(int ch) const { return _trim[ch]; }
        int  chip_type() const { return _chip_type; }
        
        void chip_type(int v) { _chip_type = v; }
        void thrs(int v) { _thrs = v; }
        void mask(int ch, bool v) { _mask[ch] = v; }
        void test(int ch, bool v) { _test[ch] = v; }
        void trim(int ch, int v) { _trim[ch] = v; }

        int load_config(const char *fnam);
        int dump_config(const char *fnam);

};

/** \ingroup ANALYSIS
 * 
 * This class is the Channel implementation for the VATA.
 * 
 */
class VATAChannel : public Channel
{
    private:
        // if true, it is an "adjacent" channel
        bool _adj;

        // nubmer of adjacent channels
        int _nadj;

        void cpy(const VATAChannel &c)
        {
            *((Channel *)this) = c;
            _adj = c._adj;
            _nadj = c._nadj;
        }
    public:
        VATAChannel(int ch = 0, double sg = 0., bool adj = false) :
            Channel(ch, sg), _adj(adj)
        {
        }
        VATAChannel(const VATAChannel &c)
        {
            cpy(c);
        }
        VATAChannel &operator=(const VATAChannel &c)
        {
            if (this!=&c)
                cpy(c);
            
            return *this;
        }
        Channel *Clone();
        
        ~VATAChannel()
        {
        }

        bool adj() const
        {
            return _adj;
        }
        int nadj() const
        {
            return _nadj;
        }
};

/** \ingroup ANALYSIS
 * 
 * An implementation of a VMEdaqObserver that decodes the VATAGP data.
 */
class VATAGP : public VMEdaqObserver
{
    private:
        /// Set to true if you want to use the data from the UP adc
        bool _use_up;

        /// Number of adjacent channels
        int _nadj;

        /// ReadOut mode
        int _romode;

        void decode_sparse_gp3(unsigned int *data, ChannelList &);
        void decode_sparse_gp7(unsigned int *data, ChannelList &);
        void decode_serial(unsigned int *data, ChannelList &);
        void (VATAGP::*decode_sparse)(unsigned int*, ChannelList &);
    public:
        VATAGP(bool up) :
            _use_up(up)
        {
            decode_sparse = &VATAGP::decode_sparse_gp7;
        }
        ~VATAGP()
        {
        }

        bool use_up() const
        {
            return _use_up;
        }
        void use_up(bool x)
        {
            _use_up = x;
        }

        int nadj() const
        {
            return _nadj;
        }
        int romode() const
        {
            return _romode;
        }
        void ConfigRecord(unsigned int *);
        void DataRecord(unsigned int *);
};

//=======================================================================
// Classes for decoding the ADC data
//

/** \ingroup ANALYSIS
 * 
 * Decodes ADC data.
 * 
 */
class ADCData : public VMEdaqObserver
{
    public:
        ADCData()
        {
        }

        void DataRecord(unsigned int *);
};

/** \ingroup ANALYSIS
 * 
 * Decodes TDC data.
 * 
 */
class TDC488Data : public VMEdaqObserver
{
    public:
        TDC488Data()
        {
        }

        void DataRecord(unsigned int *);
};

class FPGAv1495Data: public VMEdaqObserver{

public:
  FPGAv1495Data():evt_time(0){;}
  void DataRecord(unsigned int *);
protected:
  long long unsigned int evt_time;
  unsigned int coinc_array;
  unsigned int trig_reg;
};
#endif
