#include <iostream>
#include <fstream>
#include <csignal>
#include "vmedaq.h"

std::allocator<unsigned int> VMEdaq::alloc;

Channel *Channel::Clone()
{
    return new Channel(*this);
}

void ChannelList::cpy(const ChannelList &L)
{
    ChannelList::const_iterator ip;
    for (ip=L.begin(); ip!=L.end(); ip++)
        push_back( (*ip)->Clone() );
}
ChannelList::ChannelList(const ChannelList &L)
{
    cpy(L);
}
ChannelList &ChannelList::operator=(const ChannelList &L)
{
    if (this!=&L)
    {
        clear();
        cpy(L);
    }
    return *this;
}

void DataBuffer::cpy(const DataBuffer &b)
{
    *((Channel *)this) = *((Channel *)&b);
    _channels = b._channels;
}

DataBuffer::DataBuffer(const DataBuffer &b)
{
    cpy(b);
}
DataBuffer &DataBuffer::operator =(const DataBuffer &b)
{
    if (this!=&b)
        cpy(b);
    
    return *this;
}
Channel *DataBuffer::Clone()
{
    return (Channel *) new DataBuffer(*this);
}

VMEdaq::VMEdaq() :
    _in(0), _last_event(-1)
{
}

VMEdaq::~VMEdaq()
{
}

void VMEdaq::add(unsigned long type, VMEdaqObserver *o)
{
    observers[o] = type;
    o->data_type(type);
}
int VMEdaq::open_next_file()
{
    close();

    while (!_files.empty())
    {
        std::string str = _files.front();
        _files.pop();
        _in = gzopen(str.c_str(), "rb");
        if (!_in)
            continue;

        ObserverList::iterator ip;
        for (ip = observers.begin(); ip != observers.end(); ++ip)
        {
            ip->first->onNewFile(str);
        }

        return 0;
    }
    return 1;
}

void VMEdaq::close()
{
    // close any opened file
    if (_in)
    {
        gzclose(_in);
        _in = 0;
    }
}
void VMEdaq::reset()
{
    close();
    while (!_files.empty())
        _files.pop();
}

unsigned int VMEdaq::read()
{
    ObserverList::iterator ip;
    int iok, block_size, sz, id;
    unsigned int header, block_type, data_type, x, *data;
    int evt_num;
    if (!_in)
    {
        if (open_next_file() )
            return 0;
    }

    again:
    // read until we find a start of block
    do
    {
        iok = gzread(_in, &x, sizeof(unsigned int));
        if (iok <= 0)
        {
            // end of file, look for next one
            if (open_next_file() )
                return 0;

            x = 0;
            goto again;
        }
    } while (!IsDataBlock(x));

    header = x;

    if (IsBigDataRecord(x ))
    {
        unsigned int hold_data[4];
        hold_data[0] = x;
        iok = gzread(_in, hold_data + 1, sizeof(unsigned int));
        iok = gzread(_in, hold_data + 2, sizeof(unsigned int));
        iok = gzread(_in, hold_data + 3, sizeof(unsigned int));
        block_size = hold_data[3];
        sz = (block_size - 4) * sizeof(unsigned int);
        data = alloc.allocate(block_size);

        for (int i = 0; i < 4; i++)
            data[i] = hold_data[i];

        iok = gzread(_in, data + 4, sz );
        if (iok != sz)
        {
            alloc.deallocate(data, block_size);
            x = 0;
            goto again;
        }
    }
    else
    {
        block_size = x & 0xfff;
        if (block_size == 0)
            goto again;
        sz = (block_size - 1) * sizeof(unsigned int);
        data = alloc.allocate(block_size );

        data[0] = x;
        iok = gzread(_in, data + 1, sz );
        if (iok != sz )
        {
            alloc.deallocate(data, block_size );
            x = 0;
            goto again;
        }
    }

    // Now see who should we call
    block_type = (header & 0xf000);
    data_type = data[2] & 0xf0000;
    id = (data[2] & 0xfff00000L) >> 20;
    switch (block_type)
    {
    case StartOfRunRecord:
        for (ip = observers.begin(); ip != observers.end(); ++ip)
        {
            ip->first->StartOfRunRecord(data);
        }
        break;
    case DataRecord:
        evt_num = data[1] & 0xffffff;
        if (_last_event != evt_num )
        {
            if (_last_event>=0)
            {
                EndOfEvent(_last_event);
            }
            NewEvent(evt_num);
            _last_event = evt_num;
        }
        for (ip = observers.begin(); ip != observers.end(); ++ip)
        {
            if ( (ip->second == data_type) || (ip->second == 0xffffffff))
            {
                if (evt_num != ip->first->last_event())
                {
                    if (ip->first->last_event() >= 0)
                        ip->first->onEndOfEvent(ip->first->last_event(), ip->first->module(id));
                    ip->first->last_event(evt_num );
                }
                ip->first->DataRecord(data);
            }
        }
        break;
    case BigDataRecord:
        evt_num = data[1] & 0xffffff;
        if (_last_event != evt_num )
        {
            NewEvent(evt_num );
            _last_event = evt_num;
        }
        for (ip = observers.begin(); ip != observers.end(); ++ip )
        {
            if ( (ip->second == data_type ) || ( 0xffffffff == ip->second))
            {
                if (evt_num != ip->first->last_event() )
                {
                    if (ip->first->last_event() >= 0)
                        ip->first->onEndOfEvent(ip->first->last_event(), ip->first->module(id ) );
                    ip->first->last_event(evt_num );
                }
                ip->first->DataRecord(data );
            }
        }
        break;
    case EndOfRunRecord:
        for (ip = observers.begin(); ip != observers.end(); ++ip)
        {
            ip->first->EndOfRunRecord(data);
        }
        break;
    case PedestalRecord:
        for (ip = observers.begin(); ip != observers.end(); ++ip)
        {
            ip->first->PedestalRecord(data);
        }
        break;
    case NewPointRecord:
        for (ip = observers.begin(); ip != observers.end(); ++ip)
        {
            ip->first->NewPointRecord(data);
        }
        break;
    case MapRecord:
        for (ip = observers.begin(); ip != observers.end(); ++ip)
        {
            ip->first->MapRecord(data);
        }
        break;
    case ScanRecord:
        for (ip = observers.begin(); ip != observers.end(); ++ip)
        {
            ip->first->ScanRecord(data);
        }
        break;
    case ConfigRecord:
        for (ip = observers.begin(); ip != observers.end(); ++ip)
        {
            ip->first->ConfigRecord(data);
        }
        break;
    default:
        std::cout << "*** Unknown blocktype "<< std::hex << header << std::dec<< std::endl;
        break;
    }
    for (ip = observers.begin(); ip != observers.end(); ++ip)
    {
        ip->first->EndOfRecord(data);
    }
    alloc.deallocate(data, block_size);
    return header;
}

void VMEdaq::rewind()
{
    if (_in)
        gzrewind(_in);
}

static bool __stop_loop__ = false;
static void
termination_handler(int sgn)
{
    __stop_loop__ = true;
}

void VMEdaq::loop()
{
    unsigned int iok;
    struct sigaction new_action, old_action;
    
    /* Set up the structure to specify the new action. */
    new_action.sa_handler = termination_handler;
    sigemptyset (&new_action.sa_mask);
    new_action.sa_flags = 0;
    
    sigaction (SIGINT, &new_action, &old_action);
    __stop_loop__ = false;
    do
    {
        iok = read();
        if ( __stop_loop__ )
        {
            __stop_loop__ = false;
            break;
        }
    } 
    while (iok);

    // restore the SIGINT action
    sigaction (SIGINT, &old_action, NULL);
    
    // Check if we still need to call onEndPoint
    ObserverList::iterator ip;
    for (ip = observers.begin(); ip != observers.end(); ++ip)
    {
        if (ip->first->npoint() >= 0)
        {
            ip->first->onEndPoint();
        }
    }
}

/***********************************************************************
 * VMEdaqObserver stuff
 ***********************************************************************
 */
VMEdaqObserver::~VMEdaqObserver()
{
    // delete all the Modules in the list
    ModuleList::iterator ip;
    for (ip = _idlist.begin(); ip != _idlist.end(); ++ip)
        delete ip->second;

}

bool VMEdaqObserver::data_valid(unsigned int *data)
{
    unsigned int dsz, sz;
    dsz = data[0] & 0xfff;
    sz = data[2] & 0xfff;

    // Check data integrity
    if (dsz != data[dsz-1]|| dsz - sz != 4)
    {
        return false;
    }
    else
    {
        return true;
    }
}

// Called when a StartOfRun record is found
void VMEdaqObserver::StartOfRunRecord(unsigned int *data)
{
    int i, nmod = data[3];
    for (i = 0; i < nmod; i++)
    {
        int id = data[4+i] & 0xFFFFL;
        int type = (data[4+i] & 0xf0000L);
        int nchip = (data[4+i] & 0xfff00000) >> 20;

        ModuleList::iterator ip = _idlist.find(id);
        if (ip == _idlist.end())
        {
            Module *m = new Module(id, nchip, type);
            ip = _idlist.insert(ModuleList::value_type(id, m)).first;
        }
        
        DataTypeList::iterator dtp = _data_types.find(type);
        if (dtp==_data_types.end())
        {
            _data_types[type] = Idlist();
        }
        _data_types[type].push_back(id);
    }
    _run_number = data[2];
    _run_starts = data[1];

    onNewRun();
}

void VMEdaqObserver::EndOfRunRecord(unsigned int *data)
{
    _run_stops = data[1];
    onEndOfRun();
}

// Called when a PedestalRecord is found
void VMEdaqObserver::PedestalRecord(unsigned int *data)
{
    int id = data[1];
    int i, j, sz = data[0] & 0xfffL;
    int nchip = (data[2] & 0xffff0000L) >> 16;
    int nchan = (data[2] & 0xffffL);
    int ntot = nchan * nchip;

    // Check that the module has the data type that we want
    if (_type!=0xffffffff)
    {
        Idlist::iterator im = std::find(_data_types[_type].begin(), _data_types[_type].end(), id);
        if (im==_data_types[_type].end())
            return;
    }
    
    // find the module in the list. It should be there already
    Module *md = module(id);
    if (md == 0)
    {
        std::cout << "VMEdaqObserver::PedestalRecord: module "<< id 
                  << " has not been seen yet. This shouldn't happen..."
                  << std::endl;
        return;
    }

    
    // Reset the currentpedestals and noise
    md->nchip(nchip);
    md->nchan(nchan);
    md->ntot(nchip*nchan );

    std::vector<double> &ped = md->pedestals();
    std::vector<double> &noise = md->noise();

    ped.clear();
    noise.clear();

    sz--;
    for (j = 0, i = 3; i < sz && j < ntot; i++, j++)
    {
        noise.push_back( ((data[i] & 0xffff0000L) >> 16) / 50.);
        ped.push_back( (data[i] & 0xffffL) / 50.);
    }
    onPedestals(md);
}

// Called when a new point in the scan starts
void VMEdaqObserver::NewPointRecord(unsigned int *data)
{
    if (_npoint >= 0)
        onEndPoint();

    int i;
    int nvar = (data[1] & 0xffL);
    //int nevts = (data[1] & 0xffffff00L)>>8;

    // reset the counter
    _cnt = 0;

    std::vector<Point> vars;
    for (i = 0; i < nvar; i++)
    {
        int type = (data[2+i] & 0xf0000000L) >> 28;
        int val = (data[2+i] & 0xfffffffL);
        vars.push_back(Point(type, val) );
    }
    onNewPoint(vars);

    _npoint ++;
}

// Called when a MapRecord is found
void VMEdaqObserver::MapRecord(unsigned int *data)
{
    int i, j;
    int size = (data[0] & 0xfffL);
    int id = data[1];
    //int nchan = (size-4)*2;
    int nx = (data[2] >> 16);
    int ny = (data[2] & 0xffffL);
    
    // Check that the module has the data type that we want
    if (_type!=0xffffffff)
    {
        Idlist::iterator im = std::find(_data_types[_type].begin(), _data_types[_type].end(), id);
        if (im==_data_types[_type].end())
            return;
    }

    // find the module in the list. It should be there already
    Module *md = module(id);
    if (md == 0)
    {
        std::cout << "VMEdaqObserver::MapRecord: module "<< id
                  << " has not been seen yet. This shouldn't happen..."
                  << std::endl;
        return;
    }
    
    int ntot = nx*ny;
    if (ntot<md->ntot())
        ntot = md->ntot();
    
    PadMap mp(nx, ny, ntot);

    // ignore the last word
    size--;
    for (i = 0, j = 3; j < size; j++)
    {
        int ix, iy;
        ix = (data[j] >> 24) & 0xffL;
        iy = (data[j] >> 16) & 0xffL;
        mp.set(i, ix, iy);
        i++;

        ix = (data[j] >> 8) & 0xffL;
        iy = (data[j]) & 0xffL;
        mp.set(i, ix, iy);
        i++;
    }
    md->pad_map(mp);

    onMap(md);
}

// Called when a ScanRecord is found
void VMEdaqObserver::ScanRecord(unsigned int *data)
{
    union
    {
        unsigned int lval;
        float fval;
    } from, to;
    bool is_float;
    int i, j, nvar, nevt;
    int type, npts;
    double first, last, step;

    nvar = data[1];
    nevt = data[2];

    _scan_point.reset();
    for (i = 0; i < nvar; i++)
    {
        j = 3 * (i + 1);
        is_float = data[j] & 0x80000000L;
        type = ((data[j] & 0xfff0000L) >> 16);
        npts = (data[j] & 0xffffL);
        from.lval = data[j+1];
        to.lval = data[j+2];
        first = is_float ? from.fval : (double)from.lval;
        last = is_float ? to.fval : (double)to.lval;
        if (npts<=1)
            step = 1.;
        else
            step = (last - first) / (npts - 1.);
        
        ScanVar var( (ScanVar::ScanType)type, first, last, step);
        _scan_point.add_var(var);
    }

    _nevt_per_point = nevt;

    onScan();
}

int VMEdaqObserver::get_module_data_type(int id) const
{
    const Module *md = module(id);
    return md ? md->type() : 0;
}

Channel *VATAChannel::Clone()
{
    return new VATAChannel(*this);
}

VATAConfig::VATAConfig(int ntot):
    _mask(ntot, false), _test(ntot, false), _trim(ntot, false)
{
    
}

int VATAConfig::load_config(const char *fnam) 
{
    std::ifstream ifile(fnam);
    std::string dummy;
    int i, _nchip, _nchan, _ntot, _mask, _trim, _test;
    
    if (!ifile)
        return -1;
    
    ifile >> dummy >> _nchip >> dummy >> _nchan >> dummy >> _ntot;
    for (i=0; i<_ntot; i++)
    {
        ifile >> _mask >> _trim >> _test;
        mask(i, _mask);
        trim(i, _trim);
        test(i, _test);
    }
    
    ifile.close();
    return 0;
}
int VATAConfig::dump_config(const char *fnam) 
{ 
    return 0; 
}


void VATAGP::ConfigRecord(unsigned int *data)
{
    unsigned int id = data[1];
    int ich, nchip, nchan, ntot;
    
    Module *md = module(id);
    nchip = data[4]>>16;
    nchan = data[4] & 0xffff;
    ntot  = nchip*nchan;
    VATAConfig *cfg = new VATAConfig(ntot);
    
    cfg->chip_type( data[2]);
    cfg->thrs( data[3]);
    for (ich=0;ich<ntot;ich++)
    {
        unsigned int val = data[ich+5];
        cfg->mask( (val & 0xf000)>>12 );
        cfg->test( (val & 0xf00)>>8);
        cfg->trim( val & 0xff);
    }
    md->set_config(cfg);

    //this doesn't work. Do something better!
    std::cout<<"Module "<<id<<" decoding sparse set to gp7. Change in line 635 of vmedaq_cc.cc in analysis!"<<std::endl;
    decode_sparse = &VATAGP::decode_sparse_gp7;
    //while(std::cin.get()!='\n');
    

    
//     switch (cfg->chip_type())
//     {
//         case 5: // GP7
//             break;
//         default:
//             decode_sparse = &VATAGP::decode_sparse_gp3;
// 	    std::cout<<"Module "<<id<<" decoding sparse set to gp3"<<std::endl;
// 	    while(std::cin.get()!='\n');
//             break;
//     }
}

void VATAGP::DataRecord(unsigned int *data)
{
    _cnt ++;
    _ndata ++;

    ChannelList out;
    int id = (data[2] & 0xfff00000L) >> 20;
    int romode = (data[2] & 0xf000L) >> 12;
    int evtcnt = (data[1] & 0xffffff);

    // Check the data integryty
    unsigned int sz = (data[2] & 0xfffL);
    unsigned int dsz = data[0] & 0xfff;
    if (dsz != data[dsz-1]|| dsz - sz != 4)
        return;

    _romode = romode;
    if (romode == 2)
    {
        decode_serial(data, out);
        _nadj = 0;
    }
    else
    {
      (this->*decode_sparse)(data, out);
    }
    Module *md = module(id);

    onNewData(md, evtcnt, out);
}

void VATAGP::decode_serial(unsigned int *data, ChannelList &out)
{
    int sz = (data[0] & 0xfffL);
    //int ntot = sz-6;
    int j, i;

    sz--;
    for (i = 0, j = 6; j < sz; j++, i++)
    {
        int adc = data[j] & 0x3ffL;
        out.push_back(new VATAChannel(i, (double)adc, false));
    }
}

void VATAGP::decode_sparse_gp3(unsigned int *data, ChannelList &out)
{
     int dsz = (data[0] & 0xfff)-1;
    int sz = (data[2] & 0xfffL);
    int romode = (data[2] & 0xf000) >> 12;
    int nTrig, nblk, ndat;

    _nadj = romode == 8 ? (data[1] & 0xff000000L) >> 24 : 0;
    nblk = _nadj + 2;
    ndat = nblk -1;

    for (int j = 6, nTrig = 0; j < dsz; nTrig++)
    {
        /*
         * We fix here the multihit problem
         */
        if (nTrig>1)
            j++;
        
        if ((dsz-j)<ndat)
            break;
        
        //channel that gave this trigger
        int add = -1111;

        //for (i=0;i<nblk;i++) {
        for (int i = 0; i < ndat && j<dsz; i++, j++)
        {
            int dt = (data[j-3] & 0x7ff00000L) >> 20;
            int down = data[j] & 0x3ffL;
            int up = (data[j] & 0xffc00L) >> 10;

            if (i == 0)
            {
                add = dt;
                //adjacent flag is false
                out.push_back(new VATAChannel(add, down, false));
                if (_use_up )
                    out.push_back(new VATAChannel(add, up, false));
            }
            else
            {
                //set the adjacent flag to true
                out.push_back(new VATAChannel(add + i, down, true));
                if (_use_up )
                    out.push_back(new VATAChannel( add - i, up, true ));
            }
        }
    }
}

void VATAGP::decode_sparse_gp7(unsigned int *data, ChannelList &out)
{
    int dsz = (data[0] & 0xfff)-1;
    int sz = (data[2] & 0xfffL);
    int romode = (data[2] & 0xf000) >> 12;
    int nblk, ndat;

    _nadj = romode == 8 ? (data[1] & 0xff000000L) >> 24 : 0;
    nblk = _nadj + 2;
    ndat = nblk -1;

    //dump raw data
    //    for (int i=3;i<dsz;i++){
//       std::cout<<i<<' '<<std::hex<<data[i]<<std::dec<<' '
// 	       << ((data[i] >> 20) & 0x7ff)<<' '<< ((data[i]>> 10) & 0x3ff) << ' '
// 	       << (data[i] & 0x3ff ) << std::endl;
//    }



    for (int j = 6, nTrig = 0; j < dsz; nTrig++)
    {
        /*
         * We fix here the multihit problem
         */
        if (nTrig>1)
            j++;
        
        if ( (dsz-j)<ndat )
            break;
        
        //channel that gave this trigger
        int add = -1111;
        ndat = nblk;
        for (int i = 0; i < ndat && j<dsz; i++, j++)
        {
            int dt = (data[j-3] & 0x7ff00000L) >> 20;
            int down = data[j] & 0x3ffL;
            int up = (data[j] & 0xffc00L) >> 10;

            if (i == 0)
            {
                add = dt;
                //adjacent flag is false
                out.push_back(new VATAChannel(add, down, false));
                if (_use_up )
                    out.push_back(new VATAChannel(add, up, false));
                
                ndat--;
            }
            else
            {
                //set the adjacent flag to true
                out.push_back(new VATAChannel(add + i, down, true));
                if (_use_up )
                    out.push_back(new VATAChannel( add - i, up, true ));
            }
        }
    }
//     for (ChannelList::iterator it=out.begin();it!=out.end();it++){
//       VATAChannel *ch=dynamic_cast<VATAChannel *>(*it);
//       std::cout<<ch->channel()<<' '<<ch->signal()<<' '<<ch->adj()<<std::endl;
//     }
//     while(std::cin.get()!='\n');
}

void ADCData::DataRecord(unsigned int *data)
{
    int i, ip, ibuffer, nbuffer, add, val;

    if (!data_valid(data))
        return;

    _cnt ++;
    _ndata ++;

    ChannelList out;
    int id = (data[2] & 0xfff00000L) >> 20;
    int evtcnt = (data[1] & 0xffffff);
    Module *md = module(id);
    nbuffer = data[3];
    // std::cout<<"ADC("<<id<<") cnt="<<std::hex<<evtcnt<<std::dec
    //     <<" nbuffer="<<nbuffer<<std::endl;

    ip=4;
    for (ibuffer=0; ibuffer<nbuffer; ibuffer++)
    {
        DataBuffer *B = new DataBuffer();

        int header = data[ip]>>16;
        if (header!=0xacdc)
        {
            std::cout << "Wrong event format"<< std::endl;
            return;
        }
        int nchan = data[ip++] & 0xffff;
        for (i=0; i<nchan; i++, ip++)
        {
            add = (data[ip] >> 16) & 0x1F;
            val = data[ip] & 0xFFF;
            B->list().push_back(new Channel(add, val));
        }
        // skip the event counter
        B->gate_counter(data[ip++]);
        out.push_back(B);
    }
    onNewData(md, evtcnt, out);
}

void TDC488Data::DataRecord(unsigned int *data)
{
    if (!data_valid(data))
        return;

    _cnt ++;
    _ndata ++;

    ChannelList out;
    int id = (data[2] & 0xfff00000L) >> 20;
    int i,sz  = data[2] & 0xfff;
    int evtcnt = (data[1] & 0xffffff);
    int chan, val;
    Module *md = module(id);
    
    DataBuffer *B=0;
    for (i=0;i<sz;i++)
    {
        unsigned int dt = (data[3+i] & 0xffff);
        if (dt & 0x8000)
        {
            B = new DataBuffer();
            B->gate_counter( dt & 0xfff );
            out.push_back(B);
        }
        else
        {
            if ( B )
            {
                val  = dt & 0xfff;
                chan = (dt & 0x7000)>>12;
                B->list().push_back(new Channel(chan, val));
            }
            else
            {
                std::cout << "TDC488Data::DataRecord - this should not happen" << std::endl;
            }
        }
    }
    onNewData(md, evtcnt, out);
}

void FPGAv1495Data::DataRecord(unsigned int *data)
{
    int i, ip, ibuffer, nbuffer, add, val;

    if (!data_valid(data))
        return;

    _cnt ++;
    _ndata ++;

    ChannelList out;
    int id = (data[2] & 0xfff00000L) >> 20;
    int evtcnt = (data[1] & 0xffffff);
    Module *md = module(id);

    long long unsigned int hld = data[4];
    evt_time = data[3] | (hld << 32);
    
    //nbuffer = data[3];
    //ip=4;
    
    int nCoinc=9;

    unsigned int size = data[0] & 0xfff;
    coinc_array = data[size-2];
    if (size>39) trig_reg= data[size-3];
    else trig_reg=0;
    int n_trigger=32;

    DataBuffer *B = new DataBuffer();
    B->gate_counter(0);

//     std::cout<<"Coinc="<<std::hex<<coinc_array<<std::dec<<" cnt="<<std::hex<<evt_time<<std::dec
// 	     <<" trig_reg="<<std::hex<<trig_reg<<std::dec
// 	     <<std::endl;

    for (int i=5,j=0; j<n_trigger; i++,j++) {
      double sig = double(data[i]) - double(evt_time&0xffffffff);
      int channel =j;
      B->list().push_back(new Channel(j,sig));
      //std::cout<<i<<' '<<std::hex<<data[i]<<std::dec<<' '<<j<<' '<<sig<<std::endl;
    }
    out.push_back(B);
    
    onNewData(md, evtcnt, out);
}
