************************************************************************
*  
*  This is an example of how to use the VMEdaq-f77 library to read the
*  output files of VMEDAQ. 
* 
*  First a few words on how the library works. The idea is that you have
*  a "file reader" that will call user supplied routines every time it
*  finds a new data record. What is a data record ? VMEDAQ stores the 
*  data in blocks, each block corresponding to different data types. 
*  For instance, pedestals and noise are stored in the Pedestal record,
*  the pixel map in the Map record, the detector data in the Data record,
*  etc. The idea is that you only "subscribe" to the type of data you 
*  want to receive. The possible data records for which one can register
*  are:
*
*      - NewFile;
*      - StartOfRun;
*      - EndOfRun;
*      - DataRecord;
*      - Pedestal;
*      - NewPoint;
*      - EndOfPoint;
*      - PixelMap;
*      - Scan;
*
*  The interface is quite simple:
*   
*  First of all, one needs to obtain a handler for the "file reader" 
*  with a call to 
*  
*         integer function vmedaq_new()
*  
*  this will return the handler, which will have to be given to any 
*  subsequent call to the library functions. To finish the session, one
*  just call
*  
*         call vmedaq_delete(handler)
*  
*  Now, once the handler is there, one queues files to be processed with
*  as many calls to
*  
*  	call vmedaq_append_file(handler,file_name)
*  
*  as needed.
*
*  One can, in principle read more than one file (or just one file). 
*  
*  In order to process the data, one needs to register a series of hooks
*  with calls like
*  
*       call vmedaq_subscribe(handler, key, function_name)
*  
*  handler is the handler we previously acquired, key is any of
*  
*  	+ s_on_new_file
*  	+ s_start_of_run
*   + s_end_of_run
*  	+ s_data_record
*  	+ s_pedestal
*   + s_new_point
*   + s_scan
*   + s_map
*  
*  and function name is a function reference that will receive different
*  parameters according to the value of key. In tha way, when the 
*  library finds a start of run record, or a pedestal record or a data
*  record, it will call the function given with its corresponding key. 
*  If no function is registered, the library will ignore that block.
*  
*  See the examples below for the prototype of the different hooks.
*  
*  Once the hooks are registered and the data files queued, one uses
*  
*       integer function vmedaq_read(handler)
*  
*  
*  handler is the handler, and the function will read the next data
*  block, call the corresponding hooks and return the data block type
*  just read or 0 is there is no more data available. In order to figure
*  out which data we just read, one can use
*  
*      logical function is_data_block(x)
*      logical is_start_of_run(x)
*      logical is_data_record(x)
*      logical is_end_of_run(x)
*      logical is_pedestal_record(x)
*      logical is_new_point_record(x)
*      logical is_scan_record(x)
*      logical is_map_record(x)
*  
*  with x being the value returned bu vmedaq_read.
*  
*  And that is it  
***********************************************************************


        
      SUBROUTINE onNewFile(file_name)
*@(#)*******************************************************************
* onNewFile
*
*  Called every time a new file is opened
*
************************************************************************
      character *(*) file_name

      print *,'New file:',file_name

 999  RETURN
      END


      SUBROUTINE onStartOfRun(nmod,vid,vxtra,vnchip)
*@(#)********************************************************************
* onStartOfRun
*
*     Called every time a new start of run record is read. The
*     parameters are all input parameters and their meaning is as
*     follows: 
*        nmod   - number of modules
*        vid    - array of size nmod with the modules' id numbers
*        vxtra  - array of size nomd with a 'magic' bit...
*        vnchip - array of size nmod with the number of chips in each
*                 module. 
*
*************************************************************************
      include 'common_block.inc'
      integer nmod,vid(nmod),vxtra(nmod),vnchip(nmod)

      print *,'*** Start of Run ***'
      print *,'Number of modules:',nmod
      write(*,1000)
      do i=1,nmod
         write(*,1001) vid(i),vxtra(i),vnchip(i)
      enddo

      id = vid(1)
      nchip = vid(1)

 1000 format(1x,'  id   xtra  nchip')
 1001 format(1x,I4,3x,I4,5x,i2)
 999  RETURN
      END

      SUBROUTINE onPedestal(iid,inchip,inchan,intot,ped,ns)
*@(#)********************************************************************
* onPedestal
*
*     Called every time a pedestal record is found. All the parameters
*     are input parameters:
*        iid    -    id of current module
*        inchip -    number of chips of current module
*        inchan -    number of channels per chip in the module
*        intot  -    total number of channels (nchip*nchan)
*        ped    -    array with pedestals
*        ns     -    array with noise
*    
*************************************************************************
      include 'common_block.inc'
      integer i,iid,inchip,inchan,intot
      real ped(intot),ns(intot)

      integer npass
      data npass /0/
      save npass

      if (iid.ne.id) return
      if (npass.gt.0) return

      npass = npass + 1 
      nchip = inchip
      nchan = inchan
      ntot  = intot

      do i=1,ntot
         module_pedestal(i) = ped(i)
         module_noise(i)    = ns(i)
      enddo

      print *,'*** Pedestal Record ***'
      write(*,1000) id,nchip,nchan



 1000 format(1x,'id=',i3,1x,'nchip=',i3,1x,'nchan=',i3)


 999  RETURN
      END


      SUBROUTINE onScan(nvar,nevts,type,npts,from,to)
*@(#)********************************************************************
* onScan
*     Called every time a scan definition record is found.
*     All the parameteres are input parameters
*         - nvar : number of variables in the scan
*         - nevts: number of events per scan point
*         - type : array of nvar integers containing the variable type
*         - npts : array of nvar integers vontaining the number of points
*                  in the scan of this variable
*         - from : array of nvar doubles containing the starting value
*         - to   : arrat of nvar doubles containing the end value
*
*************************************************************************
      integer nvar
      integer type(nvar),npts(nvar)
      real    from(nvar),to(nvar)

      RETURN
      END


      SUBROUTINE onMap(iid,inchan,inx,iny,vx,vy)
*@(#)********************************************************************
* onMap
*     Called every time a pixel map record is found
*     All the variables are input parameters
*         - id id of the module this map belongs to
*         - nchan number of channels
*         - nx number of pixels in X
*         - ny number of pixels in Y
*         - vx: Array of size nchan with the x coordinate of the 
*               pixel associated to that channel
*         - vy: Array of size nchan with the y coordinate of the 
*               pixel associated to that channel
*
*     Note: It may happen that nchan > nx*ny, 
*           so do not expect ncha=nx*ny
*           
*           I a channel is not "connected" to a pixel, 
*           both vx[i] and vy[i] are 0xff (256)
*
*************************************************************************
      include 'common_block.inc'
#if defined(USE_ROOT)
      include 'Root.inc'
#endif
      integer i,iid,inchan,inx,iny
      integer vx(inchan),vy(inchan)

      nx = inx
      ny = iny
      write(*,1000) iid,inchan,nx,ny
      open(10,file='map.txt')
      do i=1,inchan
         chan_x(i) = vx(i)
         chan_y(i) = vy(i)
         pad_chan(vx(i),vy(i))=i
         write(10,1100) i,chan_x(i),chan_y(i)
      enddo
      close(10)
#if defined(USE_ROOT)
      if (hMap.eq.0) then
         hMap = TH2F_new('hMap','HitMap',nx,0.,real(nx),ny,0.,real(ny))
      endif
      
      if (hHit.eq.0) then
         hHit = TH1F_new('hHit','HitMap 1D',inchan,0.,real(inchan))
      endif
#endif
 1000 Format('MAP section: module',i4,1x,'number of channels',i4,1x,
     .       'nx',1x,I2,1x,'ny',1x,i2)
 1100 Format(3I6)
      RETURN
      END


      SUBROUTINE onDataRecord(iid,nevt,nadj,romode,nch,chan,val,adj)
*@(#)********************************************************************
* onDataRecord
*
*     Called every time there is a data record. Again all the parameters
*     are input parameters:
*         - iid    id number of this module
*         - nevt   event number
*         - nadj   number of adjacent channels. Having a number there does
*                  not mean that we have sparse.
*         - romode read out mode (=2 serial, =4 sparse, =8 sparse adj)
*         - nch    number of channels with data
*         - chan   array of size nch containing the channel numbers. In
*                  case of sparse, the very first one is the one triggering
*                  the system
*         - val    array of size nch containig the signal of the channels   
*         - adj    array of size nch. A non zero value means this is an
*                  adjacent channel. A zero value means the channel can 
*                  be considered for signal analysis
*************************************************************************
      include 'common_block.inc'
#if defined(USE_ROOT)
      include 'Root.inc'
#endif
      integer i,ipad
      integer iid,nevt,nadj,romode,nch,chan(nch),adj(nch)
      real    val(nch),out,raw,px,py
      
      real  xn,xv

      integer ntry
      data ntry /0/
      save ntry

      if (iid.ne.id) return
      
      ntry = ntry + 1

*
*     Remove pedestals
*
      do 100 i=1,nch
         ipad = chan(i)+1
         if ( ipad.lt.1 .or. ipad.gt.ntot )
     .        goto 100

         raw = val(i)
         out = raw - module_pedestal(ipad)


#if defined(USE_ROOT)
         if (adj(i).ne.0) then
            call TH1_Fill(hZero,out,1.)
            goto 100
         endif

         call TH1_Fill(hRaw,raw,1.)
         call TH1_Fill(hSignal,out,1.)
            
         px = chan_x(ipad)+0.5
         py = chan_y(ipad)+0.5
         if (hMap.ne.0) then
            call TH2_Fill(hMap,px,py,1.)
         endif
         if (hHit.ne.0) then
            call TH1_Fill(hHit,ipad-0.5,1.)
         endif
#endif
 100  continue

      if ( mod(ntry,10000).eq.0 ) then
         write(*,1000) ntry
      endif
 1000 format('\r',1I8,$)
 1100 format(i8,1x,i3,1x,i3,1x,i3,1x,'[',F5.0,']')
      return
      end
      


      integer function slen(line)
      character *(*) line
      j=len(line)
      do i=j,1,-1
         if (line(i:i) .ne. ' ') goto 100
         j = j-1
      enddo
 100  slen = j
      return
      end

      program PrgSpectrum
*@(#)********************************************************************
* Test program
*    
*************************************************************************
      character *256 fname
      integer slen
      include 'vmedaq.inc'
      include 'common_block.inc'
#if defined(USE_ROOT)
      include 'Root.inc'
#endif
      external onNewFile,onStartOfRun,onPedestal,onDataRecord
      external onMap
      integer vmedaq
      integer iok,i

      hMap = 0
      hHit = 0
*
*     Check the number of input files
*
      if (iargc().le.0) then
         print *,'I need at least one input file'
         goto 999
      endif

*
*     Initialize things in the common block
*
      do i=1,buffer_size
         pedestal(i) = 0.
         noise(i) = 0.
         module_pedestal(i)=0.
         module_noise(i)=0.
      enddo
*
*     Get the handler
*
      vmedaq = vmedaq_new()


*
*     Register the hooks
*
*     The first 2 are dummy
      call vmedaq_subscribe(vmedaq,s_on_new_file, onNewFile)
      call vmedaq_subscribe(vmedaq,s_start_of_run,onStartOfRun)

*     Read the map
      call vmedaq_subscribe(vmedaq,s_map,onMap)

*     
*     This one just stores the pedestals in the file
*
      call vmedaq_subscribe(vmedaq,s_pedestal,onPedestal)
*
*     Subscribe to the real data
*
      call vmedaq_subscribe(vmedaq,s_data_record,onDataRecord)



*/etc/root/system.rootmap
*     Queue one file for processing
*
      do i=1,iargc()
         call getarg(i,fname)
         il = slen(fname)
         print *,fname(1:il)
         call vmedaq_append_file(vmedaq,fname(1:il))
      enddo
      
#if defined(USE_ROOT)
      ffd     = TFile_new('output.root')
      hSignal = TH1F_new('hSignal','The signal',256,0.,1024.)
      hRaw    = TH1F_new('hRaw','The raw data',256,0.,1024.)
      hZero   = TH1F_new('hZero','No signal',128,-128.,128.)
#endif

*
*     Read while there is data
*
      call vmedaq_read_all(vmedaq)
      print *,'End of File'

#if defined(USE_ROOT)
      hPed    = TH1F_new('hPed','Pedestals',ntot,0.,real(ntot))
      do i=1,ntot
         vchan = i - 0.5
         vval  = module_pedestal(i)
         call TH1_Fill(hPed,vchan,vval)
      enddo
#endif

      call vmedaq_delete(vmedaq)

      print *,'Bye'
#if defined(USE_ROOT)
      call TFile_Write(ffd)
      call TFile_Close(ffd)
#endif

 999  return 
 1000 format(I4,1x,F8.3)
      end
