#include <iostream>
#include <queue>
#include <map>
#include <string>
#include <memory>
#include <vector>
//#include <g2c.h>
#include <zlib.h>
#include <analysis/DataFormat.h>

typedef int integer;
typedef unsigned int uinteger;
typedef char *address;
typedef short int shortint;
typedef float real;
typedef double doublereal;
typedef struct
{
    real r, i;
} complex;
typedef struct
{
    doublereal r, i;
} doublecomplex;
typedef integer logical;
typedef short int shortlogical;
typedef char logical1;
typedef char integer1;
typedef long int longint; /* system-dependent */
typedef unsigned long int ulongint; /* system-dependent */
#define qbit_clear(a,b) ((a) & ~((ulongint)1 << (b)))
#define qbit_set(a,b) ((a) |  ((ulongint)1 << (b)))

typedef int /* Unknown procedure type */(*U_fp)(...);
typedef shortint (*J_fp)(...);
typedef integer (*I_fp)(...);
typedef real (*R_fp)(...);
typedef doublereal (*D_fp)(...), (*E_fp)(...);
typedef /* Complex */void (*C_fp)(...);
typedef /* Double Complex */void (*Z_fp)(...);
typedef logical (*L_fp)(...);
typedef shortlogical (*K_fp)(...);
typedef /* Character */void (*H_fp)(...);
typedef /* Subroutine */int (*S_fp)(...);

typedef integer flag;
typedef integer ftnlen;
typedef integer ftnint;

enum BlockTypes
{
    onNewFile = 1,
    onStartOfRun,
    onDataRecord,
    onPedestal,
    onNewPoint,
    onMapRecord,
    onScanRecord,
    onEndOfPoint,
    last_block,
};

/*
 * This class controls the access to the data files
 */
class VMEdaq
{
    private:
        // opens the nes file in the queue, if any.
        // returns 0 on sucess and !=1 otherwise
        int open_next_file();

        int npoint;
    public:
        bool use_up;
        gzFile in;
        std::queue<std::string> files;

        S_fp onNewFile;
        S_fp onStartOfRun;
        S_fp onEndOfRun;
        S_fp onDataRecord;
        S_fp onPedestal;
        S_fp onNewPoint;
        S_fp onEndOfPoint;
        S_fp onPixelMap;
        S_fp onScan;

        static std::allocator<integer> int_alloc;
        static std::allocator<real> real_alloc;
        static std::allocator<uinteger> ulong_alloc;

        static std::map<integer, VMEdaq *> readers;
        static integer reader_cnt;

        /// constructor
        VMEdaq() :
            npoint(-1), use_up(false), in(0), onNewFile(0), onStartOfRun(0),
            onEndOfRun(0), onDataRecord(0), onPedestal(0), onNewPoint(0),
            onEndOfPoint(0), onPixelMap(0), onScan(0)
            {
            std::cout << "New VMEdaq"<< std::endl
            << "*** Note that I only decode VATAGP data ***"<< std::endl;
            }

        /// Destructor
        virtual ~VMEdaq()
        {
            std::cout << "Deleting VMEdaq"<< std::endl;
            if (in)
                gzclose(in);
        }

        int read();
        void close();
        void reset();

        void read_all();

        void decode_start_of_run(uinteger *data);
        void decode_data_record(uinteger *data);
        void decode_end_of_run(uinteger *data);
        void decode_pedestals(uinteger *data);
        void decode_new_point(uinteger *data);
        void decode_scan(uinteger *data);
        void decode_map(uinteger *data);

        integer decode_sparse(integer *, integer **chan, real **val, integer **adj,
                              uinteger *data);
        integer decode_serial(integer *, integer **chan, real **val, integer **adj,
                              uinteger *data);

};
std::allocator<integer> VMEdaq::int_alloc;
std::allocator<real> VMEdaq::real_alloc;
std::allocator<uinteger> VMEdaq::ulong_alloc;
std::map<integer, VMEdaq *> VMEdaq::readers;
integer VMEdaq::reader_cnt = 0;

void VMEdaq::close()
{
    // close any opened file
    if (in)
    {
        gzclose(in);
        in = 0;
    }
}

void VMEdaq::reset()
{
    close();
    while (!files.empty())
        files.pop();
}

int VMEdaq::open_next_file()
{
    close();

    while (!files.empty())
    {
        std::string str = files.front();
        files.pop();
        in = gzopen(str.c_str(), "rb");
        if (!in)
            continue;

        if (onNewFile)
            (*onNewFile)(str.c_str(), str.size());

        return 0;
    }
    return 1;
}

int VMEdaq::read()
{
    int iok, block_size, sz;
    uinteger header, x, *data;
    if (!in)
    {
        if (open_next_file() )
            return 0;
    }

    again:
    // read until we find a start of block
    do
    {
        iok = gzread(in, &x, sizeof(uinteger));
        if (iok <= 0)
        {
            // end of file, look for next one
            if (open_next_file() )
                return 0;

            x = 0;
            goto again;
        }
    } while (!IsDataBlock(x));

    header = x;

    block_size = x & 0xfff;
    sz = (block_size - 1) * sizeof(uinteger);
    data = ulong_alloc.allocate(block_size);

    data[0] = x;
    iok = gzread(in, data + 1, sz);
    if (iok != sz)
    {
        ulong_alloc.deallocate(data, block_size);
        x = 0;
        goto again;
    }

    // Now see who should we call
    x = (header & 0xf000);
    switch (x)
    {
        case StartOfRunRecord:
            if (onStartOfRun)
                decode_start_of_run(data);
            break;
        case DataRecord:
            if (onDataRecord)
                decode_data_record(data);
            break;
        case EndOfRunRecord:
            if (onEndOfRun)
                decode_end_of_run(data);
            break;
        case PedestalRecord:
            if (onPedestal)
                decode_pedestals(data);
            break;
        case NewPointRecord:
            if (onNewPoint)
                decode_new_point(data);
            break;
        case MapRecord:
            if (onPixelMap)
                decode_map(data);
            break;
        case ScanRecord:
            if (onScan)
                decode_scan(data);
            break;
        default:
            std::cout << "*** Unknown blocktype "<< std::hex << header << std::dec
            << std::endl;
            break;
    }

    ulong_alloc.deallocate(data, block_size);
    return header;
}

void VMEdaq::read_all()
{
    while (read() )
        ;
}

void VMEdaq::decode_start_of_run(uinteger *data)
{
    if (onStartOfRun == 0)
        return;

    integer i, nmod = data[3];
    integer *id = int_alloc.allocate(nmod);
    integer *xtra = int_alloc.allocate(nmod);
    integer *nchip = int_alloc.allocate(nmod);

    for (i = 0; i < nmod; i++)
    {
        id[i] = data[4+i] & 0xffffL;
        xtra[i] = (data[4+i] & 0xf0000L) >> 16;
        nchip[i] = (data[4+i] & 0xff00000) >> 20;
    }
    (*onStartOfRun)(&nmod, id, xtra, nchip);

    int_alloc.deallocate(id, nmod);
    int_alloc.deallocate(xtra, nmod);
    int_alloc.deallocate(nchip, nmod);
}
void VMEdaq::decode_data_record(uinteger *data)
{
    if (!onDataRecord)
        return;

    if (!IsNormalData(data[2]))
        return;

    integer id = (data[2] & 0xfff00000L) >> 20;
    integer nadj = (data[1] & 0xff000000L) >> 24;
    integer iromode = (data[2] & 0xf000L) >> 12;
    integer evtcnt = (data[1] & 0xffffff);
    integer bfsiz;
    integer nchan, *chan = 0, *adj = 0;
    real *val = 0;

    if (iromode == 8|| iromode == 4)
    {
        nchan = decode_sparse(&bfsiz, &chan, &val, &adj, data);
        if (iromode==4)
            nadj = 0;
    }
    else
    {
        nchan = decode_serial(&bfsiz, &chan, &val, &adj, data);
        nadj = 0;
    }
    if (nchan)
        (*onDataRecord)(&id, &evtcnt, &nadj, &iromode, &nchan, chan, val, adj);

    if (bfsiz)
    {
        int_alloc.deallocate(chan, bfsiz);
        int_alloc.deallocate(adj, bfsiz);
        real_alloc.deallocate(val, bfsiz);
    }
}
void VMEdaq::decode_end_of_run(uinteger *data)
{
}
void VMEdaq::decode_pedestals(uinteger *data)
{
    if (!onPedestal)
        return;

    integer id = data[1];
    integer i, j, sz = data[0] & 0xfffL;
    integer nchip = (data[2] & 0xffff0000L) >> 16;
    integer nchan = (data[2] & 0xffffL);
    integer ntot = nchan * nchip;

    real *pedestal = real_alloc.allocate(ntot);
    real *noise = real_alloc.allocate(ntot);

    sz--;
    for (j = 0, i = 3; i < sz && j < ntot; i++, j++)
    {
        noise[j] = ((data[i] & 0xffff0000L) >> 16) / 50.;
        pedestal[j] = (data[i] & 0xffffL) / 50.;
    }
    (*onPedestal)(&id, &nchip, &nchan, &ntot, pedestal, noise);

    real_alloc.deallocate(pedestal, ntot);
    real_alloc.deallocate(noise, ntot);
}
void VMEdaq::decode_new_point(uinteger *data)
{
    if (npoint >= 0&& onEndOfPoint)
        (*onEndOfPoint)();

    if (!onNewPoint)
        return;

    integer nvar = (data[1] & 0xffL);
    integer nevts = (data[1] & 0xffffff00L) >> 8;
    integer *ftype = int_alloc.allocate(nvar);
    integer *fval = int_alloc.allocate(nvar);

    int i;
    for (i = 0; i < nvar; i++)
    {
        ftype[i] = (data[2+i] & 0xf0000000L) >> 28;
        fval[i] = (data[2+i] & 0xfffffffL);
    }

    (*onNewPoint)(&nvar, &nevts, ftype, fval);
    npoint++;

    int_alloc.deallocate(ftype, nvar);
    int_alloc.deallocate(ftype, nvar);
}

void VMEdaq::decode_map(uinteger *data)
{
    if (!onPixelMap)
        return;

    int i, j;
    int size = (data[0] & 0xfffL);
    int id = data[1];
    integer nchan = (size - 4) * 2;
    integer nx = (data[2] >> 16);
    integer ny = (data[2] & 0xffffL);

    integer *vx = int_alloc.allocate(nchan);
    integer *vy = int_alloc.allocate(nchan);

    // ignore the last word
    size--;
    for (i = 0, j = 3; j < size; j++)
    {
        vx[i] = (data[j] >> 24) & 0xffL;
        vy[i] = (data[j] >> 16) & 0xffL;
        i++;

        vx[i] = (data[j] >> 8) & 0xffL;
        vy[i] = (data[j]) & 0xffL;
        i++;
    }

    (*onPixelMap)(&id, &nchan, &nx, &ny, vx, vy);

    int_alloc.deallocate(vx, nchan);
    int_alloc.deallocate(vy, nchan);
}

void VMEdaq::decode_scan(uinteger *data)
{
    if (!onScan)
        return;

    union
    {
        uinteger lval;
        float fval;
    } from, to;
    int j, is_float, ivar, nvar = data[1];
    int nevt = data[2];

    integer *ftype = int_alloc.allocate(nvar);
    integer *fnpts = int_alloc.allocate(nvar);
    real *ffrom = real_alloc.allocate(nvar);
    real *fto = real_alloc.allocate(nvar);

    for (ivar = 0; ivar < nvar; ivar++)
    {
        j = 3 * (ivar + 1);
        is_float = data[j] & 0x80000000L;
        ftype[ivar] = ((data[j] & 0xfff0000L) >> 16);
        fnpts[ivar] = (data[j] & 0xffffL);
        from.lval = data[j+1];
        to.lval = data[j+2];
        if (is_float)
        {
            ffrom[ivar] = from.fval;
            fto[ivar] = to.fval;
        }
        else
        {
            ffrom[ivar] = (real)from.lval;
            fto[ivar] = (real)to.lval;
        }
    }

    (*onScan)(&nvar, &nevt, ftype, fnpts, ffrom, fto);

    int_alloc.deallocate(ftype, nvar);
    int_alloc.deallocate(fnpts, nvar);
    real_alloc.deallocate(ffrom, nvar);
    real_alloc.deallocate(fto, nvar);
}

struct datahit
{
    int chan;
    int adj;
    int data;

    datahit(int c, int d, int a) :
        chan(c), adj(a), data(d)
        {
        }
};

integer VMEdaq::decode_sparse(integer *bfsiz, integer **chan, real **val, integer **adj,
                              uinteger *data)
{
    uinteger i, j;
    uinteger sz = (data[2] & 0xfffL);
    uinteger romode = (data[2] & 0xf000) >> 12;
    uinteger nadj = romode == 8 ? (data[1] & 0xff000000L) >> 24 : 0;
    uinteger nTrig;
    uinteger nblk = nadj +2;
    uinteger ndat = nblk-1;

    std::vector<datahit> out;

    *chan = 0;
    *val = 0;
    *bfsiz = 0;

    for (j = 6, nTrig = 0; j < sz; nTrig++)
    {
        if (nTrig>1)
            j++;
        
        //check if there is data enough for another channel with a trigger
        // do not care if we are missing just the last line which is useless
        //if ( sz-j< ndat )
        //    break;

        //channel that gave this trigger
        int add = -1111;

        for (i = 0; i < ndat; i++, j++)
        {
        	if (j>=sz)
			break;
            int dt = (data[j-3] & 0x7ff00000L) >> 20;
            int down = data[j] & 0x3ffL;
            int up = (data[j] & 0xffc00L) >> 10;

            if (i == 0)
            {
                add = dt;
                //adjacent flag is false
                out.push_back(datahit(add, down, 0) );
                if (use_up)
                    out.push_back(datahit(add, up, 0));
            }
            else
            {
                //set the adjacent flag to true
                out.push_back(datahit(add + i, down, 1));
                if (use_up)
                    out.push_back(datahit(add - i, up, 1));
            }
        }
    }

    int ntot = out.size();
    *chan = int_alloc.allocate(ntot);
    *adj = int_alloc.allocate(ntot);
    *val = real_alloc.allocate(ntot);
    *bfsiz = ntot;

    std::vector<datahit>::iterator ip;
    for (i=0, ip=out.begin(); ip!=out.end(); ++ip, ++i)
    {
        datahit &h = *ip;
        (*chan)[i] = h.chan;
        (*adj)[i] = h.adj;
        (*val)[i] = h.data;
    }
    return i;
}
integer VMEdaq::decode_serial(integer *bfsiz, integer **chan, real **val, integer **adj,
                              uinteger *data)
{
    uinteger i, j;
    uinteger sz = (data[2] & 0xfffL);
    real adc;

    *chan = 0;
    *val = 0;
    *bfsiz = 0;
    // check integrity
    uinteger dsz = data[0] & 0xfff;
    integer ntot = (dsz - 6);
    if (dsz != data[dsz-1]|| dsz - sz != 4)
        return 0;

    if (ntot <= 0)
        return 0;

    *chan = int_alloc.allocate(ntot);
    *adj = int_alloc.allocate(ntot);
    *val = real_alloc.allocate(ntot);
    *bfsiz = ntot;

    dsz--;
    for (i = 0, j = 6; j < dsz; j++, i++)
    {
        adc = data[j] & 0x3ffL;
        (*chan)[i] = i;
        (*val)[i] = adc;
        (*adj)[i] = 0;
    }

    return i;
}

#define check_handler(handler, name)     \
    VMEdaq *v=0;        \
    std::map<integer, VMEdaq *>::iterator ip = VMEdaq::readers.find( (*handler) ); \
    if (ip!=VMEdaq::readers.end()) {     \
        v = ip->second;       \
    }         \
    else {        \
        std::cout << #name << ": invalid handler" << std::endl;  \
        return 0;        \
    }

extern "C"
{
    integer vmedaq_new_(void);
    int vmedaq_delete_(integer *handler);
    int vmedaq_append_file_(integer *handler, char *file_name, ftnlen len);
    int vmedaq_subscribe_(integer *, integer *, S_fp);
    int vmedaq_close_(integer *handler);
    int vmedaq_reset_(integer *handler);
    integer vmedaq_read_(integer *handler);
    int vmedaq_read_all_(integer *handler);
    int vmedaq_ignore_(...);
    logical vmedaq_using_up_(integer *handler);
    int vmedaq_use_up_(integer *handler, integer &up);
    logical is_data_block_(integer *);
    logical is_start_of_run_(integer *);
    logical is_data_record_(integer *);
    logical is_end_of_run_(integer *);
    logical is_pedestal_record_(integer *);
    logical is_new_point_record_(integer *);
    logical is_scan_record_(integer *);
    logical is_map_record_(integer *);
}

integer vmedaq_new_(void)
{
    VMEdaq *v = new VMEdaq;
    integer handler = VMEdaq::reader_cnt++;
    VMEdaq::readers[ handler ] = v;
    return handler;
}
int vmedaq_delete_(integer *handler)
{
    check_handler(handler, vmedaq_delete)
    ;
    VMEdaq::readers.erase( *handler);
    delete v;
    return 0;
}

int vmedaq_ignore_(...)
{
    return 0;
}

logical vmedaq_using_up_(integer *handler)
{
    check_handler(handler, vmedaq_using_up)
    ;
    return v->use_up ? 1 : 0;
}

int vmedaq_use_up_(integer *handler, integer &up)
{
    check_handler(handler, vmedaq_use_up)
    ;
    v->use_up = (up != 0);
    return 0;
}

int vmedaq_subscribe_(integer *handler, integer *icase, S_fp func)
{
    check_handler(handler, vmedaq_subscribe)
    ;
    if (func == (S_fp)vmedaq_ignore_)
        func = (S_fp)NULL;

    switch ( *icase)
    {
        case onNewFile:
            v->onNewFile = func;
            break;
        case onStartOfRun:
            v->onStartOfRun = func;
            break;
        case onDataRecord:
            v->onDataRecord = func;
            break;
        case onPedestal:
            v->onPedestal = func;
            break;
        case onNewPoint:
            v->onNewPoint = func;
            break;
        case onEndOfPoint:
            v->onEndOfPoint = func;
            break;
        case onMapRecord:
            v->onPixelMap = func;
            break;
        case onScanRecord:
            v->onScan = func;
            break;
        default:
            break;
    }

    return 0;
}
int vmedaq_append_file_(integer *handler, char *file_name, ftnlen len)
{
    check_handler(handler, vmedaq_append_file)
    ;
    std::string file(file_name, len);

    v->files.push(file);

    return 0;
}

integer vmedaq_read_(integer *handler)
{
    check_handler(handler, vmedaq_read)
    ;
    return v->read();
}

int vmedaq_read_all_(integer *handler)
{
    check_handler(handler, vmedaq_read)
    ;
    v->read_all();
    return 0;
}

int vmedaq_close_(integer *handler)
{
    check_handler(handler, vmedaq_read)
    ;
    v->close();
    return 0;
}

int vmedaq_reset_(integer *handler)
{
    check_handler(handler, vmedaq_read)
    ;
    v->reset();
    return 0;
}

logical is_data_block_(integer *x)
{
    return IsDataBlock(*x);
}
logical is_start_of_run_(integer *x)
{
    return IsStartOfRunRecord(*x);
}
logical is_data_record_(integer *x)
{
    return IsDataRecord( *x );
}
logical is_end_of_run_(integer *x)
{
    return IsEndOfRunRecord(*x);
}
logical is_pedestal_record_(integer *x)
{
    return IsPedestalRecord (*x);
}
logical is_new_point_record_(integer *x)
{
    return IsNewPointRecord( *x );
}
logical is_scan_record_(integer *x)
{
    return IsScanRecord( *x );
}
logical is_map_record_(integer *x)
{
    return IsMapRecord( *x);
}
