#!/usr/bin/env python

"""pyvmedaq

This module defines VMEdaq as a class to read vmedaq output files
The format of the data is described below.

The idea behind the module is to register either a method or an object
that will be called for each data block found in the file. This is
done with the register method of the VMEdaq class. One can register as
many functions or objects as needed or wanted. One can even register
an object when creating the VMEdaq instance. If we register an
object, VMEdaq will call (if they exist) the methods: 
   - NewFile(self,file_name) : when a new file is opened
   - StartOfRunRecord(self,data): when a StartOfRunRecord is found
   - DataRecord(self,data): when a DataRecord is found
   - EndOfRunRecord(self,data): when an EndOfRunRecord is found
   - PedestalRecord(self,data): when a PedestalRecord is found
   - NewPointRecord(self,data): when a NewPointRecord is found
   - ScanRecord(self,data): when a ScanRecord is found
   - MapRecord(self,data): when a MapRecord
   - EndOfRecord(self, data_type, data): at the end of any record
   - ConfigRecord(self, data): When a config record is found

If we register a function, it will receive, for each data record, the
whole data buffer, and the user will have to figure out which data
record corresponds to.

The usual way of using VMEdaq is:
   1) Create an instance and, optionally, register already an object
      or function.

   2) Call the open method to open a file. This method accepts wither
      a single file name or a list of filenames.

   3) Call the read method until it returns None or, eventually, the
      loop method.

   4) Call the close method to close any file left opened.

There are several blocks in the data stream. At the beginning of each
file you should always find a StartOfRun record and a Pedestal record.

  The format of the StartOfRun record is

  Word 
   0     0xCAFE1bbb   bbb= size of data block
   1     0xtttttttt   tttttttt - time as returned by time(0)
   2     0xrrrrrrrr   rrrrrrrr - run number
   3     0xnnnnnnnn   nnnnnnnn - number of modules
   4     0x0nntiiii   nn = number of chips t = data type iiii= module id
   . 
  bbb    0x00000bbb   size of data block

  The format of the Pedestal data is the following
     Word
     0     0xCAFEibbb   i= event type bbbb= size of data block
     1     0xaaaaaaaa   aaaaaaaa = module id
     2     0xaaaabbbb   aaaa= n chips bbbb= nchan
     3     0xnnnnpppp   nnnn = 50*noise, pppp=50*pedestal
     .     .
     bbbb  bbbb         size of data block

  The format of the MapRecord is:
     Word
     0     0xCAFEibbb   i= event type bbbb= size of data block
     1     0xaaaaaaaa   aaaaaaaa = module id
     2     0xiiiijjjj   iiii = nx, jjjj=ny
     2     0xaaaabbbb   aaaa= pix. no. of chan 0, bbbb=pix no. of chan 1
     .     .
     bbbb  bbbb         size of data block

     where pixel numbers go like:
        pix. no. = y * nx + x


  The format of the ScanRecord is:
     Word
     0     0xCAFEibbb   i= event type bbbb= size of data block
     1     0xaaaaaaaa   number of variables
     2     0xaaaaaaaa   number of events per scan point
     3     0xiaaabbbb   i=value format aaaa=var type bbbb=number of points
     4     0xbbbbbbbb   first value
     5     0xbbbbbbbb   last value
     .     .
     bbbb  bbbb         size of data block

     Words 3-5 appear once per variable.
     The value format indicates how the values are stored. If it is
     one they are stored as float numbers, otherwise as integers.
     
  After that you may find Data records

  The format of the Data record is the following
  Word
   0     0xCAFEibbb   i= event type bbb= size of data block
   1     0xbbaaaaaa   bb=number of adjacent aaaaaaaa event counter
   2     0xiiitannn   iii=module id t=data type a= readout mode nnn=data size
   3     data word 0
   4     data word 1
   .     .
  bbbb    bbbb         size of data block

  -------------------------------------------------------------------------
  Scans

  Every scan has a record at the beggining of a new point. The
  structure of the  record is
  Word
   0     0xCAFEibbb   i= event type bbbb= size of data block
   1     0xbbbbbbnn   bbbbbb = number of events nn=number of variables
   2     0xtvvvvvvv   t var type, v=value
   .
  bbbb    bbbb         size of data block 


  -------------------------------------------------------------------------
  ADC data block

     Data Format
     0     0xCAFEtbbb   t = event type, bbb= size of data block
     1     0xaaaaaaaa   aaaaaaaa event counter
     2     0xiiitnnnn   iii=module id t=data type  nnnn=data size
     3     data word 0
     4     data word 1
     .     .FBTDCdata
    bbbb    bbbb         size of data block

    The data words have the format explained in the v785 manual
    data type is ADCdata
  -------------------------------------------------------------------------
  FBTDC data block

     Data Format
     0     0xCAFEtbbb   t = event type (2), bbb = size of data block
     1     0xaaaaaaaa   aaaaaaaa event counter
     2     0xiiitnnnn   iii=module id t=data type  nnnn=data size
     3     data word 0
     4     data word 1
     .     .
    bbbb    bbbb         size of data block

    The data words have the format explained in the 1877 manual
    data type is FBTDCdata

   
"""

__revision__ = "0.1"

import traceback
from _vmedaq import *

var_names = ("Channel", "Pulse", "Threshold", "Trim",
             "xposition", "yposition", "zposition",
             "angle1", "angle2", "angle3", "user1",
             "user2", "user3", "user4", "thrs_trim", "cal_dac")
(channel, pulse, threshold, trim, 
 xposition, yposition, zposition, 
 angle1, angle2, angle3, 
 user1, user2, user3,user4, thrs_trim, cal_dac ) = range(0, len(var_names))

class Item(object):
    """An item is a bunch of named values
    """
    def __init__(self, names, values):
        """
        Initializes the instance. The input parameters are:
           - names: a list of names
           - values: a list (same length as d) with values
        """
        self.__dict__.update( dict( zip(names, values) ) )

    def __str__(self):
        """String representation
        """
        return str(self.__dict__)

    def _repr__(self):
        """Representation
        """
        return str(self.__dict__)

class Module(object):
    """
    This class encapsulates the information of a Module
       - mid: the module id
       - dtype: the module data type
       - nchip: number of chips
       - nchan: number of channels per chip
       - ntot: total number of channels
       - map: the pixel map. It is a list of tuples (x,y), where the
         i-th element of the list contains the coordinates associated
         with the i-th chip channel
       - pedestals: an array of size ntot containing the pedestals
       - noise: an array of size ntot containing the pedestals
    """
    def __init__(self, mid, dtype=0, nchip=0):
        self.id        = int(mid)
        self.dtype     = long(dtype)
        self.nchip     = nchip
        self.nchan     = 0
        self.ntot      = 0
        self.map       = {}
        self.pedestals = []
        self.noise     = []
        self.nx        = 0
        self.ny        = 0


    def __str__(self):
        """ String representation
        """
        return 'Module %x [%x]: nchip %d nchan %d' % (
            self.id, self.dtype, self.nchip, self.nchan)

    def __repr__(self):
        """ Representation
        """
        return self.__str__()


class ModuleConfig(object):
    """
    This class stores the configuration information found
    in the data file
    """
    def __init__(self, mid):
        self.mid = int(mid) # module id
        
        

class SimpleVMEdaqObserver(object):
    """
    This is a simple example of a VMEdaq observer implementing all the
    possible hooks to receive the different data blocks. Each of them calls
    another method in which the data is more digested:
       onNewFile(file_name)
       onNewRun()
       onEndOfRun()
       onMap(module)
       onPedestal(module)

       onScan()
       onNewPoint(nevts, vars):
          nevts: number of expected events
          vars: a list of pairs( var type, value )
       onEndPoint()
       onNewData() : the arguments will depend on how the module data is
                     handled.
    """
    def __init__(self, dtype=None):
        """
        The main attributes of this class are:
           - data_type: This is the data type that we want to
                        analyze. Only Data blocks of this type will be
                        decoded.
                        
           - cnt:       the event counter. It is reset at the beginning
                        of any new scan point
                        
           - ndata:     It is a global counter of data blocks
           
           - npoint:    a counter of scan points
           
           - idlist:    This is a dictionary containing all the known Module
                        objects. The key of the dict is the module id
                        
           - vars:      list of variables in a scan
        """
        # error code
        self.error = None
        
        # data type
        if dtype:
            self.data_type = long(dtype)
        else:
            self.data_type = None
        # Event counter
        # In the case of a scan, it is reset at the beginning of any
        # new scan point
        self.cnt = 0

        # Global counter. Just counts the number of data records
        self.ndata = 0

        # Scan point counter
        self.npoint = -1

        # run mumber
        self.run_number = 0

        # start of run time
        self.run_starts = 0

        # end of run time
        self.run_stops = 0

        # extra byte
        self.xtra = 0
        
        # list of module ID's
        # Each element contains the module with that id
        self.idlist = {}

        # list of data types and the module ID's associated
        # to each of them
        self.typelist = {}
        self.type2id = {}

        # list of variables in a scan
        self.vars = ()

        # number of events per scan point
        self.nevt_per_point = 0

    def __dummy(self, *args):
        """
        A dummy routine to be passed with none of the standard
        names is found
        """
        pass
    
    def __getattr__(self, name):
        if name in ("onNewFile", "onNewPoint", "onEndPoint",
                    "onNewData", "onNewRun"  , "onMap",
                    "onScan"   , "onPedestal", "EndOfEvent",
                    "onEndOfRun", "onConfig"):
            return self.__dummy

        raise AttributeError, name

    def applicable(self, mid):
        """ This is a boolean function that will return true if the
            data block can be decoded by this Observer
        """
        mid = int(mid)
        if not self.data_type:
            return False
        
        if self.data_type != 0xffffffffL:
            if self.data_type not in self.type2id:
                return False
            
            if mid not in self.type2id[self.data_type]:
                return False
            
        return True

    def NewFile(self, file_name):
        """
        Called when a new file is opened. It calls
           onNewFile(file_name)
        """
        self.error = None
        self.onNewFile(file_name)

    def StartOfRunRecord(self, data):
        """
        Called when a new StartOfRun record is found.
        It calls onNewRun()
        Creates the first instance of any known module and 
        stores it in idlist
        """
        self.error = None
        nmod = data[3]
        self.idlist = {}
        for i in range(0, nmod):
            mid    = int(data[4+i] & 0xffffL);
            dtype  = long( (data[4+i] & 0xf0000L) );
            nchip  = int((data[4+i] & 0xff00000)>>20);
            module = Module(mid, dtype, nchip);
            self.idlist[mid] = module
            self.typelist.setdefault(dtype, []).append(module)
            self.type2id.setdefault(dtype, []).append(mid)

        self.run_number = data[2]
        self.run_starts = data[1]
        self.onNewRun()

    def EndOfRunRecord(self, data):
        """
        Called when a new EndOfRunRecord record is found.
        This is a dummy function
        """
        self.error = None
        self.run_stops = data[1]
        self.onEndOfRun()

    def MapRecord(self, data):
        """
        Called when a new MapRecord is found.
        It calls  onMap(module)
        Sets the map attribute of a Module. 
        """
        self.error = None
        mid = int(data[1]);
#        if not self.applicable(mid):
#            self.error = -1;
#            return
        
        m   = []
        sz  = (data[0] & 0xfffL)
        nx  = (data[2]>>16)
        ny  = (data[2]&0xffffL)
        for j in range(3, sz):
            ix = (data[j]>>24)&0xffL
            iy = (data[j]>>16)&0xffL
            m.append( Item(("x", "y"), (ix, iy)) )

            ix = (data[j]>>8)&0xffL
            iy = (data[j])&0xffL
            m.append( Item(("x", "y"), (ix, iy)) )
          
        try:
            module     = self.idlist[mid]
            module.nx  = nx
            module.ny  = ny
            module.map = tuple(m)
            self.onMap(module)
              
        except KeyError:
            self.error = -2
            print 'Module id', mid, 'not yet registered... Weird...'

        

    def ScanRecord(self, data):
        """
        Called when a new ScanRecord is found.
        It calls onScan()
        Sets the vars attribute with the list of variables in the scan
        """
        self.error = None
        nevts, scl = decode_scan(data)
        self.nevt_per_point = nevts
        self.vars = tuple(
            [ Item( ["type", "npts", "first", "last"], val) for val in scl ]
            )

        self.onScan()

    def PedestalRecord(self, data):
        """
        Called when a new PedestalRecord record is found.
        Sets the pedestals and noise arrays in the corresponding
        Module object.
        It calls onPedestal(module).
        """
        self.error = None
        mid = int(data[1]);
#        if not self.applicable(mid):
#            self.error = -1
#            return
            
        mid = decode_pedestal(data, self.idlist)
        if mid in self.idlist:
            md  = self.idlist[mid]
            if len(md.pedestals) != md.ntot:
                self.error = -3
                print 'PedestalRecord: inconsistent array size'
                return
            
            self.onPedestal(md)

    def NewPointRecord(self, data):
        """Called when a new NewPointRecord record is found.
        It calls onNewPoint when a new point starts and onEndPoint
        when finishes.
        """
        self.error = None
        if self.npoint >= 0:
            self.onEndPoint()
            
        # decode
        nvar      = (data[1] & 0xffL)
        nevts     = (data[1] & 0xffffff00L)>>8
        self.cnt  = 0
        variables = []
        for i in range(0, nvar):
            vartype = (data[2+i] & 0xf0000000L)>>28;
            val     = (data[2+i]&0xfffffffL)
            variables.append( (vartype, val) )

        self.onNewPoint(nevts, variables)
        self.npoint += 1

    def DataRecord(self, data):
        """
        Called when a new DataRecord record is found.
        """
        self.error = None

    def ConfigRecord(self, data):
        """
        Called when a ConfigRecord is found
        """
        self.error = None
        self.onConfig(data)


class VATAGP(SimpleVMEdaqObserver):
    """This class can be registered to decode the data from a VATAGP
    module
    """

    def __init__(self, dtype, use_up=0):
        """
        Takes as optional parameter use_up. When true, it will also
        return the data from the `up' ADC.
        """
        super(VATAGP, self).__init__(long(dtype))
        self.use_up = use_up
        self.nadj = 0

    def __str__(self):
        return "SiliconModule"

    def DataRecord(self, data):
        """
        Called when a new Data record is found. It will call the method:
           onData(data):
             data is a tuple of (channel, amplitude, adj) where:
                - channel is the channel number
                - amplitude is the amplitude meassured in the channel
                - adj (if !=0 indicates that this is an adjacent channel)
        """
        self.error = None
        # avoid processing events with are not of this type
        # TODO: check that This is done already in pyvmedaq::read
        mid   = int( (data[2] & 0xfff00000L)>>20 )
        dtype = long(data[0] & 0xf000)
        if not self.applicable(mid):
            self.error = -1
            return
        
        md  = self.idlist[mid]
        chip_type = 3 # A GP3
        if hasattr(md, "config"):
            chip_type = md.config.chip_type
            
        self.cnt   += 1
        self.ndata += 1
        self.nadj  = (data[1] & 0xff000000)>>24
        decode_vatagp(self, data, chip_type, self.use_up)
        
    def ConfigRecord(self, data):
        """ 
        Called when a new ConfigRecord is found
        """
        nchip = data[4]>>16
        nchan = data[4] & 0xffff
        ntot  = nchip*nchan
        mid   = int( data[1] )
        
        cfg = ModuleConfig(mid)
        md  = self.idlist[mid]
        md.config = cfg
        cfg.chip_type = data[2]
        cfg.thrs = data[3]
        cfg.nchip = nchip
        cfg.nchan = nchan
        cfg.ntot  = ntot
        trim = []
        mask = []
        test = []
        ichan = 0
        for iw,v in enumerate(data[5:6+ntot]):
            trim.append( v & 0x00ff )
            test.append( (v & 0x0f00)>>8 )
            mask.append( (v & 0xf000)>>12 ) 
            
        cfg.trim = list(trim)
        cfg.test = list(test)
        cfg.mask = list(mask)
        self.onConfig(data)
        

class ADC(SimpleVMEdaqObserver):
    """This class can be registered to decode the data from a Caen ADC
    module
    """

    def __init__(self, dtype):
        """
        """
        super(ADC, self).__init__(long(dtype))
        self.gate_counter=0

    def __str__(self):
        return "ADCModule"

    def DataRecord(self, data):
        """
        Called when a new Data record is found. It will call the method:
           onData(data):
             data is a tuple of (channel, amplitude, adj) where:
                - channel is the channel number
                - amplitude is the amplitude meassured in the channel
                - adj (if !=0 indicates that this is an adjacent channel)
        """
        self.error = None
        mid   = (data[2] & 0xfff00000L)>>20
        if not self.applicable(mid):
            self.error = -1
            return
        
        self.cnt   += 1
        self.ndata += 1
        decode_adc(self, data)
