#include <Python.h>
#include <structmember.h>
#include <pyvmedaq.h>
typedef struct
{
    PyObject_HEAD
    int own;
    int nitems;
    Py_ssize_t len;
    unsigned int *item; // memory buffer
} vmedaq_buffer;

PyObject *__get_VMEdaq_Error();
void vmedaq_register_class(PyObject *dict, const char *type_name, PyTypeObject *type,
                           PyObject *bases);

PyObject *PyVmedaq_Buffer_New(int sz, char *ptr);

static PyTypeObject DBuffertype;
#define buffer_Check(op) PyObject_TypeCheck(op, &DBuffertype)
#define buffer_CheckExact(op) ((op)->ob_type == &DBuffertype)

static PyObject * vmedaq_buffer_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
    vmedaq_buffer *self;
    self = (vmedaq_buffer *)type->tp_alloc(type, 0);
    if (!self)
        return (PyObject *)0;

    if (self!=NULL)
    {
        self->own = 0;
        self->nitems=0;
        self->len =0;
        self->item =0;
    }
    return (PyObject *)self;
}

static int vmedaq_buffer_init(vmedaq_buffer *self, PyObject *args, PyObject *kwds)
{
    PyObject *in=NULL;

    // Check that we do not get any keyword argument
    if (kwds != NULL)
    {
        Py_ssize_t i= PyObject_Length(kwds);
        if (i < 0)
            return -1;
        else if (i > 0)
        {
            PyErr_SetString(__get_VMEdaq_Error(),
                            "vmedaq.Buffer: vmedaq.Buffer constructor takes "
                             "no keyword arguments");
            return -1;
        }
    }

    // Get the Object 
    if (!PyArg_ParseTuple(args, "O:vmedaq.Buffer.__init__", &in))
        return -1;

    if (in==NULL)
    {
        PyErr_SetString(__get_VMEdaq_Error(),
                        "vmedaq.Buffer is getting a NULL object to clone");
        return -1;
    }

    // Get the size and the memory ptr of the buffer
    Py_ssize_t sz;
    char *ptr=0;
    if (PyObject_AsWriteBuffer(in, (void **)&ptr, &sz) )
    {
        PyErr_SetString(PyExc_ValueError, "vmedaq.Buffer: import object is not a buffer");
        return -1;
    }
    else
        self->own = 0;

    if (!ptr)
    {
        PyErr_SetString(PyExc_ValueError, "vmedaq.Buffer: NULL pointer");
        return -1;
    }

    if (sz%sizeof(unsigned int))
    {
        PyErr_SetString(PyExc_ValueError, "vmedaq.Buffer: buffer not aligned");
        return -1;
    }

    self->len=sz;
    self->item=(unsigned int *)ptr;
    self->nitems = sz/sizeof(unsigned int);
    return 0;
}

static void vmedaq_buffer_dealloc(vmedaq_buffer* self)
{
    if (self->own && self->item)
        PyMem_Free(self->item);
    self->ob_type->tp_free((PyObject*)self);
}

static PyObject *vmedaq_buffer_get_item(vmedaq_buffer *self, int i)
{
    PyObject *out=NULL;
    if (i>=0 && i<self->nitems)
    {
        out = PyLong_FromUnsignedLong(self->item[i]);
    }
    if (out==NULL)
    {
        PyErr_SetString(PyExc_IndexError, "vmedaq.Buffer __getitem__ index out of range");
    }
    return out;
}

static int vmedaq_buffer_length(vmedaq_buffer *self)
{
    return self->nitems;
}

static PyObject *vmedaq_buffer_concat(vmedaq_buffer *a, PyObject *bb)
{
    Py_ssize_t size;
    vmedaq_buffer *np;
    if (!buffer_Check(bb))
    {
        PyErr_Format(PyExc_TypeError,
                     "can only append buffer (not \"%.200s\") to buffer",
                     bb->ob_type->tp_name);
        return NULL;
    }

#define b ((vmedaq_buffer *)bb)
    size = a->len + b->len;
    np = (vmedaq_buffer *)PyVmedaq_Buffer_New(size, 0);
    if (np == NULL)
    {
        return NULL;
    }
    memcpy(np->item, a->item, a->len);
    memcpy(np->item + a->len, b->item, b->len);
    return (PyObject *)np;
#undef b
}

static PyObject *vmedaq_buffer_slice(vmedaq_buffer *self, Py_ssize_t ilow,
                                     Py_ssize_t ihigh)
{
    vmedaq_buffer *out;
    Py_ssize_t size;
    if (ilow<0)
        ilow=0;
    else if (ilow>self->nitems)
        ilow = self->nitems;

    if (ihigh<0)
    {
        ihigh=0;
    }
    else if (ihigh<ilow)
    {
        ihigh=ilow;
    }
    if (ihigh>self->nitems)
    {
        ihigh=self->nitems;
    }
    size = ihigh-ilow;

    out = (vmedaq_buffer *)PyVmedaq_Buffer_New(size*sizeof(unsigned int), 0);
    memcpy(out->item, self->item+ilow, out->len);

    return (PyObject *)out;
}

static PyObject* vmedaq_buffer_copy(vmedaq_buffer *self, PyObject *unused)
{
    return vmedaq_buffer_slice(self, 0, self->nitems);
}

static int vmedaq_buffer_ass_item(vmedaq_buffer *self, int i, PyObject *v)
{
    if (i<0 || i>=self->nitems)
    {
        PyErr_SetString(PyExc_IndexError, "vmedaq.Buffer assignment index out of range");
        return -1;
    }
    if (v==NULL)
    {
        PyErr_SetString(PyExc_ValueError, "vmedaq.Buffer assignment gets a NULL object");
        return -1;
    }
    if (!PyInt_Check(v))
    {
        PyErr_SetString(PyExc_TypeError, "vmedaq.Buffer: tried to assign a non integer");
        return -1;
    }
    self->item[i] = PyInt_AsLong(v);
    return 0;
}

static int vmedaq_buffer_ass_slice(vmedaq_buffer *self, int ilow, int ihigh, PyObject *v)
{
    PyErr_SetString(PyExc_SystemError, "vmedaq.Buffer: not implemented");
    return -1;
}

static PySequenceMethods vmedaq_buffer_as_sequence = { (lenfunc)vmedaq_buffer_length, /*sq_length*/
(binaryfunc)vmedaq_buffer_concat, /*sq_concat*/
(ssizeargfunc)0, /*sq_repeat*/
(ssizeargfunc)vmedaq_buffer_get_item, /*sq_item*/
(ssizessizeargfunc)vmedaq_buffer_slice, /*sq_slice*/
(ssizeobjargproc)vmedaq_buffer_ass_item, /*sq_ass_item*/
(ssizessizeobjargproc)vmedaq_buffer_ass_slice,/*sq_ass_slice*/
(objobjproc)0, /*sq_contains*/
(binaryfunc)0, /*sq_inplace_concat*/
(ssizeargfunc)0 /*sq_inplace_repeat*/
};

static int vmedaq_buffer_getreadbuf(vmedaq_buffer *self, int index, const void **ptr)
{
    if (index != 0)
    {
        PyErr_SetString(PyExc_SystemError, "Accessing non-existent array segment");
        return -1;
    }
    *ptr = (void *)self->item;
    return self->len;
}

static int vmedaq_buffer_getwritebuf(vmedaq_buffer *self, int index, const void **ptr)
{
    if (index != 0)
    {
        PyErr_SetString(PyExc_SystemError, "Accessing non-existent array segment");
        return -1;
    }
    *ptr = (void *)self->item;
    return self->len;
}

static int vmedaq_buffer_getsegcount(vmedaq_buffer *self, int *lenp)
{
    if (lenp)
        *lenp = self->len;
    return 1;
}

static PyBufferProcs
        vmedaq_buffer_as_buffer = { (readbufferproc)vmedaq_buffer_getreadbuf,
                (writebufferproc)vmedaq_buffer_getwritebuf,
                (segcountproc)vmedaq_buffer_getsegcount, 
};

static PyObject *vmedaq_buffer_get_nitems(vmedaq_buffer *self, void *closure)
{
    return PyInt_FromLong((long)self->nitems);
}

static PyObject *vmedaq_buffer_get_size(vmedaq_buffer *self, void *closure)
{
    return PyInt_FromLong((long)self->len);
}

static PyGetSetDef
        vmedaq_buffer_getsets [] = { { "nitems", (getter)vmedaq_buffer_get_nitems, NULL,
                "the number of items in the array" },
                                      { "size", (getter)vmedaq_buffer_get_size, NULL,
                                             "the size, in bytes, of the buffer" },
                                      { NULL } };

static PyObject * vmedaq_buffer_info(vmedaq_buffer *self, PyObject *unused)
{
    PyObject* retval= NULL;
    retval = PyTuple_New(3);
    if (!retval)
        return NULL;

    PyTuple_SET_ITEM(retval, 0, PyLong_FromVoidPtr(self->item));
    PyTuple_SET_ITEM(retval, 1, PyInt_FromLong((long)(self->nitems)));
    PyTuple_SET_ITEM(retval, 2, PyInt_FromLong((long)(self->len)));

    return retval;
}

static PyObject *vmedaq_buffer_tolist(vmedaq_buffer* self, PyObject *unused)
{
    PyObject *list = PyList_New(self->nitems);
    int i;

    if (list==NULL)
        return NULL;

    for (i=0; i<self->nitems; i++)
    {
        PyObject *v = vmedaq_buffer_get_item(self, i);
        if (v==NULL)
        {
            Py_DECREF(list);
            return NULL;
        }
        PyList_SetItem(list, i, v);
    }
    return list;
}

static PyObject *vmedaq_buffer_tostring(vmedaq_buffer* self, PyObject *unused)
{
    return PyString_FromStringAndSize((char *)self->item, self->len);
}

static PyObject * vmedaq_buffer_repr(vmedaq_buffer *a)
{
    PyObject *s, *t, *v= NULL;
    int len;

    len = a->nitems;
    if (len == 0)
    {
        return PyString_FromString("vmedaq.Buffer");
    }

    v = vmedaq_buffer_tolist(a, NULL);

    t = PyObject_Repr(v);
    Py_XDECREF(v);

    PyString_ConcatAndDel(&s, t);
    PyString_ConcatAndDel(&s, PyString_FromString(")"));
    return s;
}

PyMethodDef vmedaq_buffer_methods[] = { { "buffer_info",
                                           (PyCFunction)vmedaq_buffer_info, METH_NOARGS,
                                                   "." }, { "tolist",
        (PyCFunction)vmedaq_buffer_tolist, METH_NOARGS, "." }, { "tostring",
        (PyCFunction)vmedaq_buffer_tostring, METH_NOARGS, "." }, { "__copy__",
        (PyCFunction)vmedaq_buffer_copy, METH_NOARGS, "copy(buffer)" }, { "__deepcopy__",
        (PyCFunction)vmedaq_buffer_copy, METH_O, "copy(buffer)" }, { NULL, NULL } /* sentinel */
};
static PyObject *vmedaq_buffer_iter(vmedaq_buffer *self);

static PyTypeObject DBuffertype = { PyObject_HEAD_INIT(NULL)0,
"vmedaq.Buffer",
sizeof(vmedaq_buffer),
0,
(destructor)vmedaq_buffer_dealloc, /* tp_dealloc */
0, /* tp_print */
0, /* tp_getattr */
0, /* tp_setattr */
0, /* tp_compare */
(reprfunc)vmedaq_buffer_repr, /* tp_repr */
0, /* tp_as _number*/
&vmedaq_buffer_as_sequence, /* tp_as _sequence*/
0, /* tp_as _mapping*/
0, /* tp_hash */
0, /* tp_call */
0, /* tp_str */
PyObject_GenericGetAttr, /* tp_getattro */
0, /* tp_setattro */
&vmedaq_buffer_as_buffer, /* tp_as_buffer*/
Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /* tp_flags */
"Memory array wrapper", /* tp_doc */
0, /* tp_traverse */
0, /* tp_clear */
0, /* tp_richcompare */
0, /* tp_weaklistoffset */
(getiterfunc)vmedaq_buffer_iter, /* tp_iter */
0, /* tp_iternext */
vmedaq_buffer_methods, /* tp_methods */
0, /* tp_members */
vmedaq_buffer_getsets, /* tp_getset */
0, /* tp_base */
0, /* tp_dict */
0, /* tp_descr_get */
0, /* tp_descr_set */
0, /* tp_dictoffset */
(initproc)vmedaq_buffer_init, /* tp_init */
0, /* tp_alloc */
vmedaq_buffer_new, /* tp_new */
};

/*-----------------------------------------------------------
 * The Iterator
 *-----------------------------------------------------------*/

typedef struct
{
    PyObject_HEAD
    long index;
    vmedaq_buffer *buffer;
    PyObject * (*getitem)(vmedaq_buffer *, int);
} vmedaq_buffer_iter_object;

static PyObject * vmedaq_buffer_iter_next(vmedaq_buffer_iter_object *it)
{
    //	assert(PyArrayIter_Check(it));
    if (it->index < it->buffer->nitems)
        return (*it->getitem)(it->buffer, it->index++);
    return NULL;
}

static void vmedaq_buffer_iter_dealloc(vmedaq_buffer_iter_object *it)
{
    PyObject_GC_UnTrack(it);
    Py_XDECREF(it->buffer);
    PyObject_GC_Del(it);
}

static int vmedaq_buffer_iter_traverse(vmedaq_buffer_iter_object *it, visitproc visit,
                                       void *arg)
{
    if (it->buffer != NULL)
        return visit((PyObject *)(it->buffer), arg);
    return 0;
}

static PyTypeObject PyVmedaqBufferIter_Type = { PyObject_HEAD_INIT(NULL)0, /* ob_size */
"arrayiterator", /* tp_name */
sizeof(vmedaq_buffer_iter_object), /* tp_basicsize */
0, /* tp_itemsize */
/* methods */
(destructor)vmedaq_buffer_iter_dealloc, /* tp_dealloc */
0, /* tp_print */
0, /* tp_getattr */
0, /* tp_setattr */
0, /* tp_compare */
0, /* tp_repr */
0, /* tp_as_number */
0, /* tp_as_sequence */
0, /* tp_as_mapping */
0, /* tp_hash */
0, /* tp_call */
0, /* tp_str */
PyObject_GenericGetAttr, /* tp_getattro */
0, /* tp_setattro */
0, /* tp_as_buffer */
Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,/* tp_flags */
0, /* tp_doc */
(traverseproc)vmedaq_buffer_iter_traverse, /* tp_traverse */
0, /* tp_clear */
0, /* tp_richcompare */
0, /* tp_weaklistoffset */
PyObject_SelfIter, /* tp_iter */
(iternextfunc)vmedaq_buffer_iter_next, /* tp_iternext */
0, /* tp_methods */
};

static PyObject *vmedaq_buffer_iter(vmedaq_buffer *self)
{
    vmedaq_buffer_iter_object *it;
    it = PyObject_GC_New(vmedaq_buffer_iter_object, &PyVmedaqBufferIter_Type);
    if (it == NULL)
        return NULL;

    Py_INCREF(self);
    it->buffer = self;
    it->index = 0;
    it->getitem = vmedaq_buffer_get_item;
    PyObject_GC_Track(it);
    return (PyObject *)it;
}

void register_vmedaq_buffer(PyObject *dict)
{
    //    DBuffertype.ob_type = &PyType_Type;
    //    PyVmedaqBufferIter_Type.ob_type = &PyType_Type;
    vmedaq_register_class(dict, "Buffer", &DBuffertype, 0);
}
PyObject *PyVmedaq_Buffer_New(int sz, char *ptr)
{
    int own=0;
    if (ptr==0 && sz>0)
    {
        own=1;
        ptr = (char *)PyMem_Malloc(sz);
        if (ptr==NULL)
        {
            return PyErr_NoMemory();
        }
    }
    vmedaq_buffer *self= PyObject_New(vmedaq_buffer,&DBuffertype);
    self->own = own;
    self->len = sz;
    self->item=(unsigned int *)ptr;
    self->nitems = sz/sizeof(unsigned int);
    return (PyObject *)self;
}
