#!/usr/bin/env python
"""
vmedaqPipe

This is a small python script that creates a fifo in which vmedaq can
write its output. The script gets the data from the fifo and makes
some 'analysis' before dumping the data into a file
"""
import os
import sys
import select
import fcntl
import time
import struct, array
import gobject
import vmedaq
from optparse import OptionParser
from daqpp import FileIOManager


import ORBit
import CORBA

from references import find_daqpp_references
# Load the DAQcorba type library for the Observer.
ORBit.load_typelib('DAQcorba')
from DAQcorba import DAQ


def create_any(val):
    """ Creates a CORBA.Any from the input value
    """
    if isinstance(val, bool):
        return CORBA.Any(CORBA.TypeCode('IDL:omg.org/CORBA/boolean:1.0'),val)
    elif isinstance(val,int) or isinstance(val,long):
        return CORBA.Any(CORBA.TypeCode('IDL:omg.org/CORBA/long:1.0'),val)
    elif isinstance(val,float):
        return CORBA.Any(CORBA.TypeCode('IDL:omg.org/CORBA/float:1.0'),val)
    elif isinstance(val,str):
        return CORBA.Any(CORBA.TypeCode('IDL:omg.org/CORBA/string:1.0'),val)
    else:
        return CORBA.Any(CORBA.TypeCode('IDL:omg.org/CORBA/string:1.0'),str(val))


def create_pipe(fifoPath):
    """ Creates a fifo"""

    # Check if the file exists already and remove if this is the case.
    if os.access(fifoPath, os.F_OK):
        try:
            os.unlink(fifoPath)
        except OSError,excpt:
                print 'Could not remove %s\n\t%s' % (fifoPath, excpt.strerr)
                return -1

    # Now create the fifo
    try:
        os.mkfifo(fifoPath, 0666)
    except OSError, excpt:
        print 'Pipe: could not create %s\n\t%s' % (fifoPath, excpt.strerror)
        return -1

    return 0


class Data(vmedaq.VATAGP):
    """ This handles the data on the pipe """
    def __init__(self, ofile):
        """ Creates the FileIOManager and prepares the output file """
        super(Data, self).__init__(-1)
        self.ofile = ofile
        self.write = True
        self.ntot  = 0.
        self.nevts = 0.        
        self.fmgr  = FileIOManager(True, 50000000)
        self.fmgr.open( ofile )

    def close(self):
        """ Closes the file """
        print 'Closing', self.ofile
        self.fmgr.close()

    def onNewData(self,module,evtcnt,data):
        """ New data has arrived """
        self.ntot += 1
        self.write = False
        for chan,val,adj in data:
            ped = module.pedestals[chan]
            ns  = module.noise[chan]
            v   = val-ped
            
            if ns > 0. and v/ns > 3.:
                self.write = True
                self.nevts += 1
                break


    def EndOfRecord(self, header, buffer):
        """ A record has been finished """
        if self.write:
            ptr, nitems, size = buffer.buffer_info()
            self.fmgr.write( buffer.tostring(), size) 
            
        self.write = True
        
        

def get_vmedaq_data(name, pipe_name, mconf):
    """ Creates the fifo and waits until some data arrives. On arrival
        it determines if the event is worth saving by calling
        decode_data  
    """

    # Create the fifo
    if create_pipe(pipe_name):
        return

    S = Data(name)
    V = vmedaq.VMEdaq()
    V.register( S )
    V.open( (pipe_name,) )

    mconf.setParam('DAQ/log_data', create_any(True), True)
    any = mconf.getParam('DAQ/log_data')

    loop = gobject.MainLoop()
    gobject.timeout_add(1000, monitor_data, S)
    gobject.idle_add(read_data, (V, loop) )

#    gobject.timeout_add(3000, start_run, mconf)
    start_run(mconf)

    try:
        loop.run()
    except KeyboardInterrupt:
        print 'Arrrrggggg !!!!!'        

    S.close()
        

def read_data(data):
    """ Reads the data """
    V, loop = data
    rc = True
    try:
        V.read()
    except vmedaq.VMEdaqError,e:
        rc = False
        print "VMEdaq: Algo ha salido mal:",e
    except:
        rc = False
        print '?????:'

    if not rc:
        loop.quit()
        
    return rc

def start_run(mconf):
    """ Sets up vmedaq and starts the run
    """
    mconf.setParam('sync',create_any(0),True)
    any = mconf.getParam('DAQ/maxevents')
    mconf.setParam('DAQ/Start', create_any(any.value()), True)
    return False

def monitor_data(S):
    """ a GObject callback that prints some info
    """
    if S.ntot <= 0:
        return True

    sys.stdout.write('\r%8d/%8d [%.2f] evts in file%20s: %.2f  ' % (
        S.nevts,S.ntot, 100.*S.nevts/S.ntot,
        S.fmgr.file_name(),
        S.fmgr.relative_size()))
    sys.stdout.flush()
    
    return True
                

def main():
    """ The main routine
    """
    parser = OptionParser()
    parser.add_option("--run_no",
                      dest="run_no", action="store", type="int",
                      help="Run number",
                      default=0
                      )
    parser.add_option("--prefix",
                      dest="prefix", action="store", type="string",
                      help="File prefix",
                      default="run"
                      )
    parser.add_option("--odir",
                      dest="odir", action="store", type="string",
                      help="Output folder",
                      default="/tmp"
                      )
    parser.add_option("--nevts",
                      dest="nevts", action="store", type="int",
                      help="Number of events",
                      default=10000
                      )

    (options, args) = parser.parse_args()
    file_name = os.path.join(options.odir,
                             '%s%06d.data' % (options.prefix,options.run_no)
                             )
    pipe_name = os.path.join(options.odir,
                             '%s.pipe.%06d_000.data' % (options.prefix,options.run_no)
                             )
    # Now see if we can contact with vmedaq
     # Init the orb
    orb = CORBA.ORB_init(sys.argv)

    # Find the references
    daqm,observable = find_daqpp_references(orb,args[0])
    if not daqm or daqm._is_nil() :
        print 'Could not connect'
        sys.exit(1)

    mconf = daqm.findModule('VMEDAQconfig')
    if mconf._is_nil() :
        print 'Cannot find config module'
        sys.exit(1)

    mconf.setParam('IO/run_number',create_any(options.run_no),True)
    mconf.setParam('IO/compression', create_any(True), True)
    mconf.setParam('IO/prefix', create_any('%s.pipe.' % options.prefix), True)
    mconf.setParam('DAQ/trigger_type', create_any('internal'), True)
    mconf.setParam('DAQ/pedestal_run', create_any(10000), True)
    while True:
        time.sleep(0.5)
        rm = daqm.findRunManager('main')
        state = rm.getState()
        if state.status != DAQ.Running:
            break

    time.sleep(0.5)
    mconf.setParam('DAQ/maxevents',create_any(options.nevts),True)
    
    get_vmedaq_data(file_name, pipe_name, mconf)
    
if __name__ == '__main__':
    main()

#---------------------------------------------------------------
