#!/usr/bin/env python

import sys
import time
import getopt
import vmedaq

import sys, traceback, exceptions


def print_word(indx, val, msg='' ):
    print "%9d: %08x    %s" % (indx, val,msg )

def print_header(msg, length):
    sline  = "%s" % (length*'*')
    print "\n%s\n@ %s\n%s" % (sline, msg, sline)

(StartOfRun, PedestalRecord, MapRecord, ConfigRecord,
 ScanRecord, NewPointRecord, DataRecord, BigDataRecord 
 ) = tuple([ 1<<x for x in range(0,8)])
 

class CommonRecords(vmedaq.SimpleVMEdaqObserver):
    """
    Writes common records.
     
    This class will dump the Records which are not
    specific to a given module
    """
    def __init__(self, dtype=None, name=None):
        """ Initializes the instance
        """
        super(CommonRecords, self).__init__(dtype)
        self.name = name
        self.registered = {}
        self.slen=80
        
    def print_it(self, mid, data_type):
        """ Returns False if some other observer has register to
            display this block
        """
        dtype =  self.idlist[mid].dtype;
        if dtype in self.registered and (self.registered[dtype] & data_type):
            return False
        else:
            return True
    
    def register(self, dtype, dblocks):
        self.registered.setdefault(long(dtype), long(dblocks))
        
    def NewFile(self, file_name):
        self.slen = 17+len(file_name)
        print_header("%s\n* Reading file: %s" % file_name, self.slen)
        
    
    def StartOfRunRecord(self, data):
        super(CommonRecords, self).StartOfRunRecord(data)
        print_header('Start of Run', self.slen)
        print_word(0, data[0], "Block size: %d" % (data[0] & 0xfff))
        print_word(1, data[1], "Started on: %s" % ( time.ctime(data[1])))
        print_word(2, data[2], "Run no.: %d" % data[2])
        print_word(3, data[3], "Number of modules: %d" % data[3])
        for iw,v in enumerate(data[4:-1]):
            print_word(iw+4, v, "")
            
        print_word(5+iw,data[-1],"")
        
            
    def EndOfRunRecord(self, data):
        super(CommonRecords, self).EndOfRunRecord(data)
        print_header("End of Run", self.slen)
        print_word(0, data[0], "Block size: %d" % (data[0] & 0xfff))
        print_word(1, data[1], "Finished on: %s" % ( time.ctime(data[1])))
        print_word(2, data[2], "Run no.: %d" % data[2])
        print_word(3, data[3], "")
    
    def PedestalRecord(self, data):
        """ Pedestals
        """
        print '#### PedestalRecord in CommonRecords'
        print '#### ...callling parent method'
        super(CommonRecords, self).PedestalRecord(data)
        if self.error:
            print '**** Error decoding pedestals: %d' % self.error
            return
        
        if not self.print_it(data[1], PedestalRecord):
            return
        
        md  = self.idlist[ data[1] ]
        print_header("Pedestal Record", self.slen)
        print_word(0, data[0], "block type %4x size %d" % ( (data[0] & 0xf000), (data[0] & 0xfff)))
        print_word(1, data[1], "Module: %d" % data[1] )
        print_word(2, data[2], "Nchip: %d Nchan %d" % (md.nchip, md.nchan))

        for indx, val in enumerate(data[3:-1]):
            print_word(indx+3, val, "Ped: %.1f  Noise: %.1f" % (md.pedestals[indx], md.noise[indx]))
            
        print_word(indx+4, data[-1],"")
    
    def MapRecord(self, data):
        """ MapRecord
        """
        #
        # We ignore here the result of applicable and just
        # print the map record
        #
        super(CommonRecords, self).MapRecord(data)
        
        if not self.print_it(data[1], MapRecord):
            return
        
        mid = data[1]
        nx  = (data[2] >> 16);
        ny  = (data[2] & 0xffffL);
        print_header("Map Record", self.slen)
        print_word(0, data[0], "block type %4x size %d" % ( (data[0] & 0xf000), (data[0] & 0xfff)))
        print_word(1, data[1], "Module id: %d" % mid)
        print_word(2, data[2], "nx=%d ny=%d" % (nx,ny))
        for iw,v in enumerate(data[3:-1]):
            ix1 = (v >> 24) & 0xffL;
            iy1 = (v >> 16) & 0xffL;

            ix2 = (v >> 8) & 0xffL;
            iy2 = (v) & 0xffL;
            print_word(iw+3, v, "%2d.%2d - %2d.%2d" % (ix1,iy1,ix2,iy2))
            
        print_word(4+iw,data[-1],"")
    
    def ScanRecord(self, data):
        """ Prints the Scan description record
        """
        
        super(CommonRecords, self).ScanRecord(data)
        print_header("Scan Record", self.slen)
        print_word(0, data[0], "block type %4x size %d" % ( (data[0] & 0xf000), (data[0] & 0xfff)))
        print_word(1, data[1], "Number of variables: %d" % data[1])
        print_word(2, data[2], "No. of evts/point: %d" % data[2])
        
        var = None
        for iw,v in enumerate(data[3:-1]):
            jw = iw%3
            if jw == 0:
                if v & 0x80000000L:
                    is_float = True
                else:
                    is_float = False
            
             
                type = (v & 0xfff0000L)>>16
                npts = v & 0xffff
                for tv in self.vars:
                    if tv.type == type:
                        var = tv
                        break
                    
                print_word(iw+3, v,"Var. type %d (%s) No. of Points: %d" % (type, vmedaq.var_names[type], npts))
            elif jw == 1:
                first = 0.
                if var:
                    first = var.first
                    
                print_word(iw+3, v,"...from %f" % first)
            else:
                last = 0.
                if var:
                    last = var.last
                    
                print_word(iw+3, v,"...to %f" % last)
            
        print_word(4+iw,data[-1],"")
    
    def NewPointRecord(self, data):
        print_header("NewPoint", self.slen)
        print_word(0, data[0], "block type %4x size %d" % ( (data[0] & 0xf000), (data[0] & 0xfff)))
        print_word(1, data[1], "No. of variables: %d No. of events: %d" % ( (data[1]&0xff), (data[1]&0xffffff00)>>8 ))
        try:
            for iw,v in enumerate(data[2:-1]):
                vtype = (v& 0xf0000000L)>>28
                vval  = (v& 0xfffffffL)
                print_word(iw+2,v,"Var. type %d (%s) value %d" % (vtype, vmedaq.var_names[vtype], vval))
        except Exception, msg:
            print '*** Very bad:',msg
                   
        print_word(3+iw, data[-1],"")
            
    def DataRecord(self, data):
        """ Dump the data record if nobody else has registered to
            decode this module
        """
        mid = (data[2] & 0xfff00000L) >> 20
        if not self.print_it(mid, DataRecord):
            return
        
        print_header("Data Record", self.slen)
        print_word(0, data[0], "block type %4x size %d" % ( (data[0] & 0xf000), (data[0] & 0xfff)))
        print_word(1, data[1], "Event counter: %d" % (data[1] & 0xffffff))
        print_word(2, data[2], "Module id: %d" % ((data[2]&0xfff00000L)>>20))
                   
        for iw,v in enumerate(data[3:-1]):
            print_word(iw+3,v,'')
            
        print_word(4+iw, data[-1],"")

            
class GenModule(vmedaq.SimpleVMEdaqObserver):
    """ A generic module
    """
    def __init__(self, dtype, name=None):
        super(GenModule, self).__init__(dtype)
        self.name = name
        self.slen = 80
        
class TDCmodule(GenModule):
    """ Decodes a TDC
    """
    def __init__(self, dtype):
        super(TDCmodule, self).__init__(dtype, 'TDCmodule')
        
    def onNewRun(self):
        if self.data_type in self.typelist:
            print '*** TDCmodule will analyze modules with id:'
            for m in self.typelist[self.data_type]:
                print '...',m.id 

    def DataRecord(self, data):        
        mid    = (data[2] & 0xfff00000L) >> 20
        sz      = data[2] & 0xfffL
        evtcnt = (data[1] & 0xffffff)
        print_header("TDC Data Record", self.slen)
        print_word(0, data[0], "block type %4x size %d" % ( (data[0] & 0xf000), (data[0] & 0xfff)))
        print_word(1, data[1], "Event counter: %d" % (evtcnt))
        print_word(2, data[2], "Module id: %d size %d" % (mid, sz))
        
        for iw, v in enumerate(data[3:-1]):
            if v& 0x8000:
                print_word(iw+3, v, "Header: nchan %d gate cntr: %d" %( ((v & 0x7000)>>12)+1, v & 0xFFF))
            else:
                print_word(iw+3, v, "chan. %d tdc %d" % ((v & 0x7000)>>12, v & 0xFFF))
                
            
        print_word(4+iw, data[-1],"") 
        
        
class ADCmodule(GenModule):
    """ Decodes the data of an ADC module
    """
    def __init(self, dtype):
        super(ADCmodule, self).__init__(dtype,'ADCmodule')
        
    def onNewRun(self):
        if self.data_type  in self.typelist:
            print '*** ADCmodule will analyze modules with id:'
            for m in self.typelist[self.data_type]:
                print '...',m.id
                 
    def DataRecord(self, data):
        """ The ADC data block
        """
        mid    = (data[2] & 0xfff00000L) >> 20
        sz      = data[2] & 0xfffL
        evtcnt = (data[1] & 0xffffff)
        print_header("ADC Data Record", self.slen)
        print_word(0, data[0], "block type %4x size %d" % ( (data[0] & 0xf000), (data[0] & 0xfff)))
        print_word(1, data[1], "Event counter: %d" % (evtcnt))
        print_word(2, data[2], "Module id: %d size %d" % (mid, sz))
        print_word(3, data[3], "Number of blocks: %d" % data[3])
        
        nchan = 0
        for iw, v in enumerate(data[4:-1]):
            header = v>>16
            if header==0xacdc:
                nchan = v & 0xFFFF
                print_word(iw+3, v, "nchan:%d" % nchan)
            else:
                if nchan>0:
                    add = (v>>16) & 0x1f
                    val = v & 0xfff
                    nchan -= 1
                    print_word(iw+3, v,"chan %d val %d" % (add,val))
                else:
                    print_word(iw+3, v,"buffer no. %d" % v)
                    
        print_word(5+iw, data[-1],"") 
        
        
        
class VATAmodule(vmedaq.VATAGP):
    """ Print messages from a VATA module
    """
    def __init__(self, dtype):
        super(VATAmodule, self).__init__(dtype,0)
        self.name = 'VATAmodule'
        self.slen = 80
        self.data = None
        
        
    def dump_config(self, ofile, data):
        """ Dumps the configuration into a file
        """
        type  = data[2]
        nchip = data[4]>>16
        nchan = data[4] & 0xffff
        ntot  = nchip*nchan
        
        fout = open(ofile,'w')
        fout.write('nchip %d nchan %d ntot %d\n' % (nchip, nchan, ntot))
        for v in data[5:5+ntot]:
            trim = v & 0x00ff
            test = (v & 0x0f00)>>8
            mask = (v & 0xf000)>>12
            fout.write("%d %d %d\n" % (mask, trim, test))
        
        fout.close()    
        
    def ConfigRecord(self, data):
        """ config data
        """
        print_header("Configuration Data", self.slen)
        super(VATAmodule, self).ConfigRecord(data)
        
        nchip = data[4]>>16
        nchan = data[4] & 0xffff
        ntot  = nchip*nchan
        print_word(0, data[0], "block type %4x size %d" % ( (data[0] & 0xf000), (data[0] & 0xfff)))
        print_word(1, data[1], "Module: %d" % data[1] )
        print_word(2, data[2], "Chip type: %d" % data[2])
        print_word(3, data[3], "Threshold: %d" % data[3])
        print_word(4, data[4], "No. of chips: %d No. of chan./chip: %d" %(nchip, nchan))
        
        ichan=0
        for iw,v in enumerate(data[5:6+ntot]):
            print_word(iw+5, v, "Chan: %5d trim %2d test %1d mask %1d" 
                       % (ichan, 
                          (v & 0x00ff), 
                          (v & 0x0f00)>>8, 
                          (v & 0xf000)>>12)
                       )
            ichan += 1
            
        iw += 6
        print_word(iw, data[iw],"separator")
        
        iw+=1
        for ic,v in enumerate(data[iw:-1]):
            print_word(iw+ic, v,'chip id: %3d test %1d probe %1d CC %1d N %1d gain %3d gain enabled %1d' %
                       (v & 0xff,
                        (v>>8) & 1,
                        (v>>9) & 1,
                        (v>>10) & 1,
                        (v>>11) & 1,
                        (v>>12) & 7,
                        (v>>15) & 1
                        )
                       )
            
        print_word(iw+ic+1, data[-1],"")
        self.dump_config('vata.cfg', data)
        
    def decode_serial(self, data):
        """ Decodes serial data
        """
        
        for iw in range(3,6):
            print_word(iw, data[iw],"")
            
        ichan = 0
        for iw,v in enumerate(data[6:-1]):
            print_word(iw+6, v, "Chan. %4d val: %4d" % (ichan, (v&0x3FFL)))
            ichan += 1
        
        print_word(iw+7, data[-1],"")
        
    def decode_sparse(self, data):
        """ Decodes sparse data
        """
        
        for iw in range(3,6):
            print_word(iw, data[iw],"")
            
        romode = (data[2] & 0xf000L) >> 12
        nadj   = (data[1] & 0xff000000L)>>24
        sz     = (data[2] & 0xfffL)
        dsz    = data[0] & 0xfff
        if romode!=8:
            nadj = 0
            
        nblk = nadj + 2
        ndat = nblk-1
        j=6
        ihit = 0
        offs = 0
        while j<dsz:
            
            for i in range(0, ndat):
                if j>dsz-1:
                    break
                
                dt = (data[j-3] & 0x7ff00000) >> 20
                down = data[j] & 0x3FF
                up   = (data[j] & 0xFFC00)>>10
                
                if i==0:
                    print_word(j, data[j], "** chan. %4d up %3d down %3d" % (dt, up, down))
                else:
                    print_word(j, data[j], "         %4d up %3d down %3d" % (dt, up, down))
                
                j += 1
            ihit += 1
            if ihit>1:
                j += 1
                
        while j<dsz:
            print_word(j, data[j],"")
            j += 1
        
    def DataRecord(self, data):
        """ VATA data record
        """
        super(VATAmodule, self).DataRecord(data)
        if self.error:
            print '*** Error. malo'
        mid    = (data[2] & 0xfff00000L) >> 20
        romode = (data[2] & 0xf000L) >> 12
        sz      = data[2] & 0xfffL
        evtcnt = (data[1] & 0xffffff)
        nadj   = (data[1] & 0xff000000L)>>24
        print_header("VATA Data Record", self.slen)
        print_word(0, data[0], "block type %4x size %d" % ( (data[0] & 0xf000), (data[0] & 0xfff)))
        print_word(1, data[1], "Event counter: %d Nadj: %d" % (evtcnt,nadj))
        print_word(2, data[2], "Module id: %d RO mode %d size %d" % (mid, romode, sz))
        
        if romode==2:
            self.decode_serial(data)
        else:
            self.decode_sparse(data)
        
        self.print_data(self.idlist[mid])
        
    def print_data(self, module):
        print "      +-------------------------+"
        print "      |      Decoded data       |"
        print "      +-------------------------+"
        print "      | chan.   val. sign. adj. |"
        
        for chan, val, adj in self.data:
            if not adj:
                print "      +-------------------------+"
                
            try:
                vsig = '%5d' % (val-module.pedestals[chan])
            except IndexError:
                vsig = 'xxxxx'
                
            print "      | %5d %5d %5s %4d  |" % (chan, val, vsig, adj)
        
        print "      +-------------------------+"
        
    def onNewData(self, module, evtcnt, data):
        self.data = data
        
    def onNewRun(self):
        if self.data_type in self.typelist:
            print '*** VATAmodule will analyze modules with id:'
            for m in self.typelist[self.data_type]:
                print '...',m.id 
            
        
            
        
def main(ifile=None):
    filter=None
    file_list = None
    try:
        opts,args=getopt.getopt(sys.argv[1:],"f:",["filter=",])
    except getopt.GetoptError,msg:
        print msg
        sys.exit(1)

    for opt,arg in opts:
        if opt in ('-f','--filter'):
            filter=arg


    file_list = [ file for file in args ]
    if ifile:
        file_list.append(ifile)
    
    
    vata = VATAmodule(vmedaq.NormalData)
    tdc  = TDCmodule(vmedaq.TDCCaenData)
    adc  = ADCmodule(vmedaq.ADCdata)
    S = CommonRecords(name='CommonRecords')
    S.register(vata.data_type, ConfigRecord | DataRecord)
    S.register(tdc.data_type, DataRecord)
    S.register(adc.data_type, DataRecord)
    V = vmedaq.VMEdaq()
    V.register( S )
    V.register( vata )
    V.register( tdc )
    V.register( adc)
    V.open(file_list)
    
    try:
        V.loop()
    except vmedaq.VMEdaqError,e:
        print "Something went wrong:",e
    except KeyboardInterrupt:
        print 'Arrrrgggg !!!!!' 
    except:
        print 'Something very strange happened'


if __name__=="__main__":
    main()