#include <Python.h>
#include <zlib.h>
#include <glib.h>
#include <analysis/DataFormat.h>
#include "pyvmedaq.h"

#ifndef PyMODINIT_FUNC	/* declarations for DLL import/export */
#define PyMODINIT_FUNC void
#endif

#ifndef PyDoc_STRVAR
#define PyDoc_VAR(name) static char name[]
#define PyDoc_STRVAR(name,str) PyDoc_VAR(name) = PyDoc_STR(str)
#define PyDoc_STR(str) str
#endif

PyDoc_STRVAR(module_doc,
"This is amodule that provides the tools to read the VMEDAQ\n\
data files.\n\
");

static PyObject *VMEdaq_Error=NULL;
static void add_int_constant(PyObject *, char *, int);
void vmedaq_register_class(PyObject *dict, const char *type_name, PyTypeObject *type, PyObject *bases);
void register_vmedaq_buffer(PyObject *dict);

PyObject *PyVmedaq_Buffer_New(int sz, char *ptr);
PyObject *__get_VMEdaq_Error();

typedef struct
{
    PyObject_HEAD
    /* Type-specific fields go here. */
    gzFile in;
    PyObject *iter;
    PyObject *files; // Sequence of files to read
    size_t bfsiz;
    GSList *objs; // list of objects subscribed
    unsigned int *buffer;
    int last_event;
    PyObject *new_event;
    PyObject *end_of_event;
} vmedaq;

/*
 * Each object that registers has one of these structures with a
 * pointer to the different methods it is interested in
 *
 */
typedef struct
{
    PyObject *the_object;
    PyObject *new_file;
    PyObject *start_of_run;
    PyObject *data;
    PyObject *end_of_run;
    PyObject *pedestal;
    PyObject *new_point;
    PyObject *pixel_map;
    PyObject *scan;
    PyObject *end_of_record;
    PyObject *end_of_event;
    PyObject *config;
    PyObject *applicable;
    int last_evt;
    int fcn;
    int has_scan;
} observer;

PyObject *__get_VMEdaq_Error()
{
    return VMEdaq_Error;
}

static void deplete_list_of_subscribers(observer *O, gpointer user_data)
{
    Py_DECREF(O->the_object);
    if (O->fcn)
        Py_DECREF(O->new_file);
    else
    {
        Py_XDECREF(O->new_file);
        Py_XDECREF(O->start_of_run);
        Py_XDECREF(O->data);
        Py_XDECREF(O->end_of_run);
        Py_XDECREF(O->pedestal);
        Py_XDECREF(O->new_point);
        Py_XDECREF(O->pixel_map);
        Py_XDECREF(O->scan);
        Py_XDECREF(O->end_of_record);
        Py_XDECREF(O->end_of_event);
        Py_XDECREF(O->config);
        Py_XDECREF(O->applicable);
    }
    g_free(O );
}

static PyObject *vmedaq_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
    vmedaq *self;

    self = (vmedaq *)type->tp_alloc(type, 0);
    if (self!=NULL)
    {
        self->in = NULL;
        self->objs = NULL;
        self->files = NULL;
        self->buffer= NULL;
        self->bfsiz = 0;
        self->last_event = -1;
        self->new_event = NULL;
        self->end_of_event = NULL;

    }
    return (PyObject *)self;
}

static void vmedaq_dealloc(vmedaq* self)
{
    if (self->in)
    {
        gzclose(self->in);
    }

    if (self->objs)
    {
        g_slist_foreach(self->objs, (GFunc)deplete_list_of_subscribers, NULL);
        g_slist_free(self->objs);
    }

    if (self->files)
    {
        Py_DECREF(self->files);
    }

    if (self->buffer)
    {
        free(self->buffer);
        self->buffer=0;
        self->bfsiz=0;
    }
    if (self->new_event)
    {
        Py_DECREF(self->new_event);
    }
    if (self->end_of_event)
    {
        Py_DECREF(self->end_of_event);
    }
    self->ob_type->tp_free((PyObject*)self);
}

static int add_subscriber(vmedaq *self, PyObject *O)
{
    int nzero=0;
    observer *o= g_new0( observer, 1);
    Py_INCREF(O);
    o->the_object = O;
    o->has_scan=0;
    // Look now for the methods
    if ( !(o->new_file = PyObject_GetAttrString(O, "NewFile") ))
        nzero++;
    if ( !(o->start_of_run = PyObject_GetAttrString(O, "StartOfRunRecord") ))
        nzero++;
    if ( !(o->data = PyObject_GetAttrString(O, "DataRecord") ))
        nzero++;
    if ( !(o->end_of_run = PyObject_GetAttrString(O, "EndOfRunRecord") ))
        nzero++;
    if ( !(o->pedestal = PyObject_GetAttrString(O, "PedestalRecord") ))
        nzero++;
    if ( !(o->new_point = PyObject_GetAttrString(O, "NewPointRecord") ))
        nzero++;
    if ( !(o->pixel_map = PyObject_GetAttrString(O, "MapRecord") ))
        nzero++;
    if ( !(o->scan = PyObject_GetAttrString(O, "ScanRecord") ))
        nzero++;
    if ( !(o->config = PyObject_GetAttrString(O, "ConfigRecord") ))
        nzero++;
    if ( !(o->applicable = PyObject_GetAttrString(O, "applicable") ))
        nzero++;
        
    if (PyObject_HasAttrString(O, "EndOfRecord") )
        o->end_of_record = PyObject_GetAttrString(O, "EndOfRecord");

    if (PyObject_HasAttrString(O, "EndOfEvent") )
        o->end_of_event = PyObject_GetAttrString(O, "EndOfEvent");

    /*
     * Does not have any of the methods, let's see if it is callable
     */
    if (nzero==10)
    { // does not have any of the methods
        if (PyCallable_Check(O) )
        {
            o->new_file = O;
            o->start_of_run = O;
            o->end_of_run = O;
            o->pedestal = O;
            o->data = O;
            o->new_point = O;
            o->pixel_map = O;
            o->scan = O;
            o->config = 0;
            o->fcn=1;
        }
        else
        {
            // This is not a valid object
            g_free(o);
            o=NULL;
            Py_DECREF(O);
            PyErr_SetString(VMEdaq_Error, "vmedaq: Object cannot subscribe");
            return -1;
        }
    }
    else
        o->fcn=0;

    /* If we are here, this are good news */
    self->objs = g_slist_append(self->objs, o);

    return 0;
}

static int vmedaq_init(vmedaq *self, PyObject *args, PyObject *kwargs)
{
    PyObject *O=Py_None;

    if (!PyArg_ParseTuple(args, "|O:VMEdaq.__init__", &O))
        return -1;

    if (O!=Py_None)
    {
        if (add_subscriber(self, O) )
            return -1;
    }
    self->in = NULL;
    self->iter = NULL;
    self->files = NULL;
    self->buffer= NULL;
    self->bfsiz = 0;
    self->last_event = -1;
    self->new_event = PyObject_GetAttrString((PyObject *)self, "new_event");
    self->end_of_event = PyObject_GetAttrString((PyObject *)self, "end_of_event");
    return 0;
}

PyDoc_STRVAR(open_doc,"\
Opens a new file. It accepts as input either a single file name\n\
or a list of file names. It will throw an exception if any of the\n\
file names does not correspond to an existing file.\n\
");

static PyObject *vmedaq_open(vmedaq *self, PyObject *args, PyObject *kwargs)
{
    PyObject *str;
    if (!PyArg_ParseTuple(args, "O:VMEdaq.open", &str))
        return NULL;

    if ( PyString_Check(str))
    { /* We just got one file */
        if (self->files)
        {
            Py_DECREF(self->files);
        }
        self->files = PyList_New(0);
        Py_INCREF(str);
        PyList_Append(self->files, str);
    }
    else if (PySequence_Check(str) )
    { /* we got a list of files */
        int i, len= PySequence_Length(str);
        if (len<=0)
        {
            PyErr_SetString(VMEdaq_Error, "Empty file list");
            return NULL;
        }
        if (self->files)
        {
            Py_DECREF(self->files);
        }

        self->files = PyList_New(0);
        for (i=0; i<len; i++)
        {
            PyObject *O = PySequence_GetItem(str, i);
            if ( PyString_Check(O))
            {
                Py_INCREF(O);
                PyList_Append(self->files, O);
            }
            else
            {
                char line[256];
                sprintf(line, "Element %d is not a file name", i);
                PyErr_SetString(VMEdaq_Error, line);
                Py_DECREF(self->files);
                return NULL;
            }
        }
    }
    else
    {
        PyErr_SetString(VMEdaq_Error, "Parameter is neither a string nor a list of file names");
        return NULL;
    }

    self->iter = PyObject_GetIter(self->files);
    if (self->in)
        gzclose(self->in);

    self->in = NULL;

    Py_INCREF(Py_None);
    return Py_None;
}

static int call_object(PyObject *O, PyObject *val)
{
    PyObject *ret,*excp;
    ret=PyObject_CallFunction(O, "(O)", val);
    
    if (!ret)
        return -1;

    if (PyErr_Occurred())
        return -1;

    Py_DECREF(ret);
    return 0;
}

static int open_next_file(vmedaq *self)
{
    PyObject *str;
    if (self->in)
    {
        gzclose(self->in);
        self->in=NULL;
    }

    if (!self->files)
    {
        PyErr_SetString(VMEdaq_Error, "No file to read");
        return 1;
    }

    while ( (str=PyIter_Next(self->iter)))
    {
        self->in = gzopen(PyString_AsString(str), "rb");
        if (!self->in)
        {
            printf("VMEdaq: could not open %s\n", PyString_AsString(str));
            continue;
        }
        break;
    }
    if (!self->in)
        return -1;

    if (self->objs)
    {
        GSList *lst = self->objs;
        for (; lst; lst=lst->next)
        {
            int rc;
            rc = call_object(((observer *)lst->data)->new_file, str);
            if (rc)
                return rc;
        }
    }
    return 0;
}

PyDoc_STRVAR(close_doc,"\
             Closes any file left opened.\n\
             ");
static PyObject *vmedaq_close(vmedaq *self, PyObject *args, PyObject *kwargs)
{
    if (self->in)
    {
        gzclose(self->in);
        self->in = NULL;
    }

    Py_INCREF(Py_None);
    return Py_None;
}

PyDoc_STRVAR(rewind_doc,"\
             rewinds the current file.\n\
             ");
static PyObject *vmedaq_rewind(vmedaq *self, PyObject *args, PyObject *kwargs)
{
    if (self->in)
    {
        gzrewind(self->in);
    }

    Py_INCREF(Py_None);
    return Py_None;
}

PyDoc_STRVAR(register_doc,"\
             Registers a new object to be notified when new blocks are found.\n\
             Objects can be either a callable object or a class. If one registers\n\
             a function, the function will receive for each data block the whole\n\
             data buffer. If we register a class, the following methods will be called\n\
             depending of the data record type\n\
             - NewFile(self) : when a new file is opened\n\
             - StartOfRunRecord(self,data): when a StartOfRunRecord is found\n\
             - ConfigRecord(self,data): when a ConfigRecord is found\n\
             - DataRecord(self,data): when a DataRecord is found\n\
             - EndOfRunRecord(self,data): when an EndOfRunRecord is found\n\
             - PedestalRecord(self,data): when a PedestalRecord is found\n\
             - NewPointRecord(self,data): when a NewPointRecord is found\n\
             - ScanRecord(self,data): when a ScanRecord is found\n\
             - MapRecord(self,data): when a MapRecord\n\
             ");
static PyObject *vmedaq_register(vmedaq *self, PyObject *args, PyObject *kwargs)
{
    PyObject *O=Py_None;

    if (!PyArg_ParseTuple(args, "O:VMEdaq.register", &O))
        return NULL;

    if (O!=Py_None)
        add_subscriber(self, O);

    Py_INCREF(Py_None);
    return Py_None;
}

PyDoc_STRVAR(unregister_doc,"\
             Unregisters an object previously registered to receive the data\n\
             ");
static PyObject *vmedaq_unregister(vmedaq *self, PyObject *args, PyObject *kwargs)
{
    PyObject *O=Py_None;

    if (!PyArg_ParseTuple(args, "O:VMEdaq.unregister", &O))
        return NULL;

    if (O!=Py_None)
    {
        gpointer *ptr=NULL;
        GSList *lst = self->objs;
        for (; lst; lst=lst->next)
        {
            observer *o = (observer *)lst->data;
            if (O==o->the_object)
            {
                ptr = lst->data;
                break;
            }
        }
        if (ptr)
        {
            self->objs = g_slist_remove(self->objs, ptr);
            printf("Object unregistered\n");
        }
    }

    Py_INCREF(Py_None);
    return Py_None;
}

PyDoc_STRVAR(read_doc,"\
             Reads the next data block. Returns None at the end of the stream\n\
             otherwise it will return the data block type.\n\
             ");
static PyObject *vmedaq_read(vmedaq *self, PyObject *args, PyObject *kwargs)
{
    PyObject *data;
    int id, iok, block_size, evt_num;
    size_t sz;
    unsigned int header, block_type, data_type, x;
    GSList *lst;

    if ( !self->in)
    {
        if (open_next_file(self) )
            return NULL;
    }

    again: do
    {
        iok = gzread(self->in, &x, sizeof(unsigned int));
        if (iok<=0)
        {
            x = 0;
            if (open_next_file(self) )
            {
                Py_INCREF(Py_None);
                return Py_None;
            }
            continue;
        }
    } while (!IsDataBlock(x));

    header = x;

    block_size = x & 0xfff;
    sz = block_size*sizeof(unsigned int);
    if (self->bfsiz<sz )
    {
        self->buffer = realloc(self->buffer, sz);
        self->bfsiz = sz;
    }

    data = PyVmedaq_Buffer_New(sz,(char *)self->buffer);

    sz -= sizeof(unsigned int);
    self->buffer[0]=header;
    iok = gzread(self->in, self->buffer+1, sz);
    if (iok!=sz)
        goto again;


    /* Loop on registered objects */
    block_type = (header&0xf000);
    data_type = self->buffer[2] & 0xf0000;
    id = (self->buffer[2]&0xfff00000L)>>20;
    for (lst=self->objs; lst; lst=lst->next)
    {
        observer *o = (observer *)lst->data;
        iok=0;
        switch (block_type)
        {
        case StartOfRunRecord:
            if (o->start_of_run)
            {
                iok += call_object(o->start_of_run, data);
            }
            break;
        case DataRecord:
            evt_num = self->buffer[1] & 0xffffff;
            if (self->last_event !=evt_num )
            {
                if (self->last_event>=0)
                {
                    if (!PyObject_CallFunction(self->end_of_event, "i", self->last_event))
                    {
                        Py_DECREF(data);
                        return NULL;
                    }
                }
                if (self->new_event)
                {
                    if (!PyObject_CallFunction(self->new_event, "i", evt_num))
                    {
                        Py_DECREF(data);
                        return NULL;
                    }
                }
                self->last_event = evt_num;
            }

            if (o->data)
            {
                /* Check if the observer wants to get this DataBlock */
                if (o->applicable)
                {
                    int rc=0;
                    PyObject *ret = PyObject_CallFunction(o->applicable, "i", id);
                    if (ret)
                        rc = PyInt_AsLong(ret);
                    else
                    {
                        Py_DECREF(data);
                        return NULL;
                    }

                    if (!rc)
                        break;
                }

                if (evt_num!=o->last_evt)
                {
                    if (o->end_of_event>0)
                    {
                        if (!PyObject_CallFunction(o->end_of_event, "ii", id, evt_num))
                        {
                            Py_DECREF(data);
                            return NULL; 
                        }
                    }
                    o->last_evt = evt_num;
                }
                iok += call_object(o->data, data);
            }

            break;
        case EndOfRunRecord:
            if (o->end_of_run)
            {
                iok += call_object(o->end_of_run, data);
            }
            break;
        case PedestalRecord:
            if (o->pedestal)
            {
                iok += call_object(o->pedestal, data);
            }
            break;
        case NewPointRecord:
            o->has_scan++;
            if (o->new_point)
            {
                iok += call_object(o->new_point, data);
            }
            break;
        case MapRecord:
            if (o->pixel_map)
            {
                iok += call_object(o->pixel_map, data);
            }
            break;
        case ScanRecord:
            if (o->scan)
            {
                iok += call_object(o->scan, data);
            }
            break;
        case ConfigRecord:
            if (o->config)
            {
                iok += call_object(o->config, data);
            }
            break;
        default:
            printf("*** Unknown blocktype %x\n", header);
            break;
        }
        if (iok)
        {
            Py_DECREF(data);
            return NULL;
        }
        
        /* Call to the end of record method */
        if (o->end_of_record)
        {
            PyObject *ret;
            ret = PyObject_CallFunction(o->end_of_record, "iO", block_type, data);
            if (ret)
                Py_DECREF(ret);
            else
            {
                Py_DECREF(data);
                return NULL;
            }
        }
    }

    Py_DECREF(data);
    return PyInt_FromLong(block_type);
}

PyDoc_STRVAR(loop_doc,"\
             Loops until there are no more data records to read.\n\
             ");
static PyObject *vmedaq_loop(vmedaq *self, PyObject *args, PyObject *kwargs)
{

    while (1)
    {
        PyObject *ret = vmedaq_read(self, args, kwargs);
        if (!ret)
            return NULL;

        if (ret==Py_None)
        {
            Py_DECREF(ret);
            break;
        }
        Py_DECREF(ret);

        if (PyErr_Occurred())
            return NULL;
    }
    /* If we are reading scan data, call end of point after the last
     file hasbee read
     */
    if (self->objs)
    {
        GSList *lst = self->objs;
        for (; lst; lst=lst->next)
        {
            observer *O = (observer *)lst->data;
            if (!O)
                continue;

            if (O->has_scan)
            {
                PyObject *ret =PyObject_CallMethod(O->the_object, "onEndPoint", NULL);
                if (ret)
                {
                    Py_DECREF(ret);
                }
                else
                    return NULL;
            }
        }
    }

    Py_INCREF(Py_None);
    return Py_None;
}

PyDoc_STRVAR(buffer_doc,"\
             Returns a buffer with the current record's data.\n\
             ");

static PyObject *vmedaq_buffer(vmedaq *self, PyObject *unused)
{
    return PyVmedaq_Buffer_New(self->bfsiz, (char *)self->buffer);
}

PyDoc_STRVAR(new_event_doc,"\
             Called when a new event starts\n\
             ");
static PyObject *vmedaq_new_event(vmedaq *self, PyObject *unused)
{
    Py_INCREF(Py_None);
    return Py_None;
}

PyDoc_STRVAR(end_of_event_doc,"\
             Called at the end of each event\n\
             ");
static PyObject *vmedaq_end_of_event(vmedaq *self, PyObject *unused)
{
    Py_INCREF(Py_None);
    return Py_None;
}

static PyMethodDef vmedaq_methods[] = 
{ 
 { "open", (PyCFunction)vmedaq_open, METH_VARARGS, open_doc }, 
 { "close",(PyCFunction)vmedaq_close,METH_NOARGS, close_doc },
 { "rewind", (PyCFunction)vmedaq_rewind, METH_NOARGS, rewind_doc },
 { "register", (PyCFunction)vmedaq_register, METH_VARARGS, register_doc },
 { "unregister", (PyCFunction)vmedaq_unregister, METH_VARARGS, unregister_doc },
 { "read", (PyCFunction)vmedaq_read, METH_NOARGS, read_doc }, 
 { "loop",(PyCFunction)vmedaq_loop, METH_NOARGS, loop_doc },
 { "buffer", (PyCFunction)vmedaq_buffer, METH_NOARGS, buffer_doc },
 { "new_event", (PyCFunction)vmedaq_new_event, METH_VARARGS, new_event_doc }, 
 { "end_of_event", (PyCFunction)vmedaq_end_of_event, METH_VARARGS, end_of_event_doc }, 
 { NULL } /* Sentinel */
};

/*
static int vmedaq_traverse(vmedaq *self, visitproc visit, void *arg)
{
    int vret;
    GSList *lst;
    for (lst=self->objs; lst; lst=lst->next)
    {
        observer *o = (observer *)lst->data;
        vret = visit(o->the_object, arg);
        if (vret != 0)
            return vret;
    }
    vret = visit(self->files, arg);
    if (vret != 0)
        return vret;

    return 0;
}
*/

static PyTypeObject VMEdaq_Type = 
{ PyObject_HEAD_INIT(NULL)0, /*ob_size*/
  "vmedaq.VMEdaq", /*tp_name*/
  sizeof(vmedaq), /*tp_basicsize*/
  0, /*tp_itemsize*/
  (destructor)vmedaq_dealloc, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_compare*/
  (reprfunc)0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash */
  (ternaryfunc)0, /*tp_call*/
  (reprfunc)0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT| Py_TPFLAGS_BASETYPE,
  "Reads a VMEDAQ file", /* tp_doc */
  (traverseproc)0, /* tp_traverse */
  0, /* tp_clear */
  0, /* tp_richcompare */
  0, /* tp_weaklistoffset */
  0, /* tp_iter */
  0, /* tp_iternext */
  vmedaq_methods, /* tp_methods */
  0, /* tp_members */
  0, /* tp_getset */
  0, /* tp_base */
  0, /* tp_dict */
  0, /* tp_descr_get */
  0, /* tp_descr_set */
  0, /* tp_dictoffset */
  (initproc)vmedaq_init, /* tp_init */
  0, /* tp_alloc */
  vmedaq_new, /* tp_new */
};

static PyObject *pyvmedaq_NormalData(PyObject * self, PyObject * args, PyObject * kwargs)
{
    long ival;
    if (!PyArg_ParseTuple (args, "l", &ival))
    {
        return NULL;
    }
    return PyInt_FromLong (IsNormalData (ival));
}
;
static PyObject *pyvmedaq_PulseData(PyObject * self, PyObject * args, PyObject * kwargs)
{
    long ival;
    if (!PyArg_ParseTuple (args, "l", &ival))
    {
        return NULL;
    }
    return PyInt_FromLong (IsPulseData (ival));
}
;
static PyObject *pyvmedaq_ScanData(PyObject * self, PyObject * args, PyObject * kwargs)
{
    long ival;
    if (!PyArg_ParseTuple (args, "l", &ival))
    {
        return NULL;
    }
    return PyInt_FromLong (IsScanData (ival));
}
;
static PyObject *pyvmedaq_ADCdata(PyObject * self, PyObject * args, PyObject * kwargs)
{
    long ival;
    if (!PyArg_ParseTuple (args, "l", &ival))
    {
        return NULL;
    }
    return PyInt_FromLong (IsADCdata (ival));
}
;
static PyObject *pyvmedaq_FBTDCdata(PyObject * self, PyObject * args, PyObject * kwargs)
{
    long ival;
    if (!PyArg_ParseTuple (args, "l", &ival))
    {
        return NULL;
    }
    return PyInt_FromLong (IsFBTDCdata (ival));
}
;

static PyObject *pyvmedaq_SIS3300data(PyObject * self, PyObject * args, PyObject * kwargs)
{
    long ival;
    if (!PyArg_ParseTuple (args, "l", &ival))
    {
        return NULL;
    }
    return PyInt_FromLong (IsSIS3300data (ival));
}
;

static PyObject *pyvmedaq_SecDetdata(PyObject * self, PyObject * args, PyObject * kwargs)
{
    long ival;
    if (!PyArg_ParseTuple (args, "l", &ival))
    {
        return NULL;
    }
    return PyInt_FromLong (IsSecDetData (ival));
}
;

#define set_int_attribute( obj, attr, val ) {   \
    PyObject *tmp = PyInt_FromLong( (val) );	\
    PyObject_SetAttrString(obj,attr,tmp);	\
    Py_DECREF(tmp);				\
}

PyDoc_STRVAR(pedestal_doc,"\
             Called when a new PedestalRecord record is found.\n\
             It decodes the pedestals and stores the values in the\n\
             corresponding Module Object.\n\
             \n\
             Requires as input the data array and the dict of modules\n\
             ");
static PyObject *decode_PedestalRecord(PyObject * self, PyObject * args, PyObject * kwargs)
{
    unsigned int *data= NULL;
    int i, rc;
    Py_ssize_t size;
    unsigned ichan, nchip, nchan, ntot;
    PyObject *val, *module, *ped, *noise;
    PyObject *pybuf=NULL;
    PyObject *idlist=NULL;
    if (!PyArg_ParseTuple (args, "OO:PedestalRecord", &pybuf, &idlist))
    {
        return NULL;
    }
    /* Check the input */
    if (!pybuf || !idlist)
    {
        PyErr_SetString(VMEdaq_Error, "PedestalRecord: bad arguments");
        return NULL;
    }
    /* Get the buffer */
    rc = PyObject_AsReadBuffer(pybuf, (const void **)&data, &size);
    if (rc)
    {
        PyErr_SetString(VMEdaq_Error, "Pedestal record expects a buffer as its first argument");
        return NULL;
    }

    /* Get the module from the list */
    val = PyInt_FromLong(data[1]);
    module = PyDict_GetItem(idlist, val);
    if (!module)
    {
        PyErr_SetString(VMEdaq_Error, "Pedestal record: someting went very badly.\n"
            "The module is not in the list...");
        return NULL;
    }
    Py_DECREF(val);

    /* Set a number of things */
    nchan = (data[2]&0xffffL);
    nchip = (data[2]&0xffff0000L)>>16;
    ntot = nchan*nchip;

    set_int_attribute(module,"nchip",nchip);
    set_int_attribute(module,"nchan",nchan);
    set_int_attribute(module,"ntot",ntot);

    /* And finally, the pedestals and noise */
    ped = PyTuple_New(ntot );
    noise = PyTuple_New(ntot );

    size = data[0] & 0xfffL;
    if (size-4!= ntot )
    {
        printf("Inconsistent size %d [%d]\n", size-4, ntot);
    }
    size--;
    for (ichan=0, i=3; i<size; i++, ichan++)
    {
        double nv = ((data[i] & 0xffff0000L)>>16)/50.;
        double pv = (data[i] & 0xffffL)/50.;

        PyTuple_SET_ITEM(ped,ichan, PyFloat_FromDouble(pv));
        PyTuple_SET_ITEM(noise,ichan, PyFloat_FromDouble(nv));
    }
    PyObject_SetAttrString(module, "pedestals", ped);
    PyObject_SetAttrString(module, "noise", noise);

    return PyInt_FromLong(data[1]);

}

static PyObject *decode_serial(unsigned int *data)
{
    int i, j, val, sz = data[0] & 0xfffL;
    PyObject *out, *item;

    sz--;
    out = PyList_New(sz-6);
    for (i=0, j=6; j<sz; j++, i++)
    {
        val = data[j] & 0x3ffL;
        item = PyTuple_New(3);
        PyTuple_SET_ITEM(item,0,PyInt_FromLong(i));
        PyTuple_SET_ITEM(item,1,PyInt_FromLong(val));
        PyTuple_SET_ITEM(item,2,PyInt_FromLong(0));
        PyList_SET_ITEM(out, i, item);
        //	PyList_Append(out,item);
        //	Py_DECREF(item);
    }
    //    printf("sz-6 %d length %d i %d\n",sz-6,PyList_Size(out),i);
    return out;
}


static PyObject * decode_sparse_gp3(unsigned int *data, int use_up)
{
    int dsz = (data[0] & 0xfff)-1;
    int romode = (data[2] & 0xf000)>>12;
    int nadj = romode==8 ? (data[1] & 0xff000000L)>>24 : 0;
    int nblk = nadj +2, ndat = nblk-1;
    int nTrig;
    int i,j;
    
    PyObject *item, *out;
    out = PyList_New(0);
    
    for (j = 6, nTrig = 0; j < dsz; nTrig++)
    {
        /* We fix here the multiple hit problem
         */
        if (nTrig>1)
            j++;
        
        if ((dsz-j)<ndat)
            break;
        
        //channel that gave this trigger
        int add = -1111;

        for (i = 0; i < ndat && j<dsz; i++, j++)
        {
            int dt = (data[j-3] & 0x7ff00000L) >> 20;
            int down = data[j] & 0x3ffL;
            int up = (data[j] & 0xffc00L) >> 10;
            if (i == 0)
            {
                add = dt;
                item = PyTuple_New(3);
                PyTuple_SET_ITEM(item,0,PyInt_FromLong(add));
                PyTuple_SET_ITEM(item,1,PyInt_FromLong(down));
                PyTuple_SET_ITEM(item,2,PyInt_FromLong(0));
                PyList_Append(out,item);
                if (use_up!=0)
                {
                    item = PyTuple_New(3);
                    PyTuple_SET_ITEM(item,0,PyInt_FromLong(add));
                    PyTuple_SET_ITEM(item,1,PyInt_FromLong(up));
                    PyTuple_SET_ITEM(item,2,PyInt_FromLong(0));
                    PyList_Append(out,item);
                }
            }
            else
            {
                //set the adjacent flag to true
                item = PyTuple_New(3);
                PyTuple_SET_ITEM(item,0,PyInt_FromLong(add+i));
                PyTuple_SET_ITEM(item,1,PyInt_FromLong(down));
                PyTuple_SET_ITEM(item,2,PyInt_FromLong(1));
                PyList_Append(out,item);
                if (use_up!=0)
                {
                    item = PyTuple_New(3);
                    PyTuple_SET_ITEM(item,0,PyInt_FromLong(add-i));
                    PyTuple_SET_ITEM(item,1,PyInt_FromLong(up));
                    PyTuple_SET_ITEM(item,2,PyInt_FromLong(1));
                    PyList_Append(out,item);
                }
            }
        }
    }
    return out;
}

static PyObject * decode_sparse_gp7(unsigned int *data, int use_up)
{
    int dsz = (data[0] & 0xfff)-1;
    int romode = (data[2] & 0xf000)>>12;
    int nadj = romode==8 ? (data[1] & 0xff000000L)>>24 : 0;
    int nblk = nadj +2, ndat = nblk-1;
    int nTrig;
    int i,j;
    
    PyObject *item, *out;
    out = PyList_New(0);
    
    for (j = 6, nTrig = 0; j < dsz; nTrig++)
    {
        /* We fix here the multiple hit problem
         */
        if (nTrig>1)
            j++;
        
        if ((dsz-j)<ndat)
            break;
        
        //channel that gave this trigger
        int add = -1111;
        ndat = nblk;
        for (i = 0; i < ndat && j<dsz; i++, j++)
        {
            int dt = (data[j-3] & 0x7ff00000L) >> 20;
            int down = data[j] & 0x3ffL;
            int up = (data[j] & 0xffc00L) >> 10;
            if (i == 0)
            {
                add = dt;
                item = PyTuple_New(3);
                PyTuple_SET_ITEM(item,0,PyInt_FromLong(add));
                PyTuple_SET_ITEM(item,1,PyInt_FromLong(down));
                PyTuple_SET_ITEM(item,2,PyInt_FromLong(0));
                PyList_Append(out,item);
                if (use_up!=0)
                {
                    item = PyTuple_New(3);
                    PyTuple_SET_ITEM(item,0,PyInt_FromLong(add));
                    PyTuple_SET_ITEM(item,1,PyInt_FromLong(up));
                    PyTuple_SET_ITEM(item,2,PyInt_FromLong(0));
                    PyList_Append(out,item);
                }
                ndat --;
            }
            else
            {
                //set the adjacent flag to true
                item = PyTuple_New(3);
                PyTuple_SET_ITEM(item,0,PyInt_FromLong(add+i));
                PyTuple_SET_ITEM(item,1,PyInt_FromLong(down));
                PyTuple_SET_ITEM(item,2,PyInt_FromLong(1));
                PyList_Append(out,item);
                if (use_up!=0)
                {
                    item = PyTuple_New(3);
                    PyTuple_SET_ITEM(item,0,PyInt_FromLong(add-i));
                    PyTuple_SET_ITEM(item,1,PyInt_FromLong(up));
                    PyTuple_SET_ITEM(item,2,PyInt_FromLong(1));
                    PyList_Append(out,item);
                }
            }
        }
    }
    return out;
}


PyDoc_STRVAR(decode_vatagp_doc,"\
             Decodes the data of the VATAGP. It returns a tuple. Each element\n\
             of the tuple is itself another tuple where the first element is the\n\
             channel number, the second the signal and the third a boolean that,\n\
             when true, tells that this is an adjacent channel.\n\
             ");


static PyObject*decode_vatagp(PyObject * self, PyObject * args, PyObject * kwargs)
{
    int rc;
    Py_ssize_t size;
    int romode;
    unsigned int *data= NULL;
    unsigned int evtcnt;
    int use_up=0;
    int chip_type=2;
    PyObject *pybuf=NULL;
    PyObject *obs=NULL;
    PyObject *module=NULL;
    PyObject *idlist=NULL;
    PyObject *out=NULL;
    PyObject *ret;
    PyObject *fcn;

    if (!PyArg_ParseTuple (args, "OOii:decode_vatagp", &obs, &pybuf, &chip_type, &use_up))
    {
        return NULL;
    }
    /* Check the input */

    /* IF no observer, no work */
    if (!obs)
    {
        Py_INCREF(Py_None);
        return Py_None;
    }

    /* No buffer, no data */
    if (!pybuf)
    {
        PyErr_SetString(VMEdaq_Error, "decode_vatagp: bad arguments");
        return NULL;
    }

    /* No one to call...*/
    fcn = PyObject_GetAttrString(obs, "onNewData");
    if (!fcn)
    {
        Py_INCREF(Py_None);
        return Py_None;
    }

    /* Get the buffer */
    rc = PyObject_AsReadBuffer(pybuf, (const void **)&data, &size);
    if (rc)
    {
        PyErr_SetString(VMEdaq_Error, "decode_adc expects a buffer as its first argument");
        return NULL;
    }

    /** Check integrity of the data
     */
    {
        int sz = (data[2] & 0xfffL);
        int dsz = data[0] & 0xfff;
        if (dsz != data[dsz-1]|| dsz-sz != 4)
        {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }

    /* Get the module from the list */
    if (obs)
    {
        idlist = PyObject_GetAttrString(obs, "idlist");
        if (idlist)
        {
            PyObject *id = PyInt_FromLong( (data[2]&0xfff00000L)>>20);
            module = PyDict_GetItem(idlist, id);
            Py_DECREF(id);
            Py_DECREF(idlist);
        }
    }

    /* Start decoding */
    romode = (data[2] & 0xf000L)>>12;

    if (romode==2)
    { //serial
        out = decode_serial(data);
    }
    else
    {
        switch (chip_type)
        {
            case 5: // GP7
                out = decode_sparse_gp7(data, use_up);
                break;
            case 4: // GP5
                // TODO: We assume by now the GP% sparse is like GP7
                out = decode_sparse_gp3(data, use_up);
                break;
            default:
                out = decode_sparse_gp3(data, use_up);
                break;
        }
    }

    evtcnt = (data[1] & 0x00ffffffL);
    ret = PyObject_CallFunction(fcn, "OlO", module, evtcnt, out);

    Py_XDECREF(ret);
    Py_XDECREF(out);
    Py_XDECREF(fcn);

    Py_INCREF(Py_None);
    return Py_None;
}

PyDoc_STRVAR(decode_adc_doc,"\
             Decodes the data of the Caen V785 ADC. It returns a tuple. Each element\n\
             of the tuple is itself another tuple where the first element is the\n\
             channel number and the second the signal.\n\
             ");
static PyObject* decode_adc(PyObject * self, PyObject * args, PyObject * kwargs)
{
    Py_ssize_t size;
    int i, sz, rc;
    int ibuffer, nbuffer, ip;
    unsigned int *data= NULL;
    unsigned int evtcnt;
    PyObject *pybuf=NULL;
    PyObject *obs=NULL;
    PyObject *module=NULL;
    PyObject *idlist=NULL;
    PyObject *out=NULL;
    PyObject *ret;
    PyObject *fcn;
    PyObject *item, *buffer, *chanlist;

    if (!PyArg_ParseTuple (args, "OO:decode_adc", &obs, &pybuf))
    {
        return NULL;
    }
    /* Check the input */

    /* IF no observer, no work */
    if (!obs)
    {
        Py_INCREF(Py_None);
        return Py_None;
    }

    /* No buffer, no data */
    if (!pybuf)
    {
        PyErr_SetString(VMEdaq_Error, "decode_adc: bad arguments");
        return NULL;
    }

    /* No one to call...*/
    fcn = PyObject_GetAttrString(obs, "onNewData");
    if (!fcn)
    {
        Py_INCREF(Py_None);
        return Py_None;
    }

    /* Get the buffer */
    rc = PyObject_AsReadBuffer(pybuf, (const void **)&data, &size);
    if (rc)
    {
        PyErr_SetString(VMEdaq_Error, "decode_adc expects a buffer as its first argument");
        return NULL;
    }

    /** Check integrity of the data
     */
    sz = (data[2] & 0xfffL);
    {
        int dsz = data[0] & 0xfff;
        if (dsz != data[dsz-1]|| dsz-sz != 4)
        {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }

    /* Get the module from the list */
    if (obs)
    {
        idlist = PyObject_GetAttrString(obs, "idlist");
        if (idlist)
        {
            PyObject *id = PyInt_FromLong( (data[2]&0xfff00000L)>>20);
            module = PyDict_GetItem(idlist, id);
            Py_DECREF(id);
            Py_DECREF(idlist);
        }
    }

    /* Start decoding */

    nbuffer = data[3];
    ip = 4;
    out = PyTuple_New(nbuffer);
    for (ibuffer=0;ibuffer<nbuffer;ibuffer++)
    {
        int nchan, header = data[ip]>>16;
        if (header!=0xacdc)
        {
            printf("decode_adc: Wrong event format\n");
            Py_XDECREF(out);
            Py_INCREF(Py_None);
            return Py_None;
        }
        nchan = data[ip++] & 0xffff;
        buffer = PyTuple_New(2);
        chanlist = PyTuple_New(nchan);
        for (i=0; i<nchan; i++, ip++)
        {
            int add = (data[ip] >> 16) & 0x1F;
            int val = data[ip] & 0xFFF;
            item = PyTuple_New(2);
            PyTuple_SET_ITEM(item,0,PyInt_FromLong(add));
            PyTuple_SET_ITEM(item,1,PyInt_FromLong(val));
            PyTuple_SET_ITEM(chanlist, i, item);
        }
        PyTuple_SET_ITEM(buffer, 0, PyInt_FromLong(data[ip++]));
        PyTuple_SET_ITEM(buffer, 1, chanlist);
        PyTuple_SET_ITEM(out, ibuffer, buffer);
    }
    
    evtcnt = data[1];
    ret = PyObject_CallFunction(fcn, "OlO", module, evtcnt, out);

    Py_XDECREF(ret);
    Py_XDECREF(out);
    Py_XDECREF(fcn);

    Py_INCREF(Py_None);
    return Py_None;
}


PyDoc_STRVAR(decode_scan_doc,"\
             Decodes a ScanRecord with the description of a scan.\n\
             It returns a tuple with two elements:\n\
             0: the number of events per scan point\n\
             1: a tuple of tuples like (type,npoints, from, to)\n\
             ");
static PyObject*decode_scan(PyObject * self, PyObject * args, PyObject * kwargs)
{
    union
    {
        unsigned int lval;
        float fval;
    } from, to;
    Py_ssize_t size;
    int i, j;
    int nvar, nevt, type, npts, is_float;
    PyObject *pybuf=NULL;
    PyObject *out, *scan, *item;
    unsigned int *data=NULL;

    if (!PyArg_ParseTuple (args, "O:decode_scan", &pybuf))
    {
        return NULL;
    }
    /* Check the input */
    if (!pybuf)
    {
        PyErr_SetString(VMEdaq_Error, "ScanRecord: bad arguments");
        return NULL;
    }

    /* Get the buffer */
    if (PyObject_AsReadBuffer(pybuf, (const void **)&data, &size))
    {
        PyErr_SetString(VMEdaq_Error, "decode_scan expects a buffer as its first argument");
        return NULL;
    }

    nvar = data[1];
    nevt = data[2];
    scan = PyTuple_New(nvar);
    for (i=0; i<nvar; i++)
    {
        j = 3*(i+1);
        is_float = data[j] & 0x80000000L;
        type = ((data[j]&0xfff0000L)>>16);
        npts = (data[j]&0xffffL);
        from.lval = data[j+1];
        to.lval = data[j+2];
        item = PyTuple_New(4);
        PyTuple_SET_ITEM(item,0,PyInt_FromLong(type));
        PyTuple_SET_ITEM(item,1,PyInt_FromLong(npts));
        if (is_float)
        {
            PyTuple_SET_ITEM(item,2,PyFloat_FromDouble(from.fval));
            PyTuple_SET_ITEM(item,3,PyFloat_FromDouble(to.fval));
        }
        else
        {
            PyTuple_SET_ITEM(item,2,PyInt_FromLong(from.lval));
            PyTuple_SET_ITEM(item,3,PyInt_FromLong(to.lval));
        }
        PyTuple_SET_ITEM(scan,i,item);
    }
    out = PyTuple_New(2);
    PyTuple_SET_ITEM(out,0,PyInt_FromLong(nevt));
    PyTuple_SET_ITEM(out,1,scan);
    return out;
}
static PyMethodDef mpyvmedaq_functions[] = 
{
 { "IsNormalData", (PyCFunction) pyvmedaq_NormalData, METH_VARARGS,"returns trud if it is normal data" },
 { "IsPulseData", (PyCFunction) pyvmedaq_PulseData, METH_VARARGS,"returns true if it is pulse data" },
 { "IsScanData", (PyCFunction) pyvmedaq_ScanData, METH_VARARGS, "returns true if it is scan data" },
 { "IsADCdata", (PyCFunction) pyvmedaq_ADCdata, METH_VARARGS, "returns true if it is adc data" },
 { "IsFBTDCdata", (PyCFunction) pyvmedaq_FBTDCdata, METH_VARARGS,"returns true if the data is from the Fastbus TDC" },
 { "IsSIS3300data", (PyCFunction) pyvmedaq_SIS3300data, METH_VARARGS,"" },
 { "IsSecDetdata", (PyCFunction) pyvmedaq_SecDetdata, METH_VARARGS,"returns true if the data is from the Fastbus TDC" },
 { "decode_pedestal",(PyCFunction)decode_PedestalRecord ,METH_VARARGS, pedestal_doc },
 { "decode_scan", (PyCFunction)decode_scan , METH_VARARGS, decode_scan_doc },
 { "decode_vatagp", (PyCFunction)decode_vatagp , METH_VARARGS, decode_vatagp_doc }, 
 { "decode_adc", (PyCFunction)decode_adc , METH_VARARGS, decode_adc_doc }, 
 { NULL, NULL,0 }
};

PyMODINIT_FUNC
init_vmedaq(void)
{
    PyObject *m,*d;
    m = Py_InitModule3("_vmedaq",mpyvmedaq_functions,module_doc);
    d = PyModule_GetDict(m);

    if ( VMEdaq_Error == NULL ) {
        VMEdaq_Error = PyErr_NewException("_vmedaq.VMEdaqError", NULL, NULL);
        Py_INCREF(VMEdaq_Error);
        PyModule_AddObject(m, "VMEdaqError", VMEdaq_Error);
    }

    add_int_constant(d,"DataHeader",DataHeader);
    add_int_constant(d,"DataRecord",DataRecord);
    add_int_constant(d,"EndOfRunRecord",EndOfRunRecord);
    add_int_constant(d,"PedestalRecord",PedestalRecord);
    add_int_constant(d,"NewPointRecord",NewPointRecord);
    add_int_constant(d,"NewPointRecord",MapRecord);
    add_int_constant(d,"ScanRecord",ScanRecord);
    add_int_constant(d,"BigDataRecord",BigDataRecord);
    add_int_constant(d,"ConfigRecord", ConfigRecord);


    add_int_constant(d,"NormalData",NormalData);
    add_int_constant(d,"PulseData",PulseData);
    add_int_constant(d,"ScanData",ScanData);
    add_int_constant(d,"ADCdata",ADCdata);
    add_int_constant(d,"CSPRINTdata",CSPRINTdata);
    add_int_constant(d,"FBTDCdata",FBTDCdata);
    add_int_constant(d,"SIS3300data",SIS3300data);
    add_int_constant(d,"SecDetData",SecDetData);
    add_int_constant(d,"TDCCaenData",TDCCaenData);

    {
        PyObject *obj;
        obj = PyString_FromString("0.0.0");
        PyDict_SetItemString(d,"__revision__",obj);
        Py_DECREF(obj);
    }

    vmedaq_register_class(d,"VMEdaq",&VMEdaq_Type,0);
    register_vmedaq_buffer(d);
}

void add_int_constant(PyObject *dic, char *name, int value)
{
    PyObject *obj;
    obj = PyInt_FromLong(value);
    PyDict_SetItemString(dic, name, obj);
    Py_DECREF(obj);
}

void vmedaq_register_class(PyObject *dict, const char *type_name, PyTypeObject *type, PyObject *bases)
{
    const char *s, *class_name;

    class_name=type->tp_name;
    s = strrchr(class_name, '.');
    if (s!=NULL)
        class_name = s+1;

    type->ob_type = &PyType_Type;
    if (bases)
    {
        type->tp_bases = bases;
        type->tp_base = (PyTypeObject *)PyTuple_GetItem(bases, 0);
    }
    if (PyType_Ready(type) < 0)
    {
        printf("couldn't make the type %s ready\n", type->tp_name);
        return;
    }

    PyDict_SetItemString(dict, (char *)class_name, (PyObject *)type);
}
