#include <iostream>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <libxml/parserInternals.h>
#include <libxml/xmlschemas.h>
#include <libxml/xmlschemastypes.h>
#include <DAQ++/DAQmanager.h>
#include <daq/DAQcontroller.h>
#include "SchemaPath.h"
#include "Config.h"
#include "config_utils.h"


#if LIBXML_VERSION<20600

/**
 * xmlReadFile:
 * \param[in] filename  a file or URL
 * \param     encoding  the document encoding, or NULL
 * \param     options   a combination of xmlParserOption
 *
 * parse an XML file from the filesystem or the network.
 * 
 * Returns the resulting document tree
 */
xmlDocPtr
xmlReadFile(const char *filename, const char *encoding, int options)
{
    xmlParserCtxtPtr ctxt;
    xmlParserInputPtr inputStream;
    char *directory = NULL;
    const char *URL = NULL;
    int reuse = 0;
    
    ctxt = xmlNewParserCtxt();
    if (ctxt == NULL) {
       std::cout << "Could not allocate parser context" << std::endl;
       return 0;
    }
    
    inputStream = xmlLoadExternalEntity(filename, NULL, ctxt);
    if (inputStream == NULL) {
	xmlFreeParserCtxt(ctxt);
	return(NULL);
    }
    inputPush(ctxt, inputStream);
    if ((ctxt->directory == NULL) && (directory == NULL))
        directory = xmlParserGetDirectory(filename);
    if ((ctxt->directory == NULL) && (directory != NULL))
        ctxt->directory = directory;


    xmlDocPtr ret;
    if (encoding != NULL) {
        xmlCharEncodingHandlerPtr hdlr;

	hdlr = xmlFindCharEncodingHandler(encoding);
	if (hdlr != NULL)
	    xmlSwitchToEncoding(ctxt, hdlr);
    }
    if ((URL != NULL) && (ctxt->input != NULL) &&
        (ctxt->input->filename == NULL))
        ctxt->input->filename = (char *) xmlStrdup((const xmlChar *) URL);
    xmlParseDocument(ctxt);
    if ((ctxt->wellFormed) || ctxt->recovery)
        ret = ctxt->myDoc;
    else {
        ret = NULL;
	if (ctxt->myDoc != NULL) {
	    xmlFreeDoc(ctxt->myDoc);
	}
    }
    ctxt->myDoc = NULL;
    if (!reuse) {
	xmlFreeParserCtxt(ctxt);
    }

    return (ret);
}
xmlDocPtr
xmlReadMemory(const char *buffer, int size, const char *URL, const char *encoding, int options)
{
    return xmlParseMemory(buffer, size);
}

#endif

/**
 * Check that a given path points to a valid file
 * */
static int 
isfile(const char *filename)
{
    struct stat buf;
    if (stat(filename, &buf) != 0)
        return 0;
    
    if (!S_ISREG(buf.st_mode))
        return 0;
    
    return 1;
}

/**
 * Looks for a file in the given search path
 */
std::string 
search_file(const std::string &file_name,
                        const std::string &search_path,
                        const std::string pathsep)
{
    if ( isfile(file_name.c_str()) )
        return file_name;

    std::string abspath;
    std::vector<std::string> paths;
    stringtok<std::vector<std::string> > (paths, search_path, pathsep.c_str());
    if (paths.empty())
        paths.push_back(".");

    std::vector<std::string>::iterator ip;
    for (ip = paths.begin();ip != paths.end();ip++)
    {
        abspath = *ip + std::string("/") + file_name;
        if (isfile(abspath.c_str()))
            break;
    }
    return abspath;
}



static int bit_array(std::string val, int nx, std::vector<int> &vv);

template <>
bool ConfigParser::get_value(xmlNodePtr node)
{
    std::string sval = get_node_value(node);
    sval = downcase(sval);
    if (sval == "true")
    {
        return true;
    }
    else if (sval == "false")
    {
        return false;
    }
    else
    {
        std::cout << "get_value<bool>, wrong boolan " << sval << std::endl;
        std::cout << "...in node " << node->name << " @ line " << xmlGetLineNo(node) << std::endl;
        if (sval == "1")
        {
            set_value(node, "true");
            return true;
        }
        else if (sval == "0")
        {
            set_value(node, "false");
            return false;
        }
        else
        {
            std::cout << "...could not repair..." << std::endl;
        }
    }
    return false;
}

template <>
void ConfigParser::set_value(xmlNodePtr node, const bool &val)
{
    std::string sout(val ? "true" : "false");
    xmlNodeSetContent(node, (xmlChar *)sout.c_str());
}

template <>
xmlNodePtr ConfigParser::create_child(xmlNodePtr node, const std::string &name, const bool &val)
{
    std::string sout(val ? "true" : "false");
    return xmlNewTextChild (node, NULL,
                            (xmlChar *)name.c_str(),
                            (xmlChar *)sout.c_str() );
}

template <>
void ConfigParser::set_attribute(xmlNodePtr node, const std::string &att, const bool &val)
{
    std::string sout(val ? "true" : "false");
    xmlSetProp(node, (xmlChar *)att.c_str(), (xmlChar *)sout.c_str());
}

template <>
std::string ConfigParser::get_attribute(xmlNodePtr node, const std::string &name)
throw( ConfigException )
{
    xmlChar *val = xmlGetProp(node, (xmlChar *)name.c_str());
    if (val == 0)
        throw ConfigException("attribute " + name + " not present");

    std::string out( (const char *)val);

    return out;
}
template <>
bool ConfigParser::get_attribute(xmlNodePtr node, const std::string &name)
throw( ConfigException )
{
    xmlChar *val = xmlGetProp(node, (xmlChar *)name.c_str());
    if (val == 0)
        throw ConfigException("attribute " + name + " not present");

    std::string sval((const char*) val);
    sval = downcase(sval);
    if (sval == "true")
        return true;
    else if (sval == "false")
        return false;
    else
    {
        std::cout << "get_attribute<bool>, wrong boolan " << sval << std::endl;
        std::cout << "...in node " << node->name << " @ line " << xmlGetLineNo(node) << std::endl;
        return true;
    }
}



ConfigParser::ConfigParser() : 
    doc(0), vmedaq(0), do_validate(false)
{
    LIBXML_TEST_VERSION

    xmlKeepBlanksDefault(0);
    xmlLineNumbersDefault(1);
}

ConfigParser::~ConfigParser()
{
    clean();
}

void ConfigParser::clean()
{
    
    // We should first delete all the modules before closing
    // the loaded libraries, otherwise if the DAQmanager destructor
    // is called after  one we'll have a nice SIGSEGV
    DAQcontroller::instance()->reset_daq();

    // Restore the default run_manager and interrupt plugins
    PluginParser *pp = (PluginParser *)get_handler("plugins");
    if (pp)
        pp->reset_plugins();
    
    
    // free the handlers
    HandlerMap::iterator ip;
    for (ip = handlers.begin();ip != handlers.end();++ip)
        delete ip->second;

    handlers.clear();
    id_handler.clear();

    // free the document
    if (doc)
        xmlFreeDoc(doc);

    // Free global variables that may have been allocated by the
    // parser
    xmlCleanupParser();

    //PluginParser::reset_libs();
}

void ConfigParser::reset()
{
    clean();

    // register "default" handlers
    register_handler("paths", new PathConfig());
    register_handler("scans", new ScanConfig());
    register_handler("IO", new IOConfig());
    register_handler("general", new GeneralConfig());
    register_handler("plugins", new PluginParser());
    
}

xmlNodePtr ConfigParser::get_plugins_node()
{
    std::vector<xmlNodePtr> vplg;
    int nplg = get_node("//plugins", vplg);
    return nplg ? vplg[0] : 0;
}

xmlNodePtr ConfigParser::get_module_node()
{
    std::vector<xmlNodePtr> vplg;
    int nplg = get_node("//modules", vplg);
    return nplg ? vplg[0] : 0;
}

xmlNodePtr ConfigParser::get_paths_node()
{
    xmlNodePtr node = find_element(vmedaq, "paths");
    return node;
}

xmlNodePtr ConfigParser::get_io_node()
{
    xmlNodePtr node = find_element(vmedaq, "IO");
    return node;
}
xmlNodePtr ConfigParser::get_general_node()
{
    xmlNodePtr node = find_element(vmedaq, "general");
    return node;
}
xmlNodePtr ConfigParser::get_scans_node()
{
    xmlNodePtr node = find_element(vmedaq, "scans");
    return node;
}

xmlNodePtr ConfigParser::find_element(xmlNodePtr node, const std::string &name)
{
    xmlNodePtr child;
    if (!node)
        return 0;
    
    const char *element=name.c_str();
    for ( child=node->children; child; child=child->next )
    {
        if ( child->type != XML_ELEMENT_NODE )
            continue;
        
        if (!strcmp((char *)child->name, element))
            return child;
    }
    return 0;
}

xmlNodePtr ConfigParser::find_element(xmlDoc *node, const std::string &name)
{
    xmlNodePtr child;
    if (!node)
        return 0;
    
    const char *element=name.c_str();
    for ( child=node->children; child; child=child->next )
    {
        if ( child->type != XML_ELEMENT_NODE )
            continue;
        
        if (!strcmp((char *)child->name, element))
            return child;
    }
    return 0;
}


void ConfigParser::remove_child(xmlNodePtr node)
{
    if (node)
    {
        xmlUnlinkNode(node);
        xmlFreeNode(node);
    }
}

bool ConfigParser::remove_child(xmlNodePtr node, const std::string &child_name)
{
    if (!node)
        return false;
    
    xmlNodePtr nn, child;
    for (nn=0, child=node->children; child; child=child->next)
    {
        if ( node->type != XML_ELEMENT_NODE )
            continue;
        
        if (!strcmp((char *)child->name, child_name.c_str()))
        {
            nn=child;
            break;
        }
    }
    if (nn)
    {
        remove_child(nn);
        return true;
    }
    else
        return false;
}

void ConfigParser::navigate(xmlNodePtr node)
{
    for (;node;node = node->next)
    {
        if ( node->type != XML_ELEMENT_NODE )
            continue;

        // General, Plugins and Paths are already procesed
        if (!xmlStrcmp(node->name, (const xmlChar *) "paths"))
            continue;

        if (!xmlStrcmp(node->name, (const xmlChar *) "general"))
            continue;

        if (!xmlStrcmp(node->name, (const xmlChar *) "plugins"))
            continue;

        process_node(node);
    }
}

void ConfigParser::process_node(xmlNodePtr node)
{
    HandlerIterator ip;
    std::string name = (char *)node->name;
    ip = handlers.find( name );
    if ( ip != handlers.end() )
    {
        // if we find the handler, we do not descend the children
        try
        {
            ip->second->set_node(node);
            ip->second->parse_node(node);
            ModuleParser *mp;
            if ( (mp = dynamic_cast<ModuleParser *>(ip->second)) )
            {
                char *tmp;
                std::string mid = get_attribute<std::string>(node,"name");
                int v = strtol(mid.c_str(), &tmp, 0);
                if ( ! *tmp )
                    register_id_handler( v, mp);
                
                DAQpp::Module *module = DAQpp::DAQmanager::theDAQmanager()->find_module(mid);
                if (module)
                {
                    module->add_par("__section_name__", new DAQpp::Par<std::string>((char *)node->name));
                }
            }
        }
        catch ( ConfigException &e)
        {
            std::cout << "ConfigParser: node " << node->name
                      << " gives and error while parsing at line " 
                      << xmlGetLineNo(node)
                      << " : " << std::endl
                      << e.what()
                      << std::endl;
        }
    }
    else
        navigate(node->xmlChildrenNode);
}


void ConfigParser::register_handler(const std::string &s, SectionParser *t)
{
    handlers[s] = t;
}

void ConfigParser::register_id_handler(int s, ModuleParser *t)
{
    id_handler[s] = t;
}

bool ConfigParser::validate(const char *dir)
{
    int iok;

    if (!doc)
        return false;

    // Validates the document
    std::string schema_file =
        get_attribute<std::string>(xmlDocGetRootElement(doc),
                                   "noNamespaceSchemaLocation");
    std::string sschema =
        finder.find_schema(schema_file, dir);

    if ( !sschema.empty() )
    {
        // Validate
        xmlSchemaPtr schema;
        xmlSchemaParserCtxtPtr ctxt;
        std::cout << "Using schema: " << sschema << std::endl;
        ctxt = xmlSchemaNewParserCtxt(sschema.c_str());
        xmlSchemaSetParserErrors(ctxt,
                                 (xmlSchemaValidityErrorFunc) fprintf,
                                 (xmlSchemaValidityWarningFunc) fprintf,
                                 stderr);
        schema = xmlSchemaParse(ctxt);
        xmlSchemaFreeParserCtxt(ctxt);


        xmlSchemaValidCtxtPtr vctxt;
        vctxt = xmlSchemaNewValidCtxt(schema);
        xmlSchemaSetValidErrors(vctxt,
                                (xmlSchemaValidityErrorFunc) fprintf,
                                (xmlSchemaValidityWarningFunc) fprintf,
                                stderr);
        iok = xmlSchemaValidateDoc(vctxt, doc);
        xmlSchemaFreeValidCtxt(vctxt);
        if ( iok )
            return false;
    }
    else
    {
        std::cout << "Could not find schema: " << schema_file
        << std::endl;
        return false;
    }
    return true;
}

bool ConfigParser::create_doc()
{
    const char *include="<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n\
        <DAQVME xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n\
            xsi:noNamespaceSchemaLocation=\"daqvme.xsd\">\n\
            <plugins/><paths/><scans/><IO compress=\"true\"/><modules/><general/>\n</DAQVME>";
    
    std::cout << include << std::endl;
    doc = xmlReadMemory(include, strlen(include), "vmedaq.xml", NULL, 0);
    vmedaq = find_element(doc, "DAQVME");
    return (doc!=0);
}

bool ConfigParser::parse(const char *file)
{
    if ( !file ) {
        create_doc();
    }
    else
    {
        std::string ifile = file;
        ifile = search_file(ifile, config_path, ":");
        // Open and read the file
        doc = xmlReadFile(ifile.c_str(), NULL, 0);
        if (doc == 0)
        {
            std::cout << "ConfigParser::parse: Could not parse "
            << file
            << std::endl;
            return false;
        }
        vmedaq = find_element(doc, "DAQVME");
        if ( do_validate )
        {
            bool valid = validate(xmlParserGetDirectory(ifile.c_str()));
            if (!valid)
            {
                std::cout << "ConfigParser::parse -> error validating "
                << file
                << std::endl;
                xmlFreeDoc(doc);
                doc = 0;
                return false;
            }
        }
    }

    // Process first the paths section to group them according to
    // the type as specified in vmedaq.xsd
     
    int iplg, nplg;
    std::vector<xmlNodePtr> vplug;
    std::string lib, symbol;

    // Process the paths so that the PluginParser can find the stuff
    nplg = get_node("//paths", vplug);
    for (iplg = 0;iplg < nplg;iplg++)
        process_node( vplug[iplg] );
        
    // Process first the General section. This is so because we want
    // the VMEmanager loaded before loading the plugins
    nplg = get_node("//general", vplug);
    for (iplg = 0;iplg < nplg;iplg++)
        process_node( vplug[iplg] );

    // Process the plugins 
    nplg = get_node("//plugins", vplug);
    for (iplg = 0;iplg < nplg;iplg++)
        process_node( vplug[iplg] );
    
    // Now navigate the tree
    try
    {
        navigate(xmlDocGetRootElement(doc));
    }
    catch ( ConfigException &e )
    {
        std::cout << "**** Error parsing " << file << std::endl;
        std::cout << e.what() << std::endl;
    }
    return true;
}

int ConfigParser::save_config(const char *ofile)
{
    HandlerMap::iterator ip;
    for (ip = handlers.begin();ip != handlers.end();++ip)
    {
        try
        {
            ip->second->save_config();
        }
        catch (ConfigException &e)
        {
            std::cout << "*** ConfigParser: saving node " << ip->second->node_name()
                      << " gives and error: " << std::endl
                      << "*** " << e.what()
                      << std::endl;
        }
    }

    return xmlSaveFormatFileEnc(ofile, doc, "UTF-8", 1);
}


int ConfigParser::get_node(std::string xpath, std::vector<xmlNodePtr> &v)
{
    int i;
    xmlXPathContextPtr context;
    xmlXPathObjectPtr result;

    v.clear();

    context = xmlXPathNewContext(doc);
    result = xmlXPathEvalExpression((xmlChar *)xpath.c_str(), context);
    if (!result) 
    {
        return 0;
    }
    
    if (xmlXPathNodeSetIsEmpty(result->nodesetval))
    {
        return 0;
    }
    xmlNodeSetPtr nodeset = result->nodesetval;
    for (i = 0;i < nodeset->nodeNr;i++)
        v.push_back( nodeset->nodeTab[i] );

    xmlXPathFreeContext(context);
    return nodeset->nodeNr;

}


int ConfigParser::get_node(std::string xpath, xmlNodePtr node, std::vector<xmlNodePtr> &v)
{
    int i;
    xmlXPathContextPtr context;
    xmlXPathObjectPtr result;

    v.clear();

    context = xmlXPathNewContext(doc);
    context->node = node;
    result = xmlXPathEvalExpression((xmlChar *)xpath.c_str(), context);
    if (!result) 
    {
        return 0;
    }
    
    if (xmlXPathNodeSetIsEmpty(result->nodesetval))
    {
        return 0;
    }
    xmlNodeSetPtr nodeset = result->nodesetval;
    for (i = 0;i < nodeset->nodeNr;i++)
        v.push_back( nodeset->nodeTab[i] );

    xmlXPathFreeContext(context);
    return nodeset->nodeNr;

}


std::string ConfigParser::get_node_value(xmlNodePtr node)
{
    xmlChar *key = xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
    std::string ret = key ? (const char *)key : "";
    xmlFree(key);
    return ret;
}

int ConfigParser::get_bit_array(xmlNodePtr node,
                                const std::string &nam,
                                int nx, std::vector<int> &vv)
{
    std::string val = downcase(get_attribute<std::string>(node, nam));
    return bit_array(val, nx, vv);
}
int ConfigParser::get_bit_array(xmlNodePtr node,
                                int nx, std::vector<int> &vv)
{
    std::string val = get_node_value(node);
    return bit_array(val, nx, vv);
}

//=====================================================================
// Utilities
//=====================================================================



std::string get_attribute_value(xmlNodePtr node, const std::string &name, bool dc = true)
throw( ConfigException )
{
    xmlChar *val = xmlGetProp(node, (xmlChar *)name.c_str());
    if (!val)
        throw ConfigException("attribute " + name + " not present");

    std::string value((const char *)val);
    xmlFree(val);
    if ( dc )
        value = downcase(value);

    return value;
}
bool get_bool(xmlNodePtr node, const std::string &nam)
{
    bool result = false;
    std::string s = get_attribute_value(node, nam, true);
    if (s == "true")
        result = true;
    return (result);
}

int get_int(xmlNodePtr node, const std::string &nam)
{
    std::string val = get_attribute_value(node, nam);
    return strtol( val.c_str(), 0, 0 );
}


double get_double(xmlNodePtr node, const std::string &nam)
{
    std::string val = get_attribute_value(node, nam);
    return strtod( val.c_str(), 0 );
}


int bit_array(std::string val, int nx, std::vector<int> &vv)
{
    int i, j, k, n = val.size();
    char vc[2] = {0, 0};

    vv.clear();
    for (i = 0;i < nx;i++)
        vv.push_back(0);

    for (i = j = 0;i < n;i++)
    {
        vc[0] = tolower(val[i]);
        if ( vc[0] == ' ' || vc[0] == '\t' || vc[0] == '\n' || vc[0] == '\r' )
        {
            continue;
        }
        else if ( isdigit(vc[0]) || (vc[0] >= 'a' && vc[0] <= 'f') )
        {
            int kv = strtol(vc, 0, 16);
            for (k = 0;k < 4;k++)
            {
                if ( kv & (1 << k) )
                    vv[j+k] = 1;
            }
            j += 4;
        }
        else
        {
            std::cout << "Found funny character at pos. " << i << " : " << val[i]
            << std::endl
            << "-> " << val
            << std::endl;
            return -1;
        }
    }
    return 0;
}




struct free_modules :  public std::unary_function<Observer *,void>
{
    DAQpp::RunManager *rm;
    free_modules(DAQpp::RunManager *r) : rm(r)
    {}
    void operator()(Observer *o)
    {
        DAQpp::Module *m = dynamic_cast<DAQpp::Module *>(o);
        if (!m)
            return;

        rm->deleteModule( m );
        delete m;
    }
};


xmlNodePtr ConfigParser::find_plugin(const std::string &name)
{
    xmlNodePtr child, plugins = get_plugins_node();
    for (child=plugins->children;child;child=child->next)
    {
        if (xmlStrcmp(child->name, (const xmlChar *) "plugin"))
            continue;
        
        std::string id = get_attribute_value(child, "name");
        if ( id==name )
            return child;
    }
    return 0;
}
int ConfigParser::remove_plugin(const std::string &name)
{
    int rc = 0;
    xmlNodePtr node = find_plugin(name);
    if (node)
        remove_child(node);
    else
        rc = -1;
    
    return rc;
}


xmlNodePtr ConfigParser::find_module(const std::string &element, const std::string &id)
{
    xmlNodePtr modules = get_module_node();
    if (!modules)
        return 0;
    
    xmlNodePtr child;
    for (child=modules->children; child; child=child->next)
    {
        if ( child->type != XML_ELEMENT_NODE )
            continue;
        
        std::string name = (const char *)child->name;
        if ( element==name)
        {
            try
            {
                if ( get_attribute_value(child, "name")==id)
                    return child;
            }
            catch (ConfigException &e){}
        }
    }
    return 0;
}

int ConfigParser::remove_module(const std::string &element, const std::string &id)
{
    xmlNodePtr node =find_module(element, id);
    if (node)
    {
        remove_child(node);
        return 0;
    }
    return -1;
}
