#include <iostream>
#include <config/ConfigParser.h>
#include "GeneralConfig.h"
#include <daq/VMEmanager.h>

/// Default constructor
GeneralConfig::GeneralConfig() : 
    ped_nevts(10000), ped_update(true), 
    ped_method(normal), ped_min_sample(256),
    ped_node(0), corba_node(0), daq_node(0),
    tracer_color("white","#2d2d9d","black"),
    hist_color("white","#2d2d9d","black")
    {}

void GeneralConfig::cpy(const GeneralConfig &g)
{
    ped_nevts = g.ped_nevts;
    ped_update = g.ped_update;
    ped_method = g.ped_method;
    ped_min_sample = g.ped_min_sample;

    corba_use_ip = g.corba_use_ip;
    corba_port = g.corba_port;
    
    tracer_color = g.tracer_color;
    hist_color = g.hist_color;
}

void GeneralConfig::parse_node(xmlNodePtr _node) 
    throw (ConfigException)
{
    ConfigParser *cfg = ConfigParser::theConfig();
    std::string name;
    xmlNodePtr child;
    node = _node;
    child = node->xmlChildrenNode;

    for (; child; child = child->next)
    {
        if (child->type != XML_ELEMENT_NODE)
            continue;

        name = (char *)child->name;
        if (name == "pedestals")
        {
            decode_pedestals(child);
        }
        else if (name == "colors")
        {
            parse_colors(child);
        }
        else if (name == "corba")
        {

            corba_node = child;
            try
            {
                corba_use_ip = cfg->get_value<bool>(child, "use_ip");
            }
            catch ( ConfigException &e )
            {
                corba_use_ip = false;
            }
            try
            {
                corba_port = cfg->get_value<int>(child, "port");
            }
            catch ( ConfigException &e )
            {
                corba_port = 5901;
            }
        }
        else if (name == "daq")
        {
            daq_node = child;
            try
            {
                monitor_buffer = cfg->get_value<int>(child, "monitor_buffer");
            }
            catch ( ConfigException &e )
            {
                monitor_buffer = 150;
            }

            // now get the vme driver
            std::vector<xmlNodePtr> vplug;
            int nplg;
            if ( (nplg = cfg->get_node("//vme_driver", daq_node, vplug)))
            {
                xmlNode *node = vplug[0];
                for (; node; node = node->next)
                {
                    if (node->type != XML_ELEMENT_NODE)
                        continue;

                    try
                    {
                        vme_lib = cfg->get_value<std::string>(node, "lib");
                    }
                    catch ( ConfigException &e)
                    {
                        vme_lib = std::string();
                    }
                    try
                    {
                        vme_factory = cfg->get_value<std::string>(node, "hook");
                    }
                    catch ( ConfigException &e)
                    {
                        std::cout << "*** There is no routine for the VME factory ***"
                        << std::endl;
                        vme_factory = "<unknown>";
                    }
                    VMEmanager::set_factory(vme_factory, vme_lib);
                }
            }
        }
    }
}

void GeneralConfig::parse_colors(xmlNodePtr node)
{
    const char *elements[] = {"histograms","tracers",0};
    ConfigParser *cfg = ConfigParser::theConfig();
    xmlNodePtr child, color;
    
    
    for (int i=0;i<2;i++)
    {
        std::string bg, fg, axis;
        if ((child=cfg->find_element(node, elements[i])))
        {
            if ((color=cfg->find_element(child,"background")))
            {
                bg = cfg->get_node_value(color);
            }
            if ((color=cfg->find_element(child,"foreground")))
            {
                fg = cfg->get_node_value(color);
            }
            if ((color=cfg->find_element(child,"axis")))
            {
                axis = cfg->get_node_value(color);
            }
        }
        if (i==0)
            hist_color.set_colors(bg, fg, axis);
        else
            tracer_color.set_colors(bg, fg, axis);
    }
}

void GeneralConfig::decode_pedestals(xmlNodePtr _node)
{
    ConfigParser *cfg = ConfigParser::theConfig();
    ped_node = _node;
    try
    {
        ped_nevts = cfg->get_value<int>(_node, "nevts");
    }
    catch ( ConfigException &e )
    {
        ped_nevts = 10000;
    }

    try
    {
        ped_update = cfg->get_value<bool>(_node, "update");
    }
    catch ( ConfigException &e )
    {
        ped_update = true;
    }

    try
    {
        ped_min_sample = cfg->get_value<int>(_node, "min_sample");
    }
    catch ( ConfigException &e )
    {
        ped_min_sample = 256;
    }

    try
    {
        ped_period = cfg->get_value<int>(_node, "period");
    }
    catch ( ConfigException &e )
    {
        ped_period = 300;
    }

    std::string method;
    try
    {
        method = cfg->get_value<std::string>(_node, "method");
    }
    catch ( ConfigException &e )
    {
        method = "normal";
    }
    if (method == "normal")
    {
        ped_method = normal;
    }
    else if (method == "fast")
    {
        ped_method = fast;
    }
    else
    {
        std::cout << "Wrong method for pedestal calculation: " << method << std::endl;
        ped_method = normal;
    }
}

void GeneralConfig::save_config() throw (ConfigException)
{
    ConfigParser *cfg = ConfigParser::theConfig();
    node = cfg->get_general_node();
    if (node == 0)
        node = xmlNewChild(cfg->root(), NULL, (xmlChar * )"general", NULL);

    // Pedestals
    if (ped_node == 0)
    {
        ped_node = xmlNewChild(node, NULL, (xmlChar * )"pedestals", NULL);
    }
    cfg->check_value<int>(ped_node, "nevts", ped_nevts);
    cfg->check_value<bool>(ped_node, "update", ped_update);
    cfg->check_value<std::string>(ped_node, "method", (ped_method == fast ? "fast"
            : "normal"));
    cfg->check_value<int>(ped_node, "min_sample", ped_min_sample);
    cfg->check_value<int>(ped_node, "period", ped_period);

    // Corba
    if (corba_node == 0)
    {
        corba_node = xmlNewChild(node, NULL, (xmlChar * )"corba", NULL);
    }
    cfg->check_value<bool>(corba_node, "use_ip", corba_use_ip);
    cfg->check_value<int>(corba_node, "port", corba_port);

    // DAQ
    if (daq_node == 0)
        daq_node = xmlNewChild(node, NULL, (xmlChar * )"daq", NULL);

    cfg->check_value<int>(daq_node, "monitor_buffer", monitor_buffer);
    save_colors( cfg->find_element(node, "colors") );
}

void GeneralConfig::save_colors( xmlNodePtr colors )
{
    const char *elements[] = {"histograms","tracers",0};
    HstColor *color;
    xmlNodePtr child;
    ConfigParser *cfg = ConfigParser::theConfig();
    if (!colors)
    {
        colors = cfg->create_child<std::string>(node, "colors","");
    }
    for (int ic=0; ic<2; ic++)
    {
        color = ic ? &tracer_color : &hist_color;
        if (!(child=cfg->find_element(colors, elements[ic])))
        {
            child = cfg->create_child<std::string>(colors,elements[ic],"");
        }
        cfg->check_value<std::string>(child,"background", color->bg());
        cfg->check_value<std::string>(child,"foreground", color->fg());
        cfg->check_value<std::string>(child,"axis", color->axis());
    }
}
