#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>
#include <sstream>
#include <data/Analysis.h>
#include "ConfigParser.h"
#include "ModuleParser.h"
#include <DAQ++/DAQmanager.h>
#include <daq/GenModule.h>
#include <data/PadMap.h>


ModuleParser::~ModuleParser()
{
    if (map)
        delete map;
}

void ModuleParser::create_map(int nx, int ny, int ntot)
{
    if (map)
        delete map;

    map = new PadMap(nx, ny, ntot);
}

void ModuleParser::parse_monitor(const std::string &element, const std::string &id, Analysis *ana)
{
    std::ostringstream rstr;
    std::vector<xmlNodePtr> vr;
    int nr;
    xmlNodePtr nd;
    ConfigParser *cfg = ConfigParser::theConfig();

    rstr << element << "[@name=\"" << id << "\"]/monitor";
    nr = cfg->get_node(rstr.str(), vr);
    if (nr > 1)
    {
        std::cout << "ParseModule: more than one monitor set given to module "
        << id << std::endl
        << "Only the first one will be considered"
        << std::endl;
    }
    else if (nr)
    {
        nd = vr[0];
        monitor_node = nd;
        set_range( cfg->get_attribute<double>(nd, "min"),
                   cfg->get_attribute<double>(nd, "max")
        );
        set_nbins( cfg->get_attribute<int>(nd, "nbin") );
        if ( nbin()      != ana->get_nbins() ||
                range_min() != ana->get_range_min() ||
                range_max() != ana->get_range_max() )
        {
            ana->set_bins(nbin(), range_min(), range_max());
        }
        vr.clear();
        nr = cfg->get_node("weights", nd, vr);
        if (nr)
        {
            ana->SetWeights(
                            cfg->get_value<double>(vr[0], "pedestal"),
                            cfg->get_value<double>(vr[0], "noise")
            );
        }
    }
}
void ModuleParser::parse_map(const std::string &element, const std::string &id, Analysis *ana)
{
    std::ostringstream rstr;
    std::vector<xmlNodePtr> vr;
    int nr;
    xmlNodePtr nd;
    ConfigParser *cfg = ConfigParser::theConfig();

    int nx, ny, ntot;

    rstr << element << "[@name=\"" << id << "\"]/padmap";
    nr = cfg->get_node(rstr.str(), vr);
    if (nr > 1)
    {
        std::cout << "ParseModule: more than one padmap set given to module "
                  << id << std::endl
                  << "Only the first one will be considered"
                  << std::endl;
    }
    else if (nr)
    {
        int j, ix, iy;
        nd = vr[0];
        try
        {
            nx = cfg->get_attribute<int>(nd, "nx");
            ny = cfg->get_attribute<int>(nd, "ny");
        }
        catch ( ConfigException e)
        {
            std::cout << "padmap in " << id << " does not have nx or ny" << std::endl;
            throw;
        }
        try
        {
            ntot = cfg->get_attribute<int>(nd, "ntot");
        }
        catch (ConfigException e)
        {
            ntot = nx * ny;
        }

        create_map (nx, ny, ntot);
        nr = cfg->get_node("map", nd, vr);
        if (nr)
        {
            std::vector<int> vpad;
            std::vector<int>::iterator ip;
            cfg->get_array<int>(vr[0], vpad);
            for (j = 0, ip = vpad.begin();ip != vpad.end();++ip, j++)
            {
                if ( *ip < 0 )
                    continue;
                
                ix = j % nx;
                iy = j / nx;
                get_map()->set(*ip, ix, iy);
            }
        }
        ana->set_map( *get_map() );
        get_map()->print_unconnected();
    }

}


void ModuleParser::parse_common(const std::string &element, const std::string &id)
{
    std::ostringstream rstr;
    std::vector<xmlNodePtr> vr;
    int nr;
    xmlNodePtr nd;
    ConfigParser *cfg = ConfigParser::theConfig();

    GenModule *m = dynamic_cast<GenModule *>(
            DAQpp::DAQmanager::theDAQmanager()->find_module(id));

    if (!m)
        return;

    parse_monitor(element, id, m->get_analysis());
    parse_map(element, id, m->get_analysis());

    rstr << element << "[@name=\"" << id << "\"]";
    nr = cfg->get_node(rstr.str(), vr);
    if (nr)
    {
        nd = vr[0];
        try
        {
            pedestal_file = cfg->get_value<std::string>(nd, "pedestal_file");
            if (pedestal_file.empty())
                return;
            
            int i, sz;
            std::vector<double> &ped   = m->get_par_value< std::vector<double> > ("pedestal");
            std::vector<double> &noise = m->get_par_value< std::vector<double> > ("noise");
            std::ifstream ifile(pedestal_file.c_str());
            if (!ifile)
            {
                std::cout << "Could not load pedestals from " << pedestal_file << std::endl;
                pedestal_file = "";
            }
            else
            {
                std::cout << "Loading pedestals for module id " << id << " from " << pedestal_file << std::endl;
                sz = ped.size();
                for (i = 0;ifile && !ifile.eof() && i < sz;i++)
                {
                    ifile >> ped[i] >> noise[i];
                }
                if (i == sz)
                {
                    m->copy_pedestals();
                    m->get_analysis()->trust_pedestals();
                }
            }
        }
        catch ( ConfigException &e)
        {}
    }
}

void
ModuleParser::save_common()
{
    ConfigParser *cfg = ConfigParser::theConfig();

    if (!monitor_node)
        return;

    GenModule *m = dynamic_cast<GenModule *>(
            DAQpp::DAQmanager::theDAQmanager()->find_module(id));

    if (!m)
        return;

    cfg->check_value<std::string>(node,"pedestal_file",pedestal_file);
    
    cfg->check_attribute<int>(monitor_node, "nbin", nbins);
    cfg->check_attribute<double>(monitor_node, "min", rmin);
    cfg->check_attribute<double>(monitor_node, "max", rmax);
    xmlNodePtr child = monitor_node->xmlChildrenNode;
    for (;child;child = child->next)
    {
        if ( node->type != XML_ELEMENT_NODE )
            continue;

        std::string name = (char *)child->name;
        if (name == "weights")
        {
            cfg->check_value<double>(child, "pedestal", m->get_analysis()->get_ped_weight());
            cfg->check_value<double>(child, "noise", m->get_analysis()->get_noise_weight());
        }
    }

}
