#include <iostream>
#include <string>
#include <DAQ++/DAQmanager.h>
#include <config/ConfigParser.h>
#include <config/GeneralConfig.h>
#include "DummyModule.h"
#include "CORBAServer.h"

#if defined(HAS_DAQcorba)

void MyServer::prepare_to_stop()
{
    std::cout << "Shuting down the MsgManager" << std::endl;

    // Get the DAQmanager
    DAQpp::DAQmanager *daq_manager
    = DAQpp::DAQmanager::theDAQmanager();

    // Remove the MsgManager as an observer of Module parameters
    DAQpp::ModulesMap::const_iterator im;
    for (im = daq_manager->ModuleMap().begin();
            im != daq_manager->ModuleMap().end();
            ++im)
    {
        DAQpp::Module *m = im->second;
        DAQpp::ParameterHolder::const_iterator ip;
        for (ip = m->begin();ip != m->end();ip++)
        {
            ip->second->deleteObserver( msg->parameter );
        }
    }

    // Remove the MsgManager as an Observer of RunManager Parameters
    DAQpp::RunManagerMap::const_iterator ir;
    for (ir = daq_manager->RunMmap().begin();
            ir != daq_manager->RunMmap().end();
            ++ir)
    {
        DAQpp::RunManager *m = ir->second;
        DAQpp::ParameterHolder::const_iterator ip;
        for (ip = m->begin();ip != m->end();ip++)
        {
            ip->second->deleteObserver( msg->parameter );
        }
    }

    daq_manager->deleteObserver(msg->object);
    if (msg)
        delete msg;
}
void MyServer::setup_daq()
{
    msg = new DAQpp::CORBA::MsgManager(get_observable());

    // Get the DAQmanager
    DAQpp::DAQmanager *daq_manager
    = DAQpp::DAQmanager::theDAQmanager();

    // Add the MsgManager as an observer of the DAQ
    daq_manager->addObserver( msg->object );


    // Add the MsgManager as an observer of Module parameters
    DAQpp::ModulesMap::const_iterator im;
    for (im = daq_manager->ModuleMap().begin();
            im != daq_manager->ModuleMap().end();
            ++im)
    {
        DAQpp::Module *m = im->second;
        DAQpp::ParameterHolder::const_iterator ip;
        for (ip = m->begin();ip != m->end();ip++)
        {
            ip->second->addObserver( msg->parameter );
        }
    }

    // Add the MsgManager as an Observer of RunManager Parameters
    DAQpp::RunManagerMap::const_iterator ir;
    for (ir = daq_manager->RunMmap().begin();
            ir != daq_manager->RunMmap().end();
            ++ir)
    {
        DAQpp::RunManager *m = ir->second;
        m->GetStatus().addObserver( msg->state );
        DAQpp::ParameterHolder::const_iterator ip;
        for (ip = m->begin();ip != m->end();ip++)
        {
            ip->second->addObserver( msg->parameter );
        }
    }

    // Now create a dummy module to which we will add some parameters
    // that will enable us to configure vmedaqw remotely
    DAQpp::Module *dummy = new DummyModule("VMEDAQconfig");
    std::cout << "Created module: " << dummy->get_id() << std::endl;

}

/** Sets up the CORBA initializtion. checks in the command arguments
 *  for CORBA specific options and if some are missing according to
 *  what is specified in the configuration file, they will be added.
 *
 */
void setup_corba(int argc, char **argv, MyServer &theServer, const std::string &ior_loc)
{
    ConfigParser *cfg = ConfigParser::theConfig();
    GeneralConfig *gnrl = dynamic_cast<GeneralConfig *>(cfg->get_handler("general"));

    CORBA_Object  name_service = NULL;
    try
    {
        bool use_ip = false, socket = false;
        int i, narg = 1;
        for (i = 1;argv[i];i++)
        {
            if ( strstr(argv[i], "ORBIIOPIPSock") )
                socket = true;
            if ( strstr(argv[i], "ORBIIOPIPv4") )
                use_ip = true;

            narg++;
        }


        if ( !use_ip && gnrl->get_corba_use_ip() )
        {

            if (!socket)
                narg += 3;
            else
                narg += 2;
        }
        char **myarg = (char **)calloc( narg + 1, sizeof(char *));
        for (i = 0;i < argc;i++)
            myarg[i] = strdup(argv[i]);

        if ( narg > argc )
        {
            myarg[i++] = strdup("-ORBIIOPUNIX=0");
            myarg[i++] = strdup("-ORBIIOPIPv4=1");
            if ( !socket )
            {
                std::ostringstream ostr;
                ostr << "-ORBIIOPIPSock=" << gnrl->get_corba_port();
                myarg[i++] = strdup(ostr.str().c_str());
            }
        }
        myarg[i] = 0;

        CORBA_Environment  ev[1];
        CORBA_exception_init(ev);

        // init the server
        theServer.init(&narg, myarg, "orbit-local-mt-orb");

        // save the IORs
        theServer.save_iors_to_name_service();


    }
    catch ( CORBA_Exception e )
    {
        std::cout << "Got a CORBA_Exception while initializing the server: " << e.what() << std::endl;
    }
    catch ( ... )
    {
        std::cout << "Got an unknown exception" << std::endl;
    }

    // Save the iors in a file
    try
    {
        CORBA_Environment  ev[1];
        CORBA_exception_init(ev);

        theServer.save_iors(ior_loc);
        FILE *pf = fopen(ior_loc.c_str(), "a");
        fprintf(pf, "\n");
        fclose(pf);
        // save also the name service
        name_service =
            (CORBA_Object)CORBA_ORB_resolve_initial_references(theServer.get_orb(),
                    "NameService",
                    ev);
        if (ev->_major == CORBA_NO_EXCEPTION)
        {
            FILE *pf = fopen(ior_loc.c_str(), "a");
            CORBA_char *ostring =
                (CORBA_char *)CORBA_ORB_object_to_string (theServer.get_orb(), name_service, ev);
            if (ostring)
                fprintf(pf, "name-srv\t%s\n", ostring);
            fclose(pf);
        }
        else
            ignore_exception(ev);
    }
    catch ( CORBA_Exception e )
    {
        std::cout << "Got a CORBA_Exception: " << e.what() << std::endl;
    }
    catch ( ... )
    {
        std::cout << "Got an unknown exception" << std::endl;
    }


    // set up the DAQ
    try
    {
        theServer.setup_daq();
    }
    catch ( CORBA_Exception e )
    {
        std::cout << "Got a CORBA_Exception setting up de DAQ: " << e.what() << std::endl;
    }
    catch ( ... )
    {
        std::cout << "Got an unknown exception" << std::endl;
    }

}

#endif