#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <iostream>
#include <DAQ++/DAQmanager.h>
#include <DAQ++/GTimer.h>
#include <DAQ++/Thread.h>
#include "DAQinterrupt.h"
#include "VMEmanager.h"

#define RATE 1.e-3

static DAQpp::Semaphore sem(0);

/** 
 * \ingroup DAQ
 * Example of an interrupt catcher.
 * 
 * MyTrigger implements a DAQpp::DAQTrigger object. We make it derive 
 * from DAQpp::Thread also to debug and generate the data
 * in a different thread when there is no VME.
 * 
 * The idea here is to have a function registered in the VME system
 * as an interrupt handler that will unlock the global semaphore.
 *
 * When MyTrigger is called it will sleep until the semaphore
 * is unlocked by the VME interrupt handler.
 *
 */
class MyTrigger : public DAQpp::Thread, public DAQpp::DAQTrigger
{
    private:
        /// A timer
        DAQpp::GTimer timer;
        double trgtim;
        
        /// The semaphore that will stop the acquisition until
        /// some other process (the interrupt) unlocks it.
        DAQpp::Semaphore *sem;
    public:
        /// Constructor
        MyTrigger(DAQpp::Semaphore *s) :
            trgtim(1.), sem(s)
        {
        }
        
        /*
         * The DAQpp::Thread::run() method implementation.
         * When there are no real interrupts this will simulate
         * them
         */
        void run();
        
        /*
         * This is the boolean operator that will tell us whether
         * there is data or not. Here is where the semaphore waits
         * until it is unlocked.
         */
        bool operator()()
        {
            sem->wait();
            return true;
        }
};

void MyTrigger::run()
{
    timer.start();
    while (is_running())
    {
        if (timer() > trgtim)
        {
            timer.reset();
            trgtim = randflat(0., RATE);
            sem->post();
        }
        test_cancel();
    }
    std::cout << "Arrrrggggg !!!!" << std::endl;
    timer.stop();
}

MyTrigger trigger(&sem);

/**
 * We have to define here the hook for the VME libraries
 */
static void InterruptHandler(int sig)
{
    sem.post();
}

void EnableInterrupts()
{
    std::cout << "Enabling trigger interrupts" << std::endl;
    // Set the trigger function in the DAQ manager
    DAQpp::DAQmanager::theDAQmanager()->set_trigger(&trigger);

    // REgister the InterruptHandler
    VMEmanager *vme = VMEmanager::VME();
    vme->enable_interrupts(InterruptHandler, 0x7f);

    // if it is a fake manager, start a fake trigger
    if (!vme->connected())
        trigger.start();

}
void DisableInterrupts()
{
    std::cout << "Disabling trigger interrupts" << std::endl;
    // Set the default for the DAQmanager trigger function
    DAQpp::DAQmanager::theDAQmanager()->set_trigger(0);

    // Disable interrupts
    VMEmanager *vme = VMEmanager::VME();
    vme->disable_interrupts();

    // Stop the fake trigger
    if (!vme->connected())
        trigger.stop();

}
