#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sstream>
#include <zlib.h>
#include "GenModule.h"
#include "../data/Analysis.h"
#include "../data/DataFormat.h"

GenModule::GenModuleMap GenModule::_module_map;

GenModule::GenModule(const DAQpp::DAQid &id, unsigned long add) :
    Module(id), ana(0), _add(add), ievt(0)
{
    pedestal = new DAQpp::Par< std::vector<double> > ();
    add_par("pedestal", pedestal);

    noise = new DAQpp::Par< std::vector<double> > ();
    add_par("noise", noise);

    _module_map[id()] = this;

}

GenModule::~GenModule()
{
    GenModuleMap::iterator im = _module_map.find(get_id());
    if (im!=_module_map.end())
        _module_map.erase(im);
}

GenModule *GenModule::find_module(const DAQpp::DAQid &id)
{
    GenModuleMap::iterator im = _module_map.find(id());
    if (im!=_module_map.end())
    {
        return im->second;
    }
    return 0;
}

void GenModule::set_analysis(Analysis *A)
{
    if (ana)
        delete ana;
    
    ana = A;
}

void GenModule::set_header_description(unsigned int &data)
{
    unsigned int id = get_id();
    data = (get_n_chips() & 0xff) << 20| get_data_type() | (id & 0xffff);
}

int GenModule::get_n_chips()
{
    return ana ? ana->Nchip() : 0;
}
int GenModule::get_n_chan()
{
    return ana ? ana->Nchan() : 0;
}

void GenModule::reset_monitor_data()
{
    get_analysis()->Init();
    copy_pedestals();
}

bool GenModule::get_monitor_data(int &sz, char **ptr)
{
    int ssize;
    struct MonitorData
    {
            unsigned long len; // uncompressed length
            unsigned long clen; // length when compressed
            char data[1];
    };
    MonitorData *mdata;
    HstData *hd = dynamic_cast<HstData *>(get_analysis() );
    std::ostringstream ostr;

    ostr << *hd;
    const std::string &out = ostr.str();

    ssize = out.size();
    sz = ssize + sizeof(MonitorData);
    mdata = (MonitorData *)new char[ sz ];
    mdata->len = mdata->clen = ssize;
    int rc = compress2((Bytef *)mdata->data, &mdata->clen, (const Bytef *)out.c_str(),
                       mdata->len, 
                       Z_BEST_COMPRESSION);
    sz = mdata->clen + 2 * sizeof(unsigned int);
    *ptr = (char *)mdata;
    if (rc)
    {
        delete [] *ptr;
        *ptr = 0;
        sz = 0;
        std::cout << "Someting happened while compressing monitor data" << std::endl;
        return false;
    }
    return true;
}

void GenModule::update_pedestals()
{
    std::vector<double> &ped = pedestal->get_value();
    std::vector<double> &ns = noise->get_value();

    int i, nt = ana->Ntot();
    for (i = 0; i < nt; i++)
    {
        ped[i] = ana->Ped(0)[i];
        ns[i] = ana->Noise(0)[i];
    }
}

unsigned int *GenModule::write_config(int &size)
{
    size = 0;
    return 0;
}

unsigned int *GenModule::write_pedestals(int &size)
{
    int i, j;
    unsigned int *data;

    size = get_n_chips() * get_n_chan();
    data = new unsigned int[4+size];

    i = 0;
    data[i++] = 0xcafe0000 | PedestalRecord | (4 + size);
    data[i++] = get_id();
    data[i++] = (get_n_chips() & 0xffff) << 16 | get_n_chan();

    for (j = 0; j < size; j++)
        data[i++] = 0;

    data[i] = 4 + size;

    size = data[i];
    return data;

}

void GenModule::copy_pedestals()
{
    std::vector<double> &ped = pedestal->get_value();
    std::vector<double> &ns = noise->get_value();

    int i, nt = ana->Ntot();
    int n_pre = nt;
    for (i = 0; i < nt; i++)
    {
        ana->Ped(0)[i] = ped[i];
        ana->Noise(0)[i] = ns[i];
        if (ns[i] != 0. && ped[i] != 0.)
            n_pre--;
    }
    if (!n_pre)
        ana->trust_pedestals();
}
