/* -*- mode: c++ -*- */
#ifndef __GenModule_h__
#define __GenModule_h__
#include <memory>
#include <vector>
#include <DAQ++/Module.h>
#include <daq/ScanPoint.h>
#include <data/Analysis.h>
#include <daq/RunTypes.h>

/**
 * \brief class GenModule.
 * \ingroup DAQ
 *
 * This is a generic Module for vmedaq.
 * It adds some more functionality on top of DAQpp::Module like 
 * for instance having pedestals and noise as a parameter and a
 * number of functions to react to scan runs or to dump the data
 * to the output file.
 * 
 * It also has an Analysis object responsible for the data decoding
 * and monitoring
 */
class GenModule : public DAQpp::Module
{
    public:
        typedef std::map<std::string, GenModule *> GenModuleMap;
        
    protected:
        /** Data monitor and decoder.
         */
        Analysis *ana;

        /** Hardware address.
         */
        unsigned long _add;

        /** Event counter.
         */
        unsigned long ievt;

        /// Pedestals. 
        DAQpp::Par< std::vector<double> > *pedestal;
        
        /// Noise.
        DAQpp::Par< std::vector<double> > *noise;

        /** The type of run is stored here.
         *  This will help to react in different ways according
         *  to the RunType. It should not be like this since, ideally,
         *  the RunManager has to take care of that, but ideal worlds 
         *  do not exist in reallity...
         */
        RunType run_type;
        
        /**
         * The map of existing GenModules
         */
        static GenModuleMap _module_map;
    public:

        /**
         * Constructor.
         * @param[in] id  the DAQ++ identifier of the module
         * @param[in] add the hardware address of the module
         */
        GenModule(const DAQpp::DAQid &id, unsigned long add);
        
        /** Destructor
         */
        ~GenModule();

        static GenModule *find_module(const DAQpp::DAQid &id);
        
        /** @name Analysis
         *  The functons bellow are related with the Module's data 
         *  analysis and monitoring
         */
        //@{
        /**
         * Returns the Analysis object associated to this module
         */
        Analysis *get_analysis() const
        {
            return ana;
        }
        
        /**
         * Associates a new Analysis object to this module.
         * \param[in] A The new Analysis object
         */
        void set_analysis( Analysis *A);
        
        /**
         * Serializes the monitor data into the array dt
         * \param[out] sz size of data block
         * \param[out] dt the data block.
         */
        virtual bool get_monitor_data(int &sz, char **dt);
        
        /**
         * Resets to the default values the monitor data of the module
         */
        virtual void reset_monitor_data();
        
        /**
         * REturns the number of chips in the module.
         * In vmedaq the channels of a module are grouped in chips, which have
         * a fixed number of channels each.
         */
        virtual int get_n_chips();
        
        /**
         * Returns the number of channels in a chip of the module.
         * See get_n_chips()
         */
        virtual int get_n_chan();

        //@}


        /** @name DAQ++ addons
         * The methods bellow are used in the DAQ process and are
         * additions to the DAQpp::Module functionality.
         */
        //@{
        /**
         * Called at the end of each event.
         * This is a method added on top of DAQpp::Module
         */
        virtual void EndOfEvent()
        {
        }
        
        /** 
         * Returns the hardware Address
         */
        unsigned long get_address() const
        {
            return _add;
        }

        /**
         * Resets the event counter.
         * This allows each Module to handle their own event
         * counter.
         */
        void reset_evt_counter()
        {
            ievt = 0;
        }
        /**
         * Increase the event counter by one.
         */
        void increase_evt_counter()
        {
            ievt++;
            ievt &= 0xffffff;
        }
        
        /**
         * Returns the event counter.
         */
        unsigned long get_evt_counter()
        {
            return ievt;
        }

        /**
         * Sets the run type.
         * 
         * This is called from Rmanager::PrepareForRun so that
         * the module can take the appropriate actions depending
         * on the run type.
         */
        void set_run_type(RunType x)
        {
            run_type = x;
        }
        
        /**
         * Sets the module parameters corresponding to the given ScanPoint.
         * 
         * When running a ScanRunManager, vmedaq will scan over different values
         * of a number of parameters. Every time we change the scan point, that is,
         * any of the variables chages, the ScanRunManager will notify the modules
         * through this method.
         */
        virtual void set_scan_point(ScanPoint &)
        {
        }
        //@}
        
        /** @name Pedestals and Noise
         */
        //@{
        
        /** Returns the vector of pedestals.
         */
        std::vector<double> get_pedestal() const
        {
            return pedestal->get_value();
        }
        
        /** Returns the vector of noise
         */
        std::vector<double> get_noise() const
        {
            return noise->get_value();
        }

        /** Update the module pedestals and noise.
         *  This method copies the pedestals stored in the Analysis
         *  in the local vectors of the module.
         */
        void update_pedestals();

        /** Copies the module's pedestals and files to the 
         *  module's Analysis object.
         */
        void copy_pedestals();
        //@}
        
        
        /** @name File IO
         * 
         *  The methods below are used when the RManager has a data logger
         *  associated and are meant to provide the module specific information
         *  in the different data blocks.
         */
        //@{
        
        /**
         * Returns the data type.
         * 
         * This specifies the data type that will be stored in the DataRecord
         * blocks of the data file.
         * 
         */
        virtual unsigned long get_data_type() const = 0;
                
        /**
         * The module should override this method to change the default
         * description word in the StartOfRun record. The default is
         *       0xn000iiii  n=number of chips iiii=module id
         */
        virtual void set_header_description(unsigned int &data);

        /**
         * The module should override this to change the format of the
         * Pedestal record.
         * 
         * \param[out] size  it will contain the size of the returned array
         * \return An array of unsigned int. The module should allocate this
         *         array and the RunManager will delete it.  
         *
         */
        virtual unsigned int *write_pedestals(int &size);

        /**
         * Dumps the module configuration.
         * The module should override this to dump the configuration
         * in the ConfigRecord.
         * 
         * \param[out] size  it will contain the size of the returned array
         * \return An array of unsigned int. The module should allocate this
         *         array and the RunManager will delete it.  
         */
        virtual unsigned int *write_config(int &size);
        //@}
};

#endif
