#ifndef __Rmanager_h__
#define __Rmanager_h__
#include <iostream>
#include <sstream>
#include <memory>
#include <DAQ++/RunManager.h>
#include <DAQ++/FileIOManager.h>
#include <daq/RunTypes.h>
#include <daq/MyMonitor.h>

/**
 *  \brief class Rmanager.
 *  \ingroup DAQ++
 *
 *  This is the base class of the different RunManager objects.
 *  Each RunManager defines a set of settings and a type of
 *  acquisition.
 *
 */
class Rmanager : public DAQpp::RunManager
{
    protected:
        /// Pointer to the data logger for this run
        DAQpp::FileIOManager *logger;

        std::ostringstream obuf;
        
        /// The type of run
        RunType run_type;
    public:
        /// constructor
        Rmanager(DAQpp::DAQid id, DAQpp::RunCommand *rc = 0) ;
        /// destructor
        ~Rmanager();


        /*
         * DAQpp::RunManager methods
         */
        void PrepareForRun();
        void Start();
        void Stop();
        void EndOfEvent();
        
        /// Returns the type of run
        RunType get_run_type() const { return run_type; }

        // Creates a new "default" run manager from the "fixed" list
        static Rmanager *new_manager(const std::string &s, const DAQpp::DAQid &);

        // Creates a new run manager according to the config file
        static Rmanager *create_run_manager(const std::string &s, const DAQpp::DAQid &name);
        
        // Creates a new Logger
        DAQpp::FileIOManager *create_logger(bool, int);

        // removes all the modules from this RunManager
        std::vector<DAQpp::Module *>  deplete_run_manager();
        
        /**---------------------------------------------------------------------
         * IO stuff
         *---------------------------------------------------------------------
         */
        /**
         * Writes a header at the beginning of a file.
         * This is used when data is being logged. The data logge inherits from
         * DAQpp::FileIOManager and this will call this function every time a 
         * file is opened. 
         * 
         * Pedestal, Map, Config and extra records are writen here.
         */
        static int writeHeader(const char **data, Rmanager *);
        
        /**
         * Writes a block of data at the end of the file. 
         * It will produce the EndOfRun record
         */
        static int writeTail(const char** data, Rmanager *);

      
        /**
         * Starts the data logger.
         */
        void  log_data(const char *file, bool cmp, int sz);
        /**
         * Stops the data logger.
         */
        void  stop_logging();
        
        /**
         * Returns ta pointer to the data logger
         */
        DAQpp::FileIOManager *get_logger()
        {
            return logger;
        }
        
        /** Writes the data, if any, at the end of an event.
         */
        virtual int   writeData(int size, const char *data);
        
        /** Writes the Pedestals in the PedestalRecord at the beginning
         *  of each file. It is called from the writeHeader method.
         */
        virtual void  writePedestals(std::ostream &);
        
        /** Writes the MapRecord at the beginning of each file. It is
         *  called from the writeHeader method.
         */
        virtual void  writeMaps(std::ostream &);
        
        /** Writes any extra information at the beginning of each file. 
         *  It is called from the writeHeader method. 
         * 
         *  This method is implemented in RunManager objects like the 
         *  ScanRunManager, where it dumps the description of the scan. 
         */
        virtual void  writeHeaderXtra(std::ostream &)
        {}

        /** Writes thes the ConfigRecord.
         *  It loops on the associated modules and calls their write_config
         *  method.
         */
        virtual void writeConfig(std::ostream &);
        
    public:
        //---------------------------------------------------------------------
        // Monitoring
        //---------------------------------------------------------------------
        void start_monitor()
        {
            if ( get_monitor() )
                get_monitor()->start_monitor();
        }
    private:
        static void *_monitor(void *ptr);
        void monitor();
};

#endif
