#include <iostream>
#include <cstdlib>
#include <cstdio>
#include <cerrno>
#include <csignal>
#include <cstring>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if defined(HAS_VME) && defined(HAS_SIS1100)
extern "C"
{
#include "dev/pci/sis1100_var.h"
#include "sis3100_vme_calls.h"
}

#include <daq/VMEmanager.h>

static void __attribute__ ((constructor)) __init__lib__()
{
    printf("### Loading SIS1100 module...\n");
}

static void __attribute__ ((destructor)) __close__lib__()
{
    printf("### Unloading SIS1100 module...\n");
}

class SISmanager : public VMEmanager
{
protected:
    SISmanager(const SISmanager &);
    void operator=(const SISmanager &)
    {}

    // The file descriptor
    int fd;
public:
    SISmanager();
    ~SISmanager();

    bool connected() const
    {
        return true;
    }

    unsigned long readA32D32(unsigned long add );
    void writeA32D32(unsigned long add, unsigned long val);

    unsigned long readA32BLT32(unsigned long add, u_int32_t *dat, unsigned long nwords);
    unsigned long readA322EVME(unsigned long add, u_int32_t *dat, unsigned long nwords);
    unsigned char read8(unsigned long add );
    void write8(unsigned long add, unsigned char val);


    unsigned long read32(unsigned long add );
    void write32(unsigned long add, unsigned long val);
    unsigned short read16(unsigned long add );
    void write16(unsigned long add, unsigned short val);
    void enable_interrupts(VMEmanager::InterruptFunc, unsigned short);
    void disable_interrupts();

    void acknowledge(VMEmanager::IRQinfo *);
    void Reset();

};

//#define MTX_ACQ() mtx.acquire()
//#define MTX_REL() mtx.release()
#define MTX_ACQ()
#define MTX_REL()

SISmanager::SISmanager()
        : fd(0)
{
    fd = open( SIS_DEVICE, O_RDWR, 0);
    if ( fd < 0 )
    {
        std::cout << "Could not initialize VME" << std::endl;
        perror("SISmanager::SISmanager - ");
        //exit(1);
    }
}
SISmanager::~SISmanager()
{
    std::cout << "Closing VME" << std::endl;
    close(fd);
}

void SISmanager::enable_interrupts(InterruptFunc func, unsigned short irqmask)
{
    struct sigaction action;
    struct sis1100_irq_ctl irqctl;


    action.sa_handler = func;
    sigemptyset(&action.sa_mask);
    sigaddset(&action.sa_mask, SIGUSR1);
    action.sa_flags = 0;
    sigaction(SIGUSR1, &action, 0);


    MTX_ACQ();
    irqctl.irq_mask = irqmask;
    irqctl.signal = SIGUSR1;
    if (ioctl(fd, SIS1100_IRQ_CTL, &irqctl) < 0)
    {
        std::cout << "ioctl(SIS1100_IRQ_CTL): " << strerror(errno) << std::endl;
    }
    MTX_REL();
    return;
}
void SISmanager::disable_interrupts()
{
    struct sigaction action;
    action.sa_handler = SIG_IGN;
    sigemptyset(&action.sa_mask);
    action.sa_flags = 0;
    sigaction(SIGUSR1, &action, 0);

    struct sis1100_irq_ctl irqctl;
    MTX_ACQ();
    irqctl.irq_mask = 0xffffffff;
    irqctl.signal = 0;
    if (ioctl(fd, SIS1100_IRQ_CTL, &irqctl) < 0)
    {
        std::cout << "ioctl(SIS1100_IRQ_CTL): " << strerror(errno) << std::endl;
    }
    MTX_REL();
}


void SISmanager::acknowledge(IRQinfo *irq)
{
    struct sis1100_irq_get irqget;
    struct sis1100_irq_ack irqack;

    memset(&irqget, 0, sizeof(irqget));
    irqget.irq_mask = 0xffffffff;

    if (ioctl(fd, SIS1100_IRQ_GET, &irqget) < 0)
    {
        std::cout << "ioctl(SIS1100_IRQ_GET): " << strerror(errno) << std::endl;
        return;
    }
    if (irq)
    {
        irq->irqs = irqget.irqs;
        irq->level = irqget.level;
        irq->vector = irqget.vector;
    }
//    printf("irqget.irqs (doorbel)          : 0x%08x\n", irqget.irqs);
//    printf("irqget.level (VME IRQ Level)   : 0x%08x\n", irqget.level);
//    printf("irqget.vector (VME IRQ Vector) : 0x%08x\n", irqget.vector);

    irqack.irq_mask = irqget.irqs;
    if (ioctl(fd, SIS1100_IRQ_ACK, &irqack) < 0)
    {
        std::cout << "ioctl(SIS1100_IRQ_ACK): " <<  strerror(errno) << std::endl;
        return;
    }

    return;
}

void SISmanager::Reset()
{
    MTX_ACQ();
    vmesysreset(fd);
    MTX_REL();
}

unsigned long SISmanager::readA32D32(unsigned long add )
{
    u_int32_t  dat;
    MTX_ACQ();
    int rc = vme_A32D32_read(fd, add, &dat ) ;
    check_rc(rc);
    MTX_REL();
    return dat;
}

void SISmanager::writeA32D32(unsigned long add, unsigned long val)
{
    int rc;
    MTX_ACQ();
    rc = vme_A32D32_write(fd, add, val );
    check_rc(rc);
    MTX_REL();
}

unsigned long SISmanager::readA32BLT32(unsigned long add, u_int32_t *dat,
                                           unsigned long nwords)
{
    u_int32_t gwords;
    MTX_ACQ();
    int rc = vme_A32BLT32_read(fd, add, dat, nwords, &gwords) ;
    check_rc(rc);
    MTX_REL();
    return (unsigned long) gwords;
}

unsigned long SISmanager::readA322EVME(unsigned long add, u_int32_t *dat,
                                           unsigned long nwords)
{
    u_int32_t gwords;
    MTX_ACQ();
    int rc = vme_A32_2EVME_read(fd, add, dat, nwords, &gwords) ;
    check_rc(rc);
    MTX_REL();
    return (unsigned long) gwords;
}

unsigned long SISmanager::read32(unsigned long add )
{
    u_int32_t  dat;
    int rc;
    MTX_ACQ();
    rc = vme_A24D32_read(fd, add, &dat ) ;
    check_rc(rc);
    MTX_REL();
    return dat;
}

void SISmanager::write32(unsigned long add, unsigned long val)
{
    int rc;
    MTX_ACQ();
    rc = vme_A24D32_write(fd, add, val );
    check_rc(rc);
    MTX_REL();
}

unsigned short SISmanager::read16(unsigned long add )
{
    unsigned short dat;
    int rc;
    MTX_ACQ();
    rc = vme_A24D16_read(fd, add, &dat ) ;
    check_rc(rc);
    MTX_REL();
    return dat;
}

void SISmanager::write16(unsigned long add, unsigned short val)
{
    int rc;
    MTX_ACQ();
    rc = vme_A24D16_write(fd, add, val );
    check_rc(rc);
    MTX_REL();
}
unsigned char SISmanager::read8(unsigned long add )
{
    unsigned char dat;
    int rc;
    MTX_ACQ();
    rc = vme_A24D8_read(fd, add, &dat ) ;
    check_rc(rc);
    MTX_REL();
    return dat;

}
void SISmanager::write8(unsigned long add, unsigned char val)
{
    int rc;
    MTX_ACQ();
    rc = vme_A24D8_write(fd, add, val );
    check_rc(rc);
    MTX_REL();
}

extern "C"
{
    VMEmanager *create_sis_manager()
    {
        static SISmanager sis;
        return &sis;
    }
}
#endif
