#include <string>
#include <DAQ++/GTimer.h>
#include "ScanRunManager.h"
#include "../data/DataFormat.h"
#include "../data/Analysis.h"
#include "GenModule.h"

bool EventTimer::operator()()
{
    if (!has_started())
    {
        start();
        return true;
    }
    else
    {
        double t = timer();
        return t> mxtime;
    }
}

/**
 *  Special classes to handle special parameters
 */
class ParScanList : public DAQpp::Par<ScanList>
{
    public:
        ParScanList()
        {
        }

        ParScanList(const ScanList &sl) :
            DAQpp::Par<ScanList>(sl)
            {}

        void set_from_string(const std::string &s, bool notify = false);
        std::string get_as_string();
};

void ParScanList::set_from_string(const std::string &s, bool notify)
{
    std::istringstream is(s);
    get_value().load(is);
    if (notify)
    {
        Arg<ScanList> arg(get_value());
        notifyObservers(&arg);
    }
}

std::string ParScanList::get_as_string()
{
    std::ostringstream os;
    get_value().dump(os);
    return os.str();
}

class ParScanPoint : public DAQpp::Par<ScanPoint>
{
    public:
        ParScanPoint()
        {
        }
        ParScanPoint(const ScanPoint &sp) :
            DAQpp::Par<ScanPoint>(sp)
            {}

        void set_from_string(const std::string &s, bool notify = false);
        std::string get_as_string();
};

void ParScanPoint::set_from_string(const std::string &s, bool notify)
{
    std::istringstream is(s);
    get_value().load(is);
    if (notify)
    {
        Arg<ScanPoint> arg(get_value());
        notifyObservers(&arg);
    }
}

std::string ParScanPoint::get_as_string()
{
    std::ostringstream os;
    get_value().dump(os);
    return os.str();
}

//---------------------------------------------------------------------
//
//---------------------------------------------------------------------
struct reset_state : public std::unary_function<Observer *, void>
{
    reset_state()
    {
    }
    void operator()(Observer *o)
    {
        GenModule *m = dynamic_cast<GenModule *>(o);
        if (m)
        {

	  DAQpp::RunState &st = m->get_status();
            st.reset();
            st.start();
        }
    }
};


struct scan_send_trigger : public std::unary_function<Observer *, void>
{
    scan_send_trigger()
    {
    }
    void operator()(Observer *o)
    {
        GenModule *m = dynamic_cast<GenModule *>(o);
        if (m)
        {
            m->Trigger();
        }
    }
};

struct scan_reset : public std::unary_function<Observer *, void>
{
    scan_reset()
    {
    }
    void operator()(Observer *o)
    {
        GenModule *m = dynamic_cast<GenModule *>(o);
        if (m)
        {
            m->Reset();
        }
    }
};


struct change_value : public std::unary_function<Observer *, void>
{
    change_value(ScanPoint &scp) :
        point(scp)
        {
        }
    void operator()(Observer *o)
    {
        GenModule *m = dynamic_cast<GenModule *>(o);
        if (m)
            m->set_scan_point(point);
    }
    ScanPoint point;
};

//---------------------------------------------------------------------
// ScanRunManager
//---------------------------------------------------------------------
ScanRunManager::ScanRunManager(DAQpp::DAQid id) :
    PulseRunManager(id), _time(false), ipoint(-1), next_point(0)
{
    run_type = Scan;
}

void ScanRunManager::set_scan_list(int n, const std::vector<ScanVar> &v)
{
    ipoint = -1;
    vvar = v;
    nevt = n;
    ScanPoint sc(vvar);
    ScanPoint::FillList(sc.begin(), sc.end(), sc, sclist);
    current_point = sclist.begin();
    set_max_events(0);

    add_par("scanpoint", new ParScanPoint( *current_point ));
}

void ScanRunManager::set_new_values()
{
    set_par<ScanPoint>("scanpoint", *current_point, true);

    for_each_module(change_value(*current_point) );

    // If we have
    std::vector<ScanVar>::iterator ip = current_point->find_type(ScanVar::pulse);
    if (ip != current_point->end())
        set_pulse(ip->value() );
    else
        set_pulse(0.25);

    // We send the burst from here, since this is also called from
    // Prepare for run, and this is important if we run with
    // interrupts.
    if (burst())
    {
        send_pulse();
    }
}

void ScanRunManager::PrepareForRun()
{
    PulseRunManager::PrepareForRun();
    
    // Set variables to the values in the first point
    current_point = sclist.begin();
    if (next_point)
        delete next_point;

    if (_time)
    {
        next_point = new EventTimer(nevt);
    }
    else
    {
        next_point = new EventCntr(nevt);
    }
    next_point->reset();

    event_number = 0;

    // Prepares the pulse
    setup_pulse(get_pulse(), 2*nevt_per_point());

    // reset the scan point counter
    ipoint = 0;

    // call the user hook
    init_run();
}
void ScanRunManager::EndOfRun()
{
    PulseRunManager::EndOfRun();
}

void ScanRunManager::Trigger()
{
    // Check if we need to change to a new point
    if (!next_point)
        return;

    if ( (*next_point)() )
    {
        if (ipoint == 0)
            current_point = sclist.begin();
        else
            current_point++;

        ipoint++;
        set_output(false);
        if (current_point == sclist.end() )
            return;

        next_point->reset();
        write_header();

        // Wait until the monitor queue is empty
        if (get_monitor() )
        {
            DAQpp::GTimer tim;
            tim.start();
            while ( !get_monitor()->empty() )
            {
                if (tim() > 5.)
                    break;
            }
        }

        // Clear RunState of Modules
        for_each_module(reset_state() );

        set_output(true);

        // Set the new values
        set_new_values();

        // call the user hook
        new_point();
    }

    // Trigger the modules
    next_point->next();
    
    PulseRunManager::Trigger();

    // Send the pulse
    send_pulse();

    // call the user hook
    new_trigger();
}

int ScanRunManager::writeData(int size, const char *data)
{
    unsigned int *dt =const_cast<unsigned int *>((const unsigned int *)data);

    dt[0] |= DataRecord;
    dt[1] = (dt[1] & 0xff000000) | (get_event_number() % 0xffffff);
    return Rmanager::writeData(size, data);
}

void ScanRunManager::write_header()
{
    int i = 0, size = 3 + current_point->nvar();
    unsigned int *data = new unsigned int[ size ];
    data[i++] = DataHeader | NewPointRecord | size;
    data[i++] = (nevt << 8) | current_point->nvar();

    std::vector<ScanVar>::iterator ip;
    for (ip = current_point->begin(); ip != current_point->end(); ++ip)
    {
        data[i] = ip->type() << 28;
        switch (ip->type() )
        {
            case ScanVar::channel:
            case ScanVar::threshold:
            case ScanVar::trim:
	   case ScanVar::thrs_trim:
	   case ScanVar::cal_dac:
                data[i] |= int(ip->value());
                break;
            case ScanVar::pulse:
            case ScanVar::xposition:
            case ScanVar::yposition:
            case ScanVar::zposition:
            case ScanVar::angle1:
            case ScanVar::angle2:
            case ScanVar::angle3:
                data[i] |= ( static_cast<int>( 100. * ip->value() ) & 0xfffffff);
                break;
            default:
                break;
        }
        i++;
    }
    data[i] = size;
    obuf.write((const char *)data, size*sizeof(unsigned int));
    delete [] data;
}

void ScanRunManager::writeHeaderXtra(std::ostream &oheader)
{
    /**
     *  The format of the ScanRecord is:
     *  Word
     *  0     0xCAFEibbb   i= event type bbbb= size of data block
     *  1     0xaaaaaaaa   number of variables
     *  2     0xaaaaaaaa   number of events per scan point
     *  3     0xiaaabbbb   i=value format aaaa=var type bbbb=number of points
     *  4     0xbbbbbbbb   first value
     *  5     0xbbbbbbbb   last value
     *  .     .
     *  bbbb  bbbb         size of data block
     *  
     *  Words 3-5 appear once per variable.
     *  The value format indicates how the values are stored. If it is
     *  one they are stored as float numbers, otherwise as integers.
     *  
     */
    unsigned int val;
    unsigned int iword;
    int size = 4 + 3 * vvar.size();
    union
    {
        unsigned int lval;
        float fval;
    } ifirst, ilast;

    iword = 1;
    val = DataHeader | ScanRecord | size;
    oheader.write((char *)&val, sizeof(unsigned int));

    iword ++;
    val = vvar.size();
    oheader.write((char *)&val, sizeof(unsigned int));

    iword++;
    val = nevt;
    oheader.write((char *)&val, sizeof(unsigned int));

    std::vector<ScanVar>::iterator ip;
    for (ip = vvar.begin(); ip != vvar.end(); ++ip)
    {
        ScanVar &v = *ip;
        iword++;
        val = 0;
        switch (v.type() )
        {
            case ScanVar::pulse:
            case ScanVar::xposition:
            case ScanVar::yposition:
            case ScanVar::zposition:
            case ScanVar::angle1:
            case ScanVar::angle2:
            case ScanVar::angle3:
                val |= 0x80000000;
                ifirst.fval = v.first();
                ilast.fval = v.last();
                break;
            default:
                ifirst.lval = ( unsigned long )(v.first() );
                ilast.lval = ( unsigned long )(v.last() );
        }

        val |= (v.type() & 0xfff) << 16| (v.n_points()) & 0xffff;
        oheader.write((char *)&val, sizeof(unsigned int));

        iword++;
        oheader.write((char *)&ifirst, sizeof(unsigned int));

        iword++;
        oheader.write((char *)&ilast, sizeof(unsigned int));
    }
    iword++;
    oheader.write((char *)&iword, sizeof(unsigned int));
}
