/* -*- mode: c++ -*- */
#include <iostream>
#include <cstdlib>
#include <cstdio>
#include <cerrno>
#include <csignal>
#include <cstring>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if defined(HAS_VME) && defined(HAS_VMUSB)
extern "C" {
#include "libxxusb.h"
}
#include <daq/VMEmanager.h>

static void __attribute__ ((constructor)) __init__lib__()
{
  printf("### Loading VMUSB module...\n");
}

static void __attribute__ ((destructor)) __close__lib__()
{
  printf("### Unloading VMUSB module...\n");
}

class VMUSBmanager : public VMEmanager
{
public:
  VMUSBmanager();
  ~VMUSBmanager();

  void acknowledge(VMEmanager::IRQinfo *);
  void Reset();
  void enable_interrupts(VMEmanager::InterruptFunc, unsigned short);
  void disable_interrupts();

  bool connected() const { return true; }

  unsigned char read8(unsigned long add );
  void write8(unsigned long add, unsigned char val);

  unsigned short read16(unsigned long add );
  void write16(unsigned long add, unsigned short val);

  unsigned long read32(unsigned long add );
  void write32(unsigned long add, unsigned long val);

  unsigned long readA32D32(unsigned long add );
  void writeA32D32(unsigned long add, unsigned long val);

  unsigned long readA32BLT32(unsigned long add, u_int32_t *dat, unsigned long nwords);
  unsigned long readA322EVME(unsigned long add, u_int32_t *dat, unsigned long nwords);

  // stack commands
  int stack_execute(std::vector<VMEcommand> &stack, std::vector<unsigned int> &data);

  int readA24_fifo32(unsigned long add, const int nwords, unsigned int *dat);

protected:
  VMUSBmanager(const VMUSBmanager &);
  void operator=(const VMUSBmanager &)
  { }

private:
  usb_dev_handle *_udev; // Device Handle

  static const int _buffer_size = 26700; // size of buffer in bytes
  static const int _max_stack_size = 0x200001; // max stack size in 16 bit words
  static const int _max_read_size = 26616; // max safe read size in bytes

  unsigned short _stack[_max_stack_size];
  std::vector<int> _data_size;
  int _stack_size;
  int _read_size;
  bool _last_write;

  usb_dev_handle* VMUSB_device_open();
  void VMUSB_reset();
  int VMUSB_buffer_clear();

  short VMUSB_MBLT_read_32(usb_dev_handle *hdev,
                           short Address_Modifier,
                           int count,
                           unsigned long VME_Address,
                           u_int32_t *dat);

  int VMUSB_stack_execute(std::vector<unsigned int> &data);

  void stack_clear();

  void stack_add_read8(unsigned long add);
  void stack_add_write8(unsigned long add, unsigned short data);

  void stack_add_read16(unsigned long add);
  void stack_add_write16(unsigned long add, unsigned short data);

  void stack_add_read32(unsigned long add);
  void stack_add_write32(unsigned long add, unsigned int data);

  void stack_add_readA32D32(unsigned long add);
  void stack_add_writeA32D32(unsigned long add, unsigned int data);

  void MTX_ACQ() { mtx.acquire(); }
  void MTX_REL() { mtx.release(); }

//  void MTX_ACQ() { std::cout<<"acquiring mutex..."<<std::endl; mtx.acquire(); std::cout<<"mutex acquired!"<<std::endl; }
//  void MTX_REL() { std::cout<<"releasing mutex..."<<std::endl; mtx.release(); std::cout<<"mutex released!"<<std::endl; }

//  void MTX_ACQ() { }
//  void MTX_REL() { }
};


VMUSBmanager::VMUSBmanager()
  : _udev(0)
{
  _udev = VMUSB_device_open();  // Open Device and get handle

  // Make sure VMUSB opened OK
  if(!_udev)
  {
    std::cout << "  Failed to open a VMUSB device\n\n"
              << "Could not initialize VME... Exiting!\n" << std::endl;
    exit(1);
  }

  _data_size.reserve(_max_read_size);
  stack_clear();
}

VMUSBmanager::~VMUSBmanager()
{
  std::cout << "Closing VME" << std::endl;
  MTX_ACQ();
  xxusb_device_close(_udev);
  MTX_REL();
}

void
VMUSBmanager::acknowledge(IRQinfo *irq)
{
  /*
  struct sis1100_irq_get irqget;
  struct sis1100_irq_ack irqack;

  memset(&irqget, 0, sizeof(irqget));
  irqget.irq_mask = 0xffffffff;

  if (ioctl(fd, SIS1100_IRQ_GET, &irqget) < 0)
  {
    std::cout << "ioctl(SIS1100_IRQ_GET): " << strerror(errno) << std::endl;
    return;
  }
  if (irq)
  {
    irq->irqs = irqget.irqs;
    irq->level = irqget.level;
    irq->vector = irqget.vector;
  }
//  printf("irqget.irqs (doorbel)          : 0x%08x\n", irqget.irqs);
//  printf("irqget.level (VME IRQ Level)   : 0x%08x\n", irqget.level);
//  printf("irqget.vector (VME IRQ Vector) : 0x%08x\n", irqget.vector);

  irqack.irq_mask = irqget.irqs;
  if (ioctl(fd, SIS1100_IRQ_ACK, &irqack) < 0)
  {
    std::cout << "ioctl(SIS1100_IRQ_ACK): " <<  strerror(errno) << std::endl;
    return;
  }

  return;
  */
}

void
VMUSBmanager::Reset()
{
  VMUSB_buffer_clear();
  VMUSB_reset();
}

void
VMUSBmanager::enable_interrupts(InterruptFunc func, unsigned short irqmask)
{
  /*
  struct sigaction action;
  struct sis1100_irq_ctl irqctl;


  action.sa_handler = func;
  sigemptyset(&action.sa_mask);
  sigaddset(&action.sa_mask, SIGUSR1);
  action.sa_flags = 0;
  sigaction(SIGUSR1, &action, 0);


  MTX_ACQ();
  irqctl.irq_mask = irqmask;
  irqctl.signal = SIGUSR1;
  if (ioctl(fd, SIS1100_IRQ_CTL, &irqctl) < 0)
  {
    std::cout << "ioctl(SIS1100_IRQ_CTL): " << strerror(errno) << std::endl;
  }
  MTX_REL();
  return;
  */
}

void
VMUSBmanager::disable_interrupts()
{
  /*
  struct sigaction action;
  action.sa_handler = SIG_IGN;
  sigemptyset(&action.sa_mask);
  action.sa_flags = 0;
  sigaction(SIGUSR1, &action, 0);

  struct sis1100_irq_ctl irqctl;
  MTX_ACQ();
  irqctl.irq_mask = 0xffffffff;
  irqctl.signal = 0;
  if (ioctl(fd, SIS1100_IRQ_CTL, &irqctl) < 0)
  {
    std::cout << "ioctl(SIS1100_IRQ_CTL): " << strerror(errno) << std::endl;
  }
  MTX_REL();
  */
}

unsigned char
VMUSBmanager::read8(unsigned long add )
{
  // *** Warning, not tested ***
  return (unsigned char) read16(add);
}
void
VMUSBmanager::write8(unsigned long add, unsigned char val)
{
  // *** Warning, not tested ***
  write16(add, (unsigned short) val);
}

unsigned short
VMUSBmanager::read16(unsigned long add )
{
  long dat;
  MTX_ACQ();
  int rc = VME_read_16(_udev, 0x39, add, &dat);
  if (rc==2) check_rc(0);
  else check_rc(rc);
  MTX_REL();
  return (unsigned short) dat;
}
void
VMUSBmanager::write16(unsigned long add, unsigned short val)
{
  MTX_ACQ();
  int rc = VME_write_16(_udev, 0x39, add, (long) val);
  if (rc==2) check_rc(0);
  else check_rc(rc);
  MTX_REL();
}

unsigned long
VMUSBmanager::read32(unsigned long add )
{
  long dat;
  MTX_ACQ();
  int rc = VME_read_32(_udev, 0x39, add, &dat);
  if (rc==4) check_rc(0);
  else check_rc(rc);
  MTX_REL();
  return (unsigned long) dat;
}
void
VMUSBmanager::write32(unsigned long add, unsigned long val)
{
  MTX_ACQ();
  int rc = VME_write_32(_udev, 0x39, add, (long) val);
  if (rc==2) check_rc(0);
  else check_rc(rc);
  MTX_REL();
}

unsigned long
VMUSBmanager::readA32D32(unsigned long add )
{
  long dat;
  MTX_ACQ();
  int rc = VME_read_32(_udev, 0x9, add, &dat);
  if (rc==4) check_rc(0);
  else check_rc(rc);
  MTX_REL();
  return (unsigned long) dat;
}
void
VMUSBmanager::writeA32D32(unsigned long add, unsigned long val)
{
  MTX_ACQ();
  int rc = VME_write_32(_udev, 0x9, add, (long) val);
  if (rc==2) check_rc(0);
  else check_rc(rc);
  MTX_REL();
}

unsigned long
VMUSBmanager::readA32BLT32(unsigned long add, u_int32_t *dat, unsigned long nwords)
{
  int block_size = 64;
  int nblocks = nwords/block_size;
  int remain = nwords%block_size;
  int gbytes=0;

  MTX_ACQ();
  if (nblocks) {
    int rc = VMUSB_MBLT_read_32(_udev, 0x0f, nblocks, add, dat);
    if ( rc<0 )
      check_rc(rc);
    else {
      gbytes+=rc;
      check_rc(0);
    }
  }
  if (remain) {
    int rc = VME_BLT_read_32(_udev, 0x0f, remain,
                             add+nblocks*block_size*sizeof(unsigned long),
                             (long*) dat+nblocks*block_size);
    if ( rc<0 )
      check_rc(rc);
    else {
      gbytes+=rc;
      check_rc(0);
    }
  }
  MTX_REL();

  return (unsigned long) gbytes/4;
}

unsigned long
VMUSBmanager::readA322EVME(unsigned long add, u_int32_t *dat, unsigned long nwords)
{
  return readA32BLT32(add, dat, nwords);
}

int
VMUSBmanager::stack_execute(std::vector<VMEcommand> &stack, std::vector<unsigned int> &data)
{
  if (stack.empty()) return 0;
  MTX_ACQ();
  stack_clear();

  int ret,sum=0;
  std::vector<VMEcommand>::iterator it=stack.begin();
  for (;it!=stack.end();++it) {
    switch ((*it).type()) {
      case VMEcommand::read8 :
        if (_stack_size+4>_max_stack_size || _read_size+2>_max_read_size) {
          ret = VMUSB_stack_execute(data);
          if (ret<0) return ret;
          else sum+=ret;
        }
        stack_add_read8((*it).add());
        break;
      case VMEcommand::write8 :
        if (_stack_size+4>_max_stack_size) {
          ret = VMUSB_stack_execute(data);
          if (ret<0) return ret;
          else sum+=ret;
        }
        stack_add_write8((*it).add(),(*it).data());
        break;
      case VMEcommand::read16 :
        if (_stack_size+4>_max_stack_size || _read_size+2>_max_read_size) {
          ret = VMUSB_stack_execute(data);
          if (ret<0) return ret;
          else sum+=ret;
        }
        stack_add_read16((*it).add());
        break;
      case VMEcommand::write16 :
        if (_stack_size+6>_max_stack_size) {
          ret = VMUSB_stack_execute(data);
          if (ret<0) return ret;
          else sum+=ret;
        }
        stack_add_write16((*it).add(),(*it).data());
        break;
      case VMEcommand::read32 :
        if (_stack_size+4>_max_stack_size || _read_size+4>_max_read_size) {
          ret = VMUSB_stack_execute(data);
          if (ret<0) return ret;
          else sum+=ret;
        }
        stack_add_read32((*it).add());
        break;
      case VMEcommand::write32 :
        if (_stack_size+6>_max_stack_size) {
          ret = VMUSB_stack_execute(data);
          if (ret<0) return ret;
          else sum+=ret;
        }
        stack_add_write32((*it).add(),(*it).data());
        break;
      case VMEcommand::readA32D32 :
        if (_stack_size+4>_max_stack_size || _read_size+4>_max_read_size) {
          ret = VMUSB_stack_execute(data);
          if (ret<0) return ret;
          else sum+=ret;
        }
        stack_add_readA32D32((*it).add());
        break;
      case VMEcommand::writeA32D32 :
        if (_stack_size+6>_max_stack_size) {
          ret = VMUSB_stack_execute(data);
          if (ret<0) return ret;
          else sum+=ret;
        }
        stack_add_writeA32D32((*it).add(),(*it).data());
        break;
      default :
        std::cout << "VMUSBmanager::stack_execute: Error\n"
                  << "  unrecognized command type: " << (*it).type() << "\n"
                  << "  aborting stack..." << std::endl;
        stack_clear();
        return -1;
    }
  }
  if (_stack_size>3) {
    ret = VMUSB_stack_execute(data);
    if (ret<0) return ret;
    else sum+=ret;
  }
  MTX_REL();

  return sum;
}

int
VMUSBmanager::readA24_fifo32(unsigned long add,
                             const int nwords,
                             unsigned int *dat )
{
  if (nwords<1) return 0;
  MTX_ACQ();
  stack_clear();

  char *buf = (char*)_stack;
  int words,rc,sum=0,index=0;
  int words_left = nwords;
  while (words_left>0) {
    if (4*words_left>_max_read_size) words = _max_read_size/4;
    else words = words_left;

    for (int i=0;i<words;i++) stack_add_read32(add);

    rc = xxusb_longstack_execute(_udev, buf, _buffer_size, 1000);
    if (rc!=4*words) {
      check_rc(rc);
      Reset();
      MTX_REL();
      return -1;
    }
    sum += rc;

    unsigned int *stack = (unsigned int*)buf;
    for (int i=0;i<words;i++) dat[index++] = stack[i];
    if (false) {
      std::cout << "\nVMUSBmanager::readA24_fifo32 printing stack:" << std::hex;
      for (int i=0;i<words;i++) std::cout << "\n  " << stack[i];
      std::cout << std::dec << std::endl;
    }
    stack_clear();
    words_left -= words;
  }
  MTX_REL();
  return sum;
}

usb_dev_handle*
VMUSBmanager::VMUSB_device_open()
{
  usb_dev_handle *udev;
  xxusb_device_type list[100];
  MTX_ACQ();
  std::cout << "\nSearching for xxusb devices\n";
  short devices = xxusb_devices_find(list);
  std::cout << "..." << devices << " devices found\n";
  for (int i=0;i<devices;i++) {
    std::cout << "...attempting to open device " << i << "\n";
    udev = xxusb_device_open(list[i].usbdev);
    if (udev) {
      std::cout << "...Success!\n" << std::endl;
      MTX_REL();
      return udev;
    }
  }
  MTX_REL();
  return 0;
}

void
VMUSBmanager::VMUSB_reset()
{
  MTX_ACQ();
  xxusb_register_write(_udev, 10, 0x04);
  MTX_REL();
}

int
VMUSBmanager::VMUSB_buffer_clear()
{
  int timeout = 1000;
  char buffer[_buffer_size];
  MTX_ACQ();
  return xxusb_bulk_read(_udev, buffer, _buffer_size, timeout);
  MTX_REL();
}

short
VMUSBmanager::VMUSB_MBLT_read_32(usb_dev_handle *hdev,
                                 short Adress_Modifier,
                                 int count,
                                 unsigned long VME_Address,
                                 u_int32_t *dat)
{
  bool wrk_arnd = false; // work around for problem with transfers that are multiples of 4
  if (count%4==0) wrk_arnd = true;
  if (wrk_arnd) count++;
  long intbuf[64*count*2];

  intbuf[0]=7;
  intbuf[1]=0;
  intbuf[2]=Adress_Modifier +0x900;
  intbuf[3]=0;
  intbuf[4]=(count & 0xffff);
  intbuf[5]=((count >>16) & 0xffff);
  intbuf[6]=(VME_Address & 0xffff);
  intbuf[7]=((VME_Address >>16) & 0xffff);

  // check the stack
  /*
  if (false)
  {
    for (int i=0; i<8; i++)
      std::cout << std::hex << intbuf[i] << std::dec << std::endl;
  }
  */
  MTX_ACQ();
  short ret = xxusb_stack_execute(hdev, intbuf);
  MTX_REL();

  // check bytes returned
  /*
  if (false)
  {
    std::cout << "VMUSBmanager::VMUSB_MBLT_read_32 ret = " << ret << std::endl;
  }
  */

  if (wrk_arnd) {
    count--;
    ret-=256;
  }

  int j=0;
  for (int i=0; i<(64*count*2); i=i+2)
  {
    dat[j]=intbuf[i] + (intbuf[i+1] * 0x10000);
    j++;
  }

  return ret;
}

int
VMUSBmanager::VMUSB_stack_execute(std::vector<unsigned int> &data)
{
  if (_stack_size<4) return 0;

  int timeout = 1000;
  char *buf = (char*)_stack;
  int check = _read_size;
  if (_last_write) check += 2;
  int ret = xxusb_longstack_execute(_udev, buf, _buffer_size, timeout);
  if (ret!=check) {
    std::cout << "VMUSBmanager::stack_execute: Error\n"
              << "...stack read mismatch, returns: " << ret << ", expected: " << check << "\n"
              << "...attempting to clear output buffer as a precaution...\n"
              << "..." << VMUSB_buffer_clear() << " excess bytes cleared from output buffer" << std::endl;
    VMUSB_reset();
    stack_clear();
    return -1;
  }
  if (_last_write) ret -= 2;

  data.reserve(data.size()+_data_size.size());
  unsigned int dat;
  char *cdat = (char*)&dat;
  int index = 0;
  std::vector<int>::iterator it=_data_size.begin();
  for (;it!=_data_size.end();++it) {
    dat = 0;
    for (int i=0;i<(*it);i++) cdat[i] = buf[index++];
    data.push_back(dat);
  }

  stack_clear();
  return ret;
}

void
VMUSBmanager::stack_clear()
{
  _stack_size = 3;
  _read_size  = 0;
  _last_write = false;
  _data_size.clear();
}

void
VMUSBmanager::stack_add_read8(unsigned long add)
{
  int Addr_mod=0x39;
  _stack[_stack_size++]=Addr_mod+0x100;
  _stack[_stack_size++]=0;
  _stack[_stack_size++]=(add&0xffff)+0x01;
  _stack[_stack_size++]=(add>>16)&0xffff;
  _stack[1]=(_stack_size-2)&0xffff;
  _stack[2]=((_stack_size-2)>>16)&0x1f;
  _read_size+=2;
  _data_size.push_back(2);
  _last_write=false;
}
void
VMUSBmanager::stack_add_write8(unsigned long add, unsigned short data)
{
  int Addr_mod=0x39;
  _stack[_stack_size++]=Addr_mod;
  _stack[_stack_size++]=0;
  _stack[_stack_size++]=(add&0xffff)+0x01;
  _stack[_stack_size++]=(add>>16)&0xffff;
  _stack[_stack_size++]=data&0xffff;
  _stack[_stack_size++]=0;//(Data>>16)&0xffff;
  _stack[1]=(_stack_size-2)&0xffff;
  _stack[2]=((_stack_size-2)>>16)&0x1f;
  _last_write=true;
}

void
VMUSBmanager::stack_add_read16(unsigned long add)
{
  int Addr_mod=0x39;
  _stack[_stack_size++]=Addr_mod+0x100;
  _stack[_stack_size++]=0;
  _stack[_stack_size++]=(add&0xffff)+0x01;
  _stack[_stack_size++]=(add>>16)&0xffff;
  _stack[1]=(_stack_size-2)&0xffff;
  _stack[2]=((_stack_size-2)>>16)&0x1f;
  _read_size+=2;
  _data_size.push_back(2);
  _last_write=false;
}
void
VMUSBmanager::stack_add_write16(unsigned long add, unsigned short data)
{
  int Addr_mod=0x39;
  _stack[_stack_size++]=Addr_mod;
  _stack[_stack_size++]=0;
  _stack[_stack_size++]=(add&0xffff)+0x01;
  _stack[_stack_size++]=(add>>16)&0xffff;
  _stack[_stack_size++]=data&0xffff;
  _stack[_stack_size++]=0;//(Data>>16)&0xffff;
  _stack[1]=(_stack_size-2)&0xffff;
  _stack[2]=((_stack_size-2)>>16)&0x1f;
  _last_write=true;
}

void
VMUSBmanager::stack_add_read32(unsigned long add)
{
  int Addr_mod=0x39;
  _stack[_stack_size++]=Addr_mod+0x100;
  _stack[_stack_size++]=0;
  _stack[_stack_size++]=add&0xffff;
  _stack[_stack_size++]=(add>>16)&0xffff;
  _stack[1]=(_stack_size-2)&0xffff;
  _stack[2]=((_stack_size-2)>>16)&0x1f;
  _read_size+=4;
  _data_size.push_back(4);
  _last_write=false;
}
void
VMUSBmanager::stack_add_write32(unsigned long add, unsigned int data)
{
  int Addr_mod=0x39;
  _stack[_stack_size++]=Addr_mod;
  _stack[_stack_size++]=0;
  _stack[_stack_size++]=add&0xffff;
  _stack[_stack_size++]=(add>>16)&0xffff;
  _stack[_stack_size++]=data&0xffff;
  _stack[_stack_size++]=(data>>16)&0xffff;
  _stack[1]=(_stack_size-2)&0xffff;
  _stack[2]=((_stack_size-2)>>16)&0x1f;
  _last_write=true;
}

void
VMUSBmanager::stack_add_readA32D32(unsigned long add)
{
  int Addr_mod=0x09;
  _stack[_stack_size++]=Addr_mod+0x100;
  _stack[_stack_size++]=0;
  _stack[_stack_size++]=add&0xffff;
  _stack[_stack_size++]=(add>>16)&0xffff;
  _stack[1]=(_stack_size-2)&0xffff;
  _stack[2]=((_stack_size-2)>>16)&0x1f;
  _read_size+=4;
  _data_size.push_back(4);
  _last_write=false;
}
void
VMUSBmanager::stack_add_writeA32D32(unsigned long add, unsigned int data)
{
  int Addr_mod=0x09;
  _stack[_stack_size++]=Addr_mod;
  _stack[_stack_size++]=0;
  _stack[_stack_size++]=add&0xffff;
  _stack[_stack_size++]=(add>>16)&0xffff;
  _stack[_stack_size++]=data&0xffff;
  _stack[_stack_size++]=(data>>16)&0xffff;
  _stack[1]=(_stack_size-2)&0xffff;
  _stack[2]=((_stack_size-2)>>16)&0x1f;
  _last_write=true;
}

extern "C"
{
  VMEmanager *create_vmusb_manager()
  {
    static VMUSBmanager vmusb;
    return &vmusb;
  }
}

#endif
