#include <iostream>
#include <iomanip>
#include <functional>
#include <algorithm>
#include <cmath>
#include <cstring>
#include "Analysis.h"
#include "DataFormat.h"
#include <config/ConfigParser.h>
#include <config/ModuleParser.h>
#include <config/GeneralConfig.h>

/*
 * Compute the common mode when the data comes in serial mode
 */
void compute_common_mode(Analysis *A)
{
	double *chan, *out, *ped, *noise, *data;
	chan  = A->Chan();
	ped   = A->Ped();
	noise = A->Noise();
	out   = A->Event();
	data  = A->Chan();
	if (A->DoingCommonMode())
	{
		int ich,nch=A->Ntot();
		for ( ich=0; ich<nch;ich++)
		{
			data[ich] = out[ich]-ped[ich];
		}
		int ichip, nchip=A->Nchip();
		for (ichip=0;ichip<nchip;ichip++)
		{
			double cm,sn;
			A->CommonMode(A->Nchan(), data+ichip*A->Nchan(), &cm, &sn);
			A->SetCmmd(cm, ichip);
			A->SetAvNoise(sn, ichip);
		}
	}
	else
	{
		int ichip, nchip=A->Nchip();
		for (ichip=0;ichip<nchip;ichip++)
		{
			A->SetCmmd(0., ichip);
			A->SetAvNoise(0., ichip);
		}
	}	
}

struct ComputePedestals
{
    Analysis *A;
    std::vector<int> &vout;
    double *nevts, *ped, *noise, *pedtmp, *noisetmp, *out;
    ComputePedestals(Analysis *a, std::vector<int> &v)
            : A(a), vout(v)
    {
        nevts = A->Nevts();
        ped   = A->Ped();
        noise = A->Noise();
        pedtmp = A->PedMax();
        noisetmp = A->PedMin();
        out   = A->Event();
        compute_common_mode(A);
    }
    void operator()(int i)
    {
        const double d1=10.;
        const double d2 = 20.;
        double val, std;
        double n;

        if (!A->DoingPedestalUpdate())
            return;

        if (nevts[i] > A->get_min_sample())
        {
            vout.push_back(i);
            return;
        }

        val = out[i];
        std = val - ped[i] -A->Cmmd(i/A->Nchan());
        std *= std;
        n  = nevts[i];
        if ( n > d2 && fabs(std) > 16.*noisetmp[i] )
            return;

        ped[i] = (val + n * ped[i]) / (n + 1.);
        if (n > d1)
        {
            n -= d1;
            noisetmp[i] = (std + n * noisetmp[i]) / (n + 1.);
            noise[i] = sqrt(noisetmp[i]);
        }
        nevts[i]++;

        if ( nevts[i] >= A->get_min_sample())
            A->got_my_pedestal();

    }
};

struct UpdatePedestals
{
    Analysis *A;
    bool find_pulse;
    double *chan, *out, *ped, *noise, *data;
    UpdatePedestals(Analysis *a, bool pulse = false)
            : A(a), find_pulse(pulse)
    {
        chan  = A->Chan();
        ped   = A->Ped();
        noise = A->Noise();
        out   = A->Event();
        data  = A->Chan();
        compute_common_mode(A);
    }

    void get_the_pulse(int i)
    {
        double xs;
        if (noise[i] > 0.)
        {
            chan[i] = out[i] - ped[i];
            xs = chan[i] / noise[i];
            if ( xs > 5. )
                A->add_hit(i);
        }
    }
    void operator()(int i)
    {
        // IF noise is 0, set it arbitrarily to 1.
        if (noise[i] == 0.)
            noise[i] = 1.;

        // Quit if Analysis does not want to update the pedestals
        if ( !A->DoingPedestalUpdate() )
        {
            if (find_pulse)
                get_the_pulse(i);

            return;
        }

        if (ped[i] == 0.)
        {
            // If pedestal is not yet computed we assume the current
            // channel value should not be too far
            ped[i] = out[i];
        }
        else
        {
            // Do the pedestal and noise correction
            double corr;
            double xs;
            double cmmd = A->Cmmd(i/A->Nchan());
            chan[i] = out[i] - ped[i] - cmmd;
            xs = chan[i] / noise[i];

            corr = (chan[i]+cmmd) * A->get_ped_weight();
            if (corr > 1.) corr = 1.;
            if (corr < -1) corr = -1.;
            ped[i] += corr;

            if (fabs(xs) < 5.)
            {
                noise[i] *= ((1. - A->get_noise_weight()) + xs * xs * A->get_noise_weight());
            }
            if (find_pulse)
                get_the_pulse(i);
        }
    }
};


Analysis::Analysis(int n_chip, int n_chan)
        : HstData(n_chip, n_chan), n_sampled(0), trigger(0),
        pedmax(0), pedmin(0), pd2(0), nevts(0), inside(false),
        do_cmmd(true), do_update(true), emin(-1.e10), emax(1.e10)
{
    g_nevts      = 256.;
    noise_weight = 1. / 1024.;
    ped_weight   = 1. / 128.;
    Resize(Nchip(), Nchan());
}

Analysis::~Analysis()
{
    clear();
}

void Analysis::clear()
{
    if (pedmax) delete[] pedmax;
    if (pedmin) delete[] pedmin;
    if (nevts)  delete[] nevts;
    if ( pd2)   delete[] pd2;


    pedmax = 0;
    pedmin = 0;
    nevts = 0;
    pd2 = 0;
}

void Analysis::Resize(int _nchip, int _nchan)
{
    int ntot = Ntot();
    clear();

    pedmax = new double[ntot];
    pedmin = new double[ntot];
    nevts  = new double[ntot];
    pd2    = new double[ntot];
}

void Analysis::Init()
{
    int i;
    for (i = 0; i < Ntot(); i++)
    {
        pedmax[i] = 0.;
        pedmin[i] = 0.;
        nevts[i] = 0.;
        pd2[i] = 0.;
    }

    trigger = 0;
    n_sampled = 0;
    n_pre = Ntot();

    HstData::Init();

}



void Analysis::SetWeights(double a, double b)
{
    ped_weight = a;
    noise_weight = b;
}

//---------------------------------------------------------------------
// Computes common mode
//---------------------------------------------------------------------
double Analysis::CommonMode(int nn, double *dd, double *cm, double *noise)
{
    int i, ip;
    double mean, sm, xn, xx, xs, xm;
    double *ptr, tmp;
    bool use_it;
    mean = sm = 0.;

    // Start iterations
    for (ip = 0; ip < 3; ip++)
    {
        for (ptr = dd, i = 0, xn = xs = xm = 0.; i < nn; i++, ptr++)
        {
            use_it = true;
            xx = *ptr;
            if (ip)
            {
                tmp = fabs( (xx-mean)/sm );
                use_it = (tmp<2.5);
            }
            if (use_it)
            {
                xn++;
                xm += xx;
                xs += xx * xx;
            }
        }
        if (xn>0.)
        {
            mean = xm / xn;
            sm = sqrt( xs/xn - mean*mean);
        }
    }
    for (i = 0; i < nn; i++)
        dd[i] -= mean;

    if (cm)
    	*cm = mean;
    
    if (noise)
        *noise = sm;
    return mean;
}

//---------------------------------------------------------------------
// Analyses the data
//---------------------------------------------------------------------
void Analysis::Analyze()
{

    int i;
    double *chan = Chan();
    double *noise= Noise();
    double xs;

    if ( hits.empty() )
        return;
    
    /*
     * We should already have the pulse stack filled with data.
     * Fill the histograms
     */
    m_fill.acquire();
    sn_max = 0.;
    for ( i = hits.top();1; i = hits.top())
    {
        HitMap()[i]++;
        int ib = get_bin(chan[i]);
        if (noise[i]>0.)
        {
            xs = fabs( chan[i] / noise[i]);
            if (xs>sn_max)
                sn_max =xs;
            
            if ( ib >= 0 && ib < get_nbins() )
            {
                Signal(i / Nchan() )[ib]++;
                Signal(-1)[ib]++;
                Signal2D(i)[ib]++;
                if (get_map())
                {
                    int ix, iy, ip;
                    get_map()->pad(i, ix, iy);
                    ip = ix + get_map()->nx() * iy;
                    H2Map()[ ip ]++;
                    if ( chan[i] >= emin && chan[i] <= emax )
                        H2MapC()[ ip ]++;
                }
            }
        }

        hits.pop();
        if (hits.empty())
            break;
    }
    m_fill.release();
}

/**
 * Updates the pedestals and noise of a channel
 */

void Analysis::update_pedestals_fast( std::vector<int> &v, bool pulse)
{

    for_each(v.begin(), v.end(), UpdatePedestals(this, pulse));
}

//---------------------------------------------------------------------
// Updates the pedestals
//---------------------------------------------------------------------
void Analysis::update_pedestals( std::vector<int> &v, bool find_pulse)
{
    ConfigParser  *cfg  = ConfigParser::theConfig();
    GeneralConfig *gnrl = dynamic_cast<GeneralConfig *>(cfg->get_handler("general"));

    /**
     * Check at which status of the process we are.
     */
    g_nevts = gnrl->get_ped_min_sample();

    std::vector<int> clist;
//---------------------------------------------------------------------
// First get an estimate of pedestals and noise with the first g_nevts
//---------------------------------------------------------------------
    if ( n_pre )
    {
        for_each(v.begin(), v.end(), ComputePedestals(this, clist));
        if (clist.empty())
            return;
    }
    if ( clist.empty() )
        clist = v;

//---------------------------------------------------------------------
// We have already pedestals for at least some of the channels
//---------------------------------------------------------------------
    for_each(clist.begin(), clist.end(), UpdatePedestals(this, find_pulse));

    // Get Common mode noise
//    for (i=0;i<Nchip();i++)
//       cmmd[i] = do_cmmd ? CommonMode(Nchan(),chan+i*Nchan(),avnoise+i) : 0;
}


void Analysis::reset_pedestals()
{
    int i = 0;
    n_sampled = 0;
    n_pre = Ntot();
    for (i = 0;i < Ntot();i++)
    {
        Ped()[i] = Noise()[i] = 0.;
        Nevts()[i] = 0.;
        PedMax()[i] = PedMin()[i] = 0.;
    }
}


bool Analysis::data_valid(unsigned int *data) const
{
    unsigned int sz, dsz;

    dsz = data[0] & 0xfff;
    sz  = data[2] & 0xfff;

    // Check data integrity
    if ( dsz != data[dsz-1] ||  dsz - sz != 4 )
        return false;

    // We do not want to process twice the same event
    if ( trigger == data[1] )
        return false;

    return true;
}


void Analysis::Fill(unsigned int *data)
{
    unsigned int sz, dsz, idd;
    if ( inside )
        return;

    inside = true;

//---------------------------------------------------------------------
// Decode event
//      Get the data
//      The format is the following
//      Word
//        0     0xCAFEibbb   i= event type bbbb= size of data block
//        1     0xaaaaaaaa   aaaaaaaa event counter
//        2     0xiiitannn   iii=module id t=data type a= readout mode nnn=data size
//        3     data word 0
//        4     data word 1
//        .     .
//      bbbb    bbbb         size of data block
//---------------------------------------------------------------------
    if ( !IsDataBlock(data[0]) )
    {
        inside = false;
        return;
    }

    idd = (data[2] & 0xfff00000) >> 20;
    dsz = data[0] & 0xfff;
    sz  = data[2] & 0xfff;

    // Check data integrity
    if (!data_valid(data))
    {
        inside = false;
        return;
    }

    // Find the module hook
    // ConfigParser *cfg = ConfigParser::theConfig();
    // ModuleParser *hook = cfg->get_id_handler(idd);

    // Decode the data
    decode(data);
    
    // Analyze the data
    Analyze();
    
    trigger = data[1];
    n_sampled++;

    inside = false;
    return;
}

void Analysis::SetEcuts(double e_min, double e_max)
{
    int ie, ic;

    m_fill.acquire();
    emin = std::min(e_min, e_max);
    emax = std::max(e_min, e_max);

    if (emin == emax)
    {
        emin = get_range_min();
        emax = get_range_max();
    }
    if (!get_map())
    {
        m_fill.release();
        return;
    }

    int ix, iy, ip;
    for (ic = 0;ic < Ntot();ic++)
        H2MapC()[ic] = 0.;

    for (ic = 0;ic < Ntot();ic++)
    {
        double *s = Signal2D(ic);
        double *h = H2MapC();

        get_map()->pad(ic, ix, iy);
        ip = ix + get_map()->nx() * iy;
        for (ie = 0;ie < get_nbins();ie++)
        {
            double e = get_bin_center(ie);
            if (e >= emin && e <= emax)
            {
                h[ip] += s[ie];
            }
        }
    }
    m_fill.release();
}
