/* -*- mode: c++ -*- */
#ifndef __AnalysisManager_included__
#define __AnalysisManager_included__
//#include "DataFormat.h"
#include <stack>
#include <vector>
#include <DAQ++/Thread.h>
#include <data/HstData.h>

/**
 * \brief class Analysis.
 * \ingroup DATA
 *
 * This class handles the monitoring of a module. It holds the data,
 * since it derives from HstData, decodes the raw-data fed by the
 * monitoring process and produces the histograms that will be displayed.
 *
 */
class Analysis : public HstData
{
    protected:
        int n_sampled; // number of events sampled so far
        int n_pre; // number of channels with no pedestal
        double trigger; // current trigger number
        double *pedmax, *pedmin; // pedestals
        double *pd2; // pedestal rms
        double *nevts; // controls events acc. so far
        double g_nevts; // min number of events for presampling
        double noise_weight;
        double ped_weight;
        bool inside;
        double sn_max;
        bool do_cmmd;
        bool do_update;

        double emin, emax; // cuts in energy to show the hitmap
        DAQpp::Mutex m_fill;

        std::stack<int> hits; // list of raw hits

        void clear();

    public:
        Analysis(int n_chip, int n_chan);
        virtual ~Analysis();

        void Resize(int _nchip, int _nchan);
        void SetWeights(double, double);
        void Init();

        double CommonMode(int nch, double *dd, double *mean, double *noise);
        /**
         * This is the method that analyzes the data finding, for example, the
         * channel with highest signal, etc.
         */
        virtual void Analyze();
        
        /**
         * This method looks at the data chunck and decides if it has the
         * right format. It should return true if the data is valid and
         * false otherwise
         */
        virtual bool data_valid(unsigned int *data) const;
        
        /**
         * This method decodes the raw data
         */
        virtual void decode(unsigned int *data)
        {
        }
        
        
        /**
         * This is a function that can be called by GenModule::Start
         * at the beginning of a run. 
         */ 
        virtual void init_run()
        {
        }
        
        /**
         * This is a function that can be called by GenModule::Stop
         * at the end of a run
         */
        virtual void end_run()
        {
        }
        double max_pulse()
        {
            return sn_max;
        }
        void Fill(unsigned int *);
        void DoCommonMode(bool x)
        {
            do_cmmd = x;
        }
        bool DoingCommonMode() const
        {
            return do_cmmd;
        }
        void DoPedestalUpdate(bool x)
        {
            do_update = x;
        }
        bool DoingPedestalUpdate() const
        {
            return do_update;
        }
        void update_pedestals(std::vector<int> &v, bool pulse=true);
        void update_pedestals_fast(std::vector<int> &v, bool pulse = false);
        void add_hit(int address)
        {
            hits.push(address);
        }
        int nhits() const
        {
            return hits.size();
        }
        int pop_hit()
        {
            if (hits.empty())
                return -1;
            else
            {
                int chan = hits.top();
                hits.pop();
                return chan;
            }
        }

        
        double *Nevts()
        {
            return nevts;
        }
        double *PedMax()
        {
            return pedmax;
        }
        double *PedMin()
        {
            return pedmin;
        }
        double Emin() const
        {
            return emin;
        }
        double Emax() const
        {
            return emax;
        }
        void SetEcuts(double e_min, double e_max);

        int get_min_sample() const
        {
            return (int)g_nevts;
        }
        double get_noise_weight() const
        {
            return noise_weight;
        }
        double get_ped_weight() const
        {
            return ped_weight;
        }

        void reset_pedestals();
        void trust_pedestals()
        {
            n_pre = 0;
        }
        void got_my_pedestal()
        {
            n_pre--;
        }
};

#endif
