/* -*- mode: c++ -*- */
#ifndef __HstData_h__
#define __HstData_h__

#define HST_NBIN 256
#define HST_MIN  0.
#define HST_MAX  1023.

#include <iostream>
#include <data/PadMap.h>

class HstData;
std::ostream &operator<<(std::ostream &os, const HstData &);
std::istream &operator>>(std::istream &is, HstData &);

/**
 *  \brief class HstData
 *  \ingroup DATA
 *
 *  This class is just a data holder. It should be general enough to
 *  hold all the necesaary data to monitor the performance of a
 *  module.
 */
class HstData
{
    private:
        int _size;
        int _bfsiz;
        int nchip;
        int nchan;
        int nbins;
        double bwidth;
        double rmax, rmin;
        int ntot;
        double *ptr;

        double *out;                 // raw data is dumped here
        double *chan;                // processed signal
        double *ped;                 // pedestals
        double *noise;               // noise
        double *hitmap;              // hitmap
        double *hsig;                // signal histogram
        double *hsig2;               // scatter plot
        double *cmmd;                // chip common mode
        double *avnoise;             // average noise in chip
        double *h2dmap;              // 2D hitmap
        double *h2dmap_c;            // 2D hitmap with E cuts

        PadMap *map;
        void init_ptr();
    public:
        HstData(int n_chip = 1, int n_chan = 128,
                int nb = HST_NBIN, double mn = HST_MIN, double mx = HST_MAX);
        HstData &operator=(const HstData &d);
        virtual ~HstData();

        friend std::ostream &operator<<(std::ostream &os, const HstData &);
        friend std::istream &operator>>(std::istream &is, HstData &);

        virtual void Init();

        void set_map( const PadMap &m);
        const PadMap *get_map() const
        {
            return map;
        }
        PadMap *get_map()
        {
            return map;
        }


        int size() const
        {
            return _size;
        }
        int bsize() const
        {
            return 3*sizeof(int) + _size*sizeof(double);
        }

        int     Nchip() const
        {
            return nchip;
        }
        int     Nchan() const
        {
            return nchan;
        }
        int     Ntot()  const
        {
            return ntot;
        }

        double *Ped(int i = 0)      const
        {
            return ped + i*nchan;
        }
        double *Noise(int i = 0)    const
        {
            return noise + i*nchan;
        }
        double *HitMap(int i = 0)   const
        {
            return hitmap + i*nchan;
        }
        double *H2Map(int i = 0)    const
        {
            return h2dmap + i*nchan;
        }
        double *H2MapC(int i = 0)    const
        {
            return h2dmap_c + i*nchan;
        }
        double *Signal(int i = 0)   const
        {
            return hsig + (i + 1)*nbins;
        }
        double *Signal2D(int i = 0) const
        {
            return hsig2 + i*nbins;
        }
        double *Event(int i = 0)    const
        {
            return out + i*nchan;
        }
        double *Chan(int i = 0)     const
        {
            return chan + i*nchan;
        }
        double  Cmmd(int i = 0)     const
        {
            return cmmd[i];
        }
        double  AvNoise(int i = 0)  const
        {
            return avnoise[i];
        }
        double *GetCmmd()         const
        {
            return cmmd;
        }
        double *GetAvNoise()      const
        {
            return avnoise;
        }
        void SetCmmd(double val, int ichip)
        {
            cmmd[ichip] = val;
        }
        void SetAvNoise(double val, int ichip)
        {
            avnoise[ichip] = val;
        }
        void set_bins(int nb, double vmin, double vmax);
        int    get_nbins()     const
        {
            return nbins;
        }
        double get_range_min() const
        {
            return rmin;
        }
        double get_range_max() const
        {
            return rmax;
        }
        double get_range_bwidth() const
        {
            return 1. / bwidth;
        }
        int    get_bin(double x)
        {
            int ib;
            ib = int(bwidth * (x - rmin));
            return ib;
        }
        double get_bin_center(int i)
        {
            return rmin + (i + 0.5) / bwidth;
        }
};

#endif
