#include <iostream>
#include <iomanip>
#include <sstream>
#include <cstdlib>
#include <cerrno>
#include <cstring>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <gtk/gtk.h>
#include <glade/glade.h>
#include <gui/GladeApp.h>
#include <config/ConfigParser.h>
#include <config/PathConfig.h>
#include <gmodule.h>

static
gboolean toggle_activate(GtkWidget *w)
{
    if ( gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w)) )
        return FALSE;

    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
    return FALSE;
}

static
gboolean toggle_deactivate(GtkWidget *w)
{
    if ( !gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(w)) )
        return FALSE;

    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);
    return FALSE;
}

static GtkWidget *
default_custom_handler(GladeXML *xml, gchar *func_name, gchar *name,
                       gchar *string1, gchar *string2, gint int1, gint int2,
                       gpointer user_data)
{
    typedef GtkWidget *(* create_func)(gchar *name,
                                       gchar *string1, gchar *string2,
                                       gint int1, gint int2);
    GModule *allsymbols;
    create_func func;

    if (!g_module_supported())
    {
        g_error("custom_new requires gmodule to work correctly");
        return NULL;
    }
    allsymbols = g_module_open(NULL, (GModuleFlags) 0);
    if (g_module_symbol(allsymbols, func_name, (gpointer *)&func))
        return (* func)(name, string1, string2, int1, int2);
    g_warning("could not find widget creation function");
    return NULL;
}

void GladeApp::add_path(const std::string &s)
{
    ConfigParser *cfg = ConfigParser::theConfig();
    PathConfig *cpath = dynamic_cast<PathConfig *>(cfg->get_handler("paths"));
    cpath->add_path(PathConfig::gladepath, s);
}

std::string GladeApp::find_file(const std::string &s)
{
    return PathConfig::find_glade_file(s);
}

static gboolean _quit(...)
{
    gtk_main_quit();
    return FALSE;
}

static gboolean _true(...)
{
    return TRUE;
}

static gboolean _false(...)
{
    return FALSE;
}

GladeApp::GladeApp()
        : xml(0), window(0)
{

    if (connect.empty())
    {
        connect["gtk_main_quit"] = (GCallback)_quit;
        connect["gtk_true"] = (GCallback)_true;
        connect["gtk_false"] = (GCallback)_false;
        SET_CONNECT( on_text_changed );
    }


    glade_set_custom_handler(handler, this);
}


int GladeApp::init(const char *fnam, const char *root, const char *domain)
{
    set_handlers();

    open_xml_file(fnam, root, domain);
    if (!xml)
        return -1;

    auto_connect();
    glade_xml_signal_autoconnect_full(xml, connect_callback, this);

    window = glade_xml_get_widget(xml, root);
    if (!window)
        return -1;

    user_init();

    glade_set_custom_handler(default_custom_handler, NULL);

    return 0;
}

GladeApp::~GladeApp()
{
    g_object_unref(xml);

    if (window)
        gtk_widget_destroy(window);
}

void
GladeApp::block_signals()
{
    TextSet::iterator ip;
    for (ip = ctext.begin();ip != ctext.end();++ip)
        g_signal_handler_block( (gpointer)ip->second, ip->first);
}

void
GladeApp::unblock_signals()
{
    TextSet::iterator ip;
    for (ip = ctext.begin();ip != ctext.end();++ip)
        g_signal_handler_unblock( (gpointer)ip->second, ip->first);
}

void
GladeApp::open_xml_file(const char *fnam, const char *root, const char *domain)
{
    std::string path = find_file(fnam);
    if (path.empty())
    {
        std::cout << "\n####\n"
                  << "#### Cannot find glade file " << fnam << std::endl
                  << "####\n"
                  << std::endl;
        return;
    }

    xml = glade_xml_new(path.c_str(), root, domain);
    if (!xml)
    {
        std::cout << "\n####\n"
                  << "#### Faild opening " << path << std::endl
                  << "####\n"
                  << std::endl;
        return;
    }
}

void
GladeApp::connect_callback(const char *handler, GObject *obj,
                           const char *signam, const char *sig_data,
                           GObject *connect_object, gboolean after,
                           gpointer data)
{
    int id;
    GladeApp *app = (GladeApp *)data;
    CallBackMap::iterator ip;

    ip = app->connect.find(handler);
    if (ip != app->connect.end())
    {
        if (after)
            id = g_signal_connect_after(obj, signam, ip->second, app);
        else
            id = g_signal_connect(obj, signam, ip->second, app);

        if (!strcmp(handler, "on_text_changed"))
        {
            app->ctext[id] = obj;
            g_signal_connect(obj, "focus-in-event", (GCallback)on_enter_notify, app);
            g_signal_connect(obj, "focus-out-event", (GCallback)on_leave_notify, app);
        }

    }
    else
    {
        std::cout << handler << " for " << signam
                  << " of " << gtk_widget_get_name(GTK_WIDGET(obj))
                  << std::endl;
    }
}

GtkWidget *
GladeApp::handler(GladeXML *_xml,
                  gchar *func_name,
                  gchar *name,
                  gchar *string1,
                  gchar *string2,
                  gint int1,
                  gint int2,
                  gpointer user_data)
{
    GladeApp *app = (GladeApp *)user_data;
    GtkWidget *w;

    std::map<std::string, GladeHandler>::iterator ip;
    ip = app->handlers.find(func_name);
    if (ip != app->handlers.end())
        w = (ip->second)(app, name, string1, string2, int1, int2);
    else
    {
        typedef GtkWidget *(* create_func)(gchar *name,
                                           gchar *string1, gchar *string2,
                                           gint int1, gint int2);
        create_func func;
        GModule *allsymbols = g_module_open(NULL, (GModuleFlags)0);

        if ( g_module_symbol(allsymbols, func_name, (gpointer *)&func))
        {
            w = (* func)(name, string1, string2, int1, int2);
        }
        else
        {
            std::cout << "Handler missing: " << func_name << std::endl;
            w = gtk_label_new(name);
        }
    }

    gtk_widget_show(w);
    return w;
}


gboolean GladeApp::on_enter_notify(GtkWidget *w, GdkEventFocus *e, gpointer data)
{
    GladeApp *app = (GladeApp *)data;
    std::string name = gtk_widget_get_name(w);
    app->values[name] = gtk_entry_get_text(GTK_ENTRY(w));
    return FALSE;
}
gboolean GladeApp::on_leave_notify(GtkWidget *w, GdkEventFocus *e, gpointer data)
{
    GladeApp *app = (GladeApp *)data;
    std::string name = gtk_widget_get_name(w);
    ValueMap::iterator ip = app->values.find(name);
    if (ip != app->values.end())
    {
        std::string old_value = ip->second;
        std::string new_value = gtk_entry_get_text(GTK_ENTRY(w));
        if (old_value != new_value)
            app->process_text(ip->first, w);

        app->values.erase(ip);
    }
    return FALSE;
}

void GladeApp::on_text_changed (GtkWidget *w, gpointer data)
{}

GtkWidget *
GladeApp::get_widget(const std::string &s) const
{

    return xml ? glade_xml_get_widget(xml, s.c_str()) : NULL;
}


std::string GladeApp::get_text(const std::string &s) const
{
    GtkWidget *w = get_widget(s.c_str());
    std::string out;
    if (GTK_IS_ENTRY(w))
    {
        out = gtk_entry_get_text(GTK_ENTRY(w));
    }
    else if ( GTK_IS_LABEL(w))
    {
        out = gtk_label_get_text(GTK_LABEL(w));
    }
    return out;
}

void GladeApp::set_text(const std::string &wname, const std::string &the_text)
{
    GtkWidget *w = get_widget(wname.c_str());
    if ( GTK_IS_ENTRY(w))
    {
        gtk_entry_set_text(GTK_ENTRY(w), the_text.c_str());
    }
    else if ( GTK_IS_LABEL(w))
    {
        gtk_label_set_text(GTK_LABEL(w), the_text.c_str());
    }
}


int GladeApp::get_int(const std::string &s, int def)
{
    char *endptr;
    const char *str = s.c_str();
    long val = strtol(str, &endptr, 10);
    if (errno == ERANGE || errno == EINVAL)
        val = def;

    return val;
}

double GladeApp::get_double(const std::string &s, double def)
{
    char *endptr;
    const char *str = s.c_str();
    double val = strtod(str, &endptr);
    if (endptr == str || errno == ERANGE || errno == EINVAL)
        val = def;

    return val;
}

static char __GladeApp_line__[1024];
std::string GladeApp::double2str(double x, char *fmt)
{
    std::ostringstream ostr;

    if (fmt)
    {
        sprintf(__GladeApp_line__, fmt, x);
        ostr << __GladeApp_line__;
    }
    else
        ostr << x;

    return ostr.str();
}

std::string GladeApp::int2str(long x, char *fmt)
{
    std::ostringstream ostr;

    if (fmt)
    {
        sprintf(__GladeApp_line__, fmt, x);
        ostr << __GladeApp_line__;
    }
    else
        ostr << x;

    return ostr.str();
}
bool GladeApp::btn_active(const std::string &s)
{
    return gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (get_widget(s)));
}


void GladeApp::activate_button(const std::string &s, bool state)
{
    GtkWidget *w = get_widget(s);
    if (!w)
        return;
    g_return_if_fail( GTK_IS_TOGGLE_BUTTON(w) );
    if (state)
        g_idle_add( (GSourceFunc)toggle_activate, (gpointer)w);
    else
        g_idle_add( (GSourceFunc)toggle_deactivate, (gpointer)w);
}

void GladeApp::click_button( const std::string &s)
{
    GtkWidget *w = get_widget(s);
    if (!w)
        return;

    g_return_if_fail( GTK_IS_BUTTON(w) );
    gtk_button_clicked( GTK_BUTTON(w) );
}

int GladeApp::entry2int(const std::string &s, int def)
{
    return get_int(gtk_entry_get_text(GTK_ENTRY(get_widget(s))), def);
}

double GladeApp::entry2double(const std::string &s, double def)
{
    return get_double(gtk_entry_get_text(GTK_ENTRY(get_widget(s))), def);
}

void GladeApp::int2entry(const std::string &s, int v, char *fmt)
{
    GtkWidget *w = get_widget(s);
    if (!w)
        return;

    if ( GTK_IS_ENTRY(w) )
        gtk_entry_set_text( GTK_ENTRY(w), int2str(v, fmt).c_str() );
    else if ( GTK_IS_LABEL(w) )
        gtk_label_set_text( GTK_LABEL(w), int2str(v, fmt).c_str() );


}
void GladeApp::double2entry(const std::string &s, double v, char *fmt)
{
    GtkWidget *w = get_widget(s);
    if (!w)
        return;

    if (GTK_IS_ENTRY(w))
        gtk_entry_set_text( GTK_ENTRY(w), double2str(v, fmt).c_str() );
    else if (GTK_IS_LABEL(w))
        gtk_label_set_text( GTK_LABEL(w), double2str(v, fmt).c_str() );

}

int GladeApp::spin2int(const std::string &s)
{
    return gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(get_widget(s)));
}

double GladeApp::spin2double(const std::string &s)
{
    return gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(get_widget(s)));
}

void GladeApp::int2spin(const std::string &s, int v)
{
    GtkWidget *w = get_widget(s);
    if (GTK_IS_SPIN_BUTTON(w))
    {
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(w),v);
    }
    return;

}
void GladeApp::double2spin(const std::string &s, double v)
{
    GtkWidget *w = get_widget(s);
    if (GTK_IS_SPIN_BUTTON(w))
    {
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(w),v);
    }
    return;
}



