#ifndef __MainWindow_h__
#define __MainWindow_h__

/**
 * \addtogroup GUI
 * 
 * <b> The logic: </b>
 *  
 *  The logic may seem a bit "involved" but it was prepared to work with
 *  CORBA and, therefore, it is based in message passing between the
 *  different components.
 *  
 *  The idea is that MsgManager will receive all the messages from the
 *  real DAQ++ objects and distribute them. So, the state of the
 *  RunManagers and Modules gets to MsgManager and it sends the
 *  corresponding messages to the Objects subscribed, for instance
 *  MainWindow, RunMenu or ParamSlider.
 *  
 *  On the other hand RunMenu will send the DAQ commands
 *  
 *  
 *  module_gui
 *  
 *  This class controls the small module window in the main window of wmedaq
 *  
 *  ParamSlider
 *  
 *  This class binds a Module parameter with a GtkAdjustment which can
 *  come from a GtkScale, a GtkSpinButton, etc. Changes in the adjustment
 *  will be sent to the Module parameter.
 *  
 *  The class mantains a static list of existing mappings.
 *  
 *  
 *  Widget Updates
 *  
 *  This relates the updating of number of triggers, rates, throughput,
 *  etc. It is done with WidgetUpdate-based classes that will comprise
 *  labels, spin-sliders, gauges, etc.
 *  
 *  Each monitor window derives from WidgetHolder. Widgets to be updated
 *  will register and receive the update signal from the holder together
 *  with the update data. This is retreived from any of
 *  
 *  void RunManager_update(const string &, Update &);
 *  void Module_update(const string &, Update &);
 *  void Update_all(int &nRm, Update **, int &nMd, Update **);
 *  
 *  
 *  
 *  <b>DAQ state and RnnMenu</b>
 *  
 *  DAQ state is show and controlled by a RunMenu. Each RunMenu is
 *  associated with a RunManager and will send it the DAQ commands.
 *    
 */


#include <vector>
#include <gtk/gtk.h>

#include <daq/ScanPoint.h>
#include <gui/WidgetUpdate.h>
#include <gui/GladeApp.h>
#include <gui/MsgManager.h>
class module_gui;

class Rmanager;

/**
 * \ingroup GUI
 * 
 *  This window handles the main display of the DAQ program.
 *  It also controls the principal RunManager of the application
 *
 */
class MainWindow : public GladeApp, public WidgetHolder
{
private:

    /// List of DSP id's
    std::vector<std::string> module_list;

    /// List of module_gui
    std::vector<module_gui *> module_ptr;

    /// Update pedestals
    bool do_pedestal_update;

    void CreateDisplay();

    /// callback for update
    static int      onUpdate(void *);
    static void     onLoadConfig(GtkWidget *, MainWindow *);
    static void     onSaveConfig(GtkWidget *, MainWindow *);
    static void     onSaveAsConfig(GtkWidget *, MainWindow *);
    static void     onQuit(MainWindow *);
    static void     onPreferences(GtkWidget *, MainWindow *);
    static void     onLoadPedestals(GtkWidget *, MainWindow *);
    static void     onSavePedestals(GtkWidget *, MainWindow *);
    static void     onAbout(GtkWidget *, MainWindow *);
    static void     onManual(GtkWidget *, MainWindow *);
    static void     onScan(GtkWidget *, MainWindow *);
    static void     onLogData(GtkWidget *, MainWindow *);
    static void     onTrgType(GtkWidget *, MainWindow *);
    static void     onInterrupt(GtkWidget *, MainWindow *);
    static void     onPedestal(GtkWidget *, MainWindow *);
    static void     onMonitor(GtkWidget *, MainWindow *);
    static void     do_pedestal(MainWindow *);
    static void     on_plugins_activate(GtkWidget *w, MainWindow *);
    static gboolean do_add_modules(MainWindow *);
    static gboolean do_clear_modules(MainWindow *);
    
    static GtkWidget *create_gauge_rate(MainWindow *, char *, char *, char *, int , int);
    static GtkWidget *create_gauge_eff(MainWindow *, char *, char *, char *, int , int);
    static GtkWidget *create_memory_monitor(MainWindow *, char *, char *, char *, int , int);

    void add_modules_true();

    // Scan
    bool doScan;
    int  nScan;
    ScanPoint scpoint;

    /** Update on state.
    This is in fact an Observer that will be notified whenever
    the main RunManager changes its state.
    */
    StChanged<MainWindow> state;
    friend class StChanged<MainWindow>;

    int update_tmout;

    // Default trigger type
    std::string trg_type;

    // Misc.
    unsigned int nrun;


    void set_active_widgets(bool);
    
    void check_for_interrupts(bool *ptr = 0);

    Rmanager *CreateRunManager();
public:
    MainWindow();
    ~MainWindow();


    // GladeApp functions
    // GladeApp methods
    void user_init();
    void auto_connect();
    void set_handlers();
    void process_text(const std::string &s, GtkWidget *w);
    
    void set_module_list(const std::vector<std::string> &ids);
    void clear_modules();
    void add_modules();

    void start_timer();
    void stop_timer();
    
};

#endif
