#ifndef __vmedaq_MsgManager_h__
#define __vmedaq_MsgManager_h__

#include <gtk/gtkmain.h>
#include <queue>
#include <DAQ++/ObserverModel.h>
#include <DAQ++/RunState.h>

/** \ingroup GUI
 *  \brief class StChanged<T>
 *
 *  This class is an Observer that stores the notifications of the
 *  observables in an inernal queue that is emptied when the gtk loop
 *  allows a safe reaction.
 *
 *  In particular it is sensitive to changes of the state DAQ++
 *  objects.
 * 
 */
template <class T> class StChanged : public Observer
{
    protected:
        struct Data
        {
                StChanged *obj;
                DAQpp::RunState::StatusType st;
                Data(StChanged *o, const DAQpp::RunState::StatusType &s) :
                    obj(o), st(s)
                {
                }
        };
    private:
        T *obj;
        /** This is the function that gets executed in the glib loop. It
         * just calls the update_state methid
         *
         */
        static gboolean loop_update(Data *data)
        {
            data->obj->update_state(data->st);
            delete data;
            return FALSE;
        }
        virtual void update_state(DAQpp::RunState::StatusType st)
        {
        }
    public:
        StChanged(T *m) :
            obj(m)
        {
        }
        virtual ~StChanged()
        {
        }
        /** This is the Observer update function.
         *  The only thing it does is to delegate the action to the glib
         *  loop so that things are done in the same thread as the loop  
         *  
         */
        void update(Observable *o, Argument *arg)
        {
            Arg<DAQpp::RunState::StatusType> *st =
                    static_cast< Arg<DAQpp::RunState::StatusType> * > (arg);
            Data *data = new Data(this, *st);
            g_idle_add( (GSourceFunc)loop_update, (gpointer)data);
        }
};

/** \ingroup GUI
 *  \brief DAQ++ message distributor.
 *
 *  This class serves as a distributor of messages and
 *  notifications. It is designed as a singleton and register itself
 *  to the DAQmanager as an Observer so that it will be notified
 *  whenever a new RunManger or Module is created.
 *
 *  For the RunManagers, upon creation, it will try to associate the
 *  corresponding RunMenu.
 */
class MsgManager : public Observer
{

    protected:
        // Hide the constructors
        MsgManager();
        MsgManager(const MsgManager &)
        {
        }
        void operator=(const MsgManager &)
        {
        }

        /**
         * Whenever a RunManager is created, MsgManager registers as
         * an observer to any change of state and transmits that to the
         * corresponding RunMenu. RunManager notifications come through 
         * this subclass
         */
        class StateChanged : public Observer
        {
            public:
                void update(Observable *o, Argument *arg);
        };
        
        /**
         * DAQpp::Parameter objects represented by ParamSlider objects
         *  will notify their changes through this subclass
         * 
         */
        class ParameterChanged : public Observer
        {
            public:
                void update(Observable *o, Argument *arg);
        };

        class ObjectAddedRemoved : public Observer
        {
            public:
                void update(Observable *o, Argument *arg);
        };

    public:
        virtual ~MsgManager()
        {
        }

        /// Returns a pointer to the manager
        static MsgManager *theMsgManager()
        {
            static MsgManager mngr;
            return &mngr;
        }

        StateChanged state;
        ParameterChanged parameter;
        ObjectAddedRemoved object;

        /**
         * This is the routine called by DAQmanager when the MsgManager
         * registers as an Observer
         */
        void update(Observable *o, Argument *arg);
};

#endif
