#include <iostream>
#include <string>
#include <sstream>
#include <vector>
#include <cerrno>
#include <cstring>

#include <DAQ++/Thread.h>
#include <DAQ++/DAQmanager.h>
#include <daq/GenModule.h>
#include <daq/Rmanager.h>

#define FIFO_NAME "/tmp/glib_trigger.dat"

inline void err_msg(const std::string &str, const std::string &where)
{
    std::cout << "[" << where << "] " << str << std::endl << '\t' << strerror(errno)
            << std::endl;
}

/**
 * A "trigger" module should provide the following routines:
 *   enable (mandatory): take the necessary actions to enable the
 *                       triggers.
 *   disable (mandatory): take the necessary action to disable the
 *                        triggers
 *   start (optional): called at the start of each run to make any
 *	               initialization that could be needed
 *   stop  (optional): called at the end of run to do whatever might
 *                     be needed at this time
 *   reset (option): called at the end of each event.
 */

/**
 *  This is the trigger handler that DAQmanager needs.
 *  It should return true when a trigger is available and
 *  false otherwise. As it is implemented it will block
 *  until there is data
 */

class SftTrigger : public DAQpp::DAQTrigger
{
    private:
        bool have_to_stop;
        /**
         * This an ugly implemtation of a list of modules that should trigger
         * in order to accept the event
         */
        std::vector<GenModule *> trigger_list;

    public:
        SftTrigger();
        ~SftTrigger();
        void enable();
        void disable();
        bool operator()();

        void start();
        void stop();
        void reset()
        {
        }

        void add_module(GenModule *m)
        {
            trigger_list.push_back(m);
        }

        void clear_trigger_list()
        {
            trigger_list.clear();
        }
};

static SftTrigger *the_trigger=0;

SftTrigger::SftTrigger() :
    have_to_stop(false)
{
}

SftTrigger::~SftTrigger()
{
}

struct fill_trigger_list : public std::unary_function<Observer *, void>
{
        fill_trigger_list()
        {
        }
        void operator()(Observer *o)
        {
            GenModule *m = dynamic_cast<GenModule *>(o);
            if (m)
            {
                the_trigger->add_module(m);
            }
        }
};

void SftTrigger::start()
{
    Rmanager *rm = dynamic_cast<Rmanager *>(DAQpp::DAQmanager::theDAQmanager()->find_runmanager("main"));
    if (!rm)
        return;

    rm->for_each_module(fill_trigger_list() );
}

void SftTrigger::stop()
{
    clear_trigger_list();
    have_to_stop = true;
}

/**
 * This is a predicate that requires that all the modules have data.
 */
struct request_all_modules : public std::unary_function<GenModule *,void>
{
        bool _trigger;
        request_all_modules() :
            _trigger(true)
        {
        }
        void operator()(GenModule *m)
        {
            _trigger = _trigger && m->hasData();
        }
        bool trigger() const
        {
            return _trigger;
        }
};

/**
 * This is a predicate that requires that there is data in each type
 * of module. We assume here that there are 2 types of modules, for
 * instance ADCs and VATAs. We further assume that we have set the
 * module identifiers in such a way that we can differentiate the two
 * kinds.
 */
struct request_one_of_each_type : public std::unary_function<GenModule *,void>
{
        bool trg_adc, trg_vata;
        request_one_of_each_type() :
            trg_adc(false), trg_vata(false)
        {
        }
        void operator()(GenModule *m)
        {
            int id = int(m->get_id());
            if (id & 0x1000)
            { // a VATA
                trg_vata |= m->hasData();
            }
            else
            {
                trg_adc |= m->hasData();
            }
        }
        bool trigger() const
        {
            return trg_vata && trg_adc;
        }
};

struct flush_modules : public std::unary_function<GenModule *,void>
{
        void operator()(GenModule *m)
        {
            m->flush();
        }
};

bool SftTrigger::operator()()
{
    std::vector<GenModule *>::iterator ip;

    //    struct request_all_modules trigger =
    //       for_each(trigger_list.begin(),trigger_list.end(),request_all_modules());

    struct request_one_of_each_type trigger =
        for_each(trigger_list.begin(),trigger_list.end(),request_one_of_each_type());

    /**
     * If there is no trigger we may need to flush the data on the modules
     * that have data in their buffers. 
     */
    if (!trigger.trigger())
        for_each(trigger_list.begin(),trigger_list.end(),flush_modules());

    return trigger.trigger();
}

extern "C"
{
    /*
     This is the routine called to enable trigger
     interrupts. If things went smooth it should return
     0, otherwise -1.
     */
    int enable_interrupts()
    {
        std::cout << "*** Enabling SftTrigger ***" << std::endl;
        if (!the_trigger)
            the_trigger = new SftTrigger();

        // Make DAQmanager aware of this
        DAQpp::DAQmanager::theDAQmanager()->set_trigger(the_trigger);

        return 0;
    }

    /*
     This is the routine called to disable trigger interrupts
     */
    void disable_interrupts()
    {
        if (the_trigger)
        {
            std::cout << "*** Disabling SftTrigger ***" << std::endl;
            DAQpp::DAQmanager::theDAQmanager()->set_trigger(0);
            the_trigger->stop();
            delete the_trigger;
            the_trigger=0;
        }
    }

    void reset_interrupts()
    {
        if (the_trigger)
            the_trigger->reset();
    }

    void start_interrupts()
    {
        std::cout << "Start  interrupts" << std::endl;
        if (the_trigger)
            the_trigger->start();

    }
    void stop_interrupts()
    {
        std::cout << "Stop interrupts" << std::endl;
        if (the_trigger)
        {
            the_trigger->stop();
        }
    }

}

static void __attribute__ ((constructor)) __init__lib__()
{
    // init threads
    std::cout << "### Loading SftTrigger..." << std::endl;
    the_trigger = new SftTrigger;
}

static void __attribute__ ((destructor)) __close__lib__()
{
    std::cout << "### Unloading SftTrigger..." << std::endl;
    delete the_trigger;
}
