#include <iostream>
#include <string>
#include <sstream>
#include <cerrno>
#include <cstring>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/poll.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <glib.h>

#include <DAQ++/Thread.h>
#include <DAQ++/DAQmanager.h>

#define FIFO_NAME "/tmp/glib_trigger.dat"

inline void err_msg(const std::string &str, const std::string &where)
{
    std::cout << "[" << where << "] " << str << std::endl << '\t' << strerror(errno)
            << std::endl;
}

/**
 * This is the trigger handler that DAQmanager needs.
 * It should return true when a trigger is available and
 * false otherwise. As it is implemented it will block
 * until there is data
 */
class GLibTrigger : public DAQpp::DAQTrigger
{
    private:
        bool have_to_stop;
        int fd, fdr;
        int tmout;
        static gboolean action(GLibTrigger *t);
    public:
        GLibTrigger();
        ~GLibTrigger();
        void enableInterrupts();
        void disableInterrupts();
        bool operator()();

        void start()
        {
            double freq = 100.;
            int delay=int(1000./freq);
            have_to_stop=false;
            std::cout << "Starting the interrupts: " << delay << std::endl;

            tmout = g_timeout_add(delay, (GSourceFunc)action, this);
            std::cout << "...tmout=" << tmout << std::endl;
        }
        void stop()
        {
            have_to_stop=true;
            g_source_remove(tmout);
            tmout=-1;
        }

        void reset()
        {
        }
};

static GLibTrigger *the_trigger=0;

gboolean GLibTrigger::action(GLibTrigger *t)
{
    int val = 0xcaca;
    ssize_t sz;
    sz = write(t->fd, &val, sizeof(int));
    if (sz<0)
    {
        std::ostringstream ostr;
        ostr << __PRETTY_FUNCTION__ << ':' << __LINE__;
        err_msg("Error writing in FIFO", ostr.str());
    }
    return TRUE;
}

GLibTrigger::GLibTrigger() :
    have_to_stop(false), fd(-1), fdr(-1), tmout(-1)
{
    int rc;
    again: rc = mkfifo(FIFO_NAME,0666);
    if (rc<0 && errno==EEXIST)
    {
        unlink(FIFO_NAME);
        goto again;
    }
    if (rc)
    {
        std::ostringstream ostr;
        ostr << __PRETTY_FUNCTION__ << ':' << __LINE__;
        err_msg("GLibTrigger: Problems creating fifo", ostr.str());
        return;
    }
    std::cout << "FIFO created" << std::endl;
    fdr = open(FIFO_NAME,O_RDONLY|O_NONBLOCK);
    if (fdr<0)
    {
        std::ostringstream ostr;
        ostr << __PRETTY_FUNCTION__ << ':' << __LINE__;
        err_msg("GLibTrigger: Could not open the file for reading", ostr.str());
        return;
    }
    fd = open(FIFO_NAME,O_WRONLY|O_NONBLOCK);
    if (fd<=0)
    {
        std::ostringstream ostr;
        ostr << __PRETTY_FUNCTION__ << ':' << __LINE__;
        err_msg("GLibTrigger: Could not open the file for writing", ostr.str());
        return;
    }
}

GLibTrigger::~GLibTrigger()
{
    if (tmout>0)
        g_source_remove(tmout);
    close(fd);
    close(fdr);
    unlink(FIFO_NAME);
}

bool GLibTrigger::operator()()
{
    bool rc;
    int val;
    int sz;
    if (have_to_stop || fdr<0)
    return false;

    rc = false;
    do
    {
        struct pollfd ufds = { fdr,POLLIN,0 };
        int ret = poll(&ufds,1,1000);
        if (ret<0 || have_to_stop)
            rc=true;

        if (ret==0 || (ufds.revents & 0xf8) )
            continue;

        if ( errno==EINTR)
        {
            std::cout << "Got an interrupt" << std::endl;
        }

        rc=true;
        sz = read(fdr,&val,sizeof(int));
    }
    while (!rc);
    return true;
}

extern "C"
{
    /*
     This is the routine called to enable trigger
     interrupts. If things went smooth it should return
     0, otherwise -1.
     */
    int enable_interrupts()
    {
        std::cout << "*** VMEintr: Enabling Interrupts ***" << std::endl;
        if (!the_trigger)
            the_trigger = new GLibTrigger();

        // Make DAQmanager aware of this
        DAQpp::DAQmanager::theDAQmanager()->set_trigger(the_trigger);

        return 0;
    }

    /*
     This is the routine called to disable trigger interrupts
     */
    void disable_interrupts()
    {
        if (the_trigger)
        {
            std::cout << "*** VMEintr: Disabling Interrupts ***" << std::endl;
            DAQpp::DAQmanager::theDAQmanager()->set_trigger(0);
            the_trigger->stop();
            delete the_trigger;
            the_trigger=0;
        }
    }

    void reset_interrupts()
    {
        if (the_trigger)
            the_trigger->reset();
    }

    void start_interrupts()
    {
        std::cout << "Start  interrupts" << std::endl;
        if (the_trigger)
            the_trigger->start();

    }
    void stop_interrupts()
    {
        std::cout << "Stop interrupts" << std::endl;
        if (the_trigger)
        {
            the_trigger->stop();
        }
    }

}

static void __attribute__ ((constructor)) __init__lib__()
{
    // init threads
    std::cout << "### Loading glib_trigger..." << std::endl;
    the_trigger = new GLibTrigger;
}

static void __attribute__ ((destructor)) __close__lib__()
{
    std::cout << "### Unloading glib_trigger..." << std::endl;
    delete the_trigger;
}
