#include <iostream>
#include <string>
#include <vector>
#include <data/DataFormat.h>
#include "test_module.h"
#include "test_analysis.h"

#include <cmath>
static double randuniform();
static double randgauss(double mean, double stdev);


test_module::test_module( const DAQpp::DAQid &id, unsigned long address)
        : GenModule(id, address), nchip(1), nchan(64)
{
    // reset pedestals and noise to sensible values
    fill_n(back_inserter( get_par_value< std::vector<double> > ("pedestal") ), nchan*nchip, 0.);
    fill_n(back_inserter( get_par_value< std::vector<double> > ("noise") ), nchan*nchip, 0.);


    // Using DAQ++ parameters can be useful, especially if you want to
    // access them the DAQcorba
    add_par("efficiency", new DAQpp::Par<double>(0.75));
    add_par("mean", new DAQpp::Par<double>(100.));
    add_par("width", new DAQpp::Par<double>(10.));



    // now we should create the Analysis for the monitoring
    // never forget to do this !!!
    this->ana = new test_analysis();
    this->ana->Resize(nchip, nchan);
}

int 
test_module::Trigger()
{
    state = randuniform();  
    return 0;
}

bool
test_module::hasData()
{
    return state < get_par_value<double>("efficiency");
}

char *test_module::getData(int &size)
{
    int ic, ntot = nchip * nchan;
    double _mean = DAQpp::get_par_value<double>(get_par("mean"));
    double _std  = DAQpp::get_par_value<double>(get_par("width"));
    int channel  = int(ntot * randuniform());

    /*
      Get the data
      The format is the following
      Word
      0     0xCAFEibbb   i= event type bbb= size of data block
      1     0xaaaaaaaa   aaaaaaaaaa event counter
      2     0xiiitnnnn   iii=module id t=data type nnnn=data size
      3     data word 0
      4     data word 1
      .     .
      bbbb    bbbb         size of data block

      In practice, some bits of words 1 and 2 can be used. Note,
      however, that the module id should go where it is.

      the size in word 2 is the number of data words folowing this
      word. 
      
      Also, the event type byte will allow us to identify our data block
      when reading the file. Look into DataFormat.h and choose one
      different from the existing ones:
         NormalData = 0x10000,
    PulseData  = 0x20000,
    ScanData   = 0x30000,
    ADCdata    = 0x40000,
    CSPRINTdata= 0x80000,
    FBTDCdata  = 0x90000
    */

    const unsigned int data_type = 0xa0000;

    increase_evt_counter();

    // Allocate space for the data. This has to be done with the
    // allocate method of DAQpp::Module
    //
    // Note that here we assume that we have one word per channel. It
    // may well happen that we get from the module a complicated
    // structure that we do not want ot decode here, so we just dumpt
    // it into the header and the trailer words and leave the analysis
    // program with the burden of decoding.

    size =
        4 +   // the 3 words in the header plus the
        // trailing word
        ntot; // We assume one word per channel...

    unsigned int *data =
        reinterpret_cast<unsigned int *>(allocate(sizeof(unsigned int) * size));


    // Fill the data array
    int id = get_id();
    int iw = 0;
    data[iw++] = DataHeader | DataRecord | (size);
    data[iw++] = get_evt_counter();
    data[iw++] = (id & 0xfff) << 20 | data_type | (ntot);

    for (ic = 0;ic < ntot;ic++)
        data[iw++] = (unsigned int)(randgauss(50., 5.));

    data[3+channel] += (unsigned int)(randgauss(_mean, _std));
    data[iw] = size;

    size *= sizeof(unsigned int);
    return reinterpret_cast<char *>(data);
}

int
test_module::PrepareForRun ()
{
    std::cout << "\t+-> Module.PrepareForRun [" << get_id() << "]"
    << std::endl;
    return 0;
}

int
test_module::Start (int)
{
    std::cout << "\t+-> Module.Start         [" << get_id() << "]"
    << std::endl;
    if ( get_analysis() )
        get_analysis()->init_run();

    return 0;
}

int
test_module::Stop (int)
{
    std::cout << "\t+-> Module.Stop          [" << get_id() << "]"
    << std::endl;

    if ( get_analysis() )
        get_analysis()->end_run();

    return 0;
}
int
test_module::Pause ()
{
    std::cout << "\t+-> Module.Pause         [" << get_id() << "]"
    << std::endl;
    return 0;
};

int
test_module::Continue ()
{
    std::cout << "\t+-> Module.Continue      [" << get_id() << "]"
    << std::endl;
    return 0;
};

int
test_module::Reset ()
{
    std::cout << "\t+-> Module.Reset         [" << get_id() << "]"
    << std::endl;
    return 0;
};

int
test_module::GetReady ()
{
    std::cout << "\t+-> Module.GetReady      [" << get_id() << "]"
    << std::endl;
    return 0;
};

double
randuniform()
{
    double x = ((double)rand()) / ((double)RAND_MAX);
    return x;
}

double
randgauss(double mean, double stdev)
{
    static int haveCachedNormal = 0;
    static double  cachedNormal = 0.;
    static double  pMean = 0., pStdDev = 0.;

    if ( mean != pMean || pStdDev != stdev )
    {
        haveCachedNormal = 0;
        pMean = mean;
        pStdDev = stdev;
    }
    if (haveCachedNormal == 1)
    {
        haveCachedNormal = 0;
        return(cachedNormal * pStdDev + pMean );
    }
    else
    {
        for (;;)
        {
            double u1 = randuniform();
            double u2 = randuniform();
            double v1 = 2 * u1 - 1;
            double v2 = 2 * u2 - 1;
            double w = (v1 * v1) + (v2 * v2);

//
//      We actually generate two IID normal distribution variables.
//      We cache the one & return the other.
//
            if (w <= 1)
            {
                double y = sqrt( (-2 * log(w)) / w);
                double x1 = v1 * y;
                double x2 = v2 * y;

                haveCachedNormal = 1;
                cachedNormal = x2;
                return(x1 * pStdDev + pMean);
            }
        }
    }

}
