#ifndef __test_module__h__
#define __test_module_h__
/**
 * This is a module for vmedaq.
 * They all should inherit from GenModule, which adds some extra
 * functionality on top of DAQpp::Module:
 *
 *   - The module works around the concept of something with a number
 *     of chips, each having the same number of channels. So nchip
 *     (number of chips) and nchan (number of channels per chip) is
 *     something we should supply somehow. In paticular for the
 *     Analysis instance of the module. That brings me to the second
 *     point.
 *
 *   - There is an Analysis pointer (you have to create the object)
 *     which will store all the monitoring stuff. This guy needs the
 *     number of chips and channels per chip.
 *
 *   - If you plan to make scans with that module, you will need to
 *     override the set_scan_point method to set your module in a
 *     reasonable state every time you are in a new scan point.
 *
 *   - pedestals and noise are there as DAQ++ parameters
 *      + can be updated from the values in Analysis or copied into
 *        the Analysis
 *
 *      + Pedestals are stored in the data file if you override the
 *      write_pedestals method.
 *
 *
 * See the DAQ++ documentation for further information
 *
 **/
#include <daq/GenModule.h>

class test_module : public GenModule
{
    private:
        int nchip; // number of chips
        int nchan; // number of channels per chip
        double state;
    public:

        /** Constructor
         *  The constructor needs 2 parameters:
         *  id: the Module id
         *  address: the address (VME, or whatever protocol) to
         *           communicate with the module
         *
         *  Note: although DAQ++ allows for any type of identifier,
         *  however VMEDAQ only handles properly integer IDs. The only
         *  reason for that is that it needs integer values to save them
         *  properly in the data file
         *
         **/
        test_module(const DAQpp::DAQid &id, unsigned long address);
        ~test_module()
        {
        }

        /** This is the method called by the RunManager to know if there
         * is data ready in the module. If the answer if true, then it
         * will try to read the module. If there is no way to know a
         * priory about the availavility of data, one can defer the
         * decision until getData
         *
         **/
        bool hasData();

        /** This is the method called to retrieve the data. If hasData
         * could not figure out hte right answer, one can still say here
         * that there is no dat aat all by returning a NULL pointer and
         * setting size to 0.
         *
         * The memory to store the data should be allocated with the
         * allocate method. See the DAQ++ documentation...
         *
         **/
        char *getData(int &size);

        int PrepareForRun();

        int Start(int = 0);

        int Stop(int = 0);

        int Pause();

        int Continue();

        int Reset();

        int GetReady();

        int Trigger();

        void flush()
        {
        }
        
        /**
         * This returns the data type. The existing data types
         * can be found in DataFormat.h It is a number constructed
         * like: (n & 0xF)<<16;
         */
        unsigned long get_data_type() const
        {
            return 0xd000;
        }
};

#endif
