/**
 This class is responsible to parse the corresponding section in the
 configuration file to configure the module. It is also responsible
 for create the module instance according to the paremeters retrived
 from the configuration file.
 
 It derives from ModuleParser (see ModuleParser.h in the config
 directory) which, in turn, derives from SectionParser (again, have
 a look at the header file).
 
 It has, then, to implement a number of methods from SectionParser
 
 - parse_node:  this method will parse the corresponding section
 and store locally the configuration parameters
 defined in there.
 - save_config: this method will be called if the configuration
 is to be saved to an external file. The idea is
 that the class will check which parameters have
 changed since the file was loaded and update them
 in the document before saving.
 
 and a few more from ModuleParser:
 
 - config : called by the gui to change configuration parameters
 - monitor: called by the gui to create the monitor display
 - decode : called to decode the data
 - range_min, range_max: to retrieve the range of the monitor
 histograms
 
 One can also define here the factory function specified in the
 configuration file.
 
 FOr this module we will assume that we only have 2 parameters to
 appear in the configuration file:
 - the amplitude of the pulse that we simulate and,
 - the width of the peak we want to simulate.
 
 So in th emodules section we add
 
 <modules>
 .
 .
 .
 <test_module id="100" address="deadbeef">
 <!-- monitor is a section common to all the modules -->
 <monitor>
 ...
 </monitor>
 
 <!-- padmap is also common to all the modules -->
 <padmap>
 .....
 </padmap>
 
 <!-- and here come our parameters -->
 <mean>100</mean>
 <width>5</width>
 
 </test_module>
 .
 .
 .
 </modules>
 */
#include <string>
#include <sstream>
#include <vector>
#include <iterator>
#include <DAQ++/DAQmanager.h>
#include <config/ConfigParser.h>
#include <config/ModuleParser.h>
#include <config/GeneralConfig.h>
#include <gtk/gtkwidget.h>
#include <data/Analysis.h>
#include <data/DataFormat.h>

#include <gui/GladeApp.h>
#include <gui/MonitorWindow.h>

#include "test_module.h"

class Analysis;

class test_module_conf : public ModuleParser
{
    private:
        double mean;
        double width;
        double eff;

    public:
        test_module_conf() :
            mean(100.),width(10.0),eff(0.75)
            
        {
        }
        ~test_module_conf()
        {
        }

        /**
         * This is the method responsible for parsing the configuration
         * file. 
         **/
        void parse_node(xmlNodePtr node) throw (ConfigException);

        /**
         * The method that saves the current configuration. It should
         * update all the xml nodes below the module node.
         **/
        void save_config() throw(ConfigException);

        /**
         * This is called by VMEDAQ to create a window that will allow
         * to change the configuration of the module while running the
         * program. Probably the easiest is to use glade-2 to create the
         * layout of the window, define the signals (callbacks that are
         * associated to given actions) and then something derived from
         * GladeApp (see the gui folder) to implement the stuff and
         * update the module state accordingly.
         **/
        GtkWidget *config(const std::string &id);

        /**
         * This is called by VMEDAQ to create a window to display the
         * monitor data. The input parameters are:
         *    - id: the module id (that should help finding the module)
         *    - glade_file: The path of the glade file. This is, in
         *      principle the path of the standard VMEDAQ glade
         *      file. One can certainly ignore this value.
         **/
        MonitorWindow *monitor(const std::string &id, const std::string &glade_file);

        
        // Getters and Setters
        double get_mean() const
        {
            return mean;
        }
        void set_mean(double x)
        {
            mean = x;
        }

        double get_width() const
        {
            return width;
        }
        void set_width(double x)
        {
            width = x;
        }

        double get_efficiency() const
        {
            return eff;
        }
        void set_efficiency(double x)
        {
            eff = x;
        }
};

/**
 * This is the hook specified in the config file to return a pointer
 * to the test_module_conf instance that will interact with VMEDAQ.
 *
 * Note that it should b edeclared as extern "C" so that no demangling
 * mechanism is applied to its name that prevent finding easily the
 * simbol in the library
 *
 **/
extern "C"
{
    test_module_conf *create_test_module_conf()
    {
        test_module_conf *cnf = new test_module_conf();
        return cnf;
    }
}

/**
 * Now come the implementations of the class methods
 */
void test_module_conf::parse_node(xmlNodePtr _node) throw (ConfigException)
{
    unsigned long address;
    std::string saddress;

    // The configuration manager handler
    ConfigParser *cfg = ConfigParser::theConfig();

    // the node name
    name = node_name();

    // store the node pointer
    node = _node;

    // get the module id
    id = cfg->get_attribute<std::string>(node, "name");

    // get the address
    saddress = cfg->get_attribute<std::string>(node, "address");

    std::istringstream os(saddress);
    os >> std::hex >> address;

    // Check if the module is already there
    DAQpp::Module *dm;
    test_module *module = 0;

    if ( (dm = DAQpp::DAQmanager::theDAQmanager()->find_module(id)))
        module = dynamic_cast<test_module *>(dm);

    if (!module)
    {
        // The module is not there... create
        module = new test_module(id, address);
    }

    // Now parse the things which are common to all the modules, that
    // is the monitor and padmap elements
    parse_common("//"+name, id);

    // Now we have the module. From now own one should parse the
    // section from the configuration file
    try
    {
        mean = cfg->get_value<double>(node, "mean");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
        mean = module->get_par_value<double>("mean");
    }

    try
    {
        width = cfg->get_value<double>(node, "width");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
        width = module->get_par_value<double>("width");
    }

    try
    {
        eff = cfg->get_value<double>(node,"efficiency");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
        eff = module->get_par_value<double>("efficiency");
    }

    module->set_par<double>("mean", mean);
    module->set_par<double>("width", width);
    module->set_par<double>("efficiency", eff);

    // test arrays
    try
    {
        std::vector<int> V;
        cfg->get_array(node, "attr_array", V);
        std::cout << "Attr. array: <";
        std::copy(V.begin(), V.end(),
                std::ostream_iterator<int>(std::cout, " "));
        std::cout << ">" << std::endl;
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
    }

}

void test_module_conf::save_config() throw(ConfigException)
{
    // The configuration manager handler
    ConfigParser *cfg = ConfigParser::theConfig();

    // Save common stuff
    save_common();

    // Save the current values in the configuration
    cfg->check_value<double>(node, "mean", get_mean());
    cfg->check_value<double>(node, "width", get_width());
}


class monitor_test_module : public GladeApp
{
    public:
        monitor_test_module()
        {
        }
        static void on_okbutton_clicked(GtkWidget *w, monitor_test_module *m);

        void auto_connect()
        {
            SET_CONNECT(on_okbutton_clicked);
        }

};

void monitor_test_module::on_okbutton_clicked(GtkWidget *w, monitor_test_module *m)
{
    delete m;
}

MonitorWindow * test_module_conf::monitor(const std::string &name,
                                          const std::string &glade_file)
{
    DAQpp::Module *dm = DAQpp::DAQmanager::theDAQmanager()->find_module(name);
    if (!dm)
        return 0;

    test_module *m = 0;
    if ( !(m = dynamic_cast<test_module *>(dm)))
    {
        std::cout << "test_module_conf: module " << name
                << " is not of type test_module " << std::endl;
        return 0;
    }
    monitor_test_module * mon = new monitor_test_module();
    mon->init("test.glade", "monitor");
    MonitorWindow *mw = new MonitorWindow(mon->get_main_window());
    return mw;
}

GtkWidget * test_module_conf::config(const std::string &id)
{
    monitor_test_module * mon = new monitor_test_module();
    mon->init("test.glade", "configuration");
    return mon->get_main_window();
}

static void __attribute__ ((constructor)) __init__lib__()
{
    // init threads
    std::cout << "### Loading test_module..." << std::endl;
}

static void __attribute__ ((destructor)) __close__lib__()
{
    std::cout << "### Unloading test_module..." << std::endl;
}
