#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <sstream>
#include <cstring>
#include <daq/VMEmanager.h>
#include <data/DataFormat.h>
#include "ADCmodule.h"
#include "ADCmodule_analysis.h"
#include "BGOmodule_analysis.h"
#include "ADCdef.h"

//--------------------------------------------------------
// Utils
//--------------------------------------------------------
inline static void set_bit(unsigned int addr, int bit)
{
    unsigned short r =  VMEmanager::VME()->read16(addr);
    r |= (1 << bit);
    VMEmanager::VME()->write16(addr, r);
}
inline static void clear_bit(unsigned int addr, int bit)
{
    unsigned short r =  VMEmanager::VME()->read16(addr);
    r &=  ~(1 << bit);
    VMEmanager::VME()->write16(addr, r);
}
inline static void toggle_bit(unsigned int addr, int bit)
{
    unsigned short r =  VMEmanager::VME()->read16(addr);
    r ^=  (1 << bit);
    VMEmanager::VME()->write16(addr, r);
}


ADCmodule::ADCmodule( const DAQpp::DAQid &id, unsigned long add )
    : GenModule(id, 0), thrs(32,0)
{
    _add = VMEmanager::VME()->map_address(add, 0, 0);
    std::cout << "VME addr. " << std::hex << add << " -> " << _add << std::dec << std::endl;

//    ana = new ADCmodule_analysis( 32, 1 );
    ana = new BGOmodule_analysis( 8, 32 );
    int i;
    std::vector<double> &ped = pedestal->get_value();
    std::vector<double> &ns  = noise->get_value();

    add_par("mesh_path", new DAQpp::Par<std::string>("/home/gamma/Andrej/root_scripts"));
    
    for ( i = 0;i < ana->Ntot();i++ )
    {
    	  ana->Ped()[i]=0.;
    	  ana->Noise()[i]=1.;
        ped.push_back( 0. );
        ns.push_back( 1. );
    }
}

ADCmodule::~ADCmodule()
{
    std::cout << "...destroying Module " << get_id() << std::endl;
}

unsigned long ADCmodule::get_data_type() const
{
    return ADCdata;
}

unsigned int *ADCmodule::write_config(int &size)
{
    size = 1 + // BITSET 2 register
           32 + // mask and threshold
           1;  // slide constant
    
    int ichan,ichip,indx=0;
    unsigned int *data = new unsigned int[size];
    
    data[indx++] = VMEmanager::VME()->read16(_add + BITSET2);
    data[indx++] = VMEmanager::VME()->read16(_add + SLIDE_CNST);
    
    read_threshold();
    for (ichan=0;ichan<32;ichan++)
    {
        data[indx++] = thrs[ichan];
    }
    return data;
}

void ADCmodule::set_thrs_type( ADCThreshold tp )
{
    thrs_type = tp;
    switch (tp) {
        case Low:
            VMEmanager::VME()->write16(_add + BITCLEAR2, 1<<4);
            VMEmanager::VME()->write16(_add + BITCLEAR2, 1<<8);
            break;
        case High:
            VMEmanager::VME()->write16(_add + BITCLEAR2, 1<<4);
            VMEmanager::VME()->write16(_add + BITSET2, 1<<8);
            break;
        case Disable:
            VMEmanager::VME()->write16(_add + BITSET2, 1<<4);
            break;
        default:
            break;
    }
}

void ADCmodule::set_threshold(const std::vector<int> &v)
{
    thrs.clear();
    thrs = v;
    write_threshold();
}

void ADCmodule::write_threshold()
{
    int i;
    std::vector<int>::iterator ip;
    for(i=0, ip=thrs.begin();ip!=thrs.end();++ip,++i)
    {
        VMEmanager::VME()->write16(_add + THRES(i), *ip); 
    }	
}

void ADCmodule::read_threshold()
{
    int i;
    thrs.clear();
    for (i=0; i<32;i++)
    {
        thrs.push_back( VMEmanager::VME()->read16(_add + THRES(i)));
    }
}

int ADCmodule::Reset()
{
    // Issue a reset
    // On
    set_bit(_add + BITSET1, RESET_MODE_ORD);
    // TODO: find out if this is needed: usleep(500000);
    // Off
    set_bit(_add + BITCLEAR1, RESET_MODE_ORD);

    return 0;
}

int ADCmodule::GetReady()
{
    std::cout << "ADCmodule::GetReady" << std::endl;
    Reset();
    set_bit(_add + BITSET2, OFFLINE_ORD);
    //Clear the read/write pointers
    set_bit(_add + BITSET2, CLRDATA_ORD);
    //usleep(500000);
    set_bit(_add + BITCLEAR2, CLRDATA_ORD);

    BGOmodule_analysis *bAna=dynamic_cast<BGOmodule_analysis *>(ana);
    if (bAna) bAna->SetMeshPath(get_par_value<std::string>("mesh_path"));
    
    return 0;
}

int ADCmodule::PrepareForRun()
{
    VMEmanager *vme = VMEmanager::VME();
    //Clear the read/write pointers
    set_bit(_add + BITSET2, CLRDATA_ORD);
    //usleep(500000);
    set_bit(_add + BITCLEAR2, CLRDATA_ORD);
    set_bit(_add + BITCLEAR2, OFFLINE_ORD);
    set_bit(_add + BITCLEAR2, 4);
    set_bit(_add + BITCLEAR2, 8);
    set_bit(_add + BITSET2, AUTO_INCR_ORD);

    // Set the thresholds
    write_threshold();

    // Event counter reset
    vme->write16(_add + EVNT_CNT_RES, 0);
    usleep(1000000); 
    reset_evt_counter();
    return 0;
}

int  ADCmodule::Start (int)
{
    std::cout << "ADCmodule::Start" << std::endl;
    return 0;
};

bool ADCmodule::hasData()
{
    unsigned short st1; // , st2; // not used
    VMEmanager *vme = VMEmanager::VME();
    if (vme->connected())
    {
        st1 = vme->read16(_add + 0x100e);
        //st2 = vme->read16(_add + 0x1022); // unnecessary, slows daq
        return ( (st1&1) != 0 );
    }
    else
    {
        return GetStatus().isPaused() ? false : true;
    }
}

static unsigned int __gate_counter=0;
char *ADCmodule::getData(int &size)
{
    VMEmanager *vme = VMEmanager::VME();
    std::ostringstream ostr;
    int sz, nbuffers=0;
    unsigned int word, header, type;
    int id = get_id();
    char *out = 0;
    
    increase_evt_counter();
    if (vme->connected())
    {
        /*
         *  ADC data block
         *
         *     Data Format
         *     0     0xCAFEtbbb   t = event type, bbb= size of data block
         *     1     0xaaaaaaaa   aaaaaaaa event counter
         *     2     0xiiitnnnn   iii=module id t=data type  nnnn=data size
         *     3     0xnnnnnnnn   number of buffers
         *     4     acdcnnnn     nnnn number of channels in this buffer
         *     5     data word 1
         *           .
         *     n     last data word
         *     n+1   ADC event counter
         *     n+2   acdcnnnn      
         *     .     .
         *    bbbb    bbbb         size of data block
         *
         *    The data words have the format explained in the v785 manual
         *    data type is ADCdata
         */
        
        size = 0;
        do
        {
            header =  vme->read32(_add + OUTPUT_BUFFER);
            type =  (header >> 25) & 3;
            if ( type == 3 )
                return 0;
        } while ( type != 1 );
        
        /*
         * Check and write the event header
         */
        size = 4;
        unsigned int *data = reinterpret_cast<unsigned int *>(allocate(sizeof(unsigned int) * size));
        
        data[0] = DataHeader | DataRecord;
        data[1] = get_evt_counter();
        data[2] = ((id & 0xfff) << 20) | ADCdata;
        data[3] = 0;
        ostr.write(reinterpret_cast<char *>(data), sizeof(unsigned int) * size);
        
        // write the data block
        size = (header >> 8) & 0x3f; // number of channels in event
        word = 0xACDC0000 | size;
        size++; // one more line for event footer
        ostr.write((char *)&word, sizeof(unsigned int));
        deallocate((char *)data);
        data = reinterpret_cast<unsigned int *>(allocate(sizeof(unsigned int) * size));

        for ( int i=0; i<size; i++ ) data[i]=0;

        // read channels and footer
        vme->readA32BLT32(_add + OUTPUT_BUFFER, data, size);

 	word = data[size-1];
        type =  (word >> 24) & 7;
        if (type==4)
            word &= 0xffffff;
        else
            word = 0xffffffff;
        data[size-1] = word;
        ostr.write((char *)data, sizeof(unsigned int)*size);
        deallocate((char *)data);
        
        nbuffers++;
    }
    else
    {
        int i, ibuffer;
        nbuffers = int(randflat(1,4));
     
        __gate_counter++;
        unsigned int dth[4];
        dth[0] = DataHeader | DataRecord;
        dth[1] = get_evt_counter();
        dth[2] = (id & 0xfff) << 20 | ADCdata;
        dth[3] = 0;
        ostr.write(reinterpret_cast<char *>(dth), sizeof(dth));
        for (ibuffer=0;ibuffer<nbuffers;ibuffer++) 
        {
            size = int(randflat(5, 12));
            unsigned int *data = reinterpret_cast<unsigned int *>(allocate(sizeof(unsigned int) * size));

            word = 0xACDC0000 | size;    
            ostr.write((char *)&word, sizeof(unsigned int)); 

            unsigned int mask = 0;
            int chan, val;

            for (i = 0;i < size;i++)
            {
                do
                {
                    chan = int(randflat(0, 31));
                }
                while ( mask & (1 << chan) );
                mask |= (1 << chan);
                val = int(50. + randgauss(250., 5.));
                data[i] = chan << 16 | (val & 0xfff);
            }
            ostr.write((char *)data, size*sizeof(unsigned int));
            word = __gate_counter;
            ostr.write((char *)&word, sizeof(unsigned int));
            deallocate((char *)data);
        }
    }
    // Create space for the very last word
    ostr.write((char *)&word, sizeof(unsigned int));
    
    size = ostr.str().size();
    if ( size > 0 )
    {
        out = allocate( size );
        memcpy(out, ostr.str().c_str(), size);
        
        unsigned int *data = reinterpret_cast<unsigned int *>(out);
        sz = size/sizeof(unsigned int);
        data[0] |= sz & 0xfff;
        data[2] |= (sz-4) & 0xfff;
        data[3] = nbuffers;
        data[sz-1] = sz;
    }
    return  out;
}

void ADCmodule::set_header_description(unsigned int &data)
{
    unsigned int id = get_id();
    data = get_data_type() | (id & 0xffff);
}

void ADCmodule::reset_monitor_data()
{
    GenModule::reset_monitor_data();

    std::vector<double> &ped = pedestal->get_value();
    std::vector<double> &ns  = noise->get_value();
    double *pp = get_analysis()->Ped();
    double *nn = get_analysis()->Noise();
    int i, nt = get_analysis()->Ntot();
    for (i = 0;i < nt;i++)
    {
        pp[i] = ped[i];
        nn[i] = ns[i];
    }
}
