#include <iostream>
#include <sstream>
#include <iterator>
#include <vector>
#include <DAQ++/DAQmanager.h>
#include <config/ConfigParser.h>
#include <config/ModuleParser.h>
#include <config/GeneralConfig.h>
#include <data/Analysis.h>
#include "ADCmodule.h"
#include <gui/MonitorWindow.h>
#include <gui/DefaultMonitorGUI.h>
#include "ADCmoduleGuiConfig.h"
static void __attribute__ ((constructor)) __init__lib__()
{
    printf("### Loading ADCmodule module...\n");
}

static void __attribute__ ((destructor)) __close__lib__()
{
    printf("### Unloading ADCmodule...\n");
}

class ParseADC : public ModuleParser
{
    private:
        std::string name;
    public:
        ParseADC()
        {
        }
        ~ParseADC()
        {
            std::cout << "deleting ParseADC"<< std::endl;
        }

        void parse_node_id(const std::string &) throw(ConfigException);
        void parse_node(xmlNodePtr) throw(ConfigException);

        void save_config() throw(ConfigException);
        void save_config_id(const std::string &) throw(ConfigException);
        void save_config_node(xmlNodePtr) throw(ConfigException);

        _GtkWidget *config(const std::string &);
        MonitorWindow *monitor(const std::string &, const std::string &glade_file);
};

//=====================================================================
// Parsing stuff
//=====================================================================

void ParseADC::parse_node_id(const std::string &node_id) throw(ConfigException)
{
    std::ostringstream ostr;
    ConfigParser *cfg = ConfigParser::theConfig();
    DAQpp::Module *dm = DAQpp::DAQmanager::theDAQmanager()->find_module(node_id);

    if (!dm)
        return;

    if ( !dynamic_cast<ADCmodule *>(dm) )
    {
        std::cout << "ParseADC: module "<< node_id<< " is not of type ADCmodule "
                << std::endl;
        return;
    }

    int nmod;
    std::vector<xmlNodePtr> vmod;
    ostr << "//" << node_name() << "[@name=\""<< node_id << "\"]";
    nmod = cfg->get_node(ostr.str(), vmod);
    if (nmod)
        parse_node(vmod[0]);

}

void ParseADC::parse_node(xmlNodePtr _n) throw(ConfigException)
{
    ConfigParser *cfg = ConfigParser::theConfig();
    unsigned int address;
    std::string saddress;
    std::string mesh_path;
    // we quit if the node is 0
    if (_n == 0)
        return;

    // This should not be needed
    node = _n;
    name = node_name();
    id = cfg->get_attribute<std::string>(node, "name");
    saddress = cfg->get_attribute<std::string>(node, "address");

    std::istringstream os(saddress);
    os >> std::hex >> address;

    // Check if the module already exist
    ADCmodule *m = 0;
    DAQpp::Module *dm;

    if ( (dm = DAQpp::DAQmanager::theDAQmanager()->find_module(id)))
        m = dynamic_cast<ADCmodule *>(dm);

    if (!m) // Create the module
        m = new ADCmodule(id, address);

    // should not happen unless we run out of memory
    if (m == 0)
        return;

    // threshold type
    std::string ttype;
    ADCmodule::ADCThreshold tval;
    try
    {
        ttype = cfg->get_value<std::string>(node, "threshold");
    }
    catch (ConfigException e)
    {
        ttype="disable";
    }
    if (ttype=="disable")
    {
        tval = ADCmodule::Disable;
    }
    else if (ttype=="low")
    {
        tval = ADCmodule::Low;
    }
    else if (ttype=="high")
    {
        tval = ADCmodule::High;
    }
    else
    {
        std::cout << "Module "<< id << std::endl 
                  << "...unknown threshold type " << ttype
                  << std::endl << "...setting default value" << std::endl;
        tval = ADCmodule::Disable;
    }
    m->set_thrs_type(tval);


    try
    {
      mesh_path = cfg->get_value<std::string>(node, "mesh_path");
    }
    catch (ConfigException e)
    {
      mesh_path= m ->get_par_value<std::string>("mesh_path");
    }
    m->set_par<std::string>("mesh_path",mesh_path);
    
    
    // Now we start parsing
    std::ostringstream ostr;
    std::vector<xmlNodePtr> vthrs;
    std::vector<int> v;
    int nthr;

    ostr << "//"<< name << "[@name=\""<< id << "\"]/ADCthrs";
    nthr = cfg->get_node(ostr.str(), vthrs);
    if (nthr > 1)
    {
        std::cout << "ADCParse: more than one threshold set given to module " << id
                << std::endl<< "Only the first one will be considered" << std::endl;
    }
    int i, n;
    if (vthrs.empty())
    {
        std::fill_n(std::back_inserter(v), 32, 0);
    }
    else
    {
        xmlNodePtr nd = vthrs[0];
        cfg->get_array<int>(nd, v);
        n = v.size();
        if (n<32)
            fill_n(back_inserter(v), 32-n, 0);
    }

    std::ostringstream omask;
    std::vector<int> vmask;
    omask << "//"<< name << "[@name=\""<< id << "\"]/ADCmask";
    vthrs.clear();
    nthr = cfg->get_node(omask.str(), vthrs);
    if (vthrs.empty())
    {
        vmask.clear();
        fill_n(back_inserter(vmask), 32, 0);
    }
    else
    {
        if (nthr>1)
        {
            std::cout << "ADCParse: more than one mask set given to module " << id
                    << std::endl << "Only the first one will be considered" << std::endl;
        }
        xmlNodePtr nd = vthrs[0];
        if (cfg->get_bit_array(nd, 32, vmask) )
        {
            std::cout << "Module "<< id << std::endl << "... wrong syntax in mask"
                    << std::endl;
            vmask.clear();
            std::fill_n(std::back_inserter(vmask), 32, 0);
        }
        else
        {
            n = vmask.size();
            if (n<32)
                std::fill_n(std::back_inserter(vmask), 32-n, 0);
        }
    }
    for (i=0; i<32; i++)
    {
        if (vmask[i])
        {
            v[i] |= 0x100;
        }
    }

    m->set_threshold(v);

    // range
    parse_common("//"+name, id);

    // Make sure we have default pedestals and noise
    m->copy_pedestals();
}

void ParseADC::save_config() throw(ConfigException)
{
    std::vector<xmlNodePtr> vmod;
    std::vector<xmlNodePtr>::iterator ip;
    ConfigParser *cfg = ConfigParser::theConfig();
    
    cfg->get_node("//"+name, vmod);
    for (ip = vmod.begin(); ip != vmod.end(); ip++)
        save_config_node( *ip);
}

void ParseADC::save_config_id(const std::string &name) throw(ConfigException)
{
    std::ostringstream ostr;
    ConfigParser *cfg = ConfigParser::theConfig();
    DAQpp::Module *dm = DAQpp::DAQmanager::theDAQmanager()->find_module(name);

    if (!dm)
        return;

    if ( !dynamic_cast<ADCmodule *>(dm) )
    {
        std::cout << "ParseADC: module "<< name<< " is not of type ADCodule "
                << std::endl;
        return;
    }

    int nmod;
    std::vector<xmlNodePtr> vmod;
    ostr << "//"+name+"[@name=\""<< name.c_str() << "\"]";
    nmod = cfg->get_node(ostr.str(), vmod);
    if (nmod)
        save_config_node(vmod[0]);

}

void ParseADC::save_config_node(xmlNodePtr _n) throw(ConfigException)
{
    ConfigParser *cfg = ConfigParser::theConfig();

    node = _n;
    id = cfg->get_attribute<std::string>(node, "name");
    ADCmodule *m = 0;
    DAQpp::Module *dm;

    // find the module
    if ( !(dm = DAQpp::DAQmanager::theDAQmanager()->find_module(id)))
        return;

    if ( !(m = dynamic_cast<ADCmodule *>(dm)))
        return;

  
    std::ostringstream othrs, omsk;
    std::vector<int> thrs;
    std::vector<bool> msk;
    int i;
    for (i=0; i<32; i++)
    {
        thrs.push_back(m->get_thres(i));
        msk.push_back(m->get_mask(i));
    }
    std::copy(thrs.begin(), thrs.end(), std::ostream_iterator<int>(othrs, " "));
    ADCmodule::ADCThreshold tval = m->get_thrs_type();
    std::string ttype="disable";
    switch (tval)
    {
        case ADCmodule::Disable:
            ttype = "disable";
            break;
        case ADCmodule::Low:
            ttype="low";
            break;
        case ADCmodule::High:
            ttype="high";
            break;
        default:
            break;
    }
    cfg->check_value(node, "threshold", ttype);
    cfg->set_value<std::string>(node, "ADCthrs", othrs.str());

    std::string mesh_path=m->get_par_value<std::string>("mesh_path");
    cfg->check_value<std::string>(node,"mesh_path",mesh_path);
    // Mask

    int ibit,jbit,mask;
    for (ibit = 0; ibit < 32; ibit += 4)
    {
        if (ibit && !(ibit % 16))
        omsk << " ";
        for (mask = jbit = 0; jbit < 4; jbit++)
        {
            if (msk[ibit + jbit] )
            mask |= (1 << jbit);
        }
        omsk << std::setw(1) << std::hex << mask << std::dec;
    }

    cfg->set_value<std::string>(node, "ADCmask", omsk.str());

    save_common();
}


_GtkWidget *ParseADC::config(const std::string &name)
{
    DAQpp::Module *dm = DAQpp::DAQmanager::theDAQmanager()->find_module(name);
    if (!dm)
        return 0;

    ADCmodule *m = 0;
    if ( !(m = dynamic_cast<ADCmodule *>(dm)))
    {
        std::cout << "ParseADC: module "<< name<< " is not of type ADCmodule "
                << std::endl;
        return 0;
    }
    ADCmoduleGuiConfig *df = new ADCmoduleGuiConfig(m->get_id(), m->get_thrs_type(), m->get_threshold());

    df->init("vmedaq.glade", "ADC-config");
    return df->get_main_window();
}

class ADCMonitorWindow : public MonitorWindow
{
    private:
        DefaultMonitorGUI *_mon;
    public:
        ADCMonitorWindow(const std::string &name, const std::string &glade_file,
                         ADCmodule *m)
        {
            _mon = new DefaultMonitorGUI( name,
                    glade_file,
                    m->get_analysis());
            _mon->init(glade_file.c_str(), "dsp-monitor");
            _widget = _mon->get_main_window();
        }
        ~ADCMonitorWindow()
        {
            delete _mon;
        }

        DefaultMonitorGUI *mon()
        {
            return _mon;
        }
};
MonitorWindow *ParseADC::monitor(const std::string &name, const std::string &glade_file)
{
    DAQpp::Module *dm = DAQpp::DAQmanager::theDAQmanager()->find_module(name);
    if (!dm)
        return 0;

    ADCmodule *m = 0;
    if ( !(m = dynamic_cast<ADCmodule *>(dm)))
    {
        std::cout << "ParseADC: module "<< name<< " is not of type ADCmodule "
                << std::endl;
        return 0;
    }

    ADCMonitorWindow *mw = new ADCMonitorWindow(name, glade_file, m);

    mw->add_update(new UpdateLabelNtrig(mw->mon()->get_widget("lblDSPnEvts")));
    if (mw->mon()->get_widget("gauge_chip_rate") )
        mw->add_update(new UpdateGaugeRate(mw->mon()->get_widget("gauge_chip_rate") ));

    if (mw->mon()->get_widget("gauge_chip_eff") )
        mw->add_update(new UpdateGaugeEff( mw->mon()->get_widget("gauge_chip_eff")));

    return mw;

}

extern "C"
{
    ModuleParser *create_ADCmodule()
    {
        ParseADC *gp = new ParseADC();
        return gp;
    }
}

