#include <iostream>
#include <sstream>
#include <ctime>
#include <cmath>
#include <cstring>
#include <cstdio>
#include <unistd.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <daq/gpib-utils.h>
#include <daq/ScanRunManager.h>
#include "AgilentPulseGenerator.h"
#include <cstdio>

static void __attribute__ ((constructor)) __init__lib__()
{
    printf("### Loading Agilent module (%s %s)...\n",__DATE__, __TIME__);
}

static void __attribute__ ((destructor)) __close__lib__()
{
    printf("### Unloading Agilent...\n");
}

AgilentPulseGenerator::AgilentPulseGenerator(int id)
        : dev(-1)
{
    SetID(id);
//   GPIB_debug=1;
}

AgilentPulseGenerator::~AgilentPulseGenerator()
{}

int  AgilentPulseGenerator::SetID(int gid)
{
#if defined(HAS_GPIB)
    char line[128];
    if ( gid <= 0 )
    {
        dev = -1;
        return -1;
    }
    dev =  ibdev(0, gid, 0, T10s, 1, 1);
    if ( ibsta & ERR )
    {
        dev = -1;
        return -1;
    }
    if ( GPIBWriteRead(dev, 10, line, sizeof(line), "*IDN?") )
    {
        std::cout << "AgilentPulseGenerator::" << __FUNCTION__
                  << " error" << std::endl;
        return -1;
    }
    std::cout << line << std::endl;

    /* Clear the error queue */
    int i;
    for (i = 0;i < 20;i++)
    {
        if ( GPIBWriteRead(dev, 10, line, sizeof(line), "SYSTEM:ERRor?") )
        {
            std::cout << "AgilentPulseGenerator::get_error error" << std::endl;
            break;
        }
        std::istringstream istr(line);
        istr >> status;
        if (status == 0)
            break;
    }
#endif
    return 0;
}

bool AgilentPulseGenerator::check_status()
{
#if defined(HAS_GPIB)
    char line[256];
    std::string last_command = GPIBlastCommad;
    if ( GPIBWriteRead(dev, 10, line, sizeof(line), "SYSTEM:ERRor?") )
    {
        std::cout << "AgilentPulseGenerator::check_status error" << std::endl;
        GPIBlastCommad = last_command;
        return false;
    }
    GPIBlastCommad = last_command;
    std::istringstream istr(line);
    istr >> status;
    if (status)
    {
        char *p = strrchr(line, '\n');
        if (p)
            *p = 0;
        error_string = line;
        return false;
    }
    else
    {
        error_string.clear();
        return true;
    }
#else
    return true;
#endif
}

void AgilentPulseGenerator::Reset()
{
    if (dev < 0)
        return;

#if defined(HAS_GPIB)
    GPIBWrite(dev, "*CLS;*RST");
    if (!check_status())
    {
        std::cout << "AgilentPulseGenerator::Reset: bad status ("
                  << get_error() << ")" << std::endl
                  << "*** " << GPIBlastCommad
                  << std::endl;
    }
#endif
}

void AgilentPulseGenerator::set_pulse(double x)
{
    if (dev < 0)
        return;

    std::ostringstream ocmd;
    // Output pulse is AB
    if (fabs(x)>1.e-3)
    	ocmd << "VOLTAGE "<< x ;
    else
    	ocmd << "VOLTAGE MIN";

#if defined(HAS_GPIB)
    if ( GPIBWrite(dev, ocmd.str().c_str()) )
    {
        std::cout << "AgilentPulseGenerator::set_pulse error" << std::endl;
    }

    if (!check_status())
    {
        std::cout << "AgilentPulseGenerator::set_pulse: bad status ("
                  << get_error() << ")" << std::endl
                  << "*** " << GPIBlastCommad
                  << std::endl;
    }
#endif
}

void  AgilentPulseGenerator::define_pulse(double period, double width, double edge)
{
    const int ncmd = 4;
    std::ostringstream ocmd[ncmd];
    ocmd[0] << "PULse:PERiod " << period;
    ocmd[1] << "PULse:WIDTh  " << width;
    ocmd[2] << "PULse:TRANsition " << edge;
    ocmd[3] << "FUNCtion PULse";

#if defined(HAS_GPIB)
    int icmd;
    for (icmd = 0;icmd < ncmd;icmd++)
    {
        if ( GPIBWrite(dev, ocmd[icmd].str().c_str()) )
        {
            std::cout << "AgilentPulseGenerator::define_pulse error" << std::endl;
        }
        if (!check_status())
        {
            std::cout << "AgilentPulseGenerator::define_pulse : bad status ("
                      << get_error() << ")" << std::endl
                      << "*** " << GPIBlastCommad << std::endl;
        }
    }
#endif
}


void AgilentPulseGenerator::set_trigger_type(TriggerType m)
{
    if ( dev < 0 ) return;

    std::ostringstream ocmd;
    ocmd << "TRIGger:SOURce ";
    switch (m)
    {
    case Immediate:
        ocmd << "IMMediate";
        break;
    case External:
        ocmd << "EXTernal";
        break;
    case Bus:
        ocmd << "BUS";
        break;
    }

#if defined(HAS_GPIB)
    if ( GPIBWrite(dev, ocmd.str().c_str()) )
    {
        std::cout << "AgilentPulseGenerator::set_trigger_type error" << std::endl;
    }
    if (!check_status())
    {
        std::cout << "AgilentPulseGenerator::set_trigger_type - bad status ("
                  << get_error() << ")" << std::endl
                  << "*** " << GPIBlastCommad
                  << std::endl;
    }
#endif
}

void AgilentPulseGenerator::ConfigureOutput( double imp)
{
    if ( dev < 0 ) return;
    std::ostringstream ocmd;

    ocmd << "OUTPut:LOAD " << imp;

#if defined(HAS_GPIB)
    if ( GPIBWrite(dev, ocmd.str().c_str()) )
    {
        std::cout << "AgilentPulseGenerator::ConfigureOutput error" << std::endl;
    }
    if (!check_status())
    {
        std::cout << "AgilentPulseGenerator::ConfigureOutput - bad status ("
                  << get_error() << ")" << std::endl
                  << "*** " << GPIBlastCommad
                  << std::endl;
    }
#endif
}


void AgilentPulseGenerator::set_frequency(double x)
{
    if ( dev < 0 ) return;

    std::ostringstream ocmd;
    ocmd << "FREQuency " << x;

#if defined(HAS_GPIB)
    if ( GPIBWrite(dev, ocmd.str().c_str()) )
    {
        std::cout << "AgilentPulseGenerator::set_frequency error" << std::endl;
    }
    if (!check_status())
    {
        std::cout << "AgilentPulseGenerator::set_frequency - bad status ("
                  << get_error() << ")" << std::endl
                  << "*** " << GPIBlastCommad
                  << std::endl;
    }
#endif
}

void AgilentPulseGenerator::activate_burst(bool z)
{
	burst = true;
    if (dev < 0)
        return;

    std::ostringstream ocmd;

    ocmd << "BURSt:STATe " << ( z ? "ON" : "OFF");

#if defined(HAS_GPIB)
    if ( GPIBWrite(dev, ocmd.str().c_str()) )
    {
        std::cout << "AgilentPulseGenerator::activate_burst error" << std::endl;
    }
    if (!check_status())
    {
        std::cout << "AgilentPulseGenerator::activate_burst - bad status ("
                  << get_error() << ")" << std::endl
                  << "*** " << GPIBlastCommad
                  << std::endl;
    }
#endif
}

/**
 * delay: delay rom trigger to start of burst
 * freq:  frequency. Distance among pulses in the burst
 * npulse: number of pulses in the burst
 * period: width of the burst
 */
void AgilentPulseGenerator::define_burst(double delay, double freq, int npulse, double period)
{
    if ( dev < 0 )
        return;

    double raise_time = 5.e-9;
    double pulse_period = period/npulse;
    double pulse_width = 25.0e-6;
    if (pulse_period < pulse_width + 4.*raise_time)
    {
    	pulse_width = pulse_period < 4.*raise_time;
    	if (pulse_width<0.)
    	{
    		std::cout << "AgilentPulseGenerator::define_burst: "
    		          << "pulse width is negative. Check the values of the burst"
    		          << std::endl;
    		pulse_width = pulse_period/2.;
    	}
    }
    define_pulse(pulse_period, pulse_width, raise_time);
    const int ncmd = 5;
    std::ostringstream ocmd[ncmd];

    ocmd[0] << "FREQuency "      << freq;
    ocmd[1] << "BURSt:MODE TRIGgered";
    ocmd[2] << "BURSt:NCYCles " << npulse;

    // The burst period has to be long enough
    // period > (burst count)/freq + 200 ns
    //
    ocmd[3] << "BURSt:INTernal:PERiod " << period+delay+0.5;

    set_trigger_type(Immediate);
    set_delay(delay);

#if defined(HAS_GPIB)
    int icmd;
    for (icmd = 0;icmd < ncmd;icmd++)
    {
        if ( GPIBWrite(dev, ocmd[icmd].str().c_str()) )
        {
            std::cout << "AgilentPulseGenerator::define_burst error" << std::endl;
        }
        if (!check_status())
        {
            std::cout << "AgilentPulseGenerator::define_burst - bad status ("
                      << get_error() << ")" << std::endl
                      << "*** " << GPIBlastCommad
                      << std::endl;
        }
    }
#endif
}


void AgilentPulseGenerator::trigger()
{
	if (is_burst())
		return;

#if defined(HAS_GPIB)
    if ( GPIBWriteFast(dev, "*TRG") )
    {
        std::cout << "AgilentPulseGenerator::trigger error" << std::endl;
    }
    if (!check_status())
    {
        std::cout << "AgilentPulseGenerator::trigger - bad status ("
                  << get_error() << ")" << std::endl
                  << "*** " << GPIBlastCommad
                  << std::endl;
    }
#endif
}

void AgilentPulseGenerator::set_delay(double x)
{
    if (dev < 0)
        return;

    std::ostringstream ocmd;

    ocmd << "TRIGger:DELay " << x;

#if defined(HAS_GPIB)
    if ( GPIBWrite(dev, ocmd.str().c_str()) )
    {
        std::cout << "AgilentPulseGenerator::set_delay error" << std::endl;
    }
    if (!check_status())
    {
        std::cout << "AgilentPulseGenerator::set_delay - bad status ("
                  << get_error() << ")" << std::endl
                  << "*** " << GPIBlastCommad
                  << std::endl;
    }
#endif
}

void AgilentPulseGenerator::set_output(bool z)
{
    if (dev < 0)
        return;

#if defined(HAS_GPIB)
    std::ostringstream ocmd;

    ocmd << "OUTPut " << ( z ? "ON" : "OFF");

    if ( GPIBWrite(dev, ocmd.str().c_str()) )
    {
        std::cout << "AgilentPulseGenerator::set_output error" << std::endl;
    }
    if (!check_status())
    {
        std::cout << "AgilentPulseGenerator::set_output - bad status ("
                  << get_error() << ")" << std::endl
                  << "*** " << GPIBlastCommad
                  << std::endl;
    }
#endif
}

class AgilentPulseRunManager : public PulseRunManager
{
public:
    AgilentPulseRunManager(DAQpp::DAQid id, PulseGenerator *pg)
        : PulseRunManager(id, pg)
    {}
};

/*
 * The hooks
 */
extern "C"
{
    PulseRunManager *create_pulse_run_manager(const std::string &type,
                                               const std::string &name)
    {
        PulseGenerator *pg = new AgilentPulseGenerator(10);
        return new PulseRunManager(name, pg);
    }

    ScanRunManager *create_scan_run_manager(const std::string &type,
                                            const std::string &name)
    {
        ScanRunManager *rm = new ScanRunManager(name);
        PulseGenerator *pg = new AgilentPulseGenerator(10);
        rm->set_pulse_generator(pg);
        return rm;
    }
}


