#include <iostream>
#include <fstream>
#include <list>
#include <cmath>
#include "BGOmap.h"

using namespace std;

double operator*(const BGOmap::Point& a, const BGOmap::Point& b){
  return a.x*b.x+a.y*b.y;
}

BGOmap::Point operator-(const BGOmap::Point& a, const BGOmap::Point& b){
  BGOmap::Point r;
  r.x=a.x-b.x;
  r.y=a.y-b.y;
  return r;
}

BGOmap::Point operator+(const BGOmap::Point& a, const BGOmap::Point& b){
  BGOmap::Point r;
  r.x=a.x+b.x;
  r.y=a.y+b.y;
  return r;
}  

BGOmap::pointMin::pointMin(){
  for (int i=0;i<9;i++) v[i]=-1;
}

const int BGOmap::nX;
const int BGOmap::nY;
const double BGOmap::xMin;
const double BGOmap::xMax;
const double BGOmap::yMin;
const double BGOmap::yMax;

void BGOmap::GetUV(int i, int &u, int & v){
  u=i%(nX+1);
  v=i/(nX+1);
}

int BGOmap::GetI(int u, int v){
  if (u<0) return -1;
  if (v<0) return -1;
  if (u>nX) return -1;
  if (v>nY) return -1;
  return u+v*(nX+1);
}

void BGOmap::parseBoundaries(const char *fname){
  if (!fname) return;
  bMap.clear();
  int Ncol=nX+1;
  int Nrow=nY+1;
  int Npt=Ncol*Nrow;
  bMap.resize(Npt);
  ifstream inf(fname);
  if (!inf){
    //cout<<"Could not open file "<<fname<<endl;
    return;
  }
  for (vector<Point>::iterator it=bMap.begin();it!=bMap.end();it++){
    //cout<<">>>"<<endl;
    int i=distance(bMap.begin(),it);
    //cout<<"At "<<i<<endl;
    inf>>(*it).x>>(*it).y;
    if (!inf){
      cout<<"Troubles reading point "<<i<<endl;
      return;
    }
  }
  cout<<"Parsed "<<bMap.size()<<" intersticies"<<endl;
  map_set=1;
}

int BGOmap::getNearestPoint(const BGOmap::Point& p){
  if (p.x<xMin || p.x>xMax || p.y<yMin || p.y>yMax){
    //cout<<"Bad point "<<p.x<<','<<p.y<<endl;
    return -1;
  }

  int u=static_cast<int>(round((p.x-xMin)/(xMax-xMin)*nX));
  int v=static_cast<int>(round((p.y-yMin)/(yMax-yMin)*nY));
  // in non-skewed frame the nearest point is
  // (xMin+u(xMax-xMin)/nX,yMin+v*(yMax-yMin)/nY)
  
  //u runs from 0 to (inclusive!) nX, v runs from 0 to nY
  
  int icentral=v*(nX+1)+u;
  //cout<<"Guess x="<<xMin+u*(xMax-xMin)/nX<<" y="<<yMin+v*(yMax-yMin)/nY<<endl;

  //guess that iPt is closest to P
  pointMin pm;
  pm.p=p;
  while (1){
    int imin;
    if (check_that_central_is_min(pm, icentral,imin)) break;
    shift_to_min(pm,icentral, imin);
    icentral=imin;
  }
  return icentral;
  
}

void get_indices(int i, int& ix, int &iy){
  ix=iy=0;
  if (!i) return;

  if (!(i%4)) iy=0;
  else iy=2*(i/4)-1;
  int ip=(i+2)%8;
  if (!(ip%4)) ix=0;
  else ix=2*(ip/4)-1;
}

bool BGOmap::check_that_central_is_min(BGOmap::pointMin& pm, int icentral, int& imin) {
  int u,v;
  GetUV(icentral,u,v);
  
  imin=icentral;
  double fMin;
  for (int i=0;i<9;i++){
    //calculate offset from central point
    int iX,iY;
    get_indices(i,iX,iY);
    if (pm.v[i]<0){
      //calculate
      int u1=u+iX;
      int v1=v+iY;
      int i1=GetI(u1,v1);
      if (i1<0) pm.v[i]=1e30;
      else
	pm.v[i]=(bMap[i1]-pm.p)*(bMap[i1]-pm.p);
      pm.idx[i]=i1;
      //cout<<i<<' '<<iX<<' '<<iY<<" At ("<<u1<<','<<v1<<';'<<i1<<") x="<<bMap[i1].x<<" y= "<<bMap[i1].y<<endl;
      
    }
    if (i==0) fMin=pm.v[0];
    if (i && pm.v[i]<fMin) {
      imin=pm.idx[i];
      pm.imin=i;
      fMin=pm.v[i];
    }
    //cout<<"\t v="<<pm.v[i]<<" imin="<<imin<<endl;
  }
  return (imin==icentral);
}


void BGOmap::shift_to_min(BGOmap::pointMin& pm, int icentral, int imin) {
  double vtemp[9];
  for (int i=0;i<9;i++) {
    vtemp[i]=pm.v[i];
    pm.v[i]=-1;
  }
  pm.v[0]=vtemp[pm.imin];
  switch (pm.imin){
  case 0:
    cout<<"This shouldn't happen"<<endl;
    return;
  case 1: 
    pm.v[4]=vtemp[2];
    pm.v[5]=vtemp[0];
    pm.v[6]=vtemp[8];
    return;
  case 2:
    pm.v[4]=vtemp[3];
    pm.v[5]=vtemp[4];
    pm.v[6]=vtemp[0];
    pm.v[7]=vtemp[8];
    pm.v[8]=vtemp[1];
    return;
  case 3:
    pm.v[6]=vtemp[4];
    pm.v[7]=vtemp[0];
    pm.v[8]=vtemp[2];
    return;
  case 4:
    pm.v[1]=vtemp[2];
    pm.v[2]=vtemp[3];
    pm.v[6]=vtemp[5];
    pm.v[7]=vtemp[6];
    pm.v[8]=vtemp[0];
    return;
  case 5:
    pm.v[1]=vtemp[0];
    pm.v[2]=vtemp[4];
    pm.v[8]=vtemp[6];
    return;
  case 6:
    pm.v[1]=vtemp[8];
    pm.v[2]=vtemp[0];
    pm.v[3]=vtemp[4];
    pm.v[4]=vtemp[5];
    pm.v[8]=vtemp[7];
    return;
  case 7:
    pm.v[2]=vtemp[8];
    pm.v[3]=vtemp[0];
    pm.v[4]=vtemp[6];
    return;
  case 8:
    pm.v[2]=vtemp[1];
    pm.v[3]=vtemp[2];
    pm.v[4]=vtemp[0];
    pm.v[5]=vtemp[6];
    pm.v[6]=vtemp[7];
    return;
  }
}

bool BGOmap::in_quadrant(const BGOmap::Point& x, const BGOmap::Triangle& t){
  Point v1=x-t.A;
  Point b=t.B-t.A;
  Point c=t.C-t.A;

  double c2=c*c;
  double b2=b*b;
  double bc=b*c;
  double v1b=v1*b;
  double v1c=v1*c;
  
  //we are only checking for sign and c2*b2-bc*bc is always positive
  //(dot product of c cross b with itself)
  double s=1;//c2*b2-bc*bc; 
  

  //determine components of v1 along b and c
  
  double u=(c2*v1b-bc*v1c)/s;
  double v=(b2*v1c-bc*v1b)/s;

  if (u<0) return false;
  if (v<0) return false;
  return true;
}

int BGOmap::getCrystalID(const Point& p){
  int ipt=getMeshLowLeftCornerID(p);
  int u,v;
  GetUV(ipt,u,v);
  return v*nX+u;
}

int BGOmap::getMeshLowLeftCornerID(const Point& p){
  int iPt=getNearestPoint(p);
  if (iPt<0)
    return -1;
  
  int u,v;
  GetUV(iPt,u,v);
  
  
  for (int i=0;i<4;i++){
    int ix1=((i+1)%2)*(2*(i/2)-1);
    int iy1=(i%2)*(2*(i/2)-1);
    
    int ix2=(i%2)*(2*(((i+1)%4)/2)-1);
    int iy2=((i+1)%2)*(2*(i/2)-1);
    
    Triangle t;
    t.A=bMap[iPt];
    int i1=GetI(u+ix1,v+iy1);
    if (i1<0) continue;
    t.B=bMap[i1];
    int i2=GetI(u+ix2,v+iy2);
    if (i2<0) continue;
    t.C=bMap[i2];
    if (in_quadrant(p,t)){
      if (ix1<0 || ix2<0) u--;
      if (iy1<0 || iy2<0) v--;
      break;
    }
  }

  int ip1=GetI(u,v);
  //cout<<"u="<<u<<" v="<<v<<" A("<<bMap[ip1].x<<','<<bMap[ip1].y<<')'<<endl;
  return ip1;
}
  
void BGOmap::getFinePosition(const Point& p){
  int ip=getCrystalID(p);
  if (ip<0) return;
  Quadrangle qA;


  Point* ap[4];
  ap[0]=&qA.A;
  ap[1]=&qA.B;
  ap[2]=&qA.C;
  ap[3]=&qA.D;

  int u,v;
  GetUV(ip,u,v);

  for (int i=0;i<4;i++){
    int ix=((i+1)%4)/2;
    int iy=i/2;
    int ip1=GetI(u+ix,v+iy);
    if (ip1<0) {
      cout<<"Troubles making quadrangle, setting u,v to "<<u+ix<<','<<v+iy<<endl;
      return;
    }
    (*(ap[i]))=bMap[ip1];
  }

  
  Point b=qA.B-qA.A;
  Point cd=qA.C+qA.A-qA.B-qA.D;
  Point d=qA.D-qA.A;
  Point pa=p-qA.A;

  double b2=(b*b);
  double d2=(d*d);

  double y=(pa*d)/d2;
  double x=(pa*b)/b2;
  
  double delta=(d*cd)/d2;
  double beta=(b*cd)/b2;

  double eb=(b*d)/b2;
  double ed=eb*b2/d2;
  
  double aq=beta-delta*eb;
  double bq=1-ed*eb-beta*y+delta*x;
  double cq=-(y-ed*x);

  double q=0,lam=0;
  if (beta!=0){
    //y equivalent; should be between 0 and 1
    q=(sqrt(bq*bq-4*aq*cq)-bq)/(2*aq); 
    //x equivalent; same range
    lam=(beta*y-delta*x-q*aq)/(beta*ed-delta);
  }
  else{
    q=(y-ed*(x-eb))/(1-delta*(x-eb));
    lam=x-eb;
  }
//   cout<<"Doing "<<p.x<<' '<<p.y<<endl;
//   cout<<"A("<<qA.A.x<<','<<qA.A.y
//       <<") B("<<qA.B.x<<','<<qA.B.y
//       <<") C("<<qA.C.x<<','<<qA.C.y
//       <<") D("<<qA.D.x<<','<<qA.D.y<<')'<<endl;
//   cout<<"x="<<x<<" y="<<y<<" beta="<<beta<<" delta="<<delta
//       <<" e_b="<<eb<<" e_d="<<ed
//       <<"\n\ta(eq)="<<aq<<" b(eq)="<<bq<<" c(eq)="<<cq
//       <<" sqrt(D)="<<sqrt(bq*bq-4*aq*cq)
//       <<"\nlamda="<<lam<<" q="<<q<<endl;

//   cout<<"Non-sqewed position of the quadrangle("<<ip<<"):"<<endl;

//   cout<<"\tA("<<xMin+u*(xMax-xMin)/nX<<','<<yMin+v*(yMax-yMin)/nY<<") "
//       <<"B("<<xMin+(u+1)*(xMax-xMin)/nX<<','<<yMin+v*(yMax-yMin)/nY<<") "
//       <<"C("<<xMin+(u+1)*(xMax-xMin)/nX<<','<<yMin+(v+1)*(yMax-yMin)/nY<<") "
//       <<"D("<<xMin+u*(xMax-xMin)/nX<<','<<yMin+(v+1)*(yMax-yMin)/nY<<") "<<endl;
//   cout<<"Corrected position"
//       <<"\n\tx="<<xMin+(u+lam)*(xMax-xMin)/nX<<" y="<<yMin+(v+q)*(yMax-yMin)/nY<<endl;
  

}
