#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <cstring>
#include <data/DataFormat.h>
#include "CSprint.h"
#include "CSprint_analysis.h"
#define Port 0



CSPRINT::CSPRINT( const DAQpp::DAQid &id, const char *dev )
    : GenModule(id, 0), usb(dev)
{
    ana = new CSprint_analysis( NCHAN, 1 );
    int i;
    std::vector<double> &ped = pedestal->get_value();
    std::vector<double> &ns  = noise->get_value();

    for ( i = 0;i < ana->Ntot();i++ )
    {
        ped.push_back( 0. );
        ns.push_back( 1. );
    }

    Reset();
}

CSPRINT::~CSPRINT()
{
    std::cout << "...destroying Module " << get_id() << std::endl;
}

unsigned long CSPRINT::get_data_type() const
{
    return CSPRINTdata;
}

int CSPRINT::Reset()
{
    // Issue a reset
    usb.fSioReset(Port, 0, 0);

    return 0;
}

void CSPRINT::flush()
{
    while ( usb.fSioGetc(Port) < 256 );
}

int CSPRINT::GetReady()
{
    std::cout << "GetReady" << std::endl;
    return 0;
}

int CSPRINT::PrepareForRun()
{
    // enable readout
    usb.fSioPutc(Port, 0x29);
    return 0;
}

int  CSPRINT::Start (int)
{
    std::cout << "CSPRINT::Start" << std::endl;
    // Arm -> get ready to receive trigger
    usb.fSioPutc(Port, 0x27);
    return 0;
};
int  CSPRINT::Pause ()
{
    // what is the opposite of 0x27
    return 0;
}
bool CSPRINT::hasData()
{
    int rc = usb.fSioEvent(Port, 0);
    return rc != 0;
}

char *CSPRINT::getData(int &size)
{
    unsigned int i;
    unsigned int recordLength;
    unsigned int resp;
    int notSynched;
    int synchMask[] = {'C', 'S', 'P', 'T'};

    union {
        ADCRecord adc;
        char data[sizeof(ADCRecord)];
    } myRecord;

    // We should be here only if there is anything to read
    // so the below is unneeded
    // usb.fSioEvent(Port, EV_RXCHAR);

    increase_evt_counter();

    // read ID string, version, record length, synchronize
    notSynched = 1;
    i = 0;
    while (notSynched)
    {
        resp = usb.fSioGetc(Port);
        if (resp > 256)
        {
            usb.fSioEvent(Port, EV_RXCHAR);
        }
        else if (resp == synchMask[i])
        {
            myRecord.data[i] = resp;
            i++;
            if (i == 4)
                notSynched = 0;
        }
        else
        {
            i = 0;
        }
    }

    // read version number
    resp = usb.fSioGetc(Port);
    if (resp > 256)
    {
        usb.fSioEvent(Port, EV_RXCHAR);
        resp = usb.fSioGetc(Port);
    }
    myRecord.data[4] = resp;


    // read record length
    resp = usb.fSioGetc(Port);
    if (resp > 256)
    {
        usb.fSioEvent(Port, EV_RXCHAR);
        resp = usb.fSioGetc(Port);
    }
    myRecord.data[5] = resp;

    recordLength = (unsigned int)myRecord.data[5];
    //  std::cout << dec << "Recordlength " << recordLength << std::endl;

    for (i = 6;i < recordLength;i++)
    {
        resp = usb.fSioGetc(Port);
        if (resp > 256)
        {
            usb.fSioEvent(Port, EV_RXCHAR);
            sleep(2);
            resp = usb.fSioGetc(Port);
        }
        myRecord.data[i] = resp;
    }

    /*
      Now prepare the data for the run manager
      Data Format
      0     0xCAFE1bbb   bbb= size of data block
      1     0xaaaaaaaa   aaaaaaaa event counter
      2     0xiiitnnnn   iii  = module id
      t    = data type
      a    = readout mode
      nnnn = data size
      3     data word 0
      4     data word 1
      .     .
      bbbb    bbbb         size of data block
    */
    size = sizeof(ADCRecord) / sizeof(unsigned int);
    int sz = 4 + size;
    unsigned int *data = reinterpret_cast<unsigned int *>(
        allocate(sizeof(unsigned int) * sz));

    int id = get_id();

    data[0] = DataHeader | sz;
    data[1] = get_evt_counter();
    data[2] = (id & 0xfff) << 20 | CSPRINTdata | (size & 0xfff);
    memcpy(data + 3, myRecord.data, sizeof(ADCRecord));
    data[sz-1] = sz;

    size = sz * sizeof(unsigned int);
    return reinterpret_cast<char *>(data);
}

void CSPRINT::set_header_description(unsigned int &data)
{
    unsigned int id = get_id();
    data = (3 & 0xf) << 16 | (id & 0xffff);
}

