#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <sstream>
#include <cstring>
#include <data/DataFormat.h>
#include <daq/VMEmanager.h>
#include "FBTDCdef.h"
#include "FBTDCmodule.h"
#include "FBTDCmodule_analysis.h"

//--------------------------------------------------------
// Utils
//--------------------------------------------------------
#ifdef HAS_SIS1100

/*
  The following are a subset of FRITZ Fastbus routines modified to be used with the
  SIS optical link. Note that this does not work with the NI VXI link since A32 D32
  VME cycles are needed.

  Perform secondary address cycle
  returns WT, SS code. AS remains active.
  MS is set to 0, EG is cleared
*/
static unsigned long FRITZ_SAC(unsigned int data)
{
    (void) VMEmanager::VME()->writeA32D32((unsigned long int)PIO_SET_EG_MS, 0x2) ;
    (void) VMEmanager::VME()->writeA32D32((unsigned long)PIO_DAT_CYCLE, data) ;
    (void) VMEmanager::VME()->writeA32D32((unsigned long)PIO_SET_EG_MS, 0x0) ;
    return VMEmanager::VME()->readA32D32((unsigned long)PIO_RD_WT_SS);
}

/* Perform PAC to CSR space (MS=1,EG=1). MS is set to 0 after
   the PA cycle. Returns WT, SS code
*/
static unsigned long FRITZ_setCSRSpace(int slot)
{
    (void) VMEmanager::VME()->writeA32D32((unsigned long)PIO_CLR_AS, 1) ;
    (void) VMEmanager::VME()->writeA32D32((unsigned long)PIO_SET_EG_MS, 0x9) ;
    (void) VMEmanager::VME()->writeA32D32((unsigned long)PIO_PRIM_AD_CYCLE, slot) ;
    (void) VMEmanager::VME()->writeA32D32((unsigned long)PIO_SET_EG_MS, 0x0) ;
    return VMEmanager::VME()->readA32D32((unsigned long)PIO_RD_WT_SS);
}

/* Perform PAC to DATA space (MS=0,EG=1). MS is set to 0 after
   the PA cycle. Returns WT, SS code
*/
static unsigned long FRITZ_setDataSpace(int slot)
{
    (void) VMEmanager::VME()->writeA32D32((unsigned long)PIO_CLR_AS, 1) ;
    (void) VMEmanager::VME()->writeA32D32((unsigned long)PIO_SET_EG_MS, 0x8) ;
    (void) VMEmanager::VME()->writeA32D32((unsigned long)PIO_PRIM_AD_CYCLE, slot) ;
    (void) VMEmanager::VME()->writeA32D32((unsigned long)PIO_SET_EG_MS, 0x0) ;
    return  VMEmanager::VME()->readA32D32((unsigned long)PIO_RD_WT_SS);
}

/*
  Perform a Fastbus read cycle
*/
static unsigned int FRITZ_read(void)
{
    return VMEmanager::VME()->readA32D32((unsigned long)PIO_DAT_CYCLE);
}

/*
  Perform a Fastbus write cycle
  returns WT, SS code
*/
unsigned long FRITZ_write(unsigned int data)
{
    (void) VMEmanager::VME()->writeA32D32((unsigned long)PIO_DAT_CYCLE, data) ;
    return VMEmanager::VME()->readA32D32((unsigned long)PIO_RD_WT_SS);
}

/*
  Release AS line and wait for de-assertion of AK
*/
void FRITZ_releaseAS(void)
{
    (void) VMEmanager::VME()->writeA32D32((unsigned long)PIO_CLR_AS, 1) ;
}

/* -------------------------------------------------------------------------

LeCroy 1877 TDC Routines
These routines are used to configure and readout the 1877 TDC modules.
They use the FRITZ Fastbus routines defined above.
The 1877 has a 7 event ring buffer. Valid data is available when the read
pointer is one less than the write pointer. No interrupts are used.

------------------------------------------------------------------------- */

unsigned long csr_read(int slot, int csr_num)
{
    unsigned long data;
    FRITZ_setCSRSpace(slot);
    FRITZ_SAC(csr_num);
    data = FRITZ_read();
    FRITZ_releaseAS();
    return data;
}

void csr_write(int slot, int csr_num, unsigned long csr_data)
{
    FRITZ_setCSRSpace(slot);
    FRITZ_SAC(csr_num);
    FRITZ_write(csr_data);
    FRITZ_releaseAS();
}
#else
unsigned long csr_read(int slot, int csr_num)
{
    return 0;
}
#endif

FBTDCmodule::FBTDCmodule( const DAQpp::DAQid &id, unsigned long add )
    : GenModule(id, 0)
{
    _add = add;       // add is the fastbus slot number of the TDC module
    std::cout << "Fastbus Slot (TDC module) at slot " << _add << " Type "
              << std::hex << csr_read(_add, 0) << std::dec << std::endl;

    // KH: I'm not quite sure what to do with the following. Maybe Carlos can help.

    ana = new FBTDCmodule_analysis( 1, 3 );// check Analysis class (1 module, 3 channels?)
    int i;
    std::vector<double> &ped = pedestal->get_value();
    std::vector<double> &ns  = noise->get_value();

    for ( i = 0;i < ana->Ntot();i++ )
    {
        ped.push_back( 0. );
        ns.push_back( 1. );
    }

}

FBTDCmodule::~FBTDCmodule()
{
    std::cout << "...destroying Module " << get_id() << std::endl;
}

unsigned long FBTDCmodule::get_data_type() const
{
    return FBTDCdata;
}

int FBTDCmodule::Reset()
{
    // Issue a reset
#if defined(HAS_SIS1100)
    csr_write(_add, LECROY1877_CSR0, LECROY1877_RESET);
#endif

    return 0;
}

int FBTDCmodule::GetReady()
{
    std::cout << "FBTDCmodule::GetReady" << std::endl;
#if defined(HAS_SIS1100)
    Reset();
    csr_write(_add, LECROY1877_CSR18, 0x2001);  //kh: full scale (changed 0x401 to 0x2001)
    csr_write(_add, LECROY1877_CSR1, 0x24000060);
    std::cout << "Configured TDC module at slot " << _add
              << " CSR0 " << std::hex << csr_read(_add, 0) << std::endl
              << " CSR1 " << csr_read(_add, 1)  << std::endl
              << " CSR16 " << csr_read(_add, 16)  << std::endl
              << " CSR18 " << csr_read(_add, 18)
              << std::dec << std::endl;
#endif
    return 0;
}

int FBTDCmodule::PrepareForRun()
{
    std::cout << "Preparing for run" << std::endl;
#if defined(HAS_SIS1100)         // make sure TDC buffers are cleared (basically do a re-init)
    Reset();
    csr_write(_add, LECROY1877_CSR18, 0x2001); //kh: full scale (changed 0x401 to 0x2001)
    csr_write(_add, LECROY1877_CSR1, 0x24000060);
#endif
    return 0;
}

int  FBTDCmodule::Start (int)
{
    std::cout << "FBTDCmodule::Start" << std::endl;
    return 0;
};

bool FBTDCmodule::hasData()
{
#if defined(HAS_SIS1100)
    unsigned long datum;
    int rd_ptr, wr_ptr;

    datum = csr_read(_add, LECROY1877_CSR16);
    wr_ptr = datum & 0x7;
    rd_ptr = ((datum + 0x100) >> 8) & 0x7;
    if (rd_ptr == wr_ptr)
        return false;   /* no event available */
#endif
    return true;
}

char *FBTDCmodule::getData(int &size)
{
    increase_evt_counter();

#if defined(HAS_SIS1100)
    unsigned int header;
    std::ostringstream ostr;
    char *out = 0;

    /*
      1877 TDC Data Format
      0     Header
      Bits 27-31    Slot number
      Bits 11-13    Buffer number (TDC internal)
      Bits  0-10    Wordcount (including header)
      1     data word 0
      Bits 27-31    Slot number
      Bits 17-23    Channel number (0-95)
      Bit  16       Sign of edge (rising (1), falling (0))
      Bits 0-15     Data  (Time in nsec = Data * 2)
      2     data word 1


    */


    /*
      VMEDAQ Data Format
      0     0xCAFE2bbb   2= data type, bbb= size of data block
      1     0xaaaaaaaa   aaaaaaaa event counter
      2     0xiiitnnnn   iii=module id t=data type (9) nnnn=data size
      3     data word 0  (1877 header)
      4     data word 1  (1st 1877 data word)
      .     .
      bbbb    bbbb         size of data block
    */
    //first tell TDC to load next event
    csr_write(_add, LECROY1877_CSR0, LECROY1877_LNE);

    // read Header word
    FRITZ_setDataSpace(_add);
    header = FRITZ_read();
    size = header & 0x7f;
    int sz = 4 + size;

    unsigned int *data = reinterpret_cast<unsigned int *>(
        allocate(sizeof(unsigned int) * sz));

    for (int i = 0;i < sz;i++)
        data[i] = 0;

    data[0] = DataHeader | (sz);
    data[1] = get_evt_counter();
    int id = get_id();
    data[2] = (id & 0xfff) << 20 | FBTDCdata | (size & 0xfff);
    data[3] = header;
    for (int i = 0;i < (size - 1);i++)
        data[i+4] = FRITZ_read();
    data[sz-1] = sz;
    size = sz * sizeof(unsigned int);
    ostr.write(reinterpret_cast<char *>(data), size);
    deallocate((char *)data);
    size = ostr.str().size();
    if ( size > 0 )
    {
        out = allocate(size);
        memcpy(out, ostr.str().c_str(), size);
    }
    return  out;
#else
    int i, j;
    size = int(randflat(5, 12));
    unsigned int *data = reinterpret_cast<unsigned int *>(
        allocate(sizeof(unsigned int) * (4 + (size & 0xfff))));

    int id = get_id();

    for (i = 0;i < 4 + size;i++)
        data[i] = 0;

    data[0] = DataHeader | (4 + (size & 0xfff));
    data[2] = (id & 0xfff) << 20 | FBTDCdata | (size & 0xfff);


    unsigned int mask = 0;
    int chan, val;

    for (i = 0, j = 3;i < size;i++)
    {
        do
        {
            chan = int(randflat(0, 31));
        }
        while ( mask & (1 << chan) );
        mask |= (1 << chan);
        val = int(50. + randgauss(250., 5.));
        data[j++] = chan << 16 | (val & 0xfff);
    }


    data[4+(size&0xfff)-1] = (4 + (size & 0xfff));
    size = data[j] * sizeof(unsigned int);
    data[1] = get_evt_counter();
    return  reinterpret_cast<char *>(data);
    ;

#endif
}

void FBTDCmodule::set_header_description(unsigned int &data)
{
    unsigned int id = get_id();
    data = (4 & 0xf) << 16 | (id & 0xffff);  //kh: changed (2&0xf) to (4&0xf)
}

void FBTDCmodule::reset_monitor_data()
{
    GenModule::reset_monitor_data();

    std::vector<double> &ped = pedestal->get_value();
    std::vector<double> &ns  = noise->get_value();
    double *pp = get_analysis()->Ped();
    double *nn = get_analysis()->Noise();
    int i, nt = get_analysis()->Ntot();
    for (i = 0;i < nt;i++)
    {
        pp[i] = ped[i];
        nn[i] = ns[i];
    }
}
