#ifndef __FPGACaenV1495_h__
#define __FPGACaenV1495_h__

/* This is a module for vmedaq.
 * They all should inherit from GenModule, which adds some extra
 * functionality on top of DAQpp::Module:
 *
 *   - The module works around the concept of something with a number
 *     of chips, each having the same number of channels. So nchip
 *     (number of chips) and nchan (number of channels per chip) is
 *     something we should supply somehow. In paticular for the
 *     Analysis instance of the module. That brings me to the second
 *     point.
 *
 *   - There is an Analysis pointer (you have to create the object)
 *     which will store all the monitoring stuff. This guy needs the
 *     number of chips and channels per chip.
 *
 *   - If you plan to make scans with that module, you will need to
 *     override the set_scan_point method to set your module in a
 *     reasonable state every time you are in a new scan point.
 *
 *   - pedestals and noise are there as DAQ++ parameters
 *      + can be updated from the values in Analysis or copied into
 *        the Analysis
 *
 *      + Pedestals are stored in the data file if you override the
 *      write_pedestals method.
 *
 *
 * See the DAQ++ documentation for further information
 */
#include <map>

#include <daq/GenModule.h>
#include <data/DataFormat.h>


class FPGACaenV1495 : public GenModule
{
    typedef std::pair<int,int> CoincMapEntry;
    typedef std::map<int,CoincMapEntry> CoincMap;
    typedef std::vector<int> TriggerArray;
    typedef std::map<int,TriggerArray> TriggerMap;

    public:

        /* Constructor
         * The constructor needs 2 parameters:
         * id: the Module id
         * address: the address (VME, or whatever protocol) to
         *          communicate with the module
         *
         * Note: although DAQ++ allows for any type of identifier,
         * however VMEDAQ only handles properly integer IDs. The only
         * reason for that is that it needs integer values to save them
         * properly in the data file
         */
        FPGACaenV1495(const DAQpp::DAQid &id, unsigned long address);
        ~FPGACaenV1495()
        {
            delete _coincmap;
            delete _triggermap;
        }

        /* This is the method called by the RunManager to know if there
         * is data ready in the module. If the answer if true, then it
         * will try to read the module. If there is no way to know a
         * priory about the availavility of data, one can defer the
         * decision until getData
         */
        bool hasData();

        /* This is the method called to retrieve the data. If hasData
         * could not figure out the right answer, one can still say here
         * that there is no data at all by returning a NULL pointer and
         * setting size to 0.
         *
         * The memory to store the data should be allocated with the
         * allocate method. See the DAQ++ documentation...
         */
        char *getData(int &size);

        int PrepareForRun();

        int Start(int = 0);

        int Stop(int = 0);

        int Pause();

        int Continue();

        int Reset();

        int GetReady();

        int Trigger();

        void flush()
        {
        }
        
        /* This returns the data type. The existing data types
         * can be found in DataFormat.h It is a number constructed
         * like: (n & 0xF)<<16;
         */
        unsigned long get_data_type() const
        {
           return FPGACaenData;
        }

        bool hold() { return _hold; }
        void hold(bool x) { _hold = x; }

        void coincmap_add(int i, int j, int k)
        {
           _coincmap->insert(std::pair<int,CoincMapEntry>(i,CoincMapEntry(j,k)));
        }

        std::pair<int,int> coincmap_find(int i)
        {
           CoincMap::iterator it = _coincmap->find(i);
           if (it == _coincmap->end()) return std::pair<int,int>(-1,-1);
           else return it->second;
        }

        void triggermap_add(int i, std::vector<int> array)
        {
           _triggermap->insert(std::pair<int,TriggerArray>(i,array));
        }

        std::vector<int> triggermap_find(int i)
        {
           TriggerMap::iterator it = _triggermap->find(i);
           if (it == _triggermap->end()) return std::vector<int>();
           else return it->second;
        }

    private:
        bool _hold,_waiting,_paused;

        // reset coincidence logic after reading out data
        void coinc_reset();
        void wait();
        void stop_waiting();
        void init();
        void config();

        CoincMap *_coincmap;

        TriggerMap *_triggermap;

        enum Registers
        {
            // CAEN model v1495 FPGA
            // VME Register Address Map
            // A24/A32 Addressing

            // USER FPGA Access 0x1000 to 0x7FFC
            V1495_CONTROL_REGISTER  = 0x8000, // D16 R/W
            V1495_STATUS_REGISTER   = 0x8002, // D16 R 
            V1495_INTERRUPT_LEVEL   = 0x8004, // D16 R/W
            V1495_INTERRUPT_STATUS  = 0x8006, // D16 R/W
            V1495_GEO_ADDRESS       = 0x8008, // D16 R
            V1495_RESET             = 0x800A, // D16 W
            V1495_FIRMWARE_REVISION = 0x800C, // D16 R
            V1495_VME_FPGA_SELECT   = 0x800E, // D16 R/W
            V1495_VME_FPGA_MEMORY   = 0x8010, // D16 R/W
            V1495_USER_FPGA_SELECT  = 0x8012, // D16 R/W
            V1495_USER_FPGA_MEMORY  = 0x8014, // D16 R/W
            V1495_USER_FPGA_CONFIG  = 0x8016, // D16 R/W
            V1495_SCRATCH16         = 0x8018, // D16 R/W
            V1495_SCRATCH32         = 0x8020, // D32 R/W

            // Configuration ROM 0x8100 to 0x801FE

            // CAEN model v1495 FPGA
            // COIN Reference register map
            // A24/A32 Addressing

            // USER FPGA Access 0x1000 to 0x7FFC
            // as defined by CAEN supplied COIN reference user code
            V1495_COIN_REF_A_STATUS_L  = 0x1000, // D16 R
            V1495_COIN_REF_A_STATUS_H  = 0x1002, // D16 R
            V1495_COIN_REF_B_STATUS_L  = 0x1004, // D16 R
            V1495_COIN_REF_B_STATUS_H  = 0x1006, // D16 R
            V1495_COIN_REF_C_STATUS_L  = 0x1008, // D16 R
            V1495_COIN_REF_C_STATUS_H  = 0x100A, // D16 R
            V1495_COIN_REF_A_MASK_L    = 0x100C, // D16 W
            V1495_COIN_REF_A_MASK_H    = 0x100E, // D16 W
            V1495_COIN_REF_B_MASK_L    = 0x1010, // D16 W
            V1495_COIN_REF_B_MASK_H    = 0x1012, // D16 W
            V1495_COIN_REF_C_MASK_L    = 0x1014, // D16 W
            V1495_COIN_REF_C_MASK_H    = 0x1016, // D16 W
            V1495_COIN_REF_GATEWIDTH   = 0x1018, // D16 W
	    V1495_COIN_REF_C_CONTROL_L = 0x101A, // D16 W
            V1495_COIN_REF_C_CONTROL_H = 0x101C, // D16 W
            V1495_COIN_REF_MODE        = 0x101E, // D16 W
            V1495_COIN_REF_SCRATCH     = 0x1020, // D16 R/W
            V1495_COIN_REF_G_CONTROL   = 0x1022, // D16 W
            V1495_COIN_REF_D_CONTROL_L = 0x1024, // D16 R/W
            V1495_COIN_REF_D_CONTROL_H = 0x1026, // D16 R/W
            V1495_COIN_REF_D_DATA_L    = 0x1028, // D16 R/W
            V1495_COIN_REF_D_DATA_H    = 0x102A, // D16 R/W
            V1495_COIN_REF_E_CONTROL_L = 0x102C, // D16 R/W
            V1495_COIN_REF_E_CONTROL_H = 0x102E, // D16 R/W
            V1495_COIN_REF_E_DATA_L    = 0x1030, // D16 R/W
            V1495_COIN_REF_E_DATA_H    = 0x1032, // D16 R/W
            V1495_COIN_REF_F_CONTROL_L = 0x1034, // D16 R/W
            V1495_COIN_REF_F_CONTROL_H = 0x1036, // D16 R/W
            V1495_COIN_REF_F_DATA_L    = 0x1038, // D16 R/W
            V1495_COIN_REF_F_DATA_H    = 0x103A, // D16 R/W
            V1495_COIN_REF_REVISION    = 0x103C, // D16 R/W
            V1495_COIN_REF_PDL_CONTROL = 0x103E, // D16 R/W
            V1495_COIN_REF_PDL_DATA    = 0x1040, // D16 R/W
            V1495_COIN_REF_D_IDCODE    = 0x1042, // D16 R
            V1495_COIN_REF_E_IDCODE    = 0x1044, // D16 R
            V1495_COIN_REF_F_IDCODE    = 0x1046, // D16 R

            // CAEN model v1495 FPGA
            // Compton PET custom COIN Reference register map
            // A24/A32 Addressing

            // USER FPGA Access 0x1000 to 0x7FFC
            // as defined by modified COIN reference user code
            V1495_CUSTOM_REVISION     = 0x1100, // D16 R
            V1495_CUSTOM_SCRATCH      = 0x1102, // D16 R/W
            V1495_CUSTOM_CONTROL      = 0x1104, // D16 W
            V1495_CUSTOM_STATUS       = 0x1106, // D16 R
            V1495_CUSTOM_COINCRESET   = 0x1108, // D16 W
            V1495_CUSTOM_SIGATEWIDTH  = 0x110A, // D16 W
            V1495_CUSTOM_SIRESDELAY   = 0x110C, // D16 W
	    
	    //dump of trigger bits
	    V1495_CUSTOM_ATRIGREG_L   = 0x1110,  //D16 R
	    V1495_CUSTOM_ATRIGREG_H   = 0x1112,  //D16 R
	    
            V1495_CUSTOM_COINCREG_0   = 0x1120, // D16 R
//             V1495_CUSTOM_COINCREG_1   = 0x1122, // D16 R
//             V1495_CUSTOM_COINCREG_2   = 0x1124, // D16 R
//             V1495_CUSTOM_COINCREG_3   = 0x1126, // D16 R
//             V1495_CUSTOM_COINCREG_4   = 0x1128, // D16 R
//             V1495_CUSTOM_COINCREG_5   = 0x112A, // D16 R
	    V1495_CUSTOM_COINCMASK_0  = 0x1160, //D16 W
	    //each coincreg should have corresponding coinc mask
            V1495_CUSTOM_COUNTERREG_0 = 0x11F8, // D16 R
            V1495_CUSTOM_COUNTERREG_1 = 0x11FA, // D16 R
            V1495_CUSTOM_COUNTERREG_2 = 0x11FC, // D16 R
            V1495_CUSTOM_COUNTERREG_3 = 0x11FE, // D16 R

	    //delay/width pairs coded as 0xwwdd (ww are 8 bits for width, dd are 8 bits for delay)
	    V1495_CUSTOM_SIDELAY_0    = 0x1300,
	    V1495_CUSTOM_SIDELAY_1    = 0x1302,
	    V1495_CUSTOM_SIDELAY_2    = 0x1304,
	    V1495_CUSTOM_SIDELAY_3    = 0x1306,
	    V1495_CUSTOM_BGODELAY_SHORT= 0x1308,
	    V1495_CUSTOM_BGODELAY_LONG= 0x1310,
	    V1495_CUSTOM_BGOBLOCKDELAY= 0x1312,

            // Time stamp registers
            V1495_CUSTOM_A0TIME_L     = 0x1200, // D16 R
            V1495_CUSTOM_A0TIME_H     = 0x1202, // D16 R
            V1495_CUSTOM_A1TIME_L     = 0x1204, // D16 R
            V1495_CUSTOM_A1TIME_H     = 0x1206, // D16 R
            V1495_CUSTOM_A2TIME_L     = 0x1208, // D16 R
            V1495_CUSTOM_A2TIME_H     = 0x120A, // D16 R
            V1495_CUSTOM_A3TIME_L     = 0x120C, // D16 R
            V1495_CUSTOM_A3TIME_H     = 0x120E, // D16 R
            V1495_CUSTOM_A4TIME_L     = 0x1210, // D16 R
            V1495_CUSTOM_A4TIME_H     = 0x1212, // D16 R
            V1495_CUSTOM_A5TIME_L     = 0x1214, // D16 R
            V1495_CUSTOM_A5TIME_H     = 0x1216, // D16 R
            V1495_CUSTOM_A6TIME_L     = 0x1218, // D16 R
            V1495_CUSTOM_A6TIME_H     = 0x121A, // D16 R
            V1495_CUSTOM_A7TIME_L     = 0x121C, // D16 R
            V1495_CUSTOM_A7TIME_H     = 0x121E, // D16 R
            V1495_CUSTOM_A8TIME_L     = 0x1220, // D16 R
            V1495_CUSTOM_A8TIME_H     = 0x1222, // D16 R
            V1495_CUSTOM_A9TIME_L     = 0x1224, // D16 R
            V1495_CUSTOM_A9TIME_H     = 0x1226, // D16 R
            V1495_CUSTOM_A10TIME_L    = 0x1228, // D16 R
            V1495_CUSTOM_A10TIME_H    = 0x122A, // D16 R
            V1495_CUSTOM_A11TIME_L    = 0x122C, // D16 R
            V1495_CUSTOM_A11TIME_H    = 0x122E, // D16 R
            V1495_CUSTOM_A12TIME_L    = 0x1230, // D16 R
            V1495_CUSTOM_A12TIME_H    = 0x1232, // D16 R
            V1495_CUSTOM_A13TIME_L    = 0x1234, // D16 R
            V1495_CUSTOM_A13TIME_H    = 0x1236, // D16 R
            V1495_CUSTOM_A14TIME_L    = 0x1238, // D16 R
            V1495_CUSTOM_A14TIME_H    = 0x123A, // D16 R
            V1495_CUSTOM_A15TIME_L    = 0x123C, // D16 R
            V1495_CUSTOM_A15TIME_H    = 0x123E, // D16 R
            V1495_CUSTOM_A16TIME_L    = 0x1240, // D16 R
            V1495_CUSTOM_A16TIME_H    = 0x1242, // D16 R
            V1495_CUSTOM_A17TIME_L    = 0x1244, // D16 R
            V1495_CUSTOM_A17TIME_H    = 0x1246, // D16 R
            V1495_CUSTOM_A18TIME_L    = 0x1248, // D16 R
            V1495_CUSTOM_A18TIME_H    = 0x124A, // D16 R
            V1495_CUSTOM_A19TIME_L    = 0x124C, // D16 R
            V1495_CUSTOM_A19TIME_H    = 0x124E, // D16 R
            V1495_CUSTOM_A20TIME_L    = 0x1250, // D16 R
            V1495_CUSTOM_A20TIME_H    = 0x1252, // D16 R
            V1495_CUSTOM_A21TIME_L    = 0x1254, // D16 R
            V1495_CUSTOM_A21TIME_H    = 0x1256, // D16 R
            V1495_CUSTOM_A22TIME_L    = 0x1258, // D16 R
            V1495_CUSTOM_A22TIME_H    = 0x125A, // D16 R
            V1495_CUSTOM_A23TIME_L    = 0x125C, // D16 R
            V1495_CUSTOM_A23TIME_H    = 0x125E, // D16 R
            V1495_CUSTOM_A24TIME_L    = 0x1260, // D16 R
            V1495_CUSTOM_A24TIME_H    = 0x1262, // D16 R
            V1495_CUSTOM_A25TIME_L    = 0x1264, // D16 R
            V1495_CUSTOM_A25TIME_H    = 0x1266, // D16 R
            V1495_CUSTOM_A26TIME_L    = 0x1268, // D16 R
            V1495_CUSTOM_A26TIME_H    = 0x126A, // D16 R
            V1495_CUSTOM_A27TIME_L    = 0x126C, // D16 R
            V1495_CUSTOM_A27TIME_H    = 0x126E, // D16 R
            V1495_CUSTOM_A28TIME_L    = 0x1270, // D16 R
            V1495_CUSTOM_A28TIME_H    = 0x1272, // D16 R
            V1495_CUSTOM_A29TIME_L    = 0x1274, // D16 R
            V1495_CUSTOM_A29TIME_H    = 0x1276, // D16 R
            V1495_CUSTOM_A30TIME_L    = 0x1278, // D16 R
            V1495_CUSTOM_A30TIME_H    = 0x127A, // D16 R
            V1495_CUSTOM_A31TIME_L    = 0x127C, // D16 R
            V1495_CUSTOM_A31TIME_H    = 0x127E, // D16 R

            V1495_CUSTOM_B0TIME_L     = 0x1280, // D16 R
            V1495_CUSTOM_B0TIME_H     = 0x1282, // D16 R
            V1495_CUSTOM_B1TIME_L     = 0x1284, // D16 R
            V1495_CUSTOM_B1TIME_H     = 0x1286, // D16 R
            V1495_CUSTOM_B2TIME_L     = 0x1288, // D16 R
            V1495_CUSTOM_B2TIME_H     = 0x128A, // D16 R
            V1495_CUSTOM_B3TIME_L     = 0x128C, // D16 R
            V1495_CUSTOM_B3TIME_H     = 0x128E, // D16 R
            V1495_CUSTOM_B4TIME_L     = 0x1290, // D16 R
            V1495_CUSTOM_B4TIME_H     = 0x1292, // D16 R
            V1495_CUSTOM_B5TIME_L     = 0x1294, // D16 R
            V1495_CUSTOM_B5TIME_H     = 0x1296, // D16 R
            V1495_CUSTOM_B6TIME_L     = 0x1298, // D16 R
            V1495_CUSTOM_B6TIME_H     = 0x129A, // D16 R
            V1495_CUSTOM_B7TIME_L     = 0x129C, // D16 R
            V1495_CUSTOM_B7TIME_H     = 0x129E, // D16 R
            V1495_CUSTOM_B8TIME_L     = 0x12A0, // D16 R
            V1495_CUSTOM_B8TIME_H     = 0x12A2, // D16 R
            V1495_CUSTOM_B9TIME_L     = 0x12A4, // D16 R
            V1495_CUSTOM_B9TIME_H     = 0x12A6, // D16 R
            V1495_CUSTOM_B10TIME_L    = 0x12A8, // D16 R
            V1495_CUSTOM_B10TIME_H    = 0x12AA, // D16 R
            V1495_CUSTOM_B11TIME_L    = 0x12AC, // D16 R
            V1495_CUSTOM_B11TIME_H    = 0x12AE, // D16 R
            V1495_CUSTOM_B12TIME_L    = 0x12B0, // D16 R
            V1495_CUSTOM_B12TIME_H    = 0x12B2, // D16 R
            V1495_CUSTOM_B13TIME_L    = 0x12B4, // D16 R
            V1495_CUSTOM_B13TIME_H    = 0x12B6, // D16 R
            V1495_CUSTOM_B14TIME_L    = 0x12B8, // D16 R
            V1495_CUSTOM_B14TIME_H    = 0x12BA, // D16 R
            V1495_CUSTOM_B15TIME_L    = 0x12BC, // D16 R
            V1495_CUSTOM_B15TIME_H    = 0x12BE, // D16 R
            V1495_CUSTOM_B16TIME_L    = 0x12C0, // D16 R
            V1495_CUSTOM_B16TIME_H    = 0x12C2, // D16 R
            V1495_CUSTOM_B17TIME_L    = 0x12C4, // D16 R
            V1495_CUSTOM_B17TIME_H    = 0x12C6, // D16 R
            V1495_CUSTOM_B18TIME_L    = 0x12C8, // D16 R
            V1495_CUSTOM_B18TIME_H    = 0x12CA, // D16 R
            V1495_CUSTOM_B19TIME_L    = 0x12CC, // D16 R
            V1495_CUSTOM_B19TIME_H    = 0x12CE, // D16 R
            V1495_CUSTOM_B20TIME_L    = 0x12D0, // D16 R
            V1495_CUSTOM_B20TIME_H    = 0x12D2, // D16 R
            V1495_CUSTOM_B21TIME_L    = 0x12D4, // D16 R
            V1495_CUSTOM_B21TIME_H    = 0x12D6, // D16 R
            V1495_CUSTOM_B22TIME_L    = 0x12D8, // D16 R
            V1495_CUSTOM_B22TIME_H    = 0x12DA, // D16 R
            V1495_CUSTOM_B23TIME_L    = 0x12DC, // D16 R
            V1495_CUSTOM_B23TIME_H    = 0x12DE, // D16 R
            V1495_CUSTOM_B24TIME_L    = 0x12E0, // D16 R
            V1495_CUSTOM_B24TIME_H    = 0x12E2, // D16 R
            V1495_CUSTOM_B25TIME_L    = 0x12E4, // D16 R
            V1495_CUSTOM_B25TIME_H    = 0x12E6, // D16 R
            V1495_CUSTOM_B26TIME_L    = 0x12E8, // D16 R
            V1495_CUSTOM_B26TIME_H    = 0x12EA, // D16 R
            V1495_CUSTOM_B27TIME_L    = 0x12EC, // D16 R
            V1495_CUSTOM_B27TIME_H    = 0x12EE, // D16 R
            V1495_CUSTOM_B28TIME_L    = 0x12F0, // D16 R
            V1495_CUSTOM_B28TIME_H    = 0x12F2, // D16 R
            V1495_CUSTOM_B29TIME_L    = 0x12F4, // D16 R
            V1495_CUSTOM_B29TIME_H    = 0x12F6, // D16 R
            V1495_CUSTOM_B30TIME_L    = 0x12F8, // D16 R
            V1495_CUSTOM_B30TIME_H    = 0x12FA, // D16 R
            V1495_CUSTOM_B31TIME_L    = 0x12FC, // D16 R
            V1495_CUSTOM_B31TIME_H    = 0x12FE, // D16 R
        };
};

#endif
