/**
 This class is responsible to parse the corresponding section in the
 configuration file to configure the module. It is also responsible
 for create the module instance according to the paremeters retrived
 from the configuration file.
 
 It derives from ModuleParser (see ModuleParser.h in the config
 directory) which, in turn, derives from SectionParser (again, have
 a look at the header file).
 
 It has, then, to implement a number of methods from SectionParser
 
 - parse_node:  this method will parse the corresponding section
 and store locally the configuration parameters
 defined in there.
 - save_config: this method will be called if the configuration
 is to be saved to an external file. The idea is
 that the class will check which parameters have
 changed since the file was loaded and update them
 in the document before saving.
 
 and a few more from ModuleParser:
 
 - config : called by the gui to change configuration parameters
 - monitor: called by the gui to create the monitor display
 - decode : called to decode the data
 - range_min, range_max: to retrieve the range of the monitor
 histograms
 
 One can also define here the factory function specified in the
 configuration file.
 
 FOr this module we will assume that we only have 2 parameters to
 appear in the configuration file:
 - the amplitude of the pulse that we simulate and,
 - the width of the peak we want to simulate.
 
 So in th emodules section we add
 
 <modules>
 .
 .
 .
 <FPGACaenV1495 id="100" address="deadbeef">
 <!-- monitor is a section common to all the modules -->
 <monitor>
 ...
 </monitor>
 
 <!-- padmap is also common to all the modules -->
 <padmap>
 .....
 </padmap>
 
 <!-- and here come our parameters -->
 <mean>100</mean>
 <width>5</width>
 
 </FPGACaenV1495>
 .
 .
 .
 </modules>
 */
#include <string>
#include <sstream>
#include <vector>
#include <iterator>
#include <DAQ++/DAQmanager.h>
#include <config/ConfigParser.h>
#include <config/ModuleParser.h>
#include <config/GeneralConfig.h>
#include <data/Analysis.h>
#include <data/DataFormat.h>

#include "FPGACaenV1495.h"
#include "FPGACaenV1495_gui.h"

class Analysis;

class FPGACaenV1495_conf : public ModuleParser
{
    public:
        FPGACaenV1495_conf()
          : _hold(false)
          , _gatewidth(0)
          , _glevel(0)
          , _amask(0)
          , _bmask(0)
          , _cmask(0)
	  , _si_reset_delay(0x0100)
          , _flood_mode(false)
	  , _wait_delay(1)
	  , _coinc_mask(0xffff)
	  , _si_delay0(0x8000)
	  , _si_delay1(0x8000)
	  , _si_delay2(0x8000)
	  , _si_delay3(0x8000)
	  , _bgo_delay_short(0x6860)
	  , _bgo_delay_long(0x6820)
	  , _bgo_block_delay(0x0100)
        {
        }
        ~FPGACaenV1495_conf()
        {
        }

        /**
         * This is the method responsible for parsing the configuration
         * file. 
         **/
        void parse_node(xmlNodePtr node) throw (ConfigException);

        /**
         * The method that saves the current configuration. It should
         * update all the xml nodes below the module node.
         **/
        void save_config() throw(ConfigException);

        /**
         * This is called by VMEDAQ to create a window that will allow
         * to change the configuration of the module while running the
         * program. Probably the easiest is to use glade-2 to create the
         * layout of the window, define the signals (callbacks that are
         * associated to given actions) and then something derived from
         * GladeApp (see the gui folder) to implement the stuff and
         * update the module state accordingly.
         **/
        GtkWidget *config(const std::string &id);

        /**
         * This is called by VMEDAQ to create a window to display the
         * monitor data. The input parameters are:
         *    - id: the module id (that should help finding the module)
         *    - glade_file: The path of the glade file. This is, in
         *      principle the path of the standard VMEDAQ glade
         *      file. One can certainly ignore this value.
         **/
        MonitorWindow *monitor(const std::string &id, const std::string &glade_file);

        
        // Getters and Setters

    private:
        bool         _hold; // set whether module should hold onto hasData
        unsigned int _gatewidth; // width of output pulse in 5 ns intervals
        unsigned int _glevel; // LEMO output level (0: TTL, 1: NIM) 
        unsigned int _amask;
        unsigned int _bmask;
        unsigned int _cmask;
        unsigned int _si_gate; // width of output pulse for silicon in 5 ns
        unsigned int _si_reset_delay; // si reset delay length in 5 ns
        bool         _flood_mode; //set coinc routines to allow single events from BGO module (allowing for flood data) 
        double       _wait_delay; //set delay to wait for event in FPGA; once delay is passed the module exits from hasData even if no data was collected
        unsigned int _coinc_mask; //mask coincidence tags when looking for specific event types
        unsigned int _si_delay0; //pair of width/delay parameters for fine tuning of silicon 0 delay & time window
        unsigned int _si_delay1; //pair of width/delay parameters for fine tuning of silicon 1 delay & time window
        unsigned int _si_delay2; //pair of width/delay parameters for fine tuning of silicon 16 delay & time window
        unsigned int _si_delay3; //pair of width/delay parameters for fine tuning of silicon 17 delay & time window
        unsigned int _bgo_delay_short; //pair of stop/start parameters for fine tuning of bgo short window
        unsigned int _bgo_delay_long; //pair of width/delay parameters for fine tuning of bgo long window
        unsigned int _bgo_block_delay; //delay after each trigger when that bgo module is blocked for subsequent triggers (DLT in effect)
       void parse_coincmap(FPGACaenV1495 *);
};

/**
 * This is the hook specified in the config file to return a pointer
 * to the FPGACaenV1495_conf instance that will interact with VMEDAQ.
 *
 * Note that it should b edeclared as extern "C" so that no demangling
 * mechanism is applied to its name that prevent finding easily the
 * simbol in the library
 *
 **/
extern "C"
{
    FPGACaenV1495_conf *create_FPGACaenV1495_conf()
    {
        FPGACaenV1495_conf *cnf = new FPGACaenV1495_conf();
        return cnf;
    }
}

/**
 * Now come the implementations of the class methods
 */
void FPGACaenV1495_conf::parse_node(xmlNodePtr _node) throw (ConfigException)
{
    unsigned long address;
    std::string saddress;

    // The configuration manager handler
    ConfigParser *cfg = ConfigParser::theConfig();

    // the node name
    name = node_name();

    // store the node pointer
    node = _node;

    // get the module id
    id = cfg->get_attribute<std::string>(node, "name");

    // get the address
    saddress = cfg->get_attribute<std::string>(node, "address");

    std::istringstream os(saddress);
    os >> std::hex >> address;

    // Check if the module is already there
    DAQpp::Module *dm;
    FPGACaenV1495 *module = 0;

    if ( (dm = DAQpp::DAQmanager::theDAQmanager()->find_module(id)))
        module = dynamic_cast<FPGACaenV1495 *>(dm);

    if (!module)
    {
        // The module is not there... create
        module = new FPGACaenV1495(id, address);
    }

    // Now parse the things which are common to all the modules, that
    // is the monitor and padmap elements
    parse_common("//"+name, id);
    
    // Now we have the module. From now own one should parse the
    // section from the configuration file
    try
    {
        _hold = cfg->get_attribute<bool>(node, "hold");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
        _hold = module->hold();
    }

    try
    {
        _gatewidth = cfg->get_value<unsigned int>(node, "gatewidth");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
        _gatewidth = module->get_par_value<unsigned int>("gatewidth");
    }

    try
    {
        _glevel = cfg->get_value<unsigned int>(node, "glevel");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
        _glevel = module->get_par_value<unsigned int>("glevel");
    }

    try
    {
        std::string amask = cfg->get_value<std::string>(node, "amask");
        std::istringstream iss(amask);
        iss >> std::hex >> _amask >> std::dec;
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
        _amask = module->get_par_value<unsigned int>("amask");
    }

    try
    {
        std::string bmask = cfg->get_value<std::string>(node, "bmask");
        std::istringstream iss(bmask);
        iss >> std::hex >> _bmask >> std::dec;
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
        _bmask = module->get_par_value<unsigned int>("bmask");
    }

    try
    {
        std::string cmask = cfg->get_value<std::string>(node, "cmask");
        std::istringstream iss(cmask);
        iss >> std::hex >> _cmask >> std::dec;
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
        _cmask = module->get_par_value<unsigned int>("cmask");
    }

    try
    {
      std::string hmask = cfg->get_value<std::string>(node, "si_gate");
      std::istringstream iss(hmask);
      iss >> std::hex >> _si_gate >> std::dec;
      //_si_gate = cfg->get_value<unsigned int>(node, "si_gate");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
      _si_gate = module->get_par_value<unsigned int>("si_gate");
    }

    try
    {
      std::string hmask = cfg->get_value<std::string>(node, "si_reset_delay");
      std::istringstream iss(hmask);
      iss >> std::hex >> _si_reset_delay >> std::dec;
      //_si_reset_delay = cfg->get_value<unsigned int>(node, "si_reset_delay");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
        _si_reset_delay = module->get_par_value<unsigned int>("si_reset_delay");
    }
    // Now we have the module. From now own one should parse the
    // section from the configuration file
    try
    {
        _flood_mode = cfg->get_attribute<bool>(node, "flood_mode");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
      _flood_mode = module->get_par_value<bool>("flood_mode");
    }

    try
    {
      _wait_delay = cfg->get_value<double>(node, "wait_delay");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
      _wait_delay = module->get_par_value<double>("wait_delay");
    }

    try
    {
      std::string hmask = cfg->get_value<std::string>(node, "coinc_mask");
      std::istringstream iss(hmask);
      iss >> std::hex >> _coinc_mask >> std::dec;
      //_coinc_mask = cfg->get_value<unsigned int>(node, "coinc_mask");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
      _coinc_mask = module->get_par_value<unsigned int>("coinc_mask");
    }


    try
    {
     // _si_delay0 = cfg->get_value<unsigned int>(node, "si_delay0");
      std::string hmask = cfg->get_value<std::string>(node, "si_delay0");
      std::istringstream iss(hmask);
      iss >> std::hex >> _si_delay0 >> std::dec;
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
      _si_delay0 = module->get_par_value<unsigned int>("si_delay0");
    }
    try
    {
      std::string hmask = cfg->get_value<std::string>(node, "si_delay1");
      std::istringstream iss(hmask);
      iss >> std::hex >> _si_delay1 >> std::dec;
      //_si_delay1 = cfg->get_value<unsigned int>(node, "si_delay1");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
      _si_delay1 = module->get_par_value<unsigned int>("si_delay1");
    }
    try
    {
      std::string hmask = cfg->get_value<std::string>(node, "si_delay2");
      std::istringstream iss(hmask);
      iss >> std::hex >> _si_delay2 >> std::dec;
      //_si_delay2 = cfg->get_value<unsigned int>(node, "si_delay2");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
      _si_delay2 = module->get_par_value<unsigned int>("si_delay2");
    }
    try
    {
      std::string hmask = cfg->get_value<std::string>(node, "si_delay3");
      std::istringstream iss(hmask);
      iss >> std::hex >> _si_delay3 >> std::dec;
      //_si_delay3 = cfg->get_value<unsigned int>(node, "si_delay3");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
      _si_delay3 = module->get_par_value<unsigned int>("si_delay3");
    }
    try
    {
      std::string hmask = cfg->get_value<std::string>(node, "bgo_delay_short");
      std::istringstream iss(hmask);
      iss >> std::hex >> _bgo_delay_short >> std::dec;
      //_bgo_delay = cfg->get_value<unsigned int>(node, "bgo_delay");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
      _bgo_delay_short = module->get_par_value<unsigned int>("bgo_delay_short");
    }
    try
    {
      std::string hmask = cfg->get_value<std::string>(node, "bgo_delay_long");
      std::istringstream iss(hmask);
      iss >> std::hex >> _bgo_delay_long >> std::dec;
      //_bgo_delay = cfg->get_value<unsigned int>(node, "bgo_delay");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
      _bgo_delay_long = module->get_par_value<unsigned int>("bgo_delay_long");
    }
    try
    {
      std::string hmask = cfg->get_value<std::string>(node, "bgo_block_delay");
      std::istringstream iss(hmask);
      iss >> std::hex >> _bgo_block_delay >> std::dec;
      //_bgo_delay = cfg->get_value<unsigned int>(node, "bgo_delay");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
      _bgo_block_delay = module->get_par_value<unsigned int>("bgo_block_delay");
    }
    



    module->hold(_hold);
    module->set_par<unsigned int>("gatewidth",_gatewidth);
    module->set_par<unsigned int>("glevel", _glevel);
    module->set_par<unsigned int>("amask", _amask);
    module->set_par<unsigned int>("bmask", _bmask);
    module->set_par<unsigned int>("cmask", _cmask);
    module->set_par<unsigned int>("si_gate",_si_gate);
    module->set_par<unsigned int>("si_reset_delay",_si_reset_delay);
    module->set_par<double>("wait_delay",_wait_delay);
    module->set_par<bool>("flood_mode",_flood_mode);
    module->set_par<unsigned int>("coinc_mask",_coinc_mask);
    module->set_par<unsigned int>("si_delay0",_si_delay0);
    module->set_par<unsigned int>("si_delay1",_si_delay1);
    module->set_par<unsigned int>("si_delay2",_si_delay2);
    module->set_par<unsigned int>("si_delay3",_si_delay3);
    module->set_par<unsigned int>("bgo_delay_short",_bgo_delay_short);
    module->set_par<unsigned int>("bgo_delay_long",_bgo_delay_long);
    module->set_par<unsigned int>("bgo_block_delay",_bgo_block_delay);

    parse_coincmap(module);
}

void FPGACaenV1495_conf::save_config() throw(ConfigException)
{
    // The configuration manager handler
    ConfigParser *cfg = ConfigParser::theConfig();

    // Save common stuff
    save_common();

    // find the module
    DAQpp::Module *dm = 0;
    if ( !(dm = DAQpp::DAQmanager::theDAQmanager()->find_module(id)))
        return;

    FPGACaenV1495 *m = 0;
    if ( !(m = dynamic_cast<FPGACaenV1495 *>(dm)))
        return;

    _hold      = m->hold();
    _gatewidth = m->get_par_value<unsigned int>("gatewidth");
    _glevel    = m->get_par_value<unsigned int>("glevel");
    _amask     = m->get_par_value<unsigned int>("amask");
    _bmask     = m->get_par_value<unsigned int>("bmask");
    _cmask     = m->get_par_value<unsigned int>("cmask");
    _si_gate   = m->get_par_value<unsigned int>("si_gate");
    _si_reset_delay  = m->get_par_value<unsigned int>("si_reset_delay");
    _wait_delay  = m->get_par_value<double>("wait_delay");
    _si_delay0  = m->get_par_value<unsigned int>("si_delay0");
    _si_delay1  = m->get_par_value<unsigned int>("si_delay1");
    _si_delay2  = m->get_par_value<unsigned int>("si_delay2");
    _si_delay3  = m->get_par_value<unsigned int>("si_delay3");
    _bgo_delay_short  = m->get_par_value<unsigned int>("bgo_delay_short");
    _bgo_delay_long  = m->get_par_value<unsigned int>("bgo_delay_long");
    _bgo_block_delay = m->get_par_value<unsigned int>("bgo_block_delay");
    _coinc_mask  = m->get_par_value<unsigned int>("coinc_mask");

    std::ostringstream oss;
    // Save the current values in the configuration
    cfg->check_attribute<bool>(node, "hold", _hold);
    cfg->check_value<unsigned int>(node, "gatewidth", _gatewidth);
    cfg->check_value<unsigned int>(node, "glevel", _glevel);

    oss << std::hex << _amask << std::dec;
    cfg->check_value<std::string>(node, "amask", oss.str());

    oss.str("");
    oss << std::hex << _bmask << std::dec;
    cfg->check_value<std::string>(node, "bmask", oss.str());

    oss.str("");
    oss << std::hex << _cmask << std::dec;
    cfg->check_value<std::string>(node, "cmask", oss.str());

    cfg->check_value<double>(node, "wait_delay", _wait_delay);

    oss.str(std::string());
    oss << std::hex << _si_gate << std::dec;
    cfg->check_value<std::string>(node, "si_gate", oss.str());

    oss.str(std::string());
    oss << std::hex << _si_reset_delay << std::dec;
    cfg->check_value<std::string>(node, "si_reset_delay", oss.str());

    oss.str("");
    oss << std::hex << _coinc_mask << std::dec;
    cfg->check_value<std::string>(node, "coinc_mask", oss.str());

    oss.str("");
    oss << std::hex << _si_delay0 << std::dec;
    cfg->check_value<std::string>(node, "si_delay0", oss.str());

    oss.str("");
    oss << std::hex << _si_delay1 << std::dec;
    cfg->check_value<std::string>(node, "si_delay1", oss.str());

    oss.str("");
    oss << std::hex << _si_delay2 << std::dec;
    cfg->check_value<std::string>(node, "si_delay2", oss.str());

    oss.str("");
    oss << std::hex << _si_delay3 << std::dec;
    cfg->check_value<std::string>(node, "si_delay3", oss.str());

    oss.str("");
    oss << std::hex << _bgo_delay_short << std::dec;
    cfg->check_value<std::string>(node, "bgo_delay_short", oss.str());

    oss.str("");
    oss << std::hex << _bgo_delay_long << std::dec;
    cfg->check_value<std::string>(node, "bgo_delay_long", oss.str());

    oss.str("");
    oss << std::hex << _bgo_block_delay << std::dec;
    cfg->check_value<std::string>(node, "bgo_block_delay", oss.str());

    cfg->check_attribute<bool>(node, "flood_mode", _flood_mode);
}

void
FPGACaenV1495_conf::parse_coincmap(FPGACaenV1495 *m)
{
    std::vector<xmlNodePtr> vnode;
    ConfigParser *cfg = ConfigParser::theConfig();
    std::ostringstream oss;

    oss << "//"<< name << "[@name=\""<< id << "\"]/coincmap";
    int nmap = cfg->get_node(oss.str(), vnode);

    if (!nmap) {
      std::cerr << "Module "<< id
                << " does not have a coincmap array, the analysis will not display properly."
                << std::endl;
      return;
    }
    else if (nmap > 1)
    {
      std::cerr << "Module "<< id
                << " has >1 coincmap array. Ignoring the rest."<< std::endl;
    }

    xmlNodePtr nc = vnode[0];
    std::vector<int> map;
    cfg->get_array<int>(nc, map);

    int size = map.size();
    if (size%3 != 0) {
        std::cerr << "coincmap array in module "<< id << " is not properly formatted\n"
                  << "  it has " << nmap << " values, which is not a multiple of 3" << std::endl;
    }
    int nchip = m->get_analysis()->Nchip();
    if (3*nchip < size)
    {
        std::cerr << "coincmap array in module "<< id << " has more than "
                  << 3*nchip << " entries. Ignoring the rest."<< std::endl;
        size = 3*nchip;
    }
    for (int i=0; i<(size/3); i++)
        m->coincmap_add(map[3*i],map[3*i+1],map[3*i+2]);
}

MonitorWindow * FPGACaenV1495_conf::monitor(const std::string &name,
                                          const std::string &glade_file)
{
    DAQpp::Module *dm = DAQpp::DAQmanager::theDAQmanager()->find_module(name);
    if (!dm)
        return 0;

    FPGACaenV1495 *m = 0;
    if ( !(m = dynamic_cast<FPGACaenV1495 *>(dm)))
    {
        std::cout << "FPGACaenV1495_conf: module " << name
                << " is not of type FPGACaenV1495 " << std::endl;
        return 0;
    }
    FPGACaenV1495_monitor * mon = new FPGACaenV1495_monitor(m->get_id());
    mon->init("FPGACaenV1495.glade", "monitor");
    MonitorWindow *mw = new MonitorWindow(mon->get_main_window());
    return mw;
}

GtkWidget * FPGACaenV1495_conf::config(const std::string &name)
{
    DAQpp::Module *dm = DAQpp::DAQmanager::theDAQmanager()->find_module(name);
    if (!dm)
        return 0;

    FPGACaenV1495 *m = 0;
    if ( !(m = dynamic_cast<FPGACaenV1495 *>(dm)))
    {
        std::cout << "FPGACaenV1495_conf: module " << name
                << " is not of type FPGACaenV1495 " << std::endl;
        return 0;
    }
    FPGACaenV1495_config * cnf = new FPGACaenV1495_config(m->get_id());
    cnf->init("FPGACaenV1495.glade", "configuration");
    return cnf->get_main_window();
}

static void __attribute__ ((constructor)) __init__lib__()
{
    // init threads
    std::cout << "### Loading FPGACaenV1495..." << std::endl;
}

static void __attribute__ ((destructor)) __close__lib__()
{
    std::cout << "### Unloading FPGACaenV1495..." << std::endl;
}
