#include "ReconModule_analysis.h"
#include <iostream>
#include "DAQ++/DAQmanager.h"
#include "FPGACaenV1495_analysis.h"
#include "VATAmodule_analysis.h"
#include "BGOmodule_analysis.h"
#include <map>
#include "daq/GenModule.h"
#include <cmath>

using namespace std;

const double reconModule_analysis::pi2=acos(0);

reconModule_analysis::reconModule_analysis(int nchip, int nchan):Analysis(nchip, nchan)
{
  std::cout<<"Creating reconModule_analysis"<<std::endl;
  //fpgaModuleID=100;
  //add_par("fpgaModuleID", new DAQpp::Par<int>(100));
  //this might be reconfigurable once
  anaList[100]=0;
  anaList[256]=0;
  anaList[257]=0;
  anaList[300]=0;
  anaList[301]=0;
  anaList[302]=0;

  sigMap[0]=0;
  sigMap[4]=1;
  sigMap[5]=2;
  sigMap[8]=3;

}



bool verify_range(int test, int *range){
  //test against inclusive range
  if (test<range[0]) return false;
  if (test>range[1]) return false;
  return true;
}

void get_position(Analysis *ana, int module, int hit, double pos[]);

void reconModule_analysis::Analyze(){
  //for checking, remove once done
  //return;
  
  //find FPGACaenV1495 module to guess event type
  FPGACaenV1495_analysis *fpga_ana=dynamic_cast<FPGACaenV1495_analysis *>(anaList[100]);

  if (!fpga_ana) {
    cout<<"No fpga found"<<endl;
    return;
  }
  //cout<<"Found fpga analysis="<<fpga_ana<<endl;
  unsigned int coinc=fpga_ana->get_coinc();
  int nTrig=0;
  for (int i=0;i<32;i++) 
    if (coinc & (1<<i)) nTrig++;

  if (nTrig!=1){
    //cout<<"Bad event "<<fpga_ana->get_event_number()<<" nTrig="<<nTrig<<endl;
    return;
  }


  //cout<<"Processing event "<<fpga_ana->get_event_number()<<endl;
  int cmode=0;
  for (;!(coinc & (1<<cmode));cmode++) ;
  //cout<<"Coinc mode="<<cmode<<endl;
  
  vector<int> fpga_hits;
  for (int i=0;i<fpga_ana->Ntot();i++){
    if (fpga_ana->Chan()[i]!=0) {
      fpga_hits.push_back(i);
      // cout<<'\t'<<i<<' '<<fpga_ana->Chan()[i]<<endl;
    }
  }
  if (fpga_hits.size()!=2){
    //cout<<"Too many hits in fpga. Not reconstructing"<<endl;
    return;
  }

  //cross-reference hits and coinc type
  int range[4];
  switch (cmode){
  case 0:
    range[0]=range[1]=0;
    range[2]=range[3]=16;
    break;
  case 4:
    range[0]=range[1]=0;
    range[2]=18;
    range[3]=29;
    break;
  case 5:
    range[0]=2;
    range[1]=13;
    range[2]=range[3]=16;
    break;
  case 8:
    range[0]=2;
    range[1]=13;
    range[2]=18;
    range[3]=29;
    break;
  }
  for (int i=0;i<2;i++){
    if (!verify_range(fpga_hits[i],range+2*i)) {
      //cout<<"Hit "<<i<<" should be between "<<range[2*i]<<" and "<<range[2*i+1]<<endl;
      return;
    }
  }
  
  int mID[]={-1,-1};
  double pos[4];

  while (1){
    //determine analysis of the left module
    if ( fpga_hits[0] == 0 ) {
      mID[0]=256;
      break;
    }

    if ( fpga_hits[0] -2 < 8) {
      mID[0]=300;
      break;

    }
    mID[0]=302;
    break;
  }
  
  while (1) {
    //determine right hand side

    if (fpga_hits[1] == 16 ) {
      mID[1]=257;
      break;
    }

    if ( fpga_hits[1] - 18 < 8 ) {
      mID[1]=301;
      break;
    }

    mID[1]=302;
    break;
  }

  for (int i=0;i<2;i++) {
    if (mID[i]<0 ){
      //cout<<"Hit module not set for i="<<i<<endl;
      return;
    }
    
    Analysis *ana=anaList[mID[i]];
    if (!ana){
      cout<<"Analysis for module "<<mID[i]<<" not available"<<endl;
      return;
    }


    if (ana->nhits()!=1){
      //cout<<"Too many hits on "<<(i>0 ? "right" : "left")<<" ("<<ana->nhits()<<") "<<endl;
      return;
    }
    
    int ihit=-1;
    BGOmodule_analysis *bgo=dynamic_cast<BGOmodule_analysis *>(ana);
    if (bgo) {
    
      ihit=bgo->get_hits().top();

      //verify
      int bgoHead=-1;
      int ioff=fpga_hits[i]-16*i-2;
      if (ioff>7) ioff-=8;
      
      if (i) bgoHead=8-1-ioff;
      else bgoHead=ioff;
      
      if (ihit/32 != bgoHead){
	cout<<fpga_ana->get_event_number()<<" BGO["<<mID[i]<<"] crystal: "<<ihit<<" head="<<ihit/32<<" expected="<<bgoHead<<endl;
	return;
      }
      //cout<<"Match "<<fpga_ana->get_event_number()<<" BGO["<<mID[i]<<"] crystal: "<<ihit<<" head="<<ihit/32<<" expected="<<bgoHead<<endl;

    }

    //get it from vata
    VATAmodule_analysis *v=dynamic_cast<VATAmodule_analysis *>(ana);
    if (v) {
      //cout<<"Looking in VATAmodule"<<endl;
      ihit=v->get_hits().top();
    }
      
    if (ihit < 0 ){
      cout<<"Could not find hit, mID="<<mID[i]<<endl;
      return;
    }

    get_position(ana,mID[i],ihit,pos+2*i);
    
    //cout<<"Hit in module "<<mID[i]<<" ch "<<ihit<<" pos("<<pos[2*i]<<','<<pos[2*i+1]<<')'<<endl;
    

  }
  
  double dv[2];
  for (int i=0;i<2;i++) dv[i]=pos[2+i]-pos[i];
  double gamma=atan2(dv[1],dv[0]);
  //gamma is between -pi2 and pi2

  //add object rotation
  //set by set_scan_point
  
  double theta=angle*2*pi2/180.;
  
  double phi=gamma-theta;

  //tuck it to 0..2pi
  while(phi<0) phi+=4*pi2;
  while(phi>4*pi2) phi-=4*pi2;
  
  double p=pos[1]*cos(gamma)-pos[0]*sin(gamma); 
  //cout<<"Equates to sinogram ("<<phi<<','<<p<<") "<<endl;


  //it is time to fill the sinograms
  int ib=get_bin(phi);
  int ich=static_cast<int>((p+25.1)/50.2*Nchan());
  if (ich<0 || ich>Nchan()-1)
    return;

  Signal(sigMap[cmode]*Nchan()+ich)[ib]++;
  Signal(-1)[ib]++;
  
  
  
}

void reconModule_analysis::init_run(){

  cout<<"reconModule_analysis::init_run()"<<endl;
  for (std::map<int,Analysis *>::iterator it=anaList.begin();it!=anaList.end();it++){
    if (it->second) continue;

    DAQpp::DAQid id(it->first);
    GenModule *m =dynamic_cast<GenModule *>(DAQpp::DAQmanager::theDAQmanager()->find_module(id));
    if (!m) continue;
    it->second=m->get_analysis();
  }
}

void get_position(Analysis *ana, int module, int hit, double pos[]){
  int ix,iy;

  ana->get_map()->pad(hit,ix,iy);
  
  double padSize=1.4;
  double tpos[2];
  int nx2=ana->get_map()->nx()/2;
  int ny2=ana->get_map()->ny()/2;
  tpos[0]=(ix+0.5-nx2);//to [ - ( nx2 - 0.5 ) , ... , + ( nx2 - 0.5 ) ] 
  tpos[1]=(iy+0.5-ny2);//to [ - ( ny2 - 0.5 ) , ... , + ( ny2 - 0.5 ) ] 
  double angle=0;

  double anglePerI=reconModule_analysis::pi2/128.;
  double angleOff=6*8*anglePerI;
  double r0=520;
  double offFull=ana->get_map()->nx()*anglePerI;
  double offHalf=nx2*anglePerI;

  //int halfModule=ana->get_map()->ntot()/2;

  switch(module){
  case 256:
    pos[0]=-76.86-tpos[1]*padSize;
    pos[1]=-2.4-tpos[0]*padSize;
    return;
  case 257:
    pos[0]=+76.86+tpos[1]*padSize;
    pos[1]=-2.4+tpos[0]*padSize;
    return;
  case 300:
    angle=(ix+0.5)*anglePerI-angleOff+2*reconModule_analysis::pi2;
    break;
  case 301:
    angle=(ix+0.5)*anglePerI-angleOff;
    break;
  case 302:
    if (ix>nx2-1)
      angle=(ix+0.5)*anglePerI-angleOff+offHalf;//just half since they are already shifted by a half
    else
      angle=(ix+0.5)*anglePerI-angleOff+offFull+2*reconModule_analysis::pi2;//they are sitting right next to 300 modules shifted by pi
    //cout<<"\tangles="<<offHalf<<' '<<offFull<<endl;

    break;
  }
  pos[0]=r0*cos(angle);
  pos[1]=r0*sin(angle);


}
