#include <iostream>
#include <string>
#include <vector>
#include <cmath>
#include <daq/VMEmanager.h>
#include <data/Analysis.h>
#include <data/DataFormat.h>
#include "SIS3300module.h"
#include "SIS3300moduleAna.h"

#define TEST 1

SIS3300module::SIS3300module(const DAQpp::DAQid &id, unsigned long address) :
    GenModule(id, address)
{

    // Using DAQ++ parameters can be useful, especially if you want to
    // access them the DAQcorba
    // simulation parameters
    add_par("mean", new DAQpp::Par<double>(100.));
    add_par("width", new DAQpp::Par<double>(10.));
    add_par("efficiency", new DAQpp::Par<double>(0.75));
    // module parameters
    add_par("slave", new DAQpp::Par<bool>(false));
    add_par("softstart", new DAQpp::Par<bool>( true ));
    add_par("nInputs", new DAQpp::Par<int>(8));
    add_par("nSamples", new DAQpp::Par<int>(256));
    add_par("delay", new DAQpp::Par<int>(200));
    add_par("mode", new DAQpp::Par<int>(0));
    add_par("threshold1", new DAQpp::Par<int>(0));
    add_par("threshold2", new DAQpp::Par<int>(0));
    add_par("threshold3", new DAQpp::Par<int>(0));
    add_par("threshold4", new DAQpp::Par<int>(0));
    add_par("threshold5", new DAQpp::Par<int>(0));
    add_par("threshold6", new DAQpp::Par<int>(0));
    add_par("threshold7", new DAQpp::Par<int>(0));
    add_par("threshold8", new DAQpp::Par<int>(0));

    // reset pedestals and noise to sensible values
    fill_n(back_inserter(get_par_value<std::vector<double> > ("pedestal") ), maxInputs*maxSamples, 0.);
    fill_n(back_inserter(get_par_value<std::vector<double> > ("noise") ), maxInputs*maxSamples, 0.);

    // now we should create the Analysis for the monitoring
    // never forget to do this !!!
    create_analysis();
    std::cout << "SIS3300data = "<< std::hex << SIS3300data << std::dec << std::endl;
}

unsigned long SIS3300module::get_data_type() const
{
    return SIS3300data;
}

bool SIS3300module::hasData()
{
    VMEmanager *vme = VMEmanager::VME();
    if ( !vme->connected() )
        return randuniform() < get_efficiency();

    int bank = 1;
    int clock_status = 0;
    unsigned long data;

    data = vme->readA32D32(_add + SIS3300_ACQUISTION_CONTROL);
    switch (bank)
    {
    case 1:
        clock_status = !(data & ACQ_STATUS_CLOCK_TO_BANK1);
        break;
    case 2:
        clock_status = !(data & ACQ_STATUS_CLOCK_TO_BANK2);
        break;
    default:
        std::cout << "Error::hasData() bank = "<< bank<< " out of range"<< std::endl;
    }

    return clock_status ? true : false;
}

char*SIS3300module::getData(int &size)
{
    int _nInputs = get_nInputs();
    int _nSamples = get_nSamples();
    int ntot = ((_nInputs + 1) / 2) * _nSamples;
    int bank = 1;//for now always using bank 1
    int id, iw;

    unsigned int *data;
    VMEmanager *vme = VMEmanager::VME();
    increase_evt_counter();
    if ( !vme->connected() )
    {

        double _mean = get_mean();
        double _std = get_width();
        int channel = int(ntot * randuniform());

        /*
         Get the data
         The format is the following
         Word
         0     0xCAFEibbb   i= event type bbb= size of data block
         1     0xaaaaaaaa   aaaaaaaaaa event counter
         2     0xiiitnnnn   iii=module id t=data type nnnn=data size
         3     data word 0
         4     data word 1
         .     .
         bbbb    bbbb         size of data block

         In practice, some bits of words 1 and 2 can be used. Note,
         however, that the module id should go where it is.

         the size in word 2 is the number of data words folowing this
         word. 

         Also, the event type byte will allow us to identify our data block
         when reading the file. Look into DataFormat.h and choose one
         different from the existing ones:
         NormalData = 0x10000,
         PulseData  = 0x20000,
         ScanData   = 0x30000,
         ADCdata    = 0x40000,
         CSPRINTdata= 0x80000,
         FBTDCdata  = 0x90000
         */

        const unsigned int data_type = 0xa0000;

        

        // Allocate space for the data. This has to be done with the
        // allocate method of DAQpp::Module
        //
        // Note that here we assume that we have one word per channel. It
        // may well happen that we get from the module a complicated
        // structure that we do not want ot decode here, so we just dumpt
        // it into the header and the trailer words and leave the analysis
        // program with the burden of decoding.

        size = 4 + // the 3 words in the header plus the trailing word
                ntot; // We assume one word per channel...

        data =reinterpret_cast<unsigned int *>(allocate(sizeof(unsigned int) * size));

        // Fill the data array
        id = get_id();
        iw = 0;
        data[iw++] = DataHeader | DataRecord | (size);
        data[iw++] = get_evt_counter();
        data[iw++] = (id & 0xfff) << 20| data_type | (ntot);

        for (int ic = 0; ic < ntot; ic++)
            data[iw++] = (unsigned int)(randgauss(50., 5.));

        data[3+channel] += (unsigned int)(randgauss(_mean, _std));
        data[iw] = size;

        size *= sizeof(unsigned int);
        return reinterpret_cast<char *>(data);
    }
    else
    {

        // Allocate space for the data. This has to be done with the
        // allocate method of DAQpp::Module
        //
        // Note that here we assume that we have one word per channel. It
        // may well happen that we get from the module a complicated
        // structure that we do not want ot decode here, so we just dumpt
        // it into the header and the trailer words and leave the analysis
        // program with the burden of decoding.

        size = 7 + // the 5 words in the header plus the trailing word
                ntot;

        data =reinterpret_cast<unsigned int *>(allocate(sizeof(unsigned int) * size));

        /*
         SIS3300 header

         In header we store number of channels & number of samples for each
         module. 

         SIS3300 data block

         Data Format
         0      0xCAFEibbb      i=record type,  bbb=ignored
         1      0xaaaaaaaa      aaaaaaaa=event counter
         2      0xiiitnnnn      iii=module id,  t=module type, nnnn=unused
         3      0xbbbbbbbb      bbbbbbbb=size of data block
         4      0xcsssssss      c=# of inputs, sssssss=sample size
         5      0x000(w|r)rrrr  w=wrap_bit (@ bit 19), rrrrr=ring pointer[17 bit]
         6      data word 0
         .         .
         x      0xbbbbbbbb      size of data block

         Two options for data word: raw, proccessed

         RAW:
         data word = 0x0uuu0vvv uuu=even ch (0,2,4,..,ccc-2) raw value (12 bits)
         vvv=odd ch (1,3,5,... ccc-1) raw adc value (12 bits) 
         There are ssss<<4 words per channel pair and (c+1)/2 channel pairs
         giving in total N=(ssss<<4)*(c+1)/2 data words

         Size of data block is bbbbbbbb=N+6. Since N can be larger than
         0xfff (=4096) instead of size(=bbb) the first word carries record type and 
         size is given in the next data word.
         */

        // Fill the data array
        id = get_id();
        iw = 0;
        unsigned long gwords;
        VMEmanager *vme = VMEmanager::VME();
        data[iw++] = DataHeader | BigDataRecord;
        data[iw++] = get_evt_counter();
        data[iw++] = ((id & 0xfff) << 20) | SIS3300data;
        data[iw++] = size;
        data[iw++] = (_nSamples & 0xfffffff) | ((_nInputs & 0xf) << 28);
        //get ring pointer and wrap bit
        data[iw++] = vme->readA32D32(_add + SIS3300_EVENT_DIRECTORY_BANK1_ALL_ADC+(bank - 1) * event_directory_bank_shift);
        //fill buffer with data from SIS3300 databank
        for (int i = 0; i < (_nInputs + 1) / 2; i++)
        {
            gwords = vme->readA322EVME(_add + SIS3300_MEMBASE_BANK1_ADC12+(bank - 1) * addr_bank_adc_shift + i * adc_shift,
                                       (u_int32_t*) (data + iw + i * _nSamples), _nSamples);

            if (gwords != _nSamples)
            {
                std::cout << "returned gwords = "<< gwords<< " != _nSamples ( "<< _nSamples << " )"<< std::endl;
            }
        }

        data[size-1] = size;

        size *= sizeof(unsigned int);

        if (get_evt_counter() < 11)
        {
            std::cout << "Module["<< get_id() << "] "<< std::hex<< data[0]<< " "<< data[1]<< " "<< data[2]<< " "<< data[3]
                    << " ... "<< data[(data[3]) - 1]<< std::dec << std::endl << std::flush;
        }

        Arm();

        Start();

        return reinterpret_cast<char *>(data);
    }

    std::cout << "\t+-> SIS3300 Module GetData       ["<< get_id() << "]"<< std::endl;
    return "nothing";
}

int SIS3300module::PrepareForRun()
{
    VMEmanager *vme = VMEmanager::VME();
    if (vme->connected() )
    {
        Reset();
        InitEvent();
        Arm();

        if ( !get_slave() )
            get_analysis()->init_run();
    }

    std::cout << "\t+-> SIS3300 Module PrepareForRun ["<< get_id() << "]"<< std::endl;
    return 0;
}

int SIS3300module::Arm(int)
{
    VMEmanager *vme = VMEmanager::VME();
    if (vme->connected() )
    {
        unsigned long arm_banks = acq3300_set |ACQ_ENABLE_CLOCK_TO_BANK1|ACQ_ENABLE_CLOCK_TO_BANK2;

        vme->writeA32D32(_add + SIS3300_ACQUISTION_CONTROL, arm_banks );
    }

    return 0;
}

int SIS3300module::Start(int)
{
    VMEmanager *vme = VMEmanager::VME();
    if (vme->connected() )
    {
        int _mode = get_mode();
        unsigned long arm_banks = acq3300_set |ACQ_ENABLE_CLOCK_TO_BANK1|ACQ_ENABLE_CLOCK_TO_BANK2;

        vme->writeA32D32(_add + SIS3300_ACQUISTION_CONTROL, arm_banks);

        if (_mode == 0)
        {
            vme->writeA32D32(_add + SIS3300_KEY_START, 0);
        }

    }
    return 0;    
}

int SIS3300module::Stop(int)
{
    VMEmanager *vme = VMEmanager::VME();
    if (vme->connected() )
    {
        vme->writeA32D32(_add + SIS3300_KEY_STOP, 0);

        if ( !get_slave() )
            get_analysis()->end_run();
    }

    std::cout << "\t+-> SIS3300 Module Stop          ["<< get_id() << "]"<< std::endl;
    return 0;
}

int SIS3300module::Pause()
{
    std::cout << "\t+-> SIS3300 Module Pause         ["<< get_id() << "]"<< std::endl;
    return 0;
}

int SIS3300module::Continue()
{
    std::cout << "\t+-> SIS3300 Module Continue      ["<< get_id() << "]"<< std::endl;
    return 0;
}

int SIS3300module::Reset()
{
    VMEmanager *vme = VMEmanager::VME();
    if (vme->connected() )
        vme->writeA32D32(_add + SIS3300_KEY_RESET, 0);

    std::cout << "\t+-> SIS3300 Module Reset         ["<< get_id() << "]"<< std::endl;
    return 0;
}

int SIS3300module::GetReady()
{
    std::cout << "\t+-> SIS3300 Module GetReady      ["<< get_id() << "]"<< std::endl;
    return 0;
}

void SIS3300module::set_header_description(unsigned int &data)
{
    unsigned int id = get_id();
    data = (get_nInputs() & 0xff) << 20| get_data_type() | (id & 0xffff);
}

// *** module specific methods
void SIS3300module::create_analysis(int nchip, int nchan)
{
    if (this->ana)
        delete this->ana;

    this->ana = new SIS3300moduleAna(nchip, nchan);
}

int SIS3300module::InitEvent()
{
    VMEmanager *vme = VMEmanager::VME();
    if (vme->connected() )
    {
        int _nSamples = get_nSamples();
        int _delay = get_delay();
        int _mode = get_mode();

        if (_mode == 0) // LEMO start/stop mode
        {
            acq3300_set = ACQ_ENABLE_LEMO_START_STOP |
            ACQ_ENABLE_STOP_DELAY;
            vme->writeA32D32(_add + SIS3300_STOP_DELAY, _delay);
        }
        else if (_mode == 1) // gate mode
        {
            acq3300_set = ACQ_ENABLE_LEMO_START_STOP |
            ACQ_ENABLE_GATE_MODE;
        }
        else
        {
            std::cout << "\t+-> SIS3300 Module InitEvent     ["<< get_id() << "]"<< std::endl<< "\t+     Error: mode = "<< _mode
                    << " is invalid"<< std::endl;
            return 0;
        }

        vme->writeA32D32(_add + SIS3300_ACQUISTION_CONTROL, acq3300_set);

        unsigned int config_flag = 0;

        switch (_nSamples)
        {
        case 0x80:
            config_flag = CONF_PAGE_SIZE_128_WRAP;
            break;
        case 0x100:
            config_flag = CONF_PAGE_SIZE_256_WRAP;
            break;
        case 0x200:
            config_flag = CONF_PAGE_SIZE_512_WRAP;
            break;
        case 0x400:
            config_flag = CONF_PAGE_SIZE_1K_WRAP;
            break;
        case 0x800:
            config_flag = CONF_PAGE_SIZE_2K_WRAP;
            break;
        case 0x1000:
            config_flag = CONF_PAGE_SIZE_4K_WRAP;
            break;
        case 0x4000:
            config_flag = CONF_PAGE_SIZE_16K_WRAP;
            break;
        case 0x20000:
            config_flag = CONF_PAGE_SIZE_128K_WRAP;
            break;
        default:
            std::cout << "Nsamples="<< _nSamples << " is not a valid page size"<< std::endl;
        }

        vme->writeA32D32(_add + SIS3300_EVENT_CONFIG_ALL_ADC, config_flag);
    }

    std::cout << "\t+-> SIS3300 Module InitEvent     ["<< get_id() << "]"<< std::endl;
    return 0;
}

int SIS3300module::InitPedestal()
{
    VMEmanager *vme = VMEmanager::VME();
    if (vme->connected() )
    {

        unsigned long acq3300_set = 0;

        Reset();

        acq3300_set = ACQ_ENABLE_AUTOSTART |
        ACQ_ENABLE_CLOCK_TO_BANK1 |
        ACQ_ENABLE_CLOCK_TO_BANK2;

        vme->writeA32D32(_add + SIS3300_ACQUISTION_CONTROL, acq3300_set);

        unsigned long config_flag= CONF_PAGE_SIZE_128K_AUTOSTOP;

        vme->writeA32D32(_add + SIS3300_EVENT_CONFIG_ALL_ADC, config_flag);
    }

    std::cout << "\t+-> SIS3300 Module InitPedestal  ["<< get_id() << "]"<< std::endl;
    return 0;
}

int SIS3300module::ReadData(unsigned int *data)
{
    int _nInputs = get_nInputs();
    int _nSamples = get_nSamples();
    int ntot = ((_nInputs + 1) / 2) * _nSamples;
    int bank = 1;//for now always using bank 1
    int id, iw;
    int size;
    VMEmanager *vme = VMEmanager::VME();
    if (vme->connected() )
    {

        increase_evt_counter();

        // Allocate space for the data. This has to be done with the
        // allocate method of DAQpp::Module
        //
        // Note that here we assume that we have one word per channel. It
        // may well happen that we get from the module a complicated
        // structure that we do not want ot decode here, so we just dumpt
        // it into the header and the trailer words and leave the analysis
        // program with the burden of decoding.

        size = 7 + // the 5 words in the header plus the trailing word
                ntot;

        /*
         SIS3300 header

         In header we store number of channels & number of samples for each
         module. 

         SIS3300 data block

         Data Format
         0      0xCAFEibbb      i=record type,  bbb=ignored
         1      0xaaaaaaaa      aaaaaaaa=event counter
         2      0xiiitnnnn      iii=module id,  t=module type, nnnn=unused
         3      0xbbbbbbbb      bbbbbbbb=size of data block
         4      0xcsssssss      c=# of inputs, sssssss=sample size
         5      0x000(w|r)rrrr  w=wrap_bit (@ bit 19), rrrrr=ring pointer[17 bit]
         6      data word 0
         .         .
         x      0xbbbbbbbb      size of data block

         Two options for data word: raw, proccessed

         RAW:
         data word = 0x0uuu0vvv uuu=even ch (0,2,4,..,ccc-2) raw value (12 bits)
         vvv=odd ch (1,3,5,... ccc-1) raw adc value (12 bits) 
         There are sssssss words per channel pair and (c+1)/2 channel pairs
         giving in total N=(sssssss)*(c+1)/2 data words

         Size of data block is bbbbbbbb=N+6. Since N can be larger than
         0xfff (=4096) instead of size(=bbb) the first word carries record type and 
         size is given in the next data word.
         */

        // Fill the data array
        id = get_id();
        iw = 0;
        unsigned long gwords;
        VMEmanager *vme = VMEmanager::VME();
        data[iw++] = DataHeader | BigDataRecord;
        data[iw++] = get_evt_counter();
        data[iw++] = ((id & 0xfff) << 20) | SIS3300data;
        data[iw++] = size;
        data[iw++] = ((_nSamples) & 0xfffffff) | ((_nInputs & 0xf) << 28);
        //get ring pointer and wrap bit
        data[iw++] = vme->readA32D32(_add + SIS3300_EVENT_DIRECTORY_BANK1_ALL_ADC+(bank - 1) * event_directory_bank_shift);
        //fill buffer with data from SIS3300 databank
        for (int i = 0; i < (_nInputs + 1) / 2; i++)
        {
            gwords = vme->readA322EVME(_add + SIS3300_MEMBASE_BANK1_ADC12+(bank - 1) * addr_bank_adc_shift + i * adc_shift,
                                       (u_int32_t*) (data + iw + i * _nSamples), _nSamples);

            if (gwords != _nSamples)
            {
                std::cout << "returned gwords = "<< gwords<< " != _nSamples ( "<< _nSamples << " )"<< std::endl;
            }
        }

        data[size-1] = size;

        if (get_evt_counter() < 11)
        {
            std::cout << "Module["<< get_id() << "] "<< std::hex<< data[0]<< " "<< data[1]<< " "<< data[2]<< " "<< data[3]
                    << " ... "<< data[(data[3]) - 1]<< std::dec << std::endl << std::flush;
        }

        return size;
    }

    std::cout << "\t+-> SIS3300 Module ReadData      ["<< get_id() << "]"<< std::endl;
    return 0;
}

