#ifndef __SIS3300module_h__
#define __SIS3300module_h__
/**
 * This is a module for vmedaq.
 * They all should inherit from GenModule, which adds some extra
 * functionality on top of DAQpp::Module:
 *
 *   - The module works around the concept of something with a number
 *     of chips, each having the same number of channels. So nchip
 *     (number of chips) and nchan (number of channels per chip) is
 *     something we should supply somehow.
 *
 *     For the SIS3300 module we replace nchip by the number of channels (8)
 *     and nchan by the sample size (== page size)
 *
 *     In paticular for the
 *     Analysis instance of the module. That brings me to the second
 *     point.
 *
 *   - There is an Analysis pointer (you have to create the object)
 *     which will store all the monitoring stuff. This guy needs the
 *     number of chips and channels per chip.
 *
 *   - If you plan to make scans with that module, you will need to
 *     override the set_scan_point method to set your module in a
 *     reasonable state every time you are in a new scan point.
 *
 *   - pedestals and noise are there as DAQ++ parameters
 *      + can be updated from the values in Analysis or copied into
 *        the Analysis
 *
 *      + Pedestals are stored in the data file if you override the
 *      write_pedestals method.
 *
 *
 * See the DAQ++ documentation for further information
 *
 **/
#include <daq/GenModule.h>
#include "sis3300.h"

class SIS3300module : public GenModule
{
private:
    static const int maxInputs = 8;
    static const int maxSamples = 0x20000;
    static const int maxNevts = 1024;
    static const unsigned int event_directory_bank_shift = maxNevts*sizeof(u_int32_t);
    static const u_int32_t adc_shift= MEMSIZE*sizeof(u_int32_t);
    static const u_int32_t addr_bank_adc_shift = (maxInputs + 1) / 2*MEMSIZE*sizeof(u_int32_t);
    static const u_int32_t ADCmask = 0xfff;
    unsigned long acq3300_set;

public:
    /** Constructor
     *  The constructor needs 2 parameters:
     *  id: the Module id
     *  address: the address (VME, or whatever protocol) to
     *           communicate with the module
     *
     *  Note: although DAQ++ allows for any type of identifier,
     *  however VMEDAQ only handles properly integer IDs. The only
     *  reason for that is that it needs integer values to save them
     *  properly in the data file
     *
     **/
    SIS3300module(const DAQpp::DAQid &id, unsigned long address);

    /** This is the method called by the RunManager to know if there
     * is data ready in the module. If the answer if true, then it
     * will try to read the module. If there is no way to know a
     * priory about the availavility of data, one can defer the
     * decision until getData
     *
     **/
    bool hasData();

    /** This is the method called to retrieve the data. If hasData
     * could not figure out the right answer, one can still say here
     * that there is no data at all by returning a NULL pointer and
     * setting size to 0.
     *
     * The memory to store the data should be allocated with the
     * allocate method. See the DAQ++ documentation...
     *
     **/
    char *getData(int &size);

    int PrepareForRun();

    int Arm(int = 0);

    int Start(int = 0);

    int Stop(int = 0);

    int Pause();

    int Continue();

    int Reset();

    int GetReady();

    int Trigger()
    {
        return 0;
    }
    ;
    void flush()
    {
    }
    ;

    unsigned long get_data_type() const;
    
    void set_header_description(unsigned int &data);
    // *** module specific methods
    // method for creating/recreating Analysis
    void create_analysis(int nchip = 1, int nchan = 1);
    // initialize the board to run in single event mode
    int InitEvent();
    // initialize the board to run in pedestal mode
    // currently unused
    int InitPedestal();
    // pass the data as unsigned long, called by SecondDetector
    int ReadData(unsigned int *data);

    // * get and set methods
    // simulation parameters
    double get_mean()
    {
        return DAQpp::get_par_value<double>(get_par("mean"));
    }
    void set_mean (double x)
    {
        set_par<double>("mean", x);
    }
    double get_width ()
    {
        return DAQpp::get_par_value<double>(get_par("width"));
    }
    void set_width (double x)
    {
        set_par<double>("width", x);
    }
    double get_efficiency ()
    {
        return DAQpp::get_par_value<double>(get_par("efficiency"));
    }
    void set_efficiency (double x)
    {
        set_par<double>("efficiency", x);
    }
    // module parameters
    bool get_slave ()
    {
        return DAQpp::get_par_value<bool>(get_par("slave"));
    }
    void set_slave (bool x)
    {
        set_par<bool>("slave", x);
    }
    bool get_softstart ()
    {
        return DAQpp::get_par_value<bool>( get_par( "softstart" ) );
    }
    void set_softstart ( bool x )
    {
        set_par<bool>( "softstart", x );
    }
    int get_nInputs ()
    {
        return DAQpp::get_par_value<int>(get_par("nInputs"));
    }
    void set_nInputs (int x)
    {
        set_par<int>("nInputs", x);
    }
    int get_nSamples ()
    {
        return DAQpp::get_par_value<int>(get_par("nSamples"));
    }
    void set_nSamples (int x)
    {
        set_par<int>("nSamples", x);
    }
    int get_delay ()
    {
        return DAQpp::get_par_value<int>(get_par("delay"));
    }
    void set_delay (int x)
    {
        set_par<int>("delay", x);
    }
    int get_mode ()
    {
        return DAQpp::get_par_value<int>(get_par("mode"));
    }
    void set_mode (int x)
    {
        set_par<int>("mode", x);
    }
    int get_threshold1 ()
    {
        return DAQpp::get_par_value<int>(get_par("threshold1"));
    }
    void set_threshold1 (int x)
    {
        set_par<int>("threshold1", x);
    }
    int get_threshold2 ()
    {
        return DAQpp::get_par_value<int>(get_par("threshold2"));
    }
    void set_threshold2 (int x)
    {
        set_par<int>("threshold2", x);
    }
    int get_threshold3 ()
    {
        return DAQpp::get_par_value<int>(get_par("threshold3"));
    }
    void set_threshold3 (int x)
    {
        set_par<int>("threshold3", x);
    }
    int get_threshold4 ()
    {
        return DAQpp::get_par_value<int>(get_par("threshold4"));
    }
    void set_threshold4 (int x)
    {
        set_par<int>("threshold4", x);
    }
    int get_threshold5 ()
    {
        return DAQpp::get_par_value<int>(get_par("threshold5"));
    }
    void set_threshold5 (int x)
    {
        set_par<int>("threshold5", x);
    }
    int get_threshold6 ()
    {
        return DAQpp::get_par_value<int>(get_par("threshold6"));
    }
    void set_threshold6 (int x)
    {
        set_par<int>("threshold6", x);
    }
    int get_threshold7 ()
    {
        return DAQpp::get_par_value<int>(get_par("threshold7"));
    }
    void set_threshold7 (int x)
    {
        set_par<int>("threshold7", x);
    }
    int get_threshold8 ()
    {
        return DAQpp::get_par_value<int>(get_par("threshold8"));
    }
    void set_threshold8 (int x)
    {
        set_par<int>("threshold8", x);
    }
};

#endif
