#include <iostream>
#include <iomanip>
#include <cmath>
#include <config/ConfigParser.h>
#include <config/GeneralConfig.h>
#include "SIS3300moduleAna.h"

void SIS3300moduleAna::init_run()
{
    arm_WaveForms();

    cut = 30;

    ibad = 0;
    process = 1;
    simulate = false;

    std::cout << "**SIS3300moduleAna::init_run\n"<< std::endl;
}

void SIS3300moduleAna::end_run()
{
    //disarm_WaveForms();

    std::cout << "\n**SIS3300moduleAna::end_run"<< std::endl;
}

void SIS3300moduleAna::arm_WaveForms()
{
    for (int i = 0; i < Nchip(); i++)
    {
        WaveForm* wf = new WaveForm(Nchan());
        vWaveForm.push_back(wf);
    }
}

void SIS3300moduleAna::disarm_WaveForms()
{
    vWaveForm.clear();
}

bool SIS3300moduleAna::data_valid(unsigned int *data) const
{
    unsigned int _nInputs, _nSamples, sz, dsz;

    _nInputs = ((data[4] >> 28) & 0xf);
    _nSamples = (data[4] & 0xfffffff);
    dsz = data[3];
    sz = (_nInputs / 2) * _nSamples + 7;

    if ( !IsSIS3300data(data[2]) )
    {
        std::cout << "SIS3300moduleAna::data_valid  Error: DataType = "<< std::hex << (data[2] & 0xf0000)
                << "  DataType Expected = "<< SIS3300data << std::dec << std::endl << std::flush;
        return false;
    }

    // Check data integrity
    if (dsz != data[dsz-1]|| dsz - sz != 0)
    {
        std::cout << "SIS3300moduleAna::data_valid  Error: dsz = "<< std::hex << dsz << "  sz = "<< sz << std::dec<< std::endl
                << std::flush;
        return false;
    }

    // Confirm Analysis matches data
    if (_nInputs != Nchip() || _nSamples != Nchan() )
    {
        std::cout << "SIS3300moduleAna::data_valid  Error: _nInputs = "<< _nInputs<< "  Nchip() = "<< Nchip() << std::endl
                << "                                   "<< " _nSamples = "<< _nSamples << "  Nchan() = "<< Nchan()<< std::endl
                << std::flush;
        return false;
    }

    // We do not want to process twice the same event
    if (trigger >= data[1])
    {
        std::cout << "SIS3300moduleAna::data_valid  Error: trigger = "<< std::hex << trigger << "  data event = "<< data[1]
                << std::dec << std::endl << std::flush;
        return false;
    }

    return true;
}

void SIS3300moduleAna::decode(unsigned int *data)
{
    // Get some important handlers of the configuration
    // ConfigParser *cfg = ConfigParser::theConfig();

    // In order to make a fast analysis of the data, we should provide
    // the pedestal calculation routines the list of channels to be
    // considered.
    std::vector<int> list_of_channels;

    // We assume the Analysis::Fill has made all the checks on
    // the data integrity
    int _nInputs = ((data[4] >> 28) & 0xf);
    int _nSamples = (data[4] & 0xfffffff);
    u_int32_t ring_pointer = data[5];
    int i, j, nevt;

    pulse_data = Chan();
    unwrapData(data, _nInputs, _nSamples, ring_pointer, pulse_data);

    //std::cout << "SIS3300moduleAna::decode, call to unwrapData done..."
    //    << std::endl << std::flush;

    nevt = 0;
    for (i = 0; i < _nInputs; i++)
    {
        for (j = 0; j < _nSamples; j++)
        {
            Event()[nevt] = pulse_data[nevt];
            // accepting all channels
            list_of_channels.push_back(nevt);
            nevt++;
        }
    }

    // Note: we assumed here that each word corresponds to one
    // channel and that all the channels are valid. One may have the
    // scenario in which the data is perversly encoded and we only
    // find, at the end of the process a list of channels with
    // signal. In that situation, we would call
    //
    // add_hit(j)
    // Event()[j] = Chan()[j] = signal
    //
    // because we know that channel j has signal...

    if (process == 0)
    {
        //update_pedestals(list_of_channels);
        update_pedestals_fast(list_of_channels, false);
    }

}

void SIS3300moduleAna::Analyze()
{
    int i, i_use;
    double *chan = Chan();
    double nrg;

    if (process == 1)
        process_event();
    else // old routine to fill energy (works)
    {
        for (i = 0; i < Nchip(); i++)
        {
            i_use = i;

            nrg = integrate(i, 0, Nchan(), Nchan(), pulse_data);

            if (nrg > 20)
            {
                m_fill.acquire();
                HitMap()[i_use]++;
                int ib = get_bin(nrg);
                if (ib >= 0&& ib < get_nbins() )
                {
                    Signal(i_use)[ib]++;
                    Signal(-1)[ib]++;
                    Signal2D(i_use)[ib]++;
                    if (get_map())
                    {
                        int ix, iy, ip;
                        get_map()->pad(i_use, ix, iy);
                        ip = ix + get_map()->nx() * iy;
                        H2Map()[ ip ]++;
                        if (chan[i_use] >= emin && chan[i_use] <= emax )
                            H2MapC()[ ip ]++;
                    }
                }
                m_fill.release();
            }
        }
    }
    if (ibad > 0&& ibad % 10000== 0)
        std::cout << "SIS3300moduleAna:Analyze    ibad = "<< ibad << std::endl;
}

int SIS3300moduleAna::unwrapData(const unsigned int *unwrap_data, int npmts, int nsamples, u_int32_t ring_pointer,
                                 double *pulse_data)
{
    u_int32_t ring_pointer0 = (ring_pointer & 0x1ffff);
    if ((int)ring_pointer0 >= nsamples)
    {
        std::cout << "Error: ring_pointer = "<< ring_pointer0<< ", nsamples = "<< nsamples << std::endl;
    }
    //was the readout wrapped already?
    bool wrap_bit0 = (ring_pointer & 0x80000);

    //decode to pulse_pmt_shape
    //if wrap start at ringpointer, else increase to next level
    int start = !wrap_bit0 ? nsamples : ring_pointer0;
    //stop is at next level
    int stop = ring_pointer0 + nsamples;

    for (int i = 0; i < (npmts + 1) / 2; i++)
    {
        for (int j = start, k = 0; j < stop; j++, k++)
        {
            pulse_data[2*i*nsamples+k] =((unwrap_data[6+i*nsamples+j%nsamples] >> 16) & 0xfff);
            pulse_data[(2*i+1)*nsamples+k] =(unwrap_data[6+i*nsamples+j%nsamples] & 0xfff);
        }
    }

    return (wrap_bit0 ? nsamples : ring_pointer0);
}

double SIS3300moduleAna::integrate(int ipmt, int low, int high, int nsamples, double *pulse_pmt_shape)
{
    int start = low < 0 ? 0 : low;
    int stop = high > nsamples ? high : nsamples;

    double *pmt_value = pulse_pmt_shape + ipmt * nsamples;

    double sum = 0;
    for (int k = start; k < stop; k++)
    {
        sum += pmt_value[k];
    }
    sum /= nsamples;

    return sum;
}

void SIS3300moduleAna::process_event()
{
    std::vector<WaveForm*>::iterator it;

    double nrg_scale = 1023;

    for (it = vWaveForm.begin(); it != vWaveForm.end(); ++it)
        (*it)->data_start();

    double ttime = calculate_ttime();

    if (ttime == 0)
    {
        ibad++;
        return;
    }

    //propagate final event time to modules
    for (it = vWaveForm.begin(); it != vWaveForm.end(); ++it)
        (*it)->set_ttime(ttime);

    // sets weights for position calculations and sets the matrix elements
    for (it = vWaveForm.begin(); it != vWaveForm.end(); ++it)
        (*it)->set_weights(100);

    int i_use = 0;
    for (it = vWaveForm.begin(); it != vWaveForm.end(); ++it)
    {
        double nrg = nrg_scale * fabs((*it)->get_energy());

        if (nrg > cut )
        {

            if (false)
            {
                std::cout << "SIS3300moduleAna::process_event   energy = "<< nrg << std::endl;
            }

            m_fill.acquire();
            HitMap()[i_use]++;
            int ib = get_bin(nrg);
            if (ib >= 0&& ib < get_nbins() )
            {
                Signal(i_use)[ib]++;
                Signal(-1)[ib]++;
                Signal2D(i_use)[ib]++;
                if (get_map())
                {
                    int ix, iy, ip;
                    get_map()->pad(i_use, ix, iy);
                    ip = ix + get_map()->nx() * iy;
                    H2Map()[ ip ]++;
                    if (pulse_data[i_use] >= emin && pulse_data[i_use] <= emax )
                        H2MapC()[ ip ]++;
                }
            }
            else if (false)
            {
                std::cout << "SIS3300moduleAna::process_event   channel = "<< i_use << " energy = "<< nrg << " out of range"
                        << std::endl;
            }
            m_fill.release();
        }

        double pedestal = (*it)->get_pedestal();
        for (int i = 0; i < Nchan(); i++)
            Ped(i_use)[i] = pedestal;

        i_use++;
    }
}

double SIS3300moduleAna::calculate_ttime()
{
    std::vector<WaveForm*>::iterator it;
    double ttime = 0, sum = 0, sumW = 0;

    //estimate time of event
    int ifire = 0;
    for (it = vWaveForm.begin(); it != vWaveForm.end(); ++it )
    {
        double fRise = (*it)->energy_estimate();
        if (false)
        {
            std::cout << "SIS3300moduleAna::calculate_ttime   energy_estimate = "<< fRise << std::endl;
        }
        if (fRise > cut)
        {
            double tt = (*it)->improve_trigger_time();
            sum += fRise * tt;
            sumW += fRise;
            ifire++;
        }
    }

    if (ifire > 0&& sumW > 1e-7)
        ttime = sum / sumW;

    return ttime;
}

void SIS3300moduleAna::simulate_event()
{
    int i;
    std::vector<WaveForm*>::iterator it;
    for (i = 0, it = vWaveForm.begin(); it != vWaveForm.end(); i++, ++it )
        (*it)->sim_data();
}
