/**
 This class is responsible to parse the corresponding section in the
 configuration file to configure the module. It is also responsible
 for create the module instance according to the paremeters retrived
 from the configuration file.

 It derives from ModuleParser (see ModuleParser.h in the config
 directory) which, in turn, derives from SectionParser (again, have
 a look at the header file).

 It has, then, to implement a number of methods from SectionParser

 - parse_node:  this method will parse the corresponding section
 and store locally the configuration parameters
 defined in there.
 - save_config: this method will be called if the configuration
 is to be saved to an external file. The idea is
 that the class will check which parameters have
 changed since the file was loaded and update them
 in the document before saving.

 and a few more from ModuleParser:

 - config : called by the gui to change configuration parameters
 - monitor: called by the gui to create the monitor display
 - decode : called to decode the data
 - range_min, range_max: to retrieve the range of the monitor
 histograms

 One can also define here the factory function specified in the
 configuration file.

 FOr this module we will assume that we only have 2 parameters to
 appear in the configuration file:
 - the amplitude of the pulse that we simulate and,
 - the width of the peak we want to simulate.

 So in the modules section we add

 <modules>
 .
 .
 .
 <test_module id="100" address="deadbeef">
 <!-- monitor is a section common to all the modules -->
 <monitor>
 ...
 </monitor>

 <!-- padmap is also common to all the modules -->
 <padmap>
 .....
 </padmap>

 <!-- and here come our parameters -->
 <mean>100</mean>
 <width>5</width>

 </test_module>
 .
 .
 .
 </modules>
 */
#include <string>
#include <sstream>
#include <DAQ++/DAQmanager.h>
#include <config/ConfigParser.h>
#include <config/ModuleParser.h>
#include <config/GeneralConfig.h>
#include <gtk/gtkwidget.h>
#include <gui/GladeApp.h>
#include <gui/WidgetUpdate.h>
#include <gui/MonitorWindow.h>
#include <data/Analysis.h>
#include <data/DataFormat.h>
#include "SIS3300module.h"
#include "SIS3300moduleGuiConf.h"
#include "SIS3300moduleGuiMon.h"

class SIS3300moduleConf : public ModuleParser
{
private:
    std::string _glade_file;

public:
    SIS3300moduleConf()
    {
        //_glade_file =
        //"/net/babar5/tem2/ecochran/vmedaq_devel/vmedaq-0.5.1-1/gui/glade/vmedaq.glade";
    }
    ~SIS3300moduleConf()
    {
    }

    /**
     * This is the method responsible for parsing the configuration
     * file. 
     **/
    void parse_node(xmlNodePtr node) throw(ConfigException);

    /**
     * The method that saves the current configuration. It should
     * update all the xml nodes below the module node.
     **/
    void save_config() throw(ConfigException);

    /**
     * This is called by VMEDAQ to create a window that will allow
     * to change the configuration of the module while running the
     * program. Probably the easiest is to use glade-2 to create the
     * layout of the window, define the signals (callbacks that are
     * associated to given actions) and then something derived from
     * GladeApp (see the gui folder) to implement the stuff and
     * update the module state accordingly.
     **/
    GtkWidget *config(const std::string &id);

    /**
     * This is called by VMEDAQ to create a window to display the
     * monitor data. The input parameters are:
     *    - id: the module id (that should help finding the module)
     *    - glade_file: The path of the glade file. This is, in
     *      principle the path of the standard VMEDAQ glade
     *      file. One can certainly ignore this value.
     **/
    MonitorWindow *monitor(const std::string &id, const std::string &glade_file);

   
};

/**
 * This is the hook specified in the config file to return a pointer
 * to the SIS3300moduleConf instance that will interact with VMEDAQ.
 *
 * Note that it should be declared as extern "C" so that no demangling
 * mechanism is applied to its name that prevent finding easily the
 * symbol in the library
 *
 **/
extern "C"
{
SIS3300moduleConf *create_SIS3300moduleConf()
{
    SIS3300moduleConf *cnf = new SIS3300moduleConf();
    return cnf;
}
}

/**
 * Now come the implementations of the class methods
 */
void SIS3300moduleConf::parse_node(xmlNodePtr _node) throw(ConfigException)
{
    unsigned long address;
    std::string saddress;

    double mean;
    double width;
    double efficiency;
    bool slave;
    bool softstart;
    int nInputs;
    int nSamples;
    int delay;
    int mode;
    int threshold1;
    int threshold2;
    int threshold3;
    int threshold4;
    int threshold5;
    int threshold6;
    int threshold7;
    int threshold8;

    // The configuration manager handler
    ConfigParser *cfg = ConfigParser::theConfig();

    // store the node pointer
    node = _node;

    // get the module id
    id = cfg->get_attribute<std::string>(node, "name");

    // get the address
    saddress = cfg->get_attribute<std::string>(node, "address");

    std::istringstream os(saddress);
    os >> std::hex >> address;

    // Check if the module is already there
    DAQpp::Module *dm;
    SIS3300module *module = 0;

    if ( (dm = DAQpp::DAQmanager::theDAQmanager()->find_module(id)))
        module = dynamic_cast<SIS3300module *>(dm);

    if (!module)
    {
        // The module is not there... create
        module = new SIS3300module(id, address);
    }

    // Now we have the module. From now own one should parse the
    // section from the configuration file
    try
    {
        _glade_file = cfg->get_value<std::string>(node, "guiFile");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
    }

    try
    {
        mean = cfg->get_value<double>(node, "mean");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
    }

    try
    {
        width = cfg->get_value<double>(node, "width");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
    }

    try
    {
        efficiency = cfg->get_value<double>(node, "efficiency");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
    }

    try
    {
        slave = cfg->get_value<bool>(node, "slave");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
    }

    try
    {
        softstart = cfg->get_value<bool>( node, "softstart" );
    }
    catch ( ConfigException &e )
    {
        // IT was not there... leave the default or current value
    }

    try
    {
        nInputs = cfg->get_value<int>(node, "nInputs");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
    }

    try
    {
        nSamples = cfg->get_value<int>(node, "nSamples");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
    }

    try
    {
        delay = cfg->get_value<int>(node, "delay");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
    }

    try
    {
        mode = cfg->get_value<int>(node, "mode");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
    }

    try
    {
        threshold1 = cfg->get_value<int>(node, "threshold1");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
    }

    try
    {
        threshold2 = cfg->get_value<int>(node, "threshold2");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
    }

    try
    {
        threshold3 = cfg->get_value<int>(node, "threshold3");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
    }

    try
    {
        threshold4 = cfg->get_value<int>(node, "threshold4");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
    }

    try
    {
        threshold5 = cfg->get_value<int>(node, "threshold5");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
    }

    try
    {
        threshold6 = cfg->get_value<int>(node, "threshold6");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
    }

    try
    {
        threshold7 = cfg->get_value<int>(node, "threshold7");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
    }

    try
    {
        threshold8 = cfg->get_value<int>(node, "threshold8");
    }
    catch (ConfigException &e)
    {
        // IT was not there... leave the default or current value
    }

    module->set_par<double>("mean", mean);
    module->set_par<double>("width", width);
    module->set_par<double>("efficiency", efficiency);
    module->set_par<bool>("slave", slave);
    module->set_par<bool>("softstart", softstart );
    module->set_par<int>("nInputs", nInputs);
    module->set_par<int>("nSamples", nSamples);
    module->set_par<int>("delay", delay);
    module->set_par<int>("mode", mode);
    module->set_par<int>("threshold1", threshold1);
    module->set_par<int>("threshold2", threshold2);
    module->set_par<int>("threshold3", threshold3);
    module->set_par<int>("threshold4", threshold4);
    module->set_par<int>("threshold5", threshold5);
    module->set_par<int>("threshold6", threshold6);
    module->set_par<int>("threshold7", threshold7);
    module->set_par<int>("threshold8", threshold8);
    if (slave )
        module->in_local();

    module->create_analysis(nInputs, nSamples);

    // Now parse the things which are common to all the modules, that
    // is the monitor and padmap elements
    parse_common("//module", id );
}

void SIS3300moduleConf::save_config() throw(ConfigException)
{
    SIS3300module *m = 0;
    DAQpp::Module *dm;

    // find the module
    if ( !(dm = DAQpp::DAQmanager::theDAQmanager()->find_module(id)))
        return;

    if ( !(m = dynamic_cast<SIS3300module *>(dm)))
        return;

    // The configuration manager handler
    ConfigParser *cfg = ConfigParser::theConfig();

    // Save common stuff
    save_common();

    // Save the current values in the configuration
    cfg->check_value<double>(node, "mean", m->get_mean());
    cfg->check_value<double>(node, "width", m->get_width());
    cfg->check_value<double>(node, "efficiency", m->get_efficiency());
    cfg->check_value<bool>(node, "slave", m->get_slave());
    cfg->check_value<bool>(node, "softstart", m->get_softstart() );
    cfg->check_value<int>(node, "nInputs", m->get_nInputs());
    cfg->check_value<int>(node, "nSamples", m->get_nSamples());
    cfg->check_value<int>(node, "delay", m->get_delay());
    cfg->check_value<int>(node, "mode", m->get_mode());
    cfg->check_value<int>(node, "threshold1", m->get_threshold1());
    cfg->check_value<int>(node, "threshold2", m->get_threshold2());
    cfg->check_value<int>(node, "threshold3", m->get_threshold3());
    cfg->check_value<int>(node, "threshold4", m->get_threshold4());
    cfg->check_value<int>(node, "threshold5", m->get_threshold5());
    cfg->check_value<int>(node, "threshold6", m->get_threshold6());
    cfg->check_value<int>(node, "threshold7", m->get_threshold7());
    cfg->check_value<int>(node, "threshold8", m->get_threshold8());
}



//=====================================================================
// GUI
//=====================================================================
GtkWidget *SIS3300moduleConf::config(const std::string &id)
{
    SIS3300module *m = 0;
    DAQpp::Module *dm;

    if ( !(dm = DAQpp::DAQmanager::theDAQmanager()->find_module(id)))
        return 0;

    if ( !(m = dynamic_cast<SIS3300module *>(dm)))
    {
        std::cout << "SIS3300moduleConf: module "<< id<< " is not of type SIS3300module"<< std::endl;
        return 0;
    }

    SIS3300moduleGuiConf *conf = new SIS3300moduleGuiConf(m->get_id());
    conf->init(_glade_file.c_str(), "SIS3300-Config");

    return conf->get_main_window();
}

MonitorWindow *SIS3300moduleConf::monitor(const std::string &id, const std::string &glade_file)
{
    SIS3300module *m = 0;
    DAQpp::Module *dm;

    if ( !(dm = DAQpp::DAQmanager::theDAQmanager()->find_module(id)))
        return 0;

    if ( !(m = dynamic_cast<SIS3300module *>(dm)))
    {
        std::cout << "SIS3300moduleConf: module "<< id<< " is not of type SIS3300module"<< std::endl;
        return 0;
    }

    SIS3300moduleGuiMon *mon = new SIS3300moduleGuiMon(m->get_id());
    mon->init(_glade_file.c_str(), "SIS3300-Monitor");

    MonitorWindow *mw = new MonitorWindow(mon->get_main_window());
    return mw;
}
