#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <daq/VMEmanager.h>
#include <DAQ++/DAQmanager.h>
#include <data/Analysis.h>
#include <data/DataFormat.h>
#include "SecondDetector.h"
#include "SecondDetectorAna.h"
#include "SIS3300module.h"

#ifdef HAS_USB
#include "usb/pmtUSB.h"
class pmtUSB;
#endif

SecondDetector::SecondDetector(const DAQpp::DAQid &id, unsigned long address) :
    GenModule(id, address)
{

    // Using DAQ++ parameters can be useful, especially if you want to
    // access them the DAQcorba
    // module parameters
    add_par("nAmplifiers", new DAQpp::Par<int>(0));
    add_par("nModules", new DAQpp::Par<int>(1));
    add_par("nInputs", new DAQpp::Par<int>(8));
    add_par("nSamples", new DAQpp::Par<int>(256));
    add_par("delay", new DAQpp::Par<int>(200));
    add_par("threshold1", new DAQpp::Par<int>(0));
    add_par("threshold2", new DAQpp::Par<int>(0));
    add_par("threshold3", new DAQpp::Par<int>(0));
    add_par("threshold4", new DAQpp::Par<int>(0));
    add_par("threshold5", new DAQpp::Par<int>(0));
    add_par("threshold6", new DAQpp::Par<int>(0));
    add_par("threshold7", new DAQpp::Par<int>(0));
    add_par("threshold8", new DAQpp::Par<int>(0));

    // reset pedestals and noise to sensible values
    fill_n(back_inserter(get_par_value<std::vector<double> > ("pedestal") ), maxInputs*maxSamples, 0.);
    fill_n(back_inserter(get_par_value<std::vector<double> > ("noise") ), maxInputs*maxSamples, 0.);

    // now we should create the Analysis for the monitoring
    // never forget to do this !!!
    create_analysis();
    std::cout << "SecDetData = "<< std::hex << SecDetData << std::dec << std::endl;
}

unsigned long SecondDetector::get_data_type() const
{
    return SecDetData;
}

bool SecondDetector::hasData()
{
    bool data_ready = false;
    int start_id = get_id();
    int iter_id = 0;
    DAQpp::GTimer tim;
    double delay = 1.;

    int iwhile = 0;
    while ( !data_ready )
    {
        data_ready = true;

        for (int im = 0; im < get_nModules(); im++)
        {
            iter_id = start_id + im + 1;
            SIS3300module *m = 0;
            DAQpp::Module *dm;

            std::ostringstream oss(std::ostringstream::out);
            oss << iter_id;

            // find the module
            if ( !(dm = DAQpp::DAQmanager::theDAQmanager()->find_module(oss.str())))
            {
                std::cout << "Error::hasData(), module "<< iter_id << " not found"<< std::endl;
                return false;
            }

            if ( !(m = dynamic_cast<SIS3300module *>(dm)))
            {
                std::cout << "Error::hasData(), SIS3300module "<< iter_id << " not found"<< std::endl;
                return false;
            }

            data_ready = (m->hasData() && data_ready);
        }
        iwhile++;

        if ( !data_ready && tim() > delay )
        {
            std::cout << "SecondDetector::hasData Error: time exceeded delay"<< std::endl<< "    while loops completed: "<< iwhile
                    << std::endl;

            char fname[200];
            sprintf(fname, "%s/.vmedaq/log_SecDet_error.txt", getenv("HOME") );
            std::ofstream of(fname, std::ios::app );//temporary fix
            if ( !of.is_open() )
            {
                std::cout << "Could not open file "<< fname << " for logging"<< std::endl;
            }
            time_t date;
            time( &date );
            of << std::endl << ctime( &date ) << std::endl<< "  Event Number : "<< get_evt_counter() << std::endl
                    << "  While Loops  : "<< iwhile << std::endl<< "  Modules      :";

            for (int im = 0; im < get_nModules(); im++)
            {
                iter_id = start_id + im + 1;
                SIS3300module *m = 0;
                DAQpp::Module *dm;

                std::ostringstream oss(std::ostringstream::out );
                oss << iter_id;

                if ( !(dm = DAQpp::DAQmanager::theDAQmanager()->find_module(oss.str() ) ))
                {
                    std::cout << "Error::hasData(), module "<< iter_id << " not found"<< std::endl;
                    return false;
                }

                if ( !(m = dynamic_cast<SIS3300module *>(dm ) ))
                {
                    std::cout << "Error::hasData(), SIS3300module "<< iter_id << " not found"<< std::endl;
                    return false;
                }

                if ( !m->hasData() )
                {
                    of << " "<< iter_id;
                    m->Stop();
                }
            }
            of << std::endl;
            return true;
        }
    }

    return data_ready;
}

char*SecondDetector::getData(int &size)
{
    int _nAmplifiers = get_nAmplifiers();
    int _nModules = get_nModules();
    int _nInputs = get_nInputs();
    int _nSamples = get_nSamples();
    int ntot = ((_nInputs + 1) / 2) * _nSamples;
    int id, iw;

    unsigned int *data;

    increase_evt_counter();

    // Allocate space for the data. This has to be done with the
    // allocate method of DAQpp::Module
    //
    // Note that here we assume that we have one word per channel. It
    // may well happen that we get from the module a complicated
    // structure that we do not want ot decode here, so we just dumpt
    // it into the header and the trailer words and leave the analysis
    // program with the burden of decoding.

    size = 6 + // the 5 words in the header plus the trailing word
            (7 + ntot) * _nModules;

    data =reinterpret_cast<unsigned int *>(allocate(sizeof(unsigned int) * size));

    /*
     SecondDetector Data Format

     0      0xCAFEibbb      i=record type  bbb=ignored
     1      0xaaaaaaaa      aaaaaaaa=event counter
     2      0xiiitnnmm      iii=module id t=data type nn=#amp mm=#modules
     3      0xbbbbbbbb      size of data block
     4      data word 0     begin 1st amplifier information // not implemented
     .         .  
     .      SIS3300 data blocks
     .         .  
     x      0xbbbbbbbb      size of data block
     */

    // Fill the data array
    id = get_id();
    iw = 0;
    data[iw++] = DataHeader | BigDataRecord;
    data[iw++] = get_evt_counter();
    data[iw++] = ( (id & 0xfff) << 20) | SecDetData |(_nModules & 0xff)| ((_nAmplifiers & 0xff) << 8);
    data[iw++] = size;
    data[iw++] = (_nSamples & 0xfffffff) | ((_nInputs & 0xf) << 28);

    data[size-1] = size;

    int start_id = get_id();
    int iter_id = 0;

    //for (int im=0;im<7;im++)
    for (int im = 0; im < _nModules; im++)
    {
        //iter_id = start_id +1;
        iter_id = start_id + im + 1;
        SIS3300module *m = 0;
        DAQpp::Module *dm;

        std::ostringstream oss(std::ostringstream::out);
        oss << iter_id;

        // find the module
        if ( !(dm = DAQpp::DAQmanager::theDAQmanager()->find_module(oss.str())))
        {
            std::cout << "Error::getData(), module "<< iter_id << " not found"<< std::endl;
            return false;
        }

        if ( !(m = dynamic_cast<SIS3300module *>(dm)))
        {
            std::cout << "Error::getData(), SIS3300module "<< iter_id << " not found"<< std::endl;
            return false;
        }

        m->ReadData(data + iw + im*(7 + ntot));
    }

    size *= sizeof(unsigned int);

    if (get_evt_counter() < 11)
    {
        std::cout << "Module["<< get_id() << "] "<< std::hex<< data[0]<< " "<< data[1]<< " "<< data[2]<< " "<< data[3]<< " ... "
                << data[(data[3]) - 1]<< std::dec << std::endl << std::flush;
    }

    Arm();

    Start();

    return reinterpret_cast<char *>(data);
}

int SecondDetector::PrepareForRun()
{
    int start_id = get_id();
    int iter_id = 0;

    for (int im = 0; im < get_nModules(); im++)
    {
        iter_id = start_id + im + 1;
        SIS3300module *m = 0;
        DAQpp::Module *dm;

        std::ostringstream oss(std::ostringstream::out);
        oss << iter_id;

        // find the module
        if ( !(dm = DAQpp::DAQmanager::theDAQmanager()->find_module(oss.str())))
        {
            std::cout << "Error::PrepareForRun(), module "<< iter_id << " not found"<< std::endl;
            return false;
        }

        if ( !(m = dynamic_cast<SIS3300module *>(dm)))
        {
            std::cout << "Error::PrepareForRun(), SIS3300module "<< iter_id << " not found"<< std::endl;
            return false;
        }
        m->PrepareForRun();
    }
    get_analysis()->init_run();

    std::cout << "\t+-> Second Detector PrepareForRun     ["<< get_id() << "]"<< std::endl;
    return 0;
}

int SecondDetector::Arm(int)
{

    int start_id = get_id();
    int iter_id = 0;

    for (int im = 0; im < get_nModules(); im++)
    {
        iter_id = start_id + im + 1;
        SIS3300module *m = 0;
        DAQpp::Module *dm;

        std::ostringstream oss(std::ostringstream::out );
        oss << iter_id;

        // find the module
        if ( !(dm = DAQpp::DAQmanager::theDAQmanager()->find_module(oss.str() ) ))
        {
            std::cout << "Error::Arm(), module "<< iter_id << " not found"<< std::endl;
            return false;
        }

        if ( !(m = dynamic_cast<SIS3300module *>(dm ) ))
        {
            std::cout << "Error::Arm(), SIS3300module "<< iter_id << " not found"<< std::endl;
            return false;
        }

        m->Arm();
    }
    return 0;

}

int
SecondDetector::Start ( int )
{
    int start_id = get_id();
    int iter_id = 0;

    for (int im = 0;im < get_nModules();im++)
    {
        iter_id = start_id + im + 1;
        SIS3300module *m = 0;
        DAQpp::Module *dm;

        std::ostringstream oss(std::ostringstream::out);
        oss << iter_id;

        // find the module
        if ( !(dm = DAQpp::DAQmanager::theDAQmanager()->find_module(oss.str())) )
        {
            std::cout << "Error::Start(), module " << iter_id << " not found" << std::endl;
            return false;
        }

        if ( !(m = dynamic_cast<SIS3300module *>(dm)) )
        {
            std::cout << "Error::Start(), SIS3300module " << iter_id << " not found" << std::endl;
            return false;
        }

        m->Start();
    }
    return 0;
}

int SecondDetector::Stop (int)
{
    int start_id = get_id();
    int iter_id = 0;

    for (int im = 0;im < get_nModules();im++)
    {
        iter_id = start_id + im + 1;
        SIS3300module *m = 0;
        DAQpp::Module *dm;

        std::ostringstream oss(std::ostringstream::out);
        oss << iter_id;

        // find the module
        if ( !(dm = DAQpp::DAQmanager::theDAQmanager()->find_module(oss.str())) )
        {
            std::cout << "Error::Stop(), module " << iter_id << " not found" << std::endl;
            return false;
        }

        if ( !(m = dynamic_cast<SIS3300module *>(dm)) )
        {
            std::cout << "Error::Stop(), SIS3300module " << iter_id << " not found" << std::endl;
            return false;
        }

        m->Stop();
    }
    get_analysis()->end_run();
    return 0;
}

int SecondDetector::Pause ()
{
    std::cout << "\t+-> Second Detector Pause             [" << get_id() << "]"
    << std::endl;
    return 0;
}

int SecondDetector::Continue ()
{
    std::cout << "\t+-> Second Detector Continue          [" << get_id() << "]"
    << std::endl;
    return 0;
}

int SecondDetector::Reset ()
{
    int start_id = get_id();
    int iter_id = 0;

    for (int im = 0;im < get_nModules();im++)
    {
        iter_id = start_id + im + 1;
        SIS3300module *m = 0;
        DAQpp::Module *dm;

        std::ostringstream oss(std::ostringstream::out);
        oss << iter_id;

        // find the module
        if ( !(dm = DAQpp::DAQmanager::theDAQmanager()->find_module(oss.str())) )
        {
            std::cout << "Error::Reset(), module " << iter_id << " not found" << std::endl;
            return false;
        }

        if ( !(m = dynamic_cast<SIS3300module *>(dm)) )
        {
            std::cout << "Error::Reset(), SIS3300module " << iter_id << " not found" << std::endl;
            return false;
        }

        m->Reset();
    }
    return 0;
}

int
SecondDetector::GetReady ()
{
    int _nInputs = get_nInputs();
    int _nSamples = get_nSamples();
    int _delay = get_delay();

    int start_id = get_id();
    int iter_id = 0;

    for (int im = 0;im < get_nModules();im++)
    {
        iter_id = start_id + im + 1;
        std::cout << "iter_id = " << iter_id << std::endl;
        SIS3300module *m = 0;
        DAQpp::Module *dm;

        std::ostringstream oss(std::ostringstream::out);
        oss << iter_id;

        // find the module
        if ( !(dm = DAQpp::DAQmanager::theDAQmanager()->find_module(oss.str())) )
        {
            std::cout << "Error::GetReady(), module " << iter_id << " not found" << std::endl;
            return 0;
        }

        if ( !(m = dynamic_cast<SIS3300module *>(dm)) )
        {
            std::cout << "Error::GetReady(), SIS3300module " << iter_id << " not found" << std::endl;
            return 0;
        }

        m->set_nInputs(_nInputs);
        m->set_nSamples(_nSamples);
        m->set_delay(_delay);
    }

    //ConfigAmplifiers();

    std::cout << "\t+-> Second Detector GetReady          [" << get_id() << "]"
    << std::endl;
    return 0;
}

void
SecondDetector::set_header_description( unsigned int &data )
{
    unsigned int id = get_id();
    data = ( get_nModules() & 0xff ) << 20 | get_data_type() | ( id & 0xffff );
}

void
SecondDetector::create_analysis(int nchip, int nchan)
{
    if (this->ana)
    delete this->ana;

    this->ana = new SecondDetectorAna(nchip, nchan);

    std::cout << "SecondDetector::create_analysis   nchip = " << this->ana->Nchip()
    << "  nchan = " << this->ana->Nchan()
    << "  ntot = " << this->ana->Ntot() << std::endl << std::flush;
}

int
SecondDetector::ConfigAmplifiers()
{
    int _nAmp = get_nAmplifiers();
    unsigned long serial;
    int _delay;
    std::vector<int> gains;

    for ( int i = 0; i < _nAmp; i++ )
    {
        serial = get_AmpAddress(i);
        _delay = get_AmpDelay(i);
        get_AmpGains(i, gains);

        std::cout << "Amplifier USB serial = 0x"
                  << std::hex << serial << std::dec
                  << std::endl;

#ifdef HAS_USB        
        pmtUSB *pUSB = new pmtUSB(serial);

        pUSB->set_delay(_delay);

        std::vector<int>::iterator viter;
        int j;
        for ( j = 0, viter = gains.begin(); viter != gains.end(); ++viter, j++ )
        {
            pUSB->setDACi(j, *viter);
        }

        delete pUSB;
#endif
    }

    return 0;
}
