#ifndef __SecondDetector_h__
#define __SecondDetector_h__
/**
 * This is a module for vmedaq.
 * They all should inherit from GenModule, which adds some extra
 * functionality on top of DAQpp::Module:
 *
 *   - The module works around the concept of something with a number
 *     of chips, each having the same number of channels. So nchip
 *     (number of chips) and nchan (number of channels per chip) is
 *     something we should supply somehow.
 *
 *     For the SIS3300 module we replace nchip by the number of channels (8)
 *     and nchan by the sample size (== page size)
 *
 *     In paticular for the
 *     Analysis instance of the module. That brings me to the second
 *     point.
 *
 *   - There is an Analysis pointer (you have to create the object)
 *     which will store all the monitoring stuff. This guy needs the
 *     number of chips and channels per chip.
 *
 *   - If you plan to make scans with that module, you will need to
 *     override the set_scan_point method to set your module in a
 *     reasonable state every time you are in a new scan point.
 *
 *   - pedestals and noise are there as DAQ++ parameters
 *      + can be updated from the values in Analysis or copied into
 *        the Analysis
 *
 *      + Pedestals are stored in the data file if you override the
 *      write_pedestals method.
 *
 *
 * See the DAQ++ documentation for further information
 *
 **/
#include <string>
#include <sstream>
#include <daq/GenModule.h>
#include "sis3300.h"

class SecondDetector : public GenModule
{
private:
    static const int maxInputs = 8;
    static const int maxSamples = 256;
    static const int maxNevts = 1024;
    static const unsigned int event_directory_bank_shift = maxNevts
            *sizeof(u_int32_t);
    static const u_int32_t adc_shift= MEMSIZE*sizeof(u_int32_t);
    static const u_int32_t addr_bank_adc_shift = (maxInputs + 1) / 2*MEMSIZE
            *sizeof(u_int32_t);
    static const u_int32_t ADCmask = 0xfff;
    unsigned long acq3300_set;

public:
    /** Constructor
     *  The constructor needs 2 parameters:
     *  id: the Module id
     *  address: the address (VME, or whatever protocol) to
     *           communicate with the module
     *
     *  Note: although DAQ++ allows for any type of identifier,
     *  however VMEDAQ only handles properly integer IDs. The only
     *  reason for that is that it needs integer values to save them
     *  properly in the data file
     *
     **/
    SecondDetector(const DAQpp::DAQid &id, unsigned long address);

    /** This is the method called by the RunManager to know if there
     *  is data ready in the module. If the answer if true, then it
     *  will try to read the module. If there is no way to know a
     *  priory about the availavility of data, one can defer the
     *  decision until getData
     *
     **/
    bool hasData();

    /** This is the method called to retrieve the data. If hasData
     *  could not figure out the right answer, one can still say here
     *  that there is no data at all by returning a NULL pointer and
     *  setting size to 0.
     *
     *  The memory to store the data should be allocated with the
     *  allocate method. See the DAQ++ documentation...
     *
     **/
    char *getData(int &size);

    int PrepareForRun();

    int Arm(int = 0);

    int Start(int = 0);

    int Stop(int = 0);

    int Pause();

    int Continue();

    int Reset();

    int GetReady();

    int Trigger()
    {
        return 0;
    }
    ;
    void flush()
    {
    }
    ;
    void set_header_description(unsigned int &data);
    unsigned long get_data_type() const;
    
    
    // *** module specific methods
    // method for creating/recreating Analysis
    void create_analysis(int nchip = 1, int nchan = 1);
    // configure the gains on the amplifier boards via usb
    int ConfigAmplifiers();

    // * get and set methods
    // module parameters
    int get_nAmplifiers()
    {
        return DAQpp::get_par_value<int>(get_par("nAmplifiers"));
    }
    void set_nAmplifiers (int x)
    {
        set_par<int>("nAmplifiers", x);
    }
    int get_nModules ()
    {
        return DAQpp::get_par_value<int>(get_par("nModules"));
    }
    void set_nModules (int x)
    {
        set_par<int>("nModules", x);
    }
    int get_nInputs ()
    {
        return DAQpp::get_par_value<int>(get_par("nInputs"));
    }
    void set_nInputs (int x)
    {
        set_par<int>("nInputs", x);
    }
    int get_nSamples ()
    {
        return DAQpp::get_par_value<int>(get_par("nSamples"));
    }
    void set_nSamples (int x)
    {
        set_par<int>("nSamples", x);
    }
    int get_delay ()
    {
        return DAQpp::get_par_value<int>(get_par("delay"));
    }
    void set_delay (int x)
    {
        set_par<int>("delay", x);
    }
    int get_threshold1 ()
    {
        return DAQpp::get_par_value<int>(get_par("threshold1"));
    }
    void set_threshold1 (int x)
    {
        set_par<int>("threshold1", x);
    }
    int get_threshold2 ()
    {
        return DAQpp::get_par_value<int>(get_par("threshold2"));
    }
    void set_threshold2 (int x)
    {
        set_par<int>("threshold2", x);
    }
    int get_threshold3 ()
    {
        return DAQpp::get_par_value<int>(get_par("threshold3"));
    }
    void set_threshold3 (int x)
    {
        set_par<int>("threshold3", x);
    }
    int get_threshold4 ()
    {
        return DAQpp::get_par_value<int>(get_par("threshold4"));
    }
    void set_threshold4 (int x)
    {
        set_par<int>("threshold4", x);
    }
    int get_threshold5 ()
    {
        return DAQpp::get_par_value<int>(get_par("threshold5"));
    }
    void set_threshold5 (int x)
    {
        set_par<int>("threshold5", x);
    }
    int get_threshold6 ()
    {
        return DAQpp::get_par_value<int>(get_par("threshold6"));
    }
    void set_threshold6 (int x)
    {
        set_par<int>("threshold6", x);
    }
    int get_threshold7 ()
    {
        return DAQpp::get_par_value<int>(get_par("threshold7"));
    }
    void set_threshold7 (int x)
    {
        set_par<int>("threshold7", x);
    }
    int get_threshold8 ()
    {
        return DAQpp::get_par_value<int>(get_par("threshold8"));
    }
    void set_threshold8 (int x)
    {
        set_par<int>("threshold8", x);
    }

    // parameters for the amplifier boards
    unsigned long get_AmpAddress ( int amp )
    {
        std::ostringstream oss;
        oss << "amp" << amp << "address";
        return DAQpp::get_par_value<unsigned long>(get_par(oss.str().c_str()));
    }

    int get_AmpDelay ( int amp )
    {
        std::ostringstream oss;
        oss << "amp" << amp << "delay";
        return DAQpp::get_par_value<int>(get_par(oss.str().c_str()));
    }
    void set_AmpDelay ( int amp, int x )
    {
        std::ostringstream oss;
        oss << "amp" << amp << "delay";
        set_par<int>(oss.str().c_str(), x);
    }

    void get_AmpPMTs ( int amp, std::vector<int> &v )
    {
        v.clear();
        std::ostringstream oss;
        oss << "amp" << amp << "pmts";
        v = DAQpp::get_par_value< std::vector<int> > (get_par(oss.str().c_str()));
    }
    void set_AmpPMTs ( int amp, std::vector<int> &v )
    {
        std::ostringstream oss;
        oss << "amp" << amp << "pmts";
        set_par< std::vector<int> > (oss.str().c_str(), v);
    }

    void get_AmpGains ( int amp, std::vector<int> &v )
    {
        v.clear();
        std::ostringstream oss;
        oss << "amp" << amp << "gains";
        v = DAQpp::get_par_value< std::vector<int> > (get_par(oss.str().c_str()));
    }
    void set_AmpGains ( int amp, std::vector<int> &v )
    {
        std::ostringstream oss;
        oss << "amp" << amp << "gains";
        set_par< std::vector<int> > (oss.str().c_str(), v);
    }
};

#endif
