#include <iostream>
#include <iomanip>
#include <cmath>
#include <config/ConfigParser.h>
#include <config/GeneralConfig.h>
#include "SecondDetectorAna.h"

SecondDetectorAna::SecondDetectorAna(int n_chip, int n_chan) 
    : Analysis(n_chip, n_chan)
{
}

void SecondDetectorAna::init_run()
{
    arm_WaveForms();

    cut = 100;
    time_cut = 100;
    //nrg_scale = 1;
    nrg_scale = 0xfff;

    ibad = 0;
    process = 1;
    simulate = false;

    std::cout << "** SecondDetectorAna::init_run"<< std::endl;
}

void SecondDetectorAna::end_run()
{
    //disarm_WaveForms();

    std::cout << "** SecondDetectorAna::end_run"<< std::endl;
}

void SecondDetectorAna::arm_WaveForms()
{
    for (int i = 0; i < Nchip(); i++)
    {
        WaveForm* wf = new WaveForm(Nchan());
        vWaveForm.push_back(wf);
    }
}

void SecondDetectorAna::disarm_WaveForms()
{
    vWaveForm.clear();
}

bool SecondDetectorAna::data_valid(unsigned int *data) const
{
    unsigned int _nModules, _nInputs, _nSamples, sz, dsz;

    _nModules = (data[2] & 0xff);
    _nInputs = ((data[4] >> 28) & 0xf);
    _nSamples = (data[4] & 0xfffffff);
    dsz = data[3];
    sz = (_nInputs / 2) * _nSamples + 7;
    sz *= _nModules;
    sz += 6; // account for SecondDetector header

    if ( !IsSecDetData( data[2] ))
    {
        std::cout << "SecondDetectorAna::data_valid  Error: DataType = "<< std::hex << (data[2] & 0xf0000)
                << "  DataType Expected = "<< SecDetData << std::dec << std::endl << std::flush;
        return false;
    }

    // Check data integrity
    if (dsz != data[dsz-1]|| dsz - sz != 0)
    {
        std::cout << "SecondDetectorAna::data_valid  Error: dsz = "<< std::hex << dsz << "  sz = "<< sz<< "  last line = "
                << data[dsz-1]<< std::dec<< std::endl << std::flush;
        return false;
    }

    // Confirm Analysis matches data
    if ( (_nModules*_nInputs) != Nchip() || _nSamples != Nchan() )
    {
        std::cout << "SecondDetectorAna::data_valid  Error: _nModules = "<< _nModules<< "  _nInputs = "<< _nInputs
                << "  Nchip() = "<< Nchip()<< std::endl<< "                                                  "<< " _nSamples = "
                << _nSamples << "  Nchan() = "<< Nchan()<< std::endl << std::flush;
        return false;
    }

    // We do not want to process twice the same event
    if (trigger >= data[1])
    {
        std::cout << "SecondDetectorAna::data_valid  Error: trigger = "<< std::hex << trigger << "  data event = "<< data[1]
                << std::dec << std::endl << std::flush;
        return false;
    }

    return true;
}

void SecondDetectorAna::decode(unsigned int *data)
{
    // Get some important handlers of the configuration
    // ConfigParser *cfg = ConfigParser::theConfig();

    // In order to make a fast analysis of the data, we should provide
    // the pedestal calculation routines the list of channels to be
    // considered.
    std::vector<int> list_of_channels;

    // We assume the Analysis::Fill has made all the checks on
    // the data integrity
    int data_hld = 5;
    int pulse_hld = 0;
    int _nModules = (data[2] & 0xff);
    int _nInputs = ((data[4] >> 28) & 0xf);
    int _nSamples = (data[4] & 0xfffffff);
    u_int32_t ring_pointer;
    int i, j, k, nevt;

    for (i = 0; i < _nModules; i++)
    {
        ring_pointer = data[data_hld+5];
        unwrapData(data + data_hld, _nInputs, _nSamples, ring_pointer, Chan() + pulse_hld );
        data_hld += (((_nInputs + 1) / 2) * _nSamples + 7);
        pulse_hld += (_nInputs * _nSamples);
    }

    //std::cout << "SecondDetectorAna::decode, call to unwrapData done..."
    //    << std::endl << std::flush;


    std::vector<WaveForm*>::iterator it;
    for (i = 0, it = vWaveForm.begin(); it != vWaveForm.end(); i++, ++it )
        ( *it )->set_wf(Chan() + i*_nSamples );

    if (simulate )
        simulate_event();

    nevt = 0;
    for (i = 0; i < _nModules; i++)
    {
        for (j = 0; j < _nInputs; j++)
        {
            for (k = 0; k < _nSamples; k++)
            {
                Event()[nevt] = Chan()[nevt];
                // accepting all channels
                list_of_channels.push_back(nevt);
                nevt++;
            }
        }
    }

    // Note: we assumed here that each word corresponds to one
    // channel and that all the channels are valid. One may have the
    // scenario in which the data is perversly encoded and we only
    // find, at the end of the process a list of channels with
    // signal. In that situation, we would call
    //
    // add_hit(j)
    // Event()[j] = Chan()[j] = signal
    //
    // because we know that channel j has signal...

    if (process == 0)
    {
        //update_pedestals(list_of_channels);
        update_pedestals_fast(list_of_channels, false);
    }
}

/*
 void
 SecondDetectorAna::Analyze()
 {
 int i, i_use;
 double xs, xs_max;
 double nrg;
 bool   use;

 if ( process == 1 ) process_event();
 else      // old routine to fill energy (works)
 {
 for ( i = 0; i < Nchip(); i++)
 {
 i_use = i;

 nrg = integrate(i, 0, Nchan(), Nchan(), Chan());

 if ( nrg > 20 )
 {
 m_fill.acquire();
 HitMap()[i_use]++;
 int ib = get_bin(nrg);
 if ( ib >= 0 && ib < get_nbins() )
 {
 Signal(i_use)[ib]++;
 Signal(-1)[ib]++;
 Signal2D(i_use)[ib]++;
 if (get_map())
 {
 int ix, iy, ip;
 get_map()->pad(i_use, ix, iy);
 ip = ix + get_map()->nx() * iy;
 H2Map()[ ip ]++;
 if ( Chan()[i_use]>=emin && Chan()[i_use]<=emax )
 H2MapC()[ ip ]++;
 }
 }
 m_fill.release();
 }
 }
 }
 if ( ibad > 0 && ibad % 10000 == 0 )
 std::cout << "SecondDetectorAna:Analyze    ibad = " << ibad << std::endl;
 }

 */
int SecondDetectorAna::unwrapData(const unsigned int *unwrap_data, int npmts, int nsamples, u_int32_t ring_pointer,
                                  double *pulse_data)
{
    u_int32_t ring_pointer0 = (ring_pointer & 0x1ffff);
    if ((int)ring_pointer0 >= nsamples)
    {
        std::cout << "Error: ring_pointer = "<< ring_pointer0<< ", nsamples = "<< nsamples << std::endl;
    }
    //was the readout wrapped already?+
    bool wrap_bit0 = (ring_pointer & 0x80000);

    //decode to pulse_pmt_shape
    //if wrap start at ringpointer, else increase to next level
    int start = !wrap_bit0 ? nsamples : ring_pointer0;
    //stop is at next level
    int stop = ring_pointer0 + nsamples;

    for (int i = 0; i < (npmts + 1) / 2; i++)
    {
        for (int j = start, k = 0; j < stop; j++, k++)
        {
            pulse_data[2*i*nsamples+k] =((unwrap_data[6+i*nsamples+j%nsamples] >> 16) & 0xfff);
            pulse_data[(2*i+1)*nsamples+k] =(unwrap_data[6+i*nsamples+j%nsamples] & 0xfff);
        }
    }

    return (wrap_bit0 ? nsamples : ring_pointer0);
}

double SecondDetectorAna::integrate(int ipmt, int low, int high, int nsamples, double *pulse_pmt_shape)
{
    int start = low < 0 ? 0 : low;
    int stop = high > nsamples ? high : nsamples;

    double *pmt_value = pulse_pmt_shape + ipmt * nsamples;

    double sum = 0;
    for (int k = start; k < stop; k++)
    {
        sum += pmt_value[k];
    }
    sum /= nsamples;

    return sum;
}

void SecondDetectorAna::process_event()
{
    std::vector<WaveForm*>::iterator it;

    for (it = vWaveForm.begin(); it != vWaveForm.end(); ++it)
        (*it)->data_start();

    double ttime = calculate_ttime();

    if (ttime == 0)
    {
        ibad++;
        return;
    }

    //propagate final event time to modules
    for (it = vWaveForm.begin(); it != vWaveForm.end(); ++it)
        (*it)->set_ttime(ttime);

    // sets weights for position calculations and sets the matrix elements
    for (it = vWaveForm.begin(); it != vWaveForm.end(); ++it)
        ( *it )->set_weights();

    int i_use = 0;
    for (it = vWaveForm.begin(); it != vWaveForm.end(); ++it)
    {
        double nrg = nrg_scale * ( *it )->get_energy();

        if (nrg > cut )
        {

            if (false)
            {
                std::cout << "SecondDetectorAna::process_event   energy = "<< nrg << std::endl;
            }

            m_fill.acquire();
            HitMap()[i_use]++;
            int ib = get_bin(nrg);
            if (ib >= 0&& ib < get_nbins() )
            {
                Signal(i_use)[ib]++;
                Signal(-1)[ib]++;
                Signal2D(i_use)[ib]++;
                if (get_map())
                {
                    int ix, iy, ip;
                    get_map()->pad(i_use, ix, iy);
                    ip = ix + get_map()->nx() * iy;
                    H2Map()[ ip ]++;
                    if (Chan()[i_use] >= emin && Chan()[i_use] <= emax )
                        H2MapC()[ ip ]++;
                }
            }
            else if (false)
            {
                std::cout << "SecondDetectorAna::process_event   channel = "<< i_use << " energy = "<< nrg << " out of range"
                        << std::endl;
            }
            m_fill.release();
        }

        double pedestal = (*it)->get_pedestal();
        for (int i = 0; i < Nchan(); i++)
            Ped(i_use)[i] = pedestal;

        i_use++;
    }
}

double SecondDetectorAna::calculate_ttime()
{
    std::vector<WaveForm*>::iterator it;
    double ttime = 0, sum = 0, sumW = 0;

    //estimate time of event
    int ifire = 0;
    for (it = vWaveForm.begin(); it != vWaveForm.end(); ++it )
    {
        double fRise = nrg_scale * ( *it )->energy_estimate();
        if (false)
        {
            std::cout << "SecondDetectorAna::calculate_ttime   energy_estimate = "<< fRise << std::endl;
        }
        if (fRise > time_cut )
        {
            double tt = (*it)->improve_trigger_time();
            if (tt > 0)
            {
                sum += fRise * tt;
                sumW += fRise;
                ifire++;
            }
        }
    }

    if (ifire > 0&& sumW > 1e-7)
        ttime = sum / sumW;

    return ttime;
}

void SecondDetectorAna::simulate_event()
{
    int i;
    std::vector<WaveForm*>::iterator it;
    for (i = 0, it = vWaveForm.begin(); it != vWaveForm.end(); i++, ++it )
        (*it)->sim_data();
}
