#include "WaveForm.h"
#include <iostream>
#include <fstream>
#include <cmath>

WaveForm::WaveForm(int _nSamples) :
    nSamples(_nSamples)
{
    wf = new double[nSamples];

    // set avgWF values
    // may need to be moved to config file
    deltaT = 10; // 10 ns clock cycles
    tau = 230; // NaI value
    tauRise = 40; // characteristic of pmt?
    Nquad = 4; // interval for event time calculations
    Usample = 32; // avgWF sample divisions
    Mrise = 3; // rise time in clock cycles?
    Lcheck = 7; // ?
    Beta = 75; // ? used in set_weights, matrix_elements ...

    dataRange = 1;
    //dataRange = 0xfff;

    compute_avgWF();
}

WaveForm::~WaveForm()
{
    delete[] wf;

    delete[] avgWF;
    delete[] avgWFderiv;
}

void WaveForm::compute_avgWF()
{
    double Q = exp(-deltaT / tau);
    double q = exp(-deltaT / tau / Usample);
    double P = exp(-deltaT / tauRise);
    double p = exp(-deltaT / tauRise / Usample);

    double* _avgWF = new double[nSamples*Usample];
    double* _simWF = new double[nSamples];
    double* _avgWFderiv = new double[nSamples];
    double q1 = 1, p1 = 1;
    for (int i = 0; i < nSamples*Usample; i++)
    {
        _avgWF[i] = (q1 - p1);
        q1 *= q;
        p1 *= p;
    }
    avgWF = _avgWF;

    q1 = p1 = 1;
    for (int i = 0; i < nSamples; i++)
    {
        _simWF[i] = (q1 - p1);
        q1 *= Q;
        p1 *= P;
    }
    simWF = _simWF;

    q1 = p1 = 1;
    for (int i = 0; i < nSamples; i++)
    {
        _avgWFderiv[i] = tau / tauRise * (p1 - q1);
        p1 *= P;
        q1 *= Q;
    }
    avgWFderiv = _avgWFderiv;
}

void WaveForm::set_wf(double* x)
{
    for (int i = 0; i < nSamples; i++)
        wf[i] = x[i] / 0xfff;
}

void WaveForm::sim_data()
{
    for (int i = 0; i < nSamples; i++)
        wf[i] = dataRange * simWF[i];
}

int WaveForm::data_start(int offset)
{
    start = 0;
    for (; wf[start] == 0; start++)
        ;
    start += offset;
    return start;
}

int WaveForm::basic_trigger_time() const
{
    double max_der = 0;
    int imax = -1;
    int iM = nSamples - Nquad - 1;
    int jM = Nquad + 1;
    double sum;

    for (int i = start + Nquad; i < iM; i++)
    {
        sum = 0;
        for (int j = -Nquad; j < jM; j++)
            sum += j * wf[i+j];

        if (sum > max_der)
        {
            imax = i;
            max_der = sum;
        }
    }

    return imax;
}

double WaveForm::energy_estimate(int iMaxDer) const
{
    if ( !iMaxDer )
        iMaxDer = basic_trigger_time();

    double sum = 0;

    if (iMaxDer< (Mrise + Nquad) || iMaxDer > (nSamples - Mrise - Nquad - 1))
        return sum;

    for (int i = -Mrise; i < Mrise + 1; i++)
        for (int j = -Nquad; j < Nquad + 1; j++)
            sum += j * wf[iMaxDer+i+j];

    sum *= 3. / (Nquad + 1.0)/ Nquad / (2 * Nquad + 1.0);

    return sum;
}

double WaveForm::improve_trigger_time(int iMaxDer) const
{
    if ( !iMaxDer )
        iMaxDer = basic_trigger_time();

    double sum1 = 0, sum = 0;
    int ij = 0;

    //make an estimate of the base-line
    int baseLineEstimateOff = 20;
    double base = 0;
    double ttime = 0;

    if (iMaxDer < (baseLineEstimateOff + Lcheck ))
        return ttime;

    for (int i = 0; i < Lcheck; i++)
        base += wf[iMaxDer-baseLineEstimateOff-i];

    base /= Lcheck;

    ttime = (double)iMaxDer;

    for (int j = 0; j < Lcheck; j++)
    {
        sum = 0;
        ij = iMaxDer - Lcheck + j;

        for (int k = 0; k < nSamples - ij; k++)
            sum += avgWFderiv[k] * (wf[ij+k] - base);

        if (!j && sum > 0)
            return ttime;

        if (sum > 0)
            break;

        sum1 = sum;
    }
    if (sum1 == sum)
        return ttime;

    ttime = ij - sum / (sum - sum1);

    return ttime;
}

void WaveForm::set_ttime(double _ttime)
{
    ttime = _ttime;
    ti = static_cast<int>(ceil(ttime));
    joff = static_cast<int>(round(Usample * (ti - ttime)));
}

void WaveForm::reset()
{
    start = 0;
    for (int i = 0; i < nSamples; i++)
        wf[i] = 0;
}

double WaveForm::energy_final(double beta, double& err) const
{
    double A = 0, B = 0, C = 0, F0 = 0, F1 = 0, w = 0, c;

    for (int i = start; i < nSamples; i++)
    {
        w = i < ti ? 0 : avgWF[(i-ti)*Usample+joff];
        c = 1 + beta * wf[i];
        A += 1. / c;
        B += w / c;
        C += w * w / c;
        F0 += wf[i] / c;
        F1 += w * wf[i]/ c;
    }
    err = C;
    double det = A * C - B * B;
    if (det == 0)
        return 0;
    double E = (A * F1 - B * F0) / det;
    //   if (E<3*sqrt(noise))
    //     noise+=pw*((1-pw)*E*E-noise);
    //   double p=(C*F0-B*F1)/det;
    //   pedestal+=pw*(p-pedestal);
    //cout<<"mtx "<<A<<' '<<B<<' '<<C<<' '<<F0<<' '<<F1<<' '<<det<<endl;
    return E;
}

void WaveForm::matrix_elements(double beta, int flag)
{
    WY = 0;
    Y = 0;
    W = 0;
    S = 0;
    WW = 0;

    double w = 0;
    double c = 0;

    for (int i = start; i < nSamples; i++)
    {
        w = i < ti ? 0 : avgWF[(i-ti)*Usample+joff];
        c = flag == 0 ? 1. / (1 + beta * w) : 1. / (1 + beta * wf[i]/ dataRange);

        S += c;
        W += w * c;
        Y += wf[i] * c / dataRange;
        WW += w * w * c;
        WY += w * wf[i]* c / dataRange;
    }
}

double WaveForm::chi2(double beta) const
{
    double E = get_energy();
    double sum = 0;
    double w, val;
    double p = get_pedestal();

    for (int i = start; i < nSamples; i++)
    {
        w = i < ti ? 0 : avgWF[(i-ti)*Usample+joff];
        val = (wf[i] - p - E * w) / (1 + beta * E * w);
        sum += val * val;
    }
    return sum;
}

void WaveForm::set_weights()
{
    set_weights(Beta );
}

void WaveForm::set_weights(double beta)
{
    matrix_elements(beta, 1);

    double E = get_energy();

    matrix_elements(beta*E);
}
