#include <cstdio>
#include <cassert>
#include <cstring>
#include <iostream>
#include <iomanip>
#include "TAmanager.h"


int bitset::byte_reverse(int val, int nbit) const
{
    int ib, rc = 0, nb=nbit-1;
    for (ib=0;ib<=nb;ib++)
    {
       if ( val & (1<<ib) )
       {
	  rc |= (1<<(nb-ib));
       }
    }
    return rc;
}

TAmanager::TAmanager(ChipType tp, int n_chip) :
    nchip(n_chip), nchan(128), nall(0), type(tp), all(0)
{
    Resize(n_chip);
}

TAmanager::~TAmanager()
{
    clear();
    if (all)
        delete all;

}

void TAmanager::clear()
{
    ChipList::iterator ip;
    for (ip = chips.begin(); ip != chips.end(); ++ip)
        delete *ip;

    chips.clear();
}

void TAmanager::cpy(const TAmanager &ta)
{
    int i;
    //      if ( all )
    //  delete all;
    all = 0;
    nall = 0;
    type = ta.type;
    Resize(ta.nchip);
    for (i = 0; i < nchip; i++)
        (*chips[i]) = *ta.chips[i];

}

TAmanager::TAmanager(const TAmanager &ta)
{
    cpy(ta);
}

TAmanager &TAmanager::operator=(const TAmanager &ta)
{
    if ( this != &ta )
    cpy(ta);

    return *this;
}

void TAmanager::Resize(int n_chip)
{
    int i;
    nchip = n_chip;
    clear();

    switch (type)
    {
        case chipGP:
            nchan = 128;
            for (i = 0; i < nchip; i++)
            {
                chips.push_back(new VATAGPchip(nchan));
            }
            break;
        case chipGP3:
            nchan = 128;
            for (i = 0; i < nchip; i++)
            {
                chips.push_back(new VATAGP3chip(nchan));
            }
            break;
        case chipGP3p:
            nchan = 128;
            for (i = 0; i < nchip; i++)
            {
                chips.push_back(new VATAGP3pchip(nchan));
            }
            break;
        case chipTA:
            nchan = 32;
            for (i = 0; i < nchip; i++)
            {
                chips.push_back(new TA32chip(nchan));
            }
            break;
        case chipGP5:
            nchan = 128;
            for (i = 0; i < nchip; i++)
            {
                chips.push_back(new VATAGP5chip(nchan));
            }
            break;
        case chipGP7:
            nchan = 128;
            for (i = 0; i < nchip; i++)
            {
                chips.push_back(new VATAGP7chip(nchan));
            }
            break;
        case last_type:
            break;
    }
}

bitset *TAmanager::TAvector()
{
    int ic, ib;
    int nsiz, vsiz, ioff;
    if (chips.empty() )
        return 0;

    TAchip *chp = chip(0);
    nsiz = nchip * chp->nBits();
    vsiz = nsiz / 16;
    if (vsiz * 16 < nsiz)
        vsiz++;

    ioff = vsiz * 16 - nsiz;

    if (all)
    {
        delete all;
    }
    all = new bitset(vsiz * 16, 16);
    nall = vsiz;

    /*
     std::cout << "n bits/chip " <<  chips[0].nBits()
     << std::endl
     << "total number of bits " << nsiz
     << std::endl
     << "capacity " << vsiz*24
     << std::endl
     << "offset " << ioff
     << std::endl
     ;
     */

    for (ic = 0; ic < nchip; ic++)
    {
        chp = chip(ic);
        for (ib = 0; ib < chp->nBits(); ib++, ioff++)
        {
            if (chp->Bit(ib))
                all->Set(ioff);
            else
                all->Clr(ioff);
        }
    }

    /*
     std::cout << "last bit writen in " << ioff
     << std::endl
     << *all
     << std::endl;
     for (ib=0;ib<all->nWords();ib++)
     std::cout << std::setw(8) << std::hex << all->Word(ib) << std::dec << std::endl;
     */

    return all;
}

std::ostream & operator <<(std::ostream & os, TAmanager & ta)
{
    int i;
    os << "Number of chips " << ta.nchip << std::endl;
    for (i = 0; i < ta.nchip; i++)
    {
        os << "chip " << i << std::endl 
           << *(ta.chips[i]) << std::endl;
    }

    return os;
}

//---------------------------------------------------------------------
// TAchip
//---------------------------------------------------------------------
TAchip::TAchip(const TAchip &cp) :
    bitset( *dynamic_cast<const bitset *>( &cp ) )
{
    nchan = cp.nchan;
}

TAchip &TAchip::operator=(const TAchip &cp)
{
    if ( &cp != this )
    {
        nchan = cp.nchan;
        *(dynamic_cast<bitset *>(this)) = *(dynamic_cast<const bitset *>(&cp));
    }
    return *this;
}

//---------------------------------------------------------------------
// Common stuff
//---------------------------------------------------------------------
// This is valid for all the chips (for the moment)
TAchip *TAchip::test(int i, bool t)
{
    //int b = nchan - 1 - i;
    int b = nBits() - nchan - 1 + i;
    return (TAchip *) (t ? Set(b) : Clr(b));
}

//---------------------------------------------------------------------
// VATAGP
//---------------------------------------------------------------------
// Resizing
VATAGPchip::VATAGPchip(const VATAGPchip &cp) :
    TAchip( *dynamic_cast<const TAchip *>( &cp ) )
{
}

void VATAGPchip::Resize(int n_chan)
{
    nchan = n_chan;
    new_size(5*nchan + 17, 32);
}

// Setting the trim
TAchip *VATAGPchip::trim(int i, int val)
{
    //int b = 2 * nchan + 3 * (nchan - 1 - i);
    int b = 17 + 3 * i;
    return (TAchip *) SetR(val, b, 3);
}

// Setting the mask
TAchip *VATAGPchip::mask(int i, bool t)
{
    //int b = 2 * nchan - 1 - i;
    int b = 17 + 3 * nchan + i;
    return (TAchip *) (t ? Set(b) : Clr(b));
}
void VATAGPchip::print(std::ostream & os)
{
    int i;
    os << "Number of channels " << nchan << std::endl << "Bit dump " << std::endl
            << *(dynamic_cast<bitset *>(this)) << std::endl << "Test       " << Test()
            << std::endl << "Polarity   " << Polarity() << std::endl << "G. Enabled "
            << GainEnabled() << std::endl << "Gain       " << Gain() << std::endl;

    os << "Mask " << std::endl;
    for (i = 0; i < nchan; i++)
        os << (mask(i) ? '1' : '0');
    os << std::endl;

    os << "Test " << std::endl;
    for (i = 0; i < nchan; i++)
        os << (test(i) ? '1' : '0');
    os << std::endl;

    os << "Trim " << std::endl;
    for (i = 0; i < nchan; i++)
        os << std::setw(1) << trim(i);
    os << std::endl;

}

//---------------------------------------------------------------------
// VATAGP3
//---------------------------------------------------------------------
// resizing
VATAGP3chip::VATAGP3chip(const VATAGP3chip &cp) :
    TAchip( *dynamic_cast<const TAchip *>( &cp ) )
{
}
void VATAGP3chip::Resize(int n_chan)
{
    nchan = n_chan;
    new_size(5*nchan + 11, 32);
}
// Setting the trim
TAchip *VATAGP3chip::trim(int i, int val)
{
    //int b = nchan + 3 * (nchan - 1 - i);
    int b = 11 + nchan + 3 * i;
    return (TAchip *) SetR(val, b, 3);
}
// Setting the mask
TAchip *VATAGP3chip::mask(int i, bool t)
{
    //int b = 5 * nchan - 1 - i;
    int b = 11 + i;
    return (TAchip *) (t ? Set(b) : Clr(b));
}
void VATAGP3chip::print(std::ostream & os)
{
    int i;
    os << "Number of channels " << nchan << std::endl << "Bit dump " << std::endl
            << *(dynamic_cast<bitset *>(this)) << std::endl << "Curr. Comp. "
            << CC_Enable() << std::endl << "N side      " << N_side() << std::endl
            << "Test        " << Test() << std::endl << "Polarity    " << Polarity()
            << std::endl << "G. Enabled  " << GainEnabled() << std::endl
            << "Gain        " << Gain() << std::endl << "Chip Add.   " << Chip_Address()
            << std::endl;

    os << "Mask " << std::endl;
    for (i = 0; i < nchan; i++)
        os << (mask(i) ? '1' : '0');
    os << std::endl;

    os << "Test " << std::endl;
    for (i = 0; i < nchan; i++)
        os << (test(i) ? '1' : '0');
    os << std::endl;

    os << "Trim " << std::endl;
    for (i = 0; i < nchan; i++)
        os << std::setw(1) << trim(i);
    os << std::endl;

}

//---------------------------------------------------------------------
// VATAGP3_1
//---------------------------------------------------------------------
// resizing
int VATAGP3pchip::nHeader = 9; // Added by Vladimir on 15th of dec., 2009
VATAGP3pchip::VATAGP3pchip(const VATAGP3pchip &cp) :
    VATAGP3chip( *dynamic_cast<const VATAGP3chip *>( &cp ) )
{
}
void VATAGP3pchip::Resize(int n_chan)
{
    nchan = n_chan;
    //new_size(5*nchan + 9, 32);
    //new_size(5*nchan + 8, 32);    // Commented by Vladimir on 15th of Dec., 2009
    new_size(5*nchan + nHeader, 32);// Added by Vladimir on 15th of dec., 2009
}
// Setting the trim
TAchip *VATAGP3pchip::trim(int i, int val)
{
    //int b = nchan + 3 * (nchan - 1 - i);
    //int b = 8 + nchan + 3 * i; // Commented by Vladimir on 15th of Dec., 2009
    int b = nHeader + nchan + 3*i; // Added by Vladimir on 15th of dec., 2009
    return (TAchip *) SetR(val, b, 3);
}
// Setting the mask
TAchip *VATAGP3pchip::mask(int i, bool t)
{
    //int b = 5 * nchan - 1 - i;
    //int b = 8 + i;     // Commented by Vladimir on 15th of Dec., 2009
    int b = nHeader + i; // Added by Vladimir on 15th of dec., 2009
    return (TAchip *) (t ? Set(b) : Clr(b));
}


//----------------------------------------------------------------------
// VATAGP5
//----------------------------------------------------------------------
void VATAGP5chip::Resize(int n_chan)
{
    nchan = n_chan;
    new_size(778, 32);
}
void VATAGP5chip::print(std::ostream & os)
{
    int i;
    os << "Number of channels " << nchan << std::endl 
       << "Bit dump " << std::endl
       << *(dynamic_cast<bitset *>(this)) << std::endl 
       << "Curr. Comp. " << CC_Enable() << std::endl 
       << "N side      " << N_side() << std::endl
       << "Test        " << Test() << std::endl 
       << "Polarity    " << Polarity() << std::endl 
       << "Gain        " << Gain() << std::endl 
       << "Chip Add.   " << Chip_Address()
       << std::endl;

    os << "Mask " << std::endl;
    for (i = 0; i < nchan; i++)
        os << (mask(i) ? '1' : '0');
    os << std::endl;

    os << "Test " << std::endl;
    for (i = 0; i < nchan; i++)
        os << (test(i) ? '1' : '0');
    os << std::endl;

    os << "Trim " << std::endl;
    for (i = 0; i < nchan; i++)
        os << std::setw(1) << trim(i);
    os << std::endl;

}
//----------------------------------------------------------------------
// VATAGP7
//----------------------------------------------------------------------
void VATAGP7chip::Resize(int n_chan)
{
    nchan = n_chan;
    new_size(926, 32);
}
void VATAGP7chip::print(std::ostream &os)
{
    int i;
    os << "Number of channels " << nchan << std::endl 
       << "Bit dump " << std::endl
       << *(dynamic_cast<bitset *>(this)) << std::endl 
       << "Chip Add.   " << Chip_Address() << std::endl
       << "Curr. Comp. " << CC_Enable() << std::endl 
       << "N side      " << N_side() << std::endl
       << "Test        " << Test() << std::endl 
       << "Polarity    " << Polarity() << std::endl 
       << "Probe en.   " << Probe_Enable() << std::endl
       << "Vrc select  " << Vrc() << std::endl
       << "Vfsf select " << Vfsf() << std::endl
       << "Vfss select " << Vfss() << std::endl
       << "Vfp select  " << Vfp() << std::endl
       << "Calib gen   " << calib_gen() << std::endl
       << "Calib ctrl  " << calib_ctrl() << std::endl
       << "Calib DAC   " << calib_dac() << std::endl
       << "Slew        " << slew() << std::endl
       << "dltc        " << dltc() << std::endl
       << "DAC range   " << dac_range() << std::endl
       << "Thrs. DAC   " << thrs_dac() << std::endl
       << std::endl;

    os << "Mask " << std::endl;
    for (i = 0; i < nchan; i++)
        os << (mask(i) ? '1' : '0');
    os << std::endl;

    os << "Test " << std::endl;
    for (i = 0; i < nchan; i++)
        os << (test(i) ? '1' : '0');
    os << std::endl;

    os << "Trim " << std::endl;
    for (i = 0; i < nchan; i++)
        os << std::setw(1) << trim(i);
    os << std::endl;    
}

//---------------------------------------------------------------------
// TA32
//---------------------------------------------------------------------
// resizing
TA32chip::TA32chip(const TA32chip &cp) :
    TAchip( *dynamic_cast<const TAchip *>( &cp ) )
{
}
void TA32chip::Resize(int n_chan)
{
    nchan = n_chan;
    new_size(5*nchan + 5, 32);
}

TAchip *TA32chip::mask(int i, bool t)
{
    //int b = 5 * nchan - 1 - i;
    int b = 5 + i;
    return (TAchip *) (t ? Set(b) : Clr(b));
}
TAchip *TA32chip::trim(int i, int val)
{
    //int b = nchan + 3 * (nchan - 1 - i);
    int b = 5 + nchan + 3 * i;
    return (TAchip *) SetR(val, b, 3);
}

/*
int TA32chip::trim(int i)
{
    int val = 0;
    int ib, b = nchan + 3 * (nchan - i) - 1;
    for (ib = 0; ib < 3; ib++, b--)
    {
        if (Bit(b) )
            val |= (1 << ib);
    }
    return val;
}

TAchip *TA32chip::trim(int i, int val)
{
    int ib, b = nchan + 3 * (nchan - i) - 1;
    for (ib = 0; ib < 3; ib++, b--)
    {
        if ( 1 << ib & val)
            Set(b);
        else
            Clr(b);
    }
    return (TAchip *) this;
}
*/

void TA32chip::print(std::ostream &os)
{
    int i;
    os << "Number of channels " << nchan << std::endl << "Bit dump " << std::endl
            << *(dynamic_cast<bitset *>(this)) << std::endl << "Test       " << Test()
            << std::endl << "Polarity   " << Polarity() << std::endl << "G. Enabled "
            << GainEnabled() << std::endl << "Gain       " << Gain() << std::endl;

    os << "Mask " << std::endl;
    for (i = 0; i < nchan; i++)
        os << (mask(i) ? '1' : '0');
    os << std::endl;

    os << "Test " << std::endl;
    for (i = 0; i < nchan; i++)
        os << (test(i) ? '1' : '0');
    os << std::endl;

    os << "Trim " << std::endl;
    for (i = 0; i < nchan; i++)
        os << std::setw(1) << trim(i);
    os << std::endl;

}

std::ostream& operator<<(std::ostream &os, TAchip &ch)
{
    ch.print(os);
    return os;
}

//---------------------------------------------------------------------
// Constructors & destructor
//---------------------------------------------------------------------
bitset::bitset(int nn, int szW) :
    pb(0)
{
    new_size(nn, szW);
}

bitset::bitset(int nn, unsigned long *ptr) :
    pb(0)
{
    int i, j;
    new_size(nn, 8 * sizeof(unsigned long));//8 bits per byte; sizeof in bytes
    for (i = 0; i < nSlot; i++)
    {
        pb[i] = ptr[i];
        for (j = 0; j < szSlot; i++)
            if (pb[i] & (1 << j))
                nset++;
    }
}

bitset::bitset(const bitset & b)
{
    nbits = b.nbits;
    szSlot = b.szSlot;
    nSlot = b.nSlot;
    nsTrue = nSlot;
    if (nSlot > 0)
    {
        pb = new unsigned long[nSlot];
        for (int i = 0; i < nSlot; i++)
            pb[i] = b.pb[i];
        nset = b.nset;
    }
}

bitset::bitset(int nn, char *sb)
{
    nsTrue = 0;
    szSlot = 8 * sizeof(unsigned long);
    BitStr(nn, sb);
}

bitset::~bitset()
{
    if (pb)
        delete[] pb;
}

void bitset::new_size(int nn, int szW)
{
    if (pb)
        delete[] pb;
    nbits = nn;
    szSlot = szW;
    nSlot = nbits / szSlot + (nbits % szSlot ? 1 : 0);
    nsTrue = nSlot;
    nset = 0;
    if (nn == 0)
    {
        pb = 0;
        return;
    }
    pb = new unsigned long[nSlot];
    for (int i = 0; i < nSlot; i++)
        pb[i] = 0;
}

//---------------------------------------------------------------------
// Overloaded operators
//---------------------------------------------------------------------
int bitset::operator[] (int i) const
{
    return Bit(i);
}

bitset *bitset::Word(int x, unsigned long v)
{
    if (x >= 0 && x <= nsTrue)
        pb[x] = v;
    return this;
}

unsigned long bitset::Word(int x) const
{
    if (x >= 0 && x < nsTrue)
        return pb[x];
    else
        return 0;
}

bitset *bitset::operator !()
{
    // set all elements to 0
    return AllZero();
}

bitset *bitset::AllZero()
{
    // set all elements to 0
    for (int i = 0; i < nSlot; i++)
        pb[i] = 0;
    nset = 0;
    return this;
}

bitset & bitset::operator = (const bitset & b)
{
    nbits = b.nbits;
    szSlot = b.szSlot;
    if (b.nSlot> nsTrue)
    {
        delete[] pb;
        pb = new unsigned long[b.nSlot];
        nsTrue = b.nsTrue;
    }
    nSlot = b.nSlot;
    for (int i = 0; i < nSlot; i++)
    pb[i] = b.pb[i];
    nset = b.nset;

    return *this;
}

//---------------------------------------------------------------------
// Interface
//---------------------------------------------------------------------
bitset *bitset::Reset(unsigned long *ptr)
{
    int i, j;

    for (i = 0, nset = 0; i < nSlot; i++)
    {
        pb[i] = ptr[i];
        for (j = 0; j < szSlot; j++)
            if (pb[i] & (1 << j))
                nset++;
    }
    return this;
}

void bitset::BitStr(int nn, char *sb)
{
    int i, j, k, iok;
    char *p, *q, c;
    int slts = nn / szSlot + (nn % szSlot ? 1 : 0);

    if (slts > nsTrue)
    {
        delete[] pb;
        pb = new unsigned long[slts];
        nsTrue = slts;
    }
    nSlot = slts;
    nbits = nn;
    for (i = 0; i < nSlot; i++)
        pb[i] = 0;

    // Set the bits
    for (p = sb, i = 0, c = '0'; *p; c = *p++)
    {
        switch (*p)
        {
            case '1':
                Set(i++);
                break;
            case '0':
                Clr(i++);
                break;
            case '<':
                q = strchr(p, '>');
                iok = sscanf(p, "<%d>", &j);
                assert(iok == 1);
                if (c == '1')
                    for (k = 0; k < j; k++)
                        Set(i++);
                else
                    for (k = 0; k < j; k++)
                        Clr(i++);
                p = q;
                break;
            default:
                i++;
                break;
        }
    }
}

bitset *bitset::Set(int i)
{
    int j = i % szSlot, k = i / szSlot;
    pb[k] |= (1L << j);
    nset++;
    return this;
}

bitset *bitset::Set(int val, int start, int width)
{
    //   int j= start%szSlot,k=start/szSlot;
    for (int i = start, l = 0; l < width && i < nbits; l++, i++)
        if (val & (1 << l))
            Set(i);
        else
            Clr(i);

    return this;
}

bitset *bitset::Clr(int i)
{
    int j = i % szSlot, k = i / szSlot;
    if ((pb[k] & (1L << j)) != 0)
        nset--;
    pb[k] &= ~(1L << j);
    return this;
}

unsigned long bitset::Byte(int start, int width) const
{
    unsigned long rc = 0;
    width = (width > nbits ? nbits : width);

    for (int i = start, j = 0; j < width && i < nbits; j++, i++)
        if (Bit(i))
            rc |= (1 << j);

    return rc;
}

int bitset::Bit(int i) const
{
    int j = i % szSlot, k = i / szSlot;
    return (pb[k] & (1L << j)) != 0;
}

//---------------------------------------------------------------------
// I/O operations
//---------------------------------------------------------------------
void bitset::Print(int start, int width)
{
    int i, j;
    width = (width > nbits ? nbits : width);
    for (i = start, j = 0; j < width && i < nbits; i++, j++)
        if (Bit(i))
            std::cout << '1';
        else
            std::cout << '0';

    std::cout << std::endl;
}

void bitset::prt(std::ostream & os)
{
    int i;
    os << "Number of bits " << nbits << std::endl;
    for (i = 0; i < nbits; i++)
    {
        if (Bit(i))
        {
            os << '1';
        }
        else
        {
            os << '0';
        }
    }
}

std::ostream & operator <<(std::ostream & os, bitset & b)
{
    b.prt(os);
    return os;
}

void bitset::Dump()
{
    std::cout << "Number of bits : " << nbits << std::endl << "Number of slots: "
            << nSlot << std::endl << "Size of slot     " << szSlot << " bits "
            << std::endl;
    for (int i = 0; i < nSlot; i++)
#ifdef WIN32
        std::cout << std::setw(8) << std::hex << pb[i];
#else
        std::cout << std::setw(8) << std::setbase(16) << pb[i];
#endif

    std::cout << std::dec << std::endl;
}

