#include <iostream>
#include <iomanip>
#include <sstream>
#include <string>
#include <vector>
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <DAQ++/Parameter.h>
#include <data/DataFormat.h>
#include <daq/VMEmanager.h>
#include "TDCCaen488_module.h"
#include "TDCCaen488_analysis.h"

TDCCaen488_module::TDCCaen488_module(const DAQpp::DAQid &id, unsigned long address) :
    GenModule(id, VMEmanager::VME()->map_address(address, 0, 0)), 
    DAQpp::ParameterCollection<TDCCaen488_module>(this),
    tdc(get_address())
    {
        // Using DAQ++ parameters can be useful, especially if you want to
        // access them the DAQcorba
        add_par("range",
                        new_par<double>(&TDCCaen488_module::change_range,
                                        new DAQpp::Par<double>(500.)));

        add_par("high_thrs",
                        new_par<int>( &TDCCaen488_module::change_high_threshold,
                                        new DAQpp::Par<int>(0xC7)));

        add_par("low_thrs",
                        new_par<int>(&TDCCaen488_module::change_low_threshold,
                                        new DAQpp::Par<int>(0)));

        add_par("channel_mask",
                        new_par<int>(&TDCCaen488_module::change_channel_mask,
                                        new DAQpp::Par<int>(1)));

        add_par("daq_mode",
                        new_par<int>(&TDCCaen488_module::change_daq_mode,
                                        new DAQpp::Par<int>(tdc488::start)));

        add_par("buffer_depth",
                        new_par<int>(&TDCCaen488_module::change_buffer_depth,
                                        new DAQpp::Par<int>(0)));

        // just for debug
        add_par("efficiency",new DAQpp::Par<double>(0.75));
        add_par("mean",new DAQpp::Par<double>(100.));
        add_par("width",new DAQpp::Par<double>(10.));

        // now we should create the Analysis for the monitoring
        // never forget to do this !!!
        ana = new TDCCaen488_analysis();
        ana->Resize(8,1);
        
        std::vector<double> &ped = pedestal->get_value();
        std::vector<double> &ns  = noise->get_value();

        for (int i = 0;i < ana->Ntot();i++ )
        {
            ana->Ped()[i]=0.;
            ana->Noise()[i]=1.;
            ped.push_back( 0. );
            ns.push_back( 1. );
        }
    }

void TDCCaen488_module::set_header_description(unsigned int &data)
{
    unsigned int id=get_id();
    int range = tdc.range2dac(get_par_value<double>("range"));
    data = (range&0xfff)<<20| TDCCaenData | (id&0xffff);
}

unsigned long TDCCaen488_module::get_data_type() const
{
    return TDCCaenData;
}

bool TDCCaen488_module::hasData()
{
    VMEmanager *vme = VMEmanager::VME();
    if (!vme->connected())
    {
        double eff = get_par_value<double>("efficiency");
        double rnd = randuniform();
        return rnd < eff;
    }
    else
    {
        return !tdc.empty();
    }
}

char *TDCCaen488_module::getData(int &size)
{
    VMEmanager *vme = VMEmanager::VME();
    increase_evt_counter();
    if (!vme->connected())
    {
        size =4 + // the 3 words in the header plus the
                  // trailing word
              2; // TDC header plus 1 channel

        double _mean = DAQpp::get_par_value<double>(get_par("mean"));
        double _std = DAQpp::get_par_value<double>(get_par("width"));

        // Allocate space for the data. This has to be done with the
        // allocate method of DAQpp::Module
        //
        unsigned int *data =
            reinterpret_cast<unsigned int *>(allocate(sizeof(unsigned int)*size));

        // Fill the data array
        int id = get_id();
        int iw=0;
        data[iw++] = DataHeader | DataRecord | (size);
        data[iw++] = get_evt_counter();
        data[iw++] = (id&0xfff)<< 20 | TDCCaenData | 2;
        data[iw++] = 8<<12 | (get_evt_counter() & 0x7ff);
        data[iw++] = (unsigned int)(randgauss(_mean,_std));
        data[iw] = size;

        size *= sizeof(unsigned int);
        return reinterpret_cast<char *>(data);

    }
    else
    {
        std::vector<unsigned int> odata;
        int ichan,nchan;
        unsigned short xxx;
        int id = get_id();
        unsigned int *data = 0;
        size = 0;
        dataWord *val = (dataWord *)&xxx;
        while (! tdc.empty())
        {
            xxx = tdc.data();
            if (val->header)
            {
                if (odata.empty())
                {
                    odata.push_back(DataHeader | DataRecord);
                    odata.push_back( get_evt_counter() );
                    odata.push_back( (id&0xfff)<< 20 | TDCCaenData );
                }
                odata.push_back( xxx );
                nchan = val->channel+1;
                for (ichan=0;ichan<nchan;ichan++)
                {
                    xxx = tdc.data();
                    odata.push_back( xxx );
                }
            }
            else
                break;
        }
        if (! odata.empty() )
        {
            odata.push_back( odata.size()+ 1);
            size = odata.back();
            odata[0] |= size;
            odata[2] |= (0xffff & (size-4));

            size *= sizeof(unsigned int);
            data = reinterpret_cast<unsigned int *>(allocate(size));
            std::vector<unsigned int>::iterator ip;
            for (ichan=0, ip=odata.begin(); ip != odata.end(); ++ip, ichan++)
                data[ichan ] = *ip;
        }
        return reinterpret_cast<char *>(data);
    }
}

unsigned int *TDCCaen488_module::write_config(int &size)
{
    size = 1 + // daq mode
           1 + // range
           1 + // mask
           1 + // threshold high
           1;  // threshold low
    
    int ichan,ichip,indx=0;
    unsigned int *data = new unsigned int[size];
    
    data[indx++] = get_par_value<int>("daq_mode");
    data[indx++] = tdc.range2dac( get_par_value<double>("range") );
    data[indx++] = get_par_value<int>("channel_mask");
    data[indx++] = get_par_value<int>("high_thrs");
    data[indx++] = get_par_value<int>("low_thrs");
    
    return data;
}

int TDCCaen488_module::PrepareForRun()
{
    return 0;
}

int TDCCaen488_module::Start(int)
{
    if (get_analysis() )
        get_analysis()->init_run();

    return 0;
}

int TDCCaen488_module::Stop(int)
{
    if (get_analysis() )
        get_analysis()->end_run();

    return 0;
}
int TDCCaen488_module::Pause()
{
    return 0;
}
;

int TDCCaen488_module::Continue()
{
    return 0;
}
;

int TDCCaen488_module::Reset()
{
    return 0;
}
;

int TDCCaen488_module::GetReady()
{
    tdc.reset();
    tdc.setMode( get_par_value<int>("daq_mode") ? tdc488::stop : tdc488::start )
        ->setRange(get_par_value<double>("range"))
        ->enableChan(get_par_value<int>("channel_mask"))
        ->HighThreshold(get_par_value<int>("high_thrs"))
        ->LowThreshold(get_par_value<int>("low_thrs"));
    return 0;
}
;

void TDCCaen488_module::change_low_threshold(const int &v)
{
    std::cout << "Setting low threshold to "<< v << std::endl;
    tdc.LowThreshold(v);
}
void TDCCaen488_module::change_high_threshold(const int &v)
{
    std::cout << "Setting high threshold to "<< v << std::endl;
    tdc.HighThreshold(v);
}
void TDCCaen488_module::change_range(const double &v)
{
    std::cout << "Changing range to "<< v << std::endl;
    tdc.setRange(v);
}
void TDCCaen488_module::change_channel_mask(const int &v)
{
    std::cout << "Changing channel mask to "<< v << std::endl;
    tdc.enableChan(v);
}
void TDCCaen488_module::change_daq_mode(const int &v)
{
    std::cout << "Setting daq mode to "<< v << std::endl;
    tdc.setMode(v ? tdc488::stop : tdc488::start);
}
void TDCCaen488_module::change_buffer_depth(const int &v)
{
    std::cout << "Setting buffer depth to "<< v << std::endl;
    if (v)
    {
        tdc.FFmode();
    }
    else
    {
        tdc.HFmode();
    }
}
