/**
   This class is responsible to parse the corresponding section in the
   configuration file to configure the module. It is also responsible
   for create the module instance according to the paremeters retrived
   from the configuration file.

 */
#include <string>
#include <sstream>
#include <vector>
#include <iterator>
#include <DAQ++/DAQmanager.h>
#include <config/ConfigParser.h>
#include <config/ModuleParser.h>
#include <config/GeneralConfig.h>
#include <gtk/gtkwidget.h>
#include <data/Analysis.h>
#include <data/DataFormat.h>

#include <gtk-hst/histogram.h>
#include <gtk-hst/gtkhistogram.h>

#include <gui/GladeApp.h>
#include <gui/MonitorWindow.h>

#include "TDCCaen488_module.h"

class Analysis;
static void __attribute__ ((constructor)) __init__lib__()
{
    printf("### Loading TDCCaen488 module...\n");
}

static void __attribute__ ((destructor)) __close__lib__()
{
    printf("### Unloading TDCCaen488...\n");
}



class TDCCaen488_module_conf : public ModuleParser {
    private:

    public:
        TDCCaen488_module_conf();
        ~TDCCaen488_module_conf() {}

        /**
         * This is the method responsible for parsing the configuration
         * file. 
         **/
        void parse_node(xmlNodePtr node)
        throw (ConfigException);

        /**
         * The method that saves the current configuration. It should
         * update all the xml nodes below the module node.
         **/
        void save_config()
        throw (ConfigException);



        /**
         * This is called by VMEDAQ to create a window that will allow
         * to change the configuration of the module while running the
         * program. Probably the easiest is to use glade-2 to create the
         * layout of the window, define the signals (callbacks that are
         * associated to given actions) and then something derived from
         * GladeApp (see the gui folder) to implement the stuff and
         * update the module state accordingly.
         **/
        GtkWidget *config(const std::string &id);

        /**
         * This is called by VMEDAQ to create a window to display the
         * monitor data. The input parameters are:
         *    - id: the module id (that should help finding the module)
         *    - glade_file: The path of the glade file. This is, in
         *      principle the path of the standard VMEDAQ glade
         *      file. One can certainly ignore this value.
         **/
        MonitorWindow *monitor(const std::string &id, const std::string &glade_file);
};

TDCCaen488_module_conf::TDCCaen488_module_conf()
{
}

/**
 * This is the hook specified in the config file to return a pointer
 * to the TDCCaen488_module_conf instance that will interact with VMEDAQ.
 *
 * Note that it should b edeclared as extern "C" so that no demangling
 * mechanism is applied to its name that prevent finding easily the
 * simbol in the library
 *
 **/
extern "C" {
    TDCCaen488_module_conf *create_TDCCaen488_module_conf() {
        TDCCaen488_module_conf *cnf = new TDCCaen488_module_conf();
        return cnf;
    }
}


/**
 * Now come the implementations of the class methods
 */
void
TDCCaen488_module_conf::parse_node(xmlNodePtr _node)
throw (ConfigException)
{
    unsigned long address;
    std::string saddress;

    // The configuration manager handler
    ConfigParser *cfg = ConfigParser::theConfig();

    // store the node pointer
    node = _node;

    // get the module id
    id = cfg->get_attribute<std::string>(node,"name");

    // get the address
    saddress = cfg->get_attribute<std::string>(node,"address");

    std::istringstream os(saddress);
    os >> std::hex >> address;

    // Check if the module is already there
    DAQpp::Module *dm;
    TDCCaen488_module *module=0;

    if ( (dm=DAQpp::DAQmanager::theDAQmanager()->find_module(id)) )
        module=dynamic_cast<TDCCaen488_module *>(dm);

        if (!module) {
            // The module is not there... create
            module = new TDCCaen488_module(id,address);
        }

        // Now parse the things which are common to all the modules, that
        // is the monitor and padmap elements
        std::ostringstream onode;
        onode << "//" << node->name;
        parse_common( onode.str(),id );

        std::cout << "*** Analysis: nbins "
        << module->get_analysis()->get_nbins()
        << " range "
        << module->get_analysis()->get_range_min()
        << " -> "
        << module->get_analysis()->get_range_max()
        << std::endl;
        std::cout << "*** Configuration"
        << " nbins " << nbin()
        << " range "
        << range_min() << " -> " << range_max()
        << std::endl;

        // Now we have the module. From now own one should parse the
        // section from the configuration file
        double range;
        int    low_thrs, high_thrs, channel_mask,daq_mode,buffer_depth;
        try {
            range = cfg->get_value<double>(node,"range");
        }
        catch (ConfigException &e) {
            range = module->get_par_value<double>("range");
        }

        try {
            low_thrs = cfg->get_value<int>(node,"low_thrs");
        }
        catch (ConfigException &e) {
            low_thrs = module->get_par_value<int>("low_thrs");
        }

        try {
            high_thrs = cfg->get_value<int>(node,"high_thrs");
        }
        catch (ConfigException &e) {
            high_thrs = module->get_par_value<int>("high_thrs");
        }

        try {
            channel_mask = cfg->get_value<int>(node,"channel_mask");
        }
        catch (ConfigException &e) {
            channel_mask = module->get_par_value<int>("channel_mask");
        }
        try {
            daq_mode = cfg->get_value<int>(node,"daq_mode");
        }
        catch (ConfigException &e) {
            daq_mode = module->get_par_value<int>("daq_mode");
        }
        try {
            buffer_depth = cfg->get_value<int>(node,"buffer_depth");
        }
        catch (ConfigException &e) {
            buffer_depth = module->get_par_value<int>("buffer_depth");
        }


        module->set_par<double>("range", range, true);
        module->set_par<int>("low_thrs",low_thrs, true);
        module->set_par<int>("high_thrs",high_thrs, true);   
        module->set_par<int>("channel_mask", channel_mask, true);
        module->set_par<int>("daq_mode", daq_mode, true);
        module->set_par<int>("buffer_depth", buffer_depth, true);   
}

void
TDCCaen488_module_conf::save_config()
throw (ConfigException)
{
    // The configuration manager handler
    ConfigParser *cfg = ConfigParser::theConfig();

    // Save common stuff
    save_common();

    // get the module
    TDCCaen488_module *module= dynamic_cast<TDCCaen488_module *>(
            DAQpp::DAQmanager::theDAQmanager()->find_module(id));

    if (module) {
        // Save the current values in the configuration
        cfg->check_value<double>(node,
                "range"
                ,module->get_par_value<double>("range"));
        cfg->check_value<int>(node,
                "low_thrs",
                module->get_par_value<int>("low_thrs"));
        cfg->check_value<int>(node,
                "high_thrs",
                module->get_par_value<int>("high_thrs"));
        cfg->check_value<int>(node,
                "channel_mask",
                module->get_par_value<int>("channel_mask"));
        cfg->check_value<int>(node,
                "daq_mode",
                module->get_par_value<int>("daq_mode"));
        cfg->check_value<int>(node,
                "buffer_depth",
                module->get_par_value<int>("buffer_depth"));
    }
}

class config_TDCCaen488_module : public GladeApp {
    private:
        TDCCaen488_module_conf *cfg;
    public:
        config_TDCCaen488_module(TDCCaen488_module_conf *c)
        : cfg(c) {}
        static void on_ok(GtkWidget *, config_TDCCaen488_module *);
        static void on_apply(GtkWidget *, config_TDCCaen488_module *);
        static void on_cancel(GtkWidget *, config_TDCCaen488_module *);

        void user_init();
        void auto_connect() {
            SET_CONNECT( on_ok );
            SET_CONNECT( on_apply );
            SET_CONNECT( on_cancel );
        }
};

void config_TDCCaen488_module::user_init() {
    // get the module
    TDCCaen488_module *module= dynamic_cast<TDCCaen488_module *>(
            DAQpp::DAQmanager::theDAQmanager()->find_module(cfg->get_id()));

    if (module) {
        int daq_mode, buffer_depth;
        double2entry("txtRange",module->get_par_value<double>("range"));
        int2entry("txtLowThrs",module->get_par_value<int>("low_thrs"));
        int2entry("txtHighThrs",module->get_par_value<int>("high_thrs") );
        int2entry("txtMask",module->get_par_value<int>("channel_mask") );

        daq_mode = module->get_par_value<int>("daq_mode");
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(get_widget("btnStart")),
                                     daq_mode==tdc488::start);

        buffer_depth = module->get_par_value<int>("buffer_depth");
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(get_widget("btnFull")),
                                     buffer_depth);

    }
}

void config_TDCCaen488_module::on_ok(GtkWidget *w, config_TDCCaen488_module *m)
{
    on_apply(w,m);
    on_cancel(w,m);
}
void config_TDCCaen488_module::on_apply(GtkWidget *w, config_TDCCaen488_module *m)
{
    // get the module
    TDCCaen488_module *module= dynamic_cast<TDCCaen488_module *>(
            DAQpp::DAQmanager::theDAQmanager()->find_module(m->cfg->get_id()));

    if (module) {
        module->set_par<double>("range", m->entry2double("txtRange"), true );
        module->set_par<int>("low_thrs",m->entry2int("txtLowThrs"), true );
        module->set_par<int>("high_thrs",m->entry2int("txtHighThrs"),true );
        module->set_par<int>("channel_mask",m->entry2int("txtMask"), true );
        module->set_par<int>("daq_mode",
                gtk_toggle_button_get_active(
                                             GTK_TOGGLE_BUTTON(m->get_widget("btnStart")))
                                             ? 0
                                                     : 1, true);
        module->set_par<int>("buffer_depth",
                gtk_toggle_button_get_active(
                                             GTK_TOGGLE_BUTTON(m->get_widget("btnFull")))
                                             ? 1
                                                     : 0, true);

    }
}
void config_TDCCaen488_module::on_cancel(GtkWidget *w, config_TDCCaen488_module *m)
{
    delete m;
}



class monitor_TDCCaen488_module : public GladeApp {
    private:
        Analysis *_ana;
    public:
        monitor_TDCCaen488_module(Analysis *a) : _ana(a) {}
        static void on_ok(GtkWidget *w, monitor_TDCCaen488_module *m);
        static void on_refresh(GtkWidget *w, monitor_TDCCaen488_module *m);
        static GtkWidget *create_signal_histogram(monitor_TDCCaen488_module *,char *,char *, char *,int , int);
        static GtkWidget *create_hitmap_histogram(monitor_TDCCaen488_module *,char *,char *, char *,int , int);
        void auto_connect() {
            SET_CONNECT( on_ok );
            SET_CONNECT( on_refresh );
        }
        void set_handlers();

};

void monitor_TDCCaen488_module::set_handlers() {
    handlers["create_signal_histogram"] = (GladeHandler)create_signal_histogram;
    handlers["create_hitmap_histogram"] = (GladeHandler)create_hitmap_histogram;
}
void monitor_TDCCaen488_module::on_ok(GtkWidget *w, monitor_TDCCaen488_module *m)
{
    delete m;
}
void monitor_TDCCaen488_module::on_refresh(GtkWidget *w, monitor_TDCCaen488_module *m)
{
    m->_ana->Init();
}

GtkWidget *monitor_TDCCaen488_module::create_signal_histogram(
                                                              monitor_TDCCaen488_module *mon,
                                                              char *title,char *, char *,int , int)
{
    HstHistogram   *h1d;
    GtkHst1D       *wh1d;

    h1d  = hst_histogram_new(
                             mon->_ana->get_nbins(),
                             mon->_ana->get_range_min(),
                             mon->_ana->get_range_max(),0
    );
    g_object_set(G_OBJECT(h1d),"ptr",mon->_ana->Signal(-1),NULL);
    wh1d = GTK_HST1D(gtk_hst1d_new( HST_FUNCTION(h1d) ));
    g_object_set_data( G_OBJECT(wh1d),"hst",h1d);
    gtk_hst1d_set_xtitle(wh1d,"Signal");
    gtk_hst1d_monitor_start(wh1d,500);
    gtk_hst1d_set_colors(wh1d,"background","white",NULL);
    gtk_hst1d_sync_range(wh1d,TRUE);

    return GTK_WIDGET(wh1d);
}


GtkWidget *monitor_TDCCaen488_module::create_hitmap_histogram(
                                                              monitor_TDCCaen488_module *mon,
                                                              char *title,char *, char *,int , int)
{
    HstHistogram   *h1d;
    GtkHst1D       *wh1d;

    h1d  = hst_histogram_new(mon->_ana->Ntot(),0,mon->_ana->Ntot(),0);
    g_object_set(G_OBJECT(h1d),"ptr",mon->_ana->HitMap(),NULL);
    wh1d = GTK_HST1D(gtk_hst1d_new( HST_FUNCTION(h1d) ));
    g_object_set_data( G_OBJECT(wh1d),"hst",h1d);
    gtk_hst1d_set_xtitle(wh1d,"HitMap");
    gtk_hst1d_monitor_start(wh1d,500);
    gtk_hst1d_set_colors(wh1d,"background","white",NULL);
    gtk_hst1d_sync_range(wh1d,TRUE);

    return GTK_WIDGET(wh1d);
}


MonitorWindow *
TDCCaen488_module_conf::monitor(const std::string &name, const std::string &glade_file)
{
    DAQpp::Module *dm = DAQpp::DAQmanager::theDAQmanager()->find_module(name);
    if (!dm)
        return 0;

    TDCCaen488_module *m=0;
    if ( !(m=dynamic_cast<TDCCaen488_module *>(dm)) ) {
        std::cout << "TDCCaen488_module_conf: module " << name
        << " is not of type TDCCaen488_module "
        << std::endl;
        return 0;
    }
    monitor_TDCCaen488_module * mon = new monitor_TDCCaen488_module(m->get_analysis());
    mon->init("TDCCaen488.glade","monitor");
    MonitorWindow *mw = new MonitorWindow(mon->get_main_window());
    return mw;
}

GtkWidget *
TDCCaen488_module_conf::config(const std::string &id)
{
    config_TDCCaen488_module *conf = new config_TDCCaen488_module(this);   
    conf->init("TDCCaen488.glade","configuration");
    return conf->get_main_window();
}
