#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <iomanip>
#include <stdexcept>
#include <unistd.h>
#include "VATAmodule.h"
#include <DAQ++/GTimer.h>
#include <data/Analysis.h>
#include <data/DataFormat.h>
#include <daq/VMEmanager.h>

#include <daq/ScanPoint.h>
#include "VATAmodule_analysis.h"

#define VME_debug 1
#if defined(VME_debug)
#define CHECK_BUS(vme) {                                                \
    int rc = vme->get_rc();                                         \
    if (rc) {                                                       \
        std::cout << __FUNCTION__ << " in " << __FILE__ << ":" << __LINE__ \
        << ": vme returns 0x" <<std::hex << rc << std::dec << std::endl; \
    }                                                               \
}
#else
#define CHECK_BUS(vme)
#endif
//#define ACQUIRE_VME(vme) vme->acquire()
//#define RELEASE_VME(vme) vme->release()
#define ACQUIRE_VME(vme)
#define RELEASE_VME(vme)

VATAmodule::VATAmodule(const DAQpp::DAQid &id,
                       unsigned long add)
:  GenModule(id, add),_delay(10000), threshold(0), channel(0)
{


    //---------------------------------------------------------------------
    // Get VME address
    //---------------------------------------------------------------------
    _add = VMEmanager::VME()->map_address(add, 0, 0);
    std::cout << "VME addr. " << std::hex << add
    << " -> " << _add << std::dec << std::endl;


    threshold = new DAQpp::Par<int>(0);
    add_par("threshold", threshold);
    threshold_p = new MyParameter<int> (this,
            & VATAmodule::change_threshold,
            threshold);

    rout   = new DAQpp::Par<int>( Off);
    add_par("rout", rout);
    rout_p = new MyParameter<int> (this,
            & VATAmodule::change_rout,
            rout);


    nchip     =  new DAQpp::Par<int>(1);
    add_par("nchip", nchip);
    nchip_p   =  new MyParameter<int> (this,
            & VATAmodule::change_nchip,
            nchip);


    adj       =  new DAQpp::Par<int>(0);
    add_par("adj", adj);
    adj_p     =  new MyParameter<int> (this,
            & VATAmodule::change_adj,
            adj);

    ta       =  new DAQpp::Par<TAmanager>();
    add_par("ta", ta);
    ta_p     =  new MyParameter<TAmanager> (this,
            & VATAmodule::change_ta,
            ta);
    forceDataMode = new DAQpp::Par<bool>(false);
    add_par("forceDataMode", forceDataMode);
    forceDataMode_p = new MyParameter<bool>( this,
            &VATAmodule::change_forceDataMode,
            forceDataMode);

    add_par("fast_acquire",new DAQpp::Par<bool>(true));

    TAmanager &tm = ta->get_value();
    VATAmodule_analysis *vana = new VATAmodule_analysis( tm.Nchip(), tm.Nchan() ); 
    ana = vana;
    tm.SetChipType( TAmanager::chipGP3 );
    vana->set_chip_type( tm.GetChipType());

    // set pedestals and noise with proper vector size
    std::fill_n( std::back_inserter( get_par_value< std::vector<double> >("pedestal") ), ana->Ntot(), 0.);
    std::fill_n(back_inserter( get_par_value< std::vector<double> >("noise") ), ana->Ntot(), 0.);

//    int i;
//    std::vector<double> &ped = pedestal->get_value();
//    std::vector<double> &ns  = noise->get_value();
//    for ( i = 0;i < ana->Ntot();i++ )
//    {
//        ped.push_back( 0. );
//        ns.push_back( 0. );
//    }


    // The default values for the "simulation"
    _noise =   5.;
    _mean  = 500.;
    _ped   =  50.;
    _cmmd  =  1.;
    _thePulse = -1.;
    _theChannel = -1;
    _theThreshold = -1.;
    _theThresholdGain = 10. / 256.;
    _theGain = 10. / 1024.;

}

unsigned long VATAmodule::get_data_type() const
{
    return NormalData;
}

bool VATAmodule::hasData()
{
    VMEmanager *vme = VMEmanager::VME();
    if ( vme->connected() )
    {
        unsigned int dat = 0;
        double delay = _delay;
        bool rc = false;
        DAQpp::GTimer tim;

        ACQUIRE_VME(vme);

	if (get_par_value<bool>("fast_acquire")){
	  
          std::vector<VMEcommand> stack;
          std::vector<unsigned int> vdata;
          // Replaces call to Trigger()
          stack.push_back(VMEcommand(VMEcommand::write32,_add + PULSE,0));
          switch (run_type) {
            case IntTrg:
              stack.push_back(VMEcommand(VMEcommand::write32,_add + SOFTSTART,0));
              break;
            case Pedestal:
              if (get_rout() & Serial) //serial
                stack.push_back(VMEcommand(VMEcommand::write32,_add + SOFTSTART,0));
              break;
            default:
              break;
          }
          // just check once for data
          stack.push_back(VMEcommand(VMEcommand::read32,_add + TEVENT));
          // Replaces call to Trigger()
          stack.push_back(VMEcommand(VMEcommand::write32,_add + PULSE,0));
          switch (run_type) {
            case IntTrg:
              stack.push_back(VMEcommand(VMEcommand::write32,_add + SOFTSTART,0));
              break;
            case Pedestal:
              if (get_rout() & Serial) //serial
                stack.push_back(VMEcommand(VMEcommand::write32,_add + SOFTSTART,0));
              break;
            default:
              break;
          }
          // sneak in the first check to see if readout is busy
          stack.push_back(VMEcommand(VMEcommand::read32,_add + TROUT));
          // Replaces call to Trigger()
          stack.push_back(VMEcommand(VMEcommand::write32,_add + PULSE,0));
          switch (run_type) {
            case IntTrg:
              stack.push_back(VMEcommand(VMEcommand::write32,_add + SOFTSTART,0));
              break;
            case Pedestal:
              if (get_rout() & Serial) //serial
                stack.push_back(VMEcommand(VMEcommand::write32,_add + SOFTSTART,0));
              break;
            default:
              break;
          }
          // now execute the stack, if we're lucky we get it all in one fell swoop
          int check = vme->stack_execute(stack,vdata);
          CHECK_BUS(vme);
          if (check<0 || vdata.size()!=2)
            return false;
          dat = vdata[0];
          dat &= 0x80000000;
          dat >>= 31;
          if (!dat) {
            RELEASE_VME(vme);
            return false;
          }
          dat = vdata[1];
          dat &= 0x80000000;
          dat >>= 31;
          if (!dat) {
            RELEASE_VME(vme);
            return true;
          }

          // Check for readout busy (line 31 to 1)
          delay = 10.; // 10 seconds should be a good limit
          tim.reset();
          tim.start();
          for (; dat ;) {
            dat = vme->read32(_add + TROUT);
            CHECK_BUS(vme);
            dat &= 0x80000000;
            dat >>= 31;
            if ( tim() > delay ) break;
          }
          Trigger();
          RELEASE_VME(vme);
          return true;
        }
	//end fast readout
	
        // check for line 31 (if 1, event present)
        Trigger();
        while ( !dat )
        {
            dat = vme->read32(_add + TEVENT);
            CHECK_BUS(vme);
            dat &= 0x80000000;
            dat >>= 31;
            if ( tim() > delay )
            {
                //       Reset();
                goto end;
            }
        }

        RELEASE_VME(vme);
        ACQUIRE_VME(vme);
        Trigger();
        // Check for readout busy (line 31 to 1)
        delay = 10.; // 10 seconds should be a good limit
        tim.reset();
        tim.start();
        for (dat = 1; dat ;)
        {
            dat = vme->read32(_add + TROUT);
            CHECK_BUS(vme);
            dat &= 0x80000000;
            dat >>= 31;
            if ( tim() > delay )
            {
                //       Reset();        // DO not know if that helps...
                goto end;
            }
        }
        rc = true;
        end:
        //Trigger();
        RELEASE_VME(vme);
        return true;
    }
    else
    {
        if (GetStatus().isPaused())
            return false;
        else
        {
            _theThreshold = _theThresholdGain * get_threshold();
            _theThreshold = randgauss(_theThreshold, 0.05 * _theThreshold);
            _theValue = randgauss(_mean, 10) * _theGain;
            if (_theThreshold > 0. )
            {
                return _theValue > _theThreshold;
            }
            else
                return true;
        }
    }
}

char *VATAmodule::getData(int &size)
{

    VMEmanager *vme = VMEmanager::VME();
    increase_evt_counter();

    if (vme->connected())
    {
        // Read the word count
        ACQUIRE_VME(vme);
        size = vme->read32(_add + R_WC);
        CHECK_BUS(vme);
        size &= 0x7ff;

        if ( !size )
        {
            Reset();
            RELEASE_VME(vme);
            return 0;
        }


        /*
          Get the data
          The format is the following
          Word
          0     0xCAFEibbb   i= event type bbb= size of data block
          1     0xbbaaaaaa   bb=number of adjacent aaaaaaaa event counter
          2     0xiiitannn   iii=module id t=data type a= readout mode nnn=data size
          3     data word 0
          4     data word 1
          .     .
          bbbb    bbbb         size of data block
         */
        int id;

        unsigned int *data = reinterpret_cast<unsigned int *>(
                allocate(sizeof(unsigned int) * (4 + (size & 0xfff))));

        id = get_id();
        data[0] = DataHeader | DataRecord | (4 + (size & 0xfff));
        data[1] = adj->get_value() << 24 | get_evt_counter();
        data[2] = (id & 0xfff) << 20 | NormalData | (size & 0xfff) | ( get_rout() & 0xe ) << 12 ;
		vme->readA24_fifo32(_add + RW_FIFO,size,data+3);
		CHECK_BUS(vme);
	

        data[3+(size & 0xfff)] =  (4 + (size & 0xfff));
        size = data[3+(size & 0xfff)] * sizeof(unsigned int);

        Reset();
        RELEASE_VME(vme);
        return reinterpret_cast<char *>(data);

    }
    else
    {
#define ped_corr(chan,freq) (50.*((chan)%(freq))/((double)(freq)))
        int id, i, j, chan;
        int ntot = get_n_chips() * get_n_chan();
        int nadj = get_adj();
        double *event = new double[ntot];
        unsigned int *data = 0;
        int blksize;
        double cmmd = _ped + randgauss(0., _cmmd);
        double _value = _theValue;

        chan = 0;
        if (run_type != Scan)
        {
            if ( get_analysis() && get_analysis()->get_map() )
            {
                double x, y;
                int nchip = get_n_chips();
                int nchan = get_n_chan();
                int nx = get_analysis()->get_map()->nx();
                int ny = get_analysis()->get_map()->ny();
                int ppeak[2][2] = { {nx / 4, ny / 2}, { 3*nx / 4, ny / 2} };
                double value[2] = { _theValue, _theValue / 3. };
                int peak = (int)(2. * randuniform()); // Which of the 2 peaks

                _value = value[peak];
                again:
                do
                {
                    randgauss2d(ppeak[peak][0], 4., ppeak[peak][1], 4., 0., &x, &y);
                }
                while ( x < 0 || x >= nx || y < 0 || y >= ny );
                chan = get_analysis()->get_map()->channel((int)x, (int)y);
                if (chan < 0 || chan > nchip*nchan)
                {
                    std::cout << "Expect problems" << "chan " << chan << "[x " << x << " y " << y << "]" << std::endl;
                    goto again;
                }
            }
            else
            {
                chan = int(get_n_chan() * get_n_chips() * randuniform());
            }
        }
        else
        {
            if (_theChannel < 0)
                chan = int(ntot * randuniform());
        }

        if ( _theChannel > 0 )
            chan = _theChannel;

        /* Generate the event signal */
        for (i = 0;i < ntot;i++)
        {
            event[i] = cmmd + randgauss(0., _noise) + ped_corr(i, get_n_chan() / 2);
        }
        event[chan] += _value / _theGain;

        size = 0;
        id = get_id();
        switch ( get_rout() & 0xe )
        {
            case 2: // serial
                size = 3;
                size += get_n_chan() * get_n_chips();
                blksize = 4 + (size & 0xfff);
                data = reinterpret_cast<unsigned int *>(
                        allocate(sizeof(unsigned int) * (blksize)));

                data[0] = DataHeader | DataRecord | (blksize);

                data[1] = adj->get_value() << 24 |
                get_evt_counter();

                data[2] = (id & 0xfff) << 20 | NormalData | (size & 0xfff)  | ( get_rout() & 0xe ) << 12 ;
                for (j = 6, i = 0;j < blksize - 1;i++, j++)
                {
                    data[j] = ((int)(event[i] + 0.5) & 0x3ff);
                }
                data[j] =  (blksize);
                size = data[j] * sizeof(unsigned int);
                break;

            case 4:
                nadj = 0;
            case 8:
                size = 5 + nadj;
                blksize = 4 + (size & 0xfff);
                data = reinterpret_cast<unsigned int *>(
                        allocate(sizeof(unsigned int) * (blksize)));
                data[0] = DataHeader | DataRecord | (blksize);
                data[1] = adj->get_value() << 24 | get_evt_counter();
                data[2] = ((id & 0xfff) << 20 | NormalData | size & 0xfff) | ( get_rout() & 0xe ) << 12 ;
                for (j = i = 0;i < 3;j++, i++)
                    data[3+j] = chan << 20;

                for (i = 0;j < size;j++, i++)
                {
                    if ( (chan-i)>=0 && (chan+i)<ntot)
                    {
                        data[3+j] =
                            chan << 20 |
                            ( int(event[chan-i] + 0.5) & 0x3ff ) << 10 |
                            ( int(event[chan+i] + 0.5) & 0x3ff );
                    }
                }
                data[3+j] = (blksize);

                size = data[3+j] * sizeof(unsigned int);
                break;

        }
        delete [] event;
        if ( !size )
            return 0;
        else
            return reinterpret_cast<char *>(data);
    }
}

int VATAmodule::PrepareForRun()
{
    int romode;

    _theChannel = -1;

    if ( get_forceDataMode() )
    {
        set_run_type( Data );
        std::cout << "VATAmodule::PrepareForRun()  "
        << "Setting run_type to Data..."
        << std::endl;
    }

    switch (run_type)
    {
        case Data:
        case IntTrg:
        case Pedestal:
            romode = get_rout();
            set_rout( 0 );
            Reset();
            load_register();
            // In sparse SHIFTREG must be zero, in Serial true
            if ( romode & VATAmodule::Serial )
                romode |= VATAmodule::On;
            else
                romode &= ~VATAmodule::On;

            romode &= ~VATAmodule::FORoff;
            set_rout( romode );
            reset_evt_counter();
            break;
            /*
          case Pedestal:
          romode = get_rout();
          set_rout( 0 );
          Reset();
          select_all();

          // pedestals should be in serial mode
          // In serial SHIFTREG must be true
          romode = VATAmodule::Serial | VATAmodule::On |  VATAmodule::FORoff;
          set_rout( romode );
          break;
             */
        case Pulse:
            romode = get_rout();
            Reset();
            set_rout( VATAmodule::On );
            select_channel( channel );
            load_register();
            // In sparse SHIFTREG must be zero, in Serial true
            if ( romode & VATAmodule::Serial )
                romode |= VATAmodule::On;
            else
                romode &= ~VATAmodule::On;

            set_rout( romode );
            break;
        default:
            load_register();
            set_rout( get_rout() );
    }

    if (run_type == Pedestal)
    {
        // clear noise and pedestals
        std::vector<double> &ped = pedestal->get_value();
        std::vector<double> &ns  = noise->get_value();
        double *pp = get_analysis()->Ped();
        double *nn = get_analysis()->Noise();
        int i;
        for ( i = 0;i < ana->Ntot();i++ )
        {
            pp[i] = ped[i] = 0;
            nn[i] = ns[i] = 0.;
        }
    }

    //    load_register();
    //    set_rout( get_rout() );
    return 0;
}

unsigned int *VATAmodule::write_config(int &size)
{
    TAmanager &tm = ta->get_value();
    size = 
        1 +                      // Chip type
        1 +                      // threshold
        1 +                      // number of chips and chans
        tm.Nchan()*tm.Nchip() +  // all the channels
        tm.Nchip() +             // the chip config
        1                         // separator  
        ;

    int ichan,ichip,indx=0;
    unsigned int *data = new unsigned int[size];

    indx=0;
    data[indx++] = tm.GetChipType();
    data[indx++] = get_threshold();
    data[indx++] = tm.Nchip()<< 16 | tm.Nchan();
    for (ichip=0; ichip<tm.Nchip(); ichip++)
    {
        TAchip *chip = tm.chip(ichip);
        for (ichan=0;ichan<tm.Nchan(); ichan++)
        {
            data[indx++] = chip->mask(ichan) << 12 |
                           chip->test(ichan) <<  8 |
                           chip->trim(ichan);
        }
    }
    data[indx++] = 0x100AC001;
    for (ichip=0; ichip<tm.Nchip(); ichip++)
    {
        TAchip *chip = tm.chip(ichip);
        data[indx++] = ( chip->Chip_Address() & 0xff) |
        (chip->Test() ? 1 : 0) << 8 |
        (chip->Probe_Enable() ? 1 : 0) << 9 |
        (chip->CC_Enable() ? 1 : 0) << 10 |
        (chip->N_side() ? 1 : 0) << 11 |
        (chip->Gain() & 7) << 12 |
        (chip->GainEnabled() ? 1 : 0) << 15;

    }
    return data;
}


int VATAmodule::GetReady()
{
    std::cout << "VATAmodule::GetReady" << std::endl;
    return 0;
}

void VATAmodule::set_scan_point(ScanPoint &scp)
{
    // Set variables to the values in the current point
    // For all the modules
    if (run_type != Scan )
        return;

    int romode = get_rout();
    Reset();
    set_rout( VATAmodule::On );
    std::vector<ScanVar>::iterator ip;


    ip = scp.find_type(ScanVar::channel);
    if ( ip != scp.end() )
    {
        int ch = int(ip->value());
        if (ch >= 0 && ch < get_analysis()->Ntot())
        {
            select_channel( int(ip->value()) );
            _theChannel = int(ip->value());
        }
        else
        {
            select_channel(0);
            _theChannel = 0;
        }
    }
    ip = scp.find_type(ScanVar::threshold);
    if ( ip != scp.end() )
    {
        set_threshold( int(ip->value()) );
    }
    ip = scp.find_type(ScanVar::thrs_trim);
    if ( ip != scp.end() )
    {
        TAmanager &tm = ta->get_value();
        int ichip, nchip = tm.Nchip();
        for (ichip=0;ichip<nchip;ichip++)
        {
            tm.chip(ichip)->thrs_dac( int(ip->value()) );
        }
    }

    ip = scp.find_type(ScanVar::trim);
    if ( ip != scp.end() )
        trim_channel( channel, int(ip->value()) );

    ip = scp.find_type(ScanVar::pulse);
    if ( ip != scp.end() )
    {
        // This is an external calibration pulse
        TAmanager &tm = ta->get_value();
        int ichip, nchip = tm.Nchip();
        for (ichip=0;ichip<nchip;ichip++)
        {
            tm.chip(ichip)->calib_ctrl(false)->calib_gen(false);
        }
        _thePulse = ip->value();
        _mean = _thePulse / _theGain;
    }
    
    ip = scp.find_type(ScanVar::cal_dac);
    if ( ip != scp.end() )
    {
        // This is an internal calibration pulse
        TAmanager &tm = ta->get_value();
        int ichip, nchip = tm.Nchip();
        for (ichip=0;ichip<nchip;ichip++)
        {
            tm.chip(ichip)->calib_ctrl(true)->calib_gen(true)->calib_dac( int(ip->value()) );
        }
    }
    
    
    
    // In sparse SHIFTREG must be zero, in Serial true
    if ( romode & VATAmodule::Serial )
        romode |= VATAmodule::On;
    else
        romode &= ~VATAmodule::On;

    set_rout( romode );
    load_register();
}

int VATAmodule::Reset()
{
    VMEmanager *vme = VMEmanager::VME();
    int itry;
    for (itry = 0;itry < 10;itry++)
    {
        vme->write32(_add + RST, 0);
        CHECK_BUS(vme);
        if (!vme->get_rc())
            break;
    }
    return 0;
}

int VATAmodule::Start(int mode)
{
    return 0;
}

void VATAmodule::SoftStart()
{
    VMEmanager::VME()->write32(_add + SOFTSTART, 0);
}

int VATAmodule::Trigger()
{
    // Pulse
    VMEmanager::VME()->write32(_add + PULSE, 0);
    switch (run_type)
    {
        case IntTrg:
            SoftStart();
            break;
        case Pedestal:
            if (get_rout() & Serial)
            { //serial
                SoftStart();
            }
            else
            {
            }
            break;
        default:
            break;
    }
    return 0;
}

void VATAmodule::load_register()
{
    VMEmanager *vme = VMEmanager::VME();
    ACQUIRE_VME(vme);
    TAmanager &tm = ta->get_value();
    int ichip, ibit, i;
    unsigned int ival, jval;
    bool reg_error;

    std::vector<VMEcommand> stack;
    std::vector<unsigned int> data;
    std::vector<unsigned int> reg_check;

    int ktot = 0;
    for (i=0; i<tm.Nchip(); i++) ktot+=tm.chip(i)->nBits();
    data.reserve(ktot);
    reg_check.reserve(ktot);
    stack.reserve(4*ktot);

    bool _check_register = true;
    if (_check_register) std::cout << "\nPrinting Control Register Bits..." << std::endl;

    for (int itry=0; itry<5; itry++) {
      reg_error = false;

      for (ichip=tm.Nchip()-1; ichip>=0; ichip--) {
        TAchip *chip = tm.chip(ichip);

        if (_check_register) std::cout << "Chip: " << ichip << "  nBits = " << chip->nBits() << "\n";

        for (ibit=chip->nBits()-1; ibit>=0; ibit--) {
          ival = (*chip)[ibit];
          if (_check_register) std::cout << ival;
          stack.push_back(VMEcommand(VMEcommand::write32,_add+RWREG,ival));
          stack.push_back(VMEcommand(VMEcommand::write32,_add+REGCLK,0));
          stack.push_back(VMEcommand(VMEcommand::write32,_add+REGCLK,0));
        }
        if (_check_register) std::cout << std::endl;
      }
      vme->stack_execute(stack,data);
      CHECK_BUS(vme);
      stack.clear();
      data.clear();

      if (_check_register) std::cout << "Checking Control Register..." << std::endl;

      for (ichip=tm.Nchip()-1; ichip>=0; ichip--) {
        TAchip *chip = tm.chip(ichip);
        for (ibit=chip->nBits()-1; ibit>=0; ibit--) {
          ival = (*chip)[ibit];
          reg_check.push_back(ival);
          stack.push_back(VMEcommand(VMEcommand::write32,_add+RWREG,ival));
          stack.push_back(VMEcommand(VMEcommand::read32,_add+RWREG));
          stack.push_back(VMEcommand(VMEcommand::write32,_add+REGCLK,0));
          stack.push_back(VMEcommand(VMEcommand::write32,_add+REGCLK,0));
        }
      }
      vme->stack_execute(stack,data);
      CHECK_BUS(vme);

      if ((int)reg_check.size()!=ktot || (int)data.size()!=ktot) {
        reg_error = true;
      }
      else {
        for (ichip=tm.Nchip()-1,i=0; ichip>=0; ichip--) {
          TAchip *chip = tm.chip(ichip);
          for (ibit=chip->nBits()-1; ibit>=0; ibit--) {
            ival = reg_check[i];
            jval = (data[i]&0x80000000)>>31;
            if (jval!=ival) reg_error = true;
            i++;
          }
        }
      }
      RELEASE_VME(vme);

      if (!reg_error) break;
    }

    if (reg_error) {
      std::cout << "Control Register Error..." << std::endl;
      if (_check_register) {
        for (ichip=tm.Nchip()-1,i=0; ichip>=0; ichip--) {
          TAchip *chip = tm.chip(ichip);
          std::cout << "Chip: " << ichip << "  nBits = " << chip->nBits() << "\n";
          for (ibit=chip->nBits()-1; ibit>=0; ibit--) {
            try {
              std::cout << ((data.at(i)&0x80000000)>>31);
            } catch (std::out_of_range& e) {
              std::cout << -1;
            }
            i++;
          }
          std::cout << std::endl;
        }
      }
    }
    else {
      std::cout << "Control Register OK!" << std::endl;
    }

}

/* old load_register method
void VATAmodule::load_register()
{
    VMEmanager *vme = VMEmanager::VME();
    ACQUIRE_VME(vme);
    TAmanager &tm = ta->get_value();
    int ichip, ibit, ktot;
    unsigned int ival, jval;
    bool reg_error;
    
    for (int itry=0; itry<5; itry++)
    {
       reg_error = false;

    for (ichip = tm.Nchip() - 1, ktot = 0;ichip >= 0;ichip--)
    {
        TAchip *chip = tm.chip(ichip);
//        std::cout << "Chip " << ichip << std::endl
//		          << "Bits " << chip->nBits() << std::endl;
        for ( ibit = chip->nBits() - 1; ibit >=0; ibit--)
        {
            ival = (*chip)[ibit];
//            std::cout << ival;
            vme->write32(_add + RWREG, ival);
            CHECK_BUS(vme);
            vme->write32(_add + REGCLK, 0);
            CHECK_BUS(vme);
            vme->write32(_add + REGCLK, 0);
            CHECK_BUS(vme);
            ktot++;
        }
//	std::cout << std::endl;
    }

    unsigned int reg_check[ktot];

//    std::cout << "Checking Control Register..." << std::endl;
    for (ichip = tm.Nchip() - 1, ktot = 0;ichip >= 0;ichip--)
    {
        TAchip *chip = tm.chip(ichip);
        for ( ibit = chip->nBits() - 1; ibit >=0; ibit--)
        {
            ival = (*chip)[ibit];
            vme->write32(_add + RWREG, ival);
            CHECK_BUS(vme);
            jval = vme->read32(_add + RWREG);
            CHECK_BUS(vme);
            jval &= 0x80000000;
            jval = jval >> 31;
            reg_check[ktot] = jval;

            if ( jval != ival ) 
                reg_error = true;

            vme->write32(_add + REGCLK, 0);
            CHECK_BUS(vme);
            vme->write32(_add + REGCLK, 0);
            CHECK_BUS(vme);
            ktot++;
        }
    }
    RELEASE_VME(vme);

       if (!reg_error)
	  break;
    }
    if (reg_error)
    {
        std::cout << "Control Register Error..." << std::endl;
//        for (ichip = tm.Nchip() - 1, ktot = 0; ichip >= 0; ichip--)
//        {
//            TAchip *chip = tm.chip(ichip);
//            std::cout << "Chip " << ichip << std::endl << "Bits " << chip->nBits()
//                    << std::endl;
//            for (ibit = chip->nBits() - 1; ibit >=0; ibit--)
//            {
//                std::cout << reg_check[ktot];
//                ktot++;
//            }
//            std::cout << std::endl;
//        }
    }
    else
    {
//        std::cout << "Control Register OK!" << std::endl << std::flush;
    }

//    if (false) {
//        Reset();
//        read_register();
//    }
    //   std::cout << std::endl << "Loaded " << ktot << " bits" << std::endl;
}
*/
void VATAmodule::read_register()
{
    VMEmanager *vme = VMEmanager::VME();
    ACQUIRE_VME(vme);
    TAmanager &tm = ta->get_value();
    int ichip, ibit, ktot;
    unsigned int ival;
    //   std::cout << tm << std::endl;
    //   for (ichip=0,ktot=0;ichip<tm.Nchip();ichip++) {
    for (ichip = tm.Nchip() - 1, ktot = 0;ichip >= 0;ichip--)
    {
        TAchip *chip = tm.chip(ichip);
	std::cout << "Chip " << ichip << std::endl
		  << "Bits " << chip->nBits() << std::endl;
        for ( ibit = 0; ibit < chip->nBits();ibit++)
        {
            ival = vme->read32(_add + RWREG);
            ival &= 0x80000000;
            ival = ival >> 31;
	    std::cout << ival;
            CHECK_BUS(vme);
            vme->write32(_add + REGCLK, 0);
            CHECK_BUS(vme);
            vme->write32(_add + REGCLK, 0);
            CHECK_BUS(vme);
            ktot++;
        }
	std::cout << std::endl;
    }
    RELEASE_VME(vme);
    //   std::cout << std::endl << "Loaded " << ktot << " bits" << std::endl;
}

void VATAmodule::set_threshold(int x, bool nt)
{
    int v = static_cast<int>(255. * x * 13. / 2550. );
    VMEmanager *vme = VMEmanager::VME();
    ACQUIRE_VME(vme);
    vme->write32(_add + WDAC, v);
    CHECK_BUS(vme);
    RELEASE_VME(vme);
    threshold->set_value(x, nt);
}

void VATAmodule::set_forceDataMode( bool x, bool nt )
{
    forceDataMode->set_value( x, nt );
}

void VATAmodule::set_nchip(int x, bool nt)
{
    if ( x > 16 )
    {
        std::cout << "Max. number of chips is 16" << std::endl;
        return;
    }
    unsigned int v = x * 128 + 3;
    VMEmanager *vme = VMEmanager::VME();
    ACQUIRE_VME(vme);
    vme->write32(_add + WCHAN, v);
    CHECK_BUS(vme);
    RELEASE_VME(vme);
    nchip->set_value(x, nt);
}

void VATAmodule::set_adj(int x, bool nt)
{
    x &= 0xff;

    if ( x > 2047 )
    {
        std::cout << "Max. number of adjacent channels is 2047" << std::endl;
        return;
    }
    VMEmanager *vme = VMEmanager::VME();
    ACQUIRE_VME(vme);
    vme->write32(_add + WADJCHAN, x);
    CHECK_BUS(vme);
    RELEASE_VME(vme);
    adj->set_value(x, nt);
}

void VATAmodule::set_rout( int x, bool nt)
{
    VMEmanager *vme = VMEmanager::VME();
    ACQUIRE_VME(vme);
    vme->write32(_add + WCNTREG, x);
    CHECK_BUS(vme);
    RELEASE_VME(vme);
    rout->set_value(x, nt);
}

void VATAmodule::set_ta( TAmanager &x, bool nt)
{
    TAmanager &tm = ta->get_value();
    if ( tm.Nchip() != x.Nchip() || tm.Nchan() != x.Nchan() )
    {
        if ( ana )
            delete ana;
        ana = new VATAmodule_analysis(x.Nchip(), x.Nchan());
        ((VATAmodule_analysis *)ana)->set_chip_type(x.GetChipType());
        std::vector<double> &ped = pedestal->get_value();
        std::vector<double> &ns  = noise->get_value();
        ped.clear();
        ns.clear();
        int i;
        for ( i = 0;i < ana->Ntot();i++ )
        {
            ped.push_back( 0. );
            ns.push_back( 0. );
        }
    }
    ta->set_value(x, nt);
    //load_register();
}



void VATAmodule::change_threshold(int &x)
{
    set_threshold(x, false);
}

void VATAmodule::change_forceDataMode( bool &x )
{
    set_forceDataMode( x, false );
}

void VATAmodule::change_nchip(int &x)
{
    set_nchip(x, false);
}

void VATAmodule::change_rout(int &x)
{
    set_rout(x, false);
}

void VATAmodule::change_adj(int &x)
{
    set_adj(x, false);
}

void VATAmodule::change_ta(TAmanager &x)
{
    std::cout << "**** Changing TA ****" << std::endl;
}


// Sets the gain
void VATAmodule::set_gain(int ichip, int gain)
{
    TAmanager &tm = ta->get_value();
    TAchip *chip = tm.chip(ichip);
    if ( chip == 0 )
        return;

    chip->EnableGain( gain >= 0 );
    chip->SetGain( gain >= 0 ? gain : 0);
}

// maks all the channels
void VATAmodule::mask_all()
{
    TAmanager &tm = ta->get_value();
    int ichip, ichan, ic;

    int nchan =  tm.Nchan();
    int nchip = tm.Nchip();
    TAchip *chip;
    for (ic = ichip = 0;ichip < nchip;ichip++)
    {
        chip = tm.chip(ichip);
        for (ichan = 0;ichan < nchan;ichan++, ic++)
        {
            chip
            ->mask(ichan, true)
            ->test(ichan, false)
            ;
        }
    }
}

// masks a given channel
void VATAmodule::mask_channel(int ch)
{
    TAmanager &tm = ta->get_value();
    int nchan =  tm.Nchan();
    // Get the channel we are testing
    int wchip = ch / nchan;
    int wchan = ch % nchan;
    tm.chip(wchip)->mask(wchan, true);
}

// selects one channel for triggering
void VATAmodule::select_channel(int ch)
{
    channel = ch;
    TAmanager &tm = ta->get_value();
    int wchan, nchan =  tm.Nchan();
    int wchip, nchip = tm.Nchip();

    // mask all the channels
    mask_all();

    // Set to test all the chips
    for (wchip = 0;wchip < nchip;wchip++)
        tm.chip(wchip)->Test(true);

    // Get the channel we are testing
    //   wchip = nchip-ch/nchan-1;
    wchip = ch / nchan;
    wchan = ch % nchan;
    tm.chip(wchip)->mask(wchan, false);
    tm.chip(wchip)->test(wchan, true);// Id.
}

void VATAmodule::trim_channel(int ch, int tr)
{
    TAmanager &tm = ta->get_value();
    int nchan =  tm.Nchan();
    // Get the channel we are testing
    int wchip = ch / nchan;
    int wchan = ch % nchan;

    TAchip *chip = tm.chip(wchip);
    if ( chip )
        chip->trim(wchan, tr);
}
// enable all channels for triggering
void VATAmodule::select_all()
{
    TAmanager &tm = ta->get_value();
    int ichip, ichan, ic;
    int nchan =  tm.Nchan();
    int nchip = tm.Nchip();
    TAchip *chip;
    for (ic = ichip = 0;ichip < nchip;ichip++)
    {
        chip = tm.chip(ichip);
        chip->Test(false);
        for (ichan = 0;ichan < nchan;ichan++, ic++)
        {
            chip
            ->mask(ichan, false)
            ->test(ichan, false)
            ;
        }
    }
}

void VATAmodule::set_header_description(unsigned int &data)
{
    unsigned int id = get_id();
    data = (get_nchip() & 0xff) << 20 | NormalData | (id & 0xffff);
}

unsigned int *VATAmodule::write_pedestals(int &size)
{
    int i, j;
    unsigned int *data;

    update_pedestals();

    size = get_ta().Nchip() * get_ta().Nchan();
    data = new unsigned int[4+size];

    i = 0;
    data[i++] = 0xcafe0000 | PedestalRecord | (4 + size);
    data[i++] = get_id();
    data[i++] = (get_ta().Nchip() & 0xffff) << 16 | get_ta().Nchan();

    const std::vector<double> &ped = get_pedestal();
    const std::vector<double> &ns  = get_noise();
    for ( j = 0;j < size;j++)
        data[i++] = int(50. * ns[j]) << 16 | int(50 * ped[j]);
    data[i] = 4 + size;

    size = data[i];
    return data;
}

void VATAmodule::step_advance(int ch)
{
    int i;
    VMEmanager *vme = VMEmanager::VME();
    Reset();
    Trigger();
    /*
    TAmanager &tm = ta->get_value();
    int nchan =  tm.Nchan();
    int nchip = tm.Nchip();

    vme->write32(_add + SHIFTIN_CPU, 0); // make sure SHIFTIN_D is 0
    CHECK_BUS(vme);
    for (i = 0;i < nchip*nchan;i++)
    {
        if (i == ch)
        {
            vme->write32(_add + SHIFTIN_CPU, 1); // Write 1 to SHIFTIN_D
            CHECK_BUS(vme);
            vme->write32(_add + SOFTCLK, 0);  // Give clock to VATA
            CHECK_BUS(vme);
            vme->write32(_add + SOFTCLK, 0);
            CHECK_BUS(vme);
            vme->write32(_add + SHIFTIN_CPU, 0); // SHIFTIN_D to 0 again
            CHECK_BUS(vme);
        }
        else
        {
            vme->write32(_add + SOFTCLK, 0);  // Give clock to VATA
            CHECK_BUS(vme);
            vme->write32(_add + SOFTCLK, 0);
            CHECK_BUS(vme);
        }
    }
    */

    for (i = 0;i < ch+1;i++)
    {
        if (i == 0)
        {
            vme->write32(_add + SHIFTIN_CPU, 1); // Write 1 to SHIFTIN_D
            CHECK_BUS(vme);
            vme->write32(_add + SOFTCLK, 0);  // Give clock to VATA
            CHECK_BUS(vme);
            vme->write32(_add + SOFTCLK, 0);
            CHECK_BUS(vme);
            vme->write32(_add + SHIFTIN_CPU, 0); // SHIFTIN_D to 0 again
            CHECK_BUS(vme);
        }
        else
        {
            vme->write32(_add + SOFTCLK, 0);  // Give clock to VATA
            CHECK_BUS(vme);
            vme->write32(_add + SOFTCLK, 0);
            CHECK_BUS(vme);
        }
    }
    
}

void VATAmodule::test_channel(int ch)
{
    set_rout(1);
    select_channel(ch);
    Reset();
    load_register();
    step_advance(ch);
}
