#include <iostream>
#include <iomanip>
#include <sstream>
#include <DAQ++/DAQmanager.h>
#include <config/ConfigParser.h>
#include <config/ModuleParser.h>
#include <config/GeneralConfig.h>
#include <data/Analysis.h>
#include <gui/MonitorWindow.h>
#include <gui/DefaultMonitorGUI.h>

static void __attribute__ ((constructor)) __init__lib__()
{
    printf("### Loading VATAmodule module...\n");
}

static void __attribute__ ((destructor)) __close__lib__()
{
    printf("### Unloading VATAmodule...\n");
}

#include "VATAmodule.h"
#include "VATAmoduleGuiConfig.h"
class ParseVATAmodule : public ModuleParser
{
    private:
        std::string name;
    public:
        ParseVATAmodule()
        {
        }
        ~ParseVATAmodule()
        {
            std::cout << "deleting ParseVATAmodule"<< std::endl;
        }
        void parse_node_id(const std::string &) throw(ConfigException);
        void parse_node(xmlNodePtr) throw(ConfigException);

        void parse_chip(TAchip *, xmlNodePtr);
        void parse_trim(TAchip *);
        void parse_mask(TAchip *);
        void parse_test(TAchip *);

        void save_config() throw(ConfigException);
        void save_config_id(const std::string&) throw(ConfigException);
        void save_config_node(xmlNodePtr) throw(ConfigException);

        void save_chip(TAchip *, xmlNodePtr);

        _GtkWidget *config(const std::string &);
        MonitorWindow *monitor(const std::string &, const std::string &);

};

//=====================================================================
// parsing stuff
//=====================================================================
void ParseVATAmodule::parse_trim(TAchip *chip)
{
    int i, n, ntrim;
    std::vector<xmlNodePtr> vtrim;
    ConfigParser *cfg = ConfigParser::theConfig();
    std::ostringstream otrim;

    otrim << "//"<< name << "[@name=\""<< id << "\"]/chip[@id=\""<< chip->Chip_Address()
          << "\"]/trim";

    ntrim = cfg->get_node(otrim.str(), vtrim);
    if (!ntrim)
        return;

    if (ntrim > 1)
    {
        std::cerr << "Chip "<< chip->Chip_Address() << " in module "<< id
                  << " has >1 trim array. Ignoring the rest."<< std::endl;
    }
    xmlNodePtr nc = vtrim[0];
    std::vector<int> trim;
    cfg->get_array<int>(nc, trim);
    n = trim.size();
    if (chip->Nchan() < n)
    {
        std::cerr << "Trim array in chip "<< chip->Chip_Address()<< " at module "<< id
                  << " has more than "<< chip->Nchan() << " channels."<< std::endl;
        n = chip->Nchan();
    }
    for (i = 0; i < n; i++)
        chip->trim(i, trim[i]);
}

void ParseVATAmodule::parse_mask(TAchip *chip)
{
    int i, n;
    int nmask;
    std::vector<xmlNodePtr> vmask;
    ConfigParser *cfg = ConfigParser::theConfig();
    std::ostringstream omask;

    omask << "//"<< name << "[@name=\""<< id << "\"]/chip[@id=\""<< chip->Chip_Address()
          << "\"]/mask";

    nmask = cfg->get_node(omask.str(), vmask);
    if (!nmask)
        return;

    // take only the first one
    if (nmask > 1)
    {
        std::cerr << "Chip "<< chip->Chip_Address() << " in module "<< id
                  << " has >1 mask array. Ignoring the rest."<< std::endl;
    }
    xmlNodePtr nc = vmask[0];
    std::vector<int> vv;
    if (cfg->get_bit_array(nc, chip->Nchan(), vv) )
    {
        std::cerr << "Module "<< id<< " chip "<< chip->Chip_Address() << std::endl
                  << "... wrong syntax in mask"<< std::endl;
    }
    else
    {
        n = vv.size();
        if (n > chip->Nchan() )
        {
            std::cerr << "Module "<< id<< " chip "<< chip->Chip_Address() << std::endl
                      << "... number of mask bits is " << "bigger than number of channels"
                      << std::endl;
            n = chip->Nchan();
        }
        for (i = 0; i < n; i++)
            chip->mask(i, vv[i] != 0);
    }
}

void ParseVATAmodule::parse_test(TAchip *chip)
{
    int i, n;
    int ntest;
    std::vector<xmlNodePtr> vtest;
    ConfigParser *cfg = ConfigParser::theConfig();
    std::ostringstream otest;

    otest << "//"<< name << "[@name=\""<< id << "\"]/chip[@id=\""<< chip->Chip_Address()
          << "\"]/test";
    ntest = cfg->get_node(otest.str(), vtest);
    if ( !ntest)
        return;

    // take only the first one
    if (ntest > 1)
    {
        std::cerr << "Chip "<< chip->Chip_Address() << " in module "<< id
                  << " has >1 test array. Ignoring the rest."<< std::endl;
    }
    xmlNodePtr nc = vtest[0];
    std::vector<int> vv;
    if (cfg->get_bit_array(nc, chip->Nchan(), vv) )
    {
        std::cerr << "Module "<< id<< " chip "<< chip->Chip_Address() << std::endl
                  << "... wrong syntax in test"<< std::endl;
    }
    else
    {
        n = vv.size();
        if (n > chip->Nchan() )
        {
            std::cerr << "Module "<< id<< " chip "<< chip->Chip_Address() << std::endl
                      << "... number of test bits is " << "bigger than number of channels"
                      << std::endl;
            n = chip->Nchan();
        }
        for (i = 0; i < n; i++)
            chip->test(i, vv[i] != 0);
    }

}

void ParseVATAmodule::parse_chip(TAchip *chip, xmlNodePtr nc)
{
    int ival;
    ConfigParser *cfg = ConfigParser::theConfig();

    // get the chip id
    chip->Chip_Address(cfg->get_attribute<int>(nc, "id") );

    // Gain
    try
    {
        ival = cfg->get_value<int>(nc, "gain");
    }
    catch ( ConfigException &e )
    {
        ival = -1;
    }
    if (ival < 0)
        chip->SetGain(0)->EnableGain(false);
    else
        chip->SetGain(ival)->EnableGain(true);

    try
    {
        chip->Probe_Enable( cfg->get_value<bool>(nc, "probe") );
    }
    catch ( ConfigException &e )
    {
        chip->Probe_Enable( false );
    }
    try
    {
        chip->CC_Enable( cfg->get_value<bool>(nc, "cc_enable") );
    }
    catch ( ConfigException &e )
    {
        chip->CC_Enable( false );
    }
    try
    {
        chip->N_side( cfg->get_value<bool>(nc, "n_side") );
    }
    catch ( ConfigException &e )
    {
        chip->N_side( false );
    }
    try
    {
        chip->Polarity( cfg->get_value<bool>(nc, "polarity") );
    }
    catch ( ConfigException &e )
    {
        chip->Polarity( true );
    }
    try
    {
        chip->Test( cfg->get_value<bool>(nc, "test_bit") );
    }
    catch ( ConfigException &e )
    {
        chip->Test( false );
    }
    
    try 
    {
        chip->Vrc( cfg->get_value<bool>(nc, "Vrc"));
    }
    catch (ConfigException &e)
    {
        chip->Vrc( false );
    }
    
    try 
    {
        chip->Vfsf( cfg->get_value<bool>(nc, "Vfsf"));
    }
    catch (ConfigException &e)
    {
        chip->Vfsf( false );
    }

    try 
    {
        chip->Vfss( cfg->get_value<bool>(nc, "Vfss"));
    }
    catch (ConfigException &e)
    {
        chip->Vfss( false );
    }
    
    try 
    {
        chip->Vfp( cfg->get_value<bool>(nc, "Vfp"));
    }
    catch (ConfigException &e)
    {
        chip->Vfp( false );
    }

    try 
    {
        chip->dac_range( cfg->get_value<bool>(nc, "dac_range"));
    }
    catch (ConfigException &e)
    {
        chip->dac_range( false );
    }
    
    try 
    {
        chip->thrs_dac( cfg->get_value<int>(nc, "thrs_dac"));
    }
    catch (ConfigException &e)
    {
        chip->thrs_dac( 0 );
    }
    
    try 
    {
        chip->calib_gen( cfg->get_value<bool>(nc, "calib_gen"));
    }
    catch (ConfigException &e)
    {
        chip->calib_gen( false );
    }
    try 
    {
        chip->calib_ctrl( cfg->get_value<bool>(nc, "calib_ctrl"));
    }
    catch (ConfigException &e)
    {
        chip->calib_ctrl( false );
    }
    
    try 
    {
        chip->calib_dac( cfg->get_value<int>(nc, "calib_dac"));
    }
    catch (ConfigException &e)
    {
        chip->calib_dac( 0 );
    }
    try 
    {
        chip->vrc_neg( cfg->get_value<bool>(nc, "vrc_neg"));
    }
    catch (ConfigException &e)
    {
        chip->vrc_neg( false );
    }
    try 
    {
        chip->slew( cfg->get_value<bool>(nc, "slew"));
    }
    catch (ConfigException &e)
    {
        chip->slew( false );
    }
    try 
    {
        chip->dltc( cfg->get_value<bool>(nc, "dltc"));
    }
    catch (ConfigException &e)
    {
        chip->dltc( false );
    }
    
    
    //-----------------------------------------------------------
    // Trim
    //-----------------------------------------------------------
    parse_trim(chip);
    //-----------------------------------------------------------
    // Mask
    //-----------------------------------------------------------
    parse_mask(chip);
    //-----------------------------------------------------------
    // test
    //-----------------------------------------------------------
    parse_test(chip);

}

void ParseVATAmodule::parse_node_id(const std::string &name) throw(ConfigException)
{
    std::ostringstream ostr;
    ConfigParser *cfg = ConfigParser::theConfig();
    DAQpp::Module *dm = DAQpp::DAQmanager::theDAQmanager()->find_module(name);

    if (!dm)
        return;

    if ( !dynamic_cast<VATAmodule *>(dm) )
    {
        std::cout << "ParseVATAmodule: module "<< name<< " is not of type VATAmodule "
                << std::endl;
        return;
    }

    int nmod;
    std::vector<xmlNodePtr> vmod;
    ostr << "//"<< name;
    nmod = cfg->get_node(ostr.str(), vmod);
    if (nmod)
        parse_node(vmod[0]);

}
void ParseVATAmodule::parse_node(xmlNodePtr _n) throw(ConfigException)
{
    int address;
    std::string saddress;
    ConfigParser *cfg = ConfigParser::theConfig();

    // we quit if the node is 0
    if (_n == 0)
        return;

    node = _n;
    name = node_name();
    id = cfg->get_attribute<std::string>(node, "name");

    saddress = cfg->get_attribute<std::string>(node, "address");

    std::istringstream os(saddress);
    os >> std::hex >> address;

    // Check if the module already exist
    VATAmodule *m = 0;
    DAQpp::Module *dm;

    if ( (dm = DAQpp::DAQmanager::theDAQmanager()->find_module(id)))
        m = dynamic_cast<VATAmodule *>(dm);

    if (!m) // Create the module
        m = new VATAmodule(id, address);

    // should not happen unless we run out of memory
    if (m == 0)
        return;

    // Now we start parsing
    std::string styp;
    TAmanager tam;

    int nchip, threshold, nadj;
    double rodelay;
    bool forceDataMode;
    bool fast_acquire;

    try
    {
        rodelay = cfg->get_value<double>(node, "delay");
    }
    catch ( ConfigException &e )
    {
        rodelay = 1;
    }
    try
    {
        nchip = cfg->get_value<int>(node, "nchips");
    }
    catch (ConfigException &e)
    {
        nchip = 1;
    }
    try
    {
        threshold = cfg->get_value<int>(node, "threshold");
    }
    catch ( ConfigException &e )
    {
        threshold = 50;
    }
    try
    {
        forceDataMode = cfg->get_value<bool>( node, "forceDataMode" );
    }
    catch ( ConfigException &e )
    {
        forceDataMode = false;
    }
    try
    {
        nadj = cfg->get_value<int>(node, "n_adj");
    }
    catch ( ConfigException &e )
    {
        nadj = 10;
    }

    
    try{
      fast_acquire= cfg->get_value<bool>(node,"fast_acquire");
    }
    catch (ConfigException &e){
	fast_acquire=m->get_par_value<bool>("fast_acquire");
    }
    m->set_par<bool>("fast_acquire",fast_acquire);
    
    m->set_delay(rodelay);
    m->set_nchip(nchip);
    m->set_threshold(threshold);
    m->set_forceDataMode(forceDataMode);
    m->set_adj(nadj);
    try
    {
        styp = cfg->get_value<std::string>(node, "chip_type");
    }
    catch ( ConfigException &e )
    {
        styp = "gp3";
    }
    tam.Resize(m->get_nchip() );
    if (styp == "ta32")
    {
        tam.SetChipType(TAmanager::chipTA);
    }
    else if (styp == "gp3p")
    {
        tam.SetChipType(TAmanager::chipGP3p);
    }
    else if (styp == "gp3")
    {
        tam.SetChipType(TAmanager::chipGP3);
    }
    else if (styp=="gp5")
    {
        tam.SetChipType(TAmanager::chipGP5);
    }
    else if (styp=="gp7")
    {
        tam.SetChipType(TAmanager::chipGP7);
    }
    else if (styp == "gp")
    {
        tam.SetChipType(TAmanager::chipGP);
    }
    m->set_ta(tam);

    try
    {
        styp = cfg->get_value<std::string>(node, "ro_mode");
    }
    catch ( ConfigException &e )
    {
        styp = "serial";
    }
    if (styp == "serial")
    {
        m->set_rout(VATAmodule::Serial);
    }
    else if (styp == "sparse")
    {
        m->set_rout(VATAmodule::Sparse);
    }
    else if (styp == "adjacent")
    {
        m->set_rout(VATAmodule::SparseAdj);
    }
    //---------------------------------------------------------------------
    // Look for chips
    //---------------------------------------------------------------------
    int ichip;
    std::vector<xmlNodePtr> vchip;
    std::ostringstream ostr;
    TAmanager &ta = m->get_ta();

    ostr << "//"<< name << "[@name=\""<< id << "\"]/chip";

    nchip = cfg->get_node(ostr.str(), vchip);
    if (nchip)
    {
        if (m->get_nchip() < nchip)
        {
            std::cerr << "Module "<< id<< " found more chips than specified."<< std::endl
                      << "... accepting only the first "<< nchip << "."<< std::endl;
        }
        for (ichip = 0; ichip < nchip; ichip++)
        {
            xmlNode *nc = vchip[ichip];
            TAchip *chip = ta.chip(ichip);
            parse_chip(chip, nc);
        }
    }
    m->get_analysis()->Resize(nchip, ta.Nchan() );

    // monitor and PadMap
    parse_common("//" + name, id);
    m->load_register();
    node = 0;
}

void ParseVATAmodule::save_config_id(const std::string &name) throw(ConfigException)
{
    ConfigParser *cfg = ConfigParser::theConfig();
    DAQpp::Module *dm = DAQpp::DAQmanager::theDAQmanager()->find_module(id);

    if (!dm)
        return;

    if ( !dynamic_cast<VATAmodule *>(dm) )
    {
        std::cout << "ParseVATAmodule: module "<< name<< " is not of type VATAmodule "
                  << std::endl;
        return;
    }

    std::ostringstream ostr;
    int nmod;
    std::vector<xmlNodePtr> vmod;
    ostr << "//"<< name << "[@name=\""<< id << "\"]";
    nmod = cfg->get_node(ostr.str(), vmod);
    if (nmod)
        save_config_node(vmod[0]);

}

void ParseVATAmodule::save_config() throw(ConfigException)
{
    std::vector<xmlNodePtr> vmod;
    std::vector<xmlNodePtr>::iterator ip;
    ConfigParser *cfg = ConfigParser::theConfig();

    cfg->get_node("//" + name, vmod);
    for (ip = vmod.begin(); ip != vmod.end(); ip++)
        save_config_node( *ip);
}

void ParseVATAmodule::save_config_node(xmlNodePtr _n) throw(ConfigException)
{
    ConfigParser *cfg = ConfigParser::theConfig();

    node = _n;
    id = cfg->get_attribute<std::string>(node, "name");
    VATAmodule *m = 0;
    DAQpp::Module *dm;

    // find the module
    if ( !(dm = DAQpp::DAQmanager::theDAQmanager()->find_module(id)))
        return;

    if ( !(m = dynamic_cast<VATAmodule *>(dm)))
        return;

    std::string styp[] = { "ta32", "gp", "gp3", "gp3p", "gp5", "gp7" };
    TAmanager &ta = m->get_ta();

    int romode = m->get_rout();
    int ctyp = ta.GetChipType();
    std::string ssro;
    std::string sst = styp[ctyp];

    if (romode & VATAmodule::Serial)
        ssro = "serial";
    if (romode & VATAmodule::Sparse)
        ssro = "sparse";
    if (romode & VATAmodule::SparseAdj)
        ssro = "adjacent";

    cfg->check_value<double>(node, "delay", m->get_delay());
    cfg->check_value<int>(node, "nchips", m->get_nchip());
    cfg->check_value<int>(node, "n_adj", m->get_adj());
    cfg->check_value<bool>(node, "fast_acquire",m->get_par_value<bool>("fast_acquire"));
    cfg->check_value<int>(node, "threshold", m->get_threshold());
    cfg->check_value<bool>(node, "forceDataMode", m->get_forceDataMode() );
    cfg->check_value<std::string>(node, "ro_mode", ssro);
    cfg->check_value<std::string>(node, "chip_type", sst);



    int ichip, nchip;
    std::vector<xmlNodePtr> vchip;
    for (ichip = 0; ichip < m->get_nchip(); ichip++)
    {
        std::ostringstream ochip;
        TAchip *chip = ta.chip(ichip);
        ochip << "//"<< name << "[@name=\""<< id<< "\"]/chip[@id=\""
                << chip->Chip_Address() << "\"]";
        nchip = cfg->get_node(ochip.str(), vchip);
        if (!nchip)
        {
            xmlNodePtr val = cfg->create_child(node,"chip","");
            cfg->set_attribute<int>(val,"id",ichip);
            vchip.push_back(val);
        }
        save_chip(chip, vchip[0]);
    }
    save_common();
}
void ParseVATAmodule::save_chip(TAchip *chip, xmlNodePtr n)
{
    ConfigParser *cfg = ConfigParser::theConfig();
    int gain = chip->GainEnabled() ? chip->Gain() : -1;
    cfg->check_value<bool>(n, "probe", chip->Probe_Enable());
    cfg->check_value<bool>(n, "cc_enable", chip->CC_Enable());
    cfg->check_value<bool>(n, "n_side", chip->N_side());
    cfg->check_value<bool>(n, "polarity", chip->Polarity());
    cfg->check_value<bool>(n, "test_bit", chip->Test());
    cfg->check_value<int>(n, "gain", gain);
    cfg->check_value<bool>(n,"Vrc", chip->Vrc());
    cfg->check_value<bool>(n, "Vfsf", chip->Vfsf());
    cfg->check_value<bool>(n, "Vfp", chip->Vfp());
    cfg->check_value<bool>(n, "Vfss", chip->Vfss());
    cfg->check_value<bool>(n, "dac_range", chip->dac_range());
    cfg->check_value<int>(n,"thrs_dac", chip->thrs_dac());
    cfg->check_value<bool>(n, "calib_gen", chip->calib_gen());
    cfg->check_value<bool>(n,"calib_ctrl", chip->calib_ctrl());
    cfg->check_value<int>(n,"calib_dac", chip->calib_dac());
    cfg->check_value<bool>(n,"vrc_neg", chip->vrc_neg());
    cfg->check_value<bool>(n,"slew", chip->slew());
    cfg->check_value<bool>(n,"dltc", chip->dltc());

    int ival;

    // trim
    std::ostringstream ovtrim;
    for (ival = 0; ival < chip->Nchan(); ival++)
    {
        ovtrim << std::setw(2) << chip->trim(ival);
    }

    cfg->set_value<std::string>(n, "trim", ovtrim.str());

    // mask
    std::ostringstream ovmask;

    int ibit, nbit, jbit;
    unsigned short msk;
    nbit = chip->Nchan();
    for (ibit = 0; ibit < nbit; ibit += 4)
    {
        if (ibit && !(ibit % 16))
            ovmask << " ";
        for (msk = jbit = 0; jbit < 4; jbit++)
            if (chip->mask(ibit + jbit) )
                msk |= (1 << jbit);
        ovmask << std::setw(1) << std::hex << msk << std::dec;
    }
    cfg->set_value<std::string>(n, "mask", ovmask.str());

    // test
    std::ostringstream ovtest;
    nbit = chip->Nchan();
    for (ibit = 0; ibit < nbit; ibit += 4)
    {
        if (ibit && !(ibit % 16))
            ovtest << " ";
        for (msk = jbit = 0; jbit < 4; jbit++)
            if (chip->test(ibit + jbit) )
                msk |= (1 << jbit);
        ovtest << std::setw(1) << std::hex << msk << std::dec;
    }
    cfg->set_value<std::string>(n, "test", ovtest.str());

}

//=====================================================================
// GUI
//=====================================================================
_GtkWidget *ParseVATAmodule::config(const std::string &name)
{
    DAQpp::Module *dm = DAQpp::DAQmanager::theDAQmanager()->find_module(name);
    if (!dm)
        return 0;

    VATAmodule *m = 0;
    if ( !(m = dynamic_cast<VATAmodule *>(dm)))
    {
        std::cout << "ParseVATAmodule: module "<< name<< " is not of type VATAmodule "
                << std::endl;
        return 0;
    }
    TAmanager *ta = new TAmanager( DAQpp::get_par_value<TAmanager>(m->get_par("ta")) );
    VATAmoduleGuiConfig *df = new VATAmoduleGuiConfig(m->get_id(), ta, "vmedaq.glade");

    df->init("vmedaq.glade", "DSP-config");
    return df->get_main_window();
}

class MyMonitorWindow : public MonitorWindow
{
    private:
        DefaultMonitorGUI *_mon;
    public:
        MyMonitorWindow(const std::string &name, const std::string &glade_file,
                        VATAmodule *m)
        {
            _mon = new DefaultMonitorGUI( name,
                    glade_file,
                    m->get_analysis());
            _mon->init(glade_file.c_str(), "dsp-monitor");
            _widget = _mon->get_main_window();
        }
        ~MyMonitorWindow()
        {
            delete _mon;
        }

        DefaultMonitorGUI *mon()
        {
            return _mon;
        }
};
MonitorWindow *ParseVATAmodule::monitor(const std::string &name,
                                        const std::string &glade_file)
{
    DAQpp::Module *dm = DAQpp::DAQmanager::theDAQmanager()->find_module(name);
    if (!dm)
        return 0;

    VATAmodule *m = 0;
    if ( !(m = dynamic_cast<VATAmodule *>(dm)))
    {
        std::cout << "ParseVATAmodule: module "<< name<< " is not of type VATAmodule "
                << std::endl;
        return 0;
    }

    MyMonitorWindow *mw = new MyMonitorWindow(name, glade_file, m);

    mw->add_update(new UpdateLabelNtrig(mw->mon()->get_widget("lblDSPnEvts")));
    if (mw->mon()->get_widget("gauge_chip_rate") )
        mw->add_update(new UpdateGaugeRate(mw->mon()->get_widget("gauge_chip_rate") ));

    if (mw->mon()->get_widget("gauge_chip_eff") )
        mw->add_update(new UpdateGaugeEff( mw->mon()->get_widget("gauge_chip_eff")));

    return mw;

}

extern "C"
{
    ModuleParser *create_VATAmodule()
    {
        ParseVATAmodule *gp = new ParseVATAmodule();
        return gp;
    }
}

