#include <cmath>
#include <config/Config.h>
#include "VATAmodule_analysis.h"

VATAmodule_analysis::VATAmodule_analysis(int n_chip, int n_chan) :
    Analysis(n_chip, n_chan)
{
    decode_sparse = &VATAmodule_analysis::decode_sparse_gp3;
}

VATAmodule_analysis::~VATAmodule_analysis()
{
}

void VATAmodule_analysis::set_chip_type(TAmanager::ChipType chip_type)
{
    switch (chip_type)
    {
        case TAmanager::chipGP7:
            decode_sparse =&VATAmodule_analysis::decode_sparse_gp7;
            break;
        case TAmanager::chipGP5:
            decode_sparse = &VATAmodule_analysis::decode_sparse_gp5;
            break;
        default:
            decode_sparse =&VATAmodule_analysis::decode_sparse_gp3;
            break;
    }
}

//---------------------------------------------------------------------
// Analyses the data
//---------------------------------------------------------------------
void VATAmodule_analysis::Analyze()
{

    int i;
    double *chan = Chan();
    double *noise= Noise();
    double xs;

    if ( hits.empty() )
        return;
    
    /*
     * We should already have the pulse stack filled with data.
     * Fill the histograms
     */
    m_fill.acquire();
    sn_max = 0.;
    if (_romode==2)
    {
        // Look for the highest pulse
        double xs_max = -1.e10;
        int i_max = -1;
        for ( i = hits.top();1;i = hits.top())
        {
            if ( noise[i] > 0. )
            {
                xs = chan[i] / noise[i];
                if ( xs > xs_max )
                {
                    i_max = i;
                    xs_max = xs;
                }
            }

            hits.pop();
            if ( hits.empty() )
                break;
        }

        // If found, fill histograms
        if ( i_max < 0 )
            return;

        sn_max = xs_max;
        HitMap()[i_max]++;
        int ib = get_bin(chan[i_max]);
        if ( ib >= 0 && ib < get_nbins() )
        {
            Signal(i_max / Nchan() )[ib]++;
            Signal(-1)[ib]++;
            Signal2D(i_max)[ib]++;
            if (get_map())
            {
                int ix, iy, ip;
                get_map()->pad(i_max, ix, iy);
                ip = ix + get_map()->nx() * iy;
                H2Map()[ ip ]++;
                if ( chan[i_max] >= emin && chan[i_max] <= emax )
                    H2MapC()[ ip ]++;
            }
        }
    }
    else
    {

      //make a copy of hits
      std::stack<int> hits_copy=hits;
      //now analysis will clear up the copy of hits rather than hits themselves
      
      for ( i = hits_copy.top();1; i = hits_copy.top())
	  

        {
            HitMap()[i]++;
            int ib = get_bin(chan[i]);
            if (noise[i]>0.)
            {
                xs = fabs( chan[i] / noise[i]);
                if (xs>sn_max)
                    sn_max =xs;
            
                if ( ib >= 0 && ib < get_nbins() )
                {
                    Signal(i / Nchan() )[ib]++;
                    Signal(-1)[ib]++;
                    Signal2D(i)[ib]++;
                    if (get_map())
                    {
                        int ix, iy, ip;
                        get_map()->pad(i, ix, iy);
                        ip = ix + get_map()->nx() * iy;
                        H2Map()[ ip ]++;
                        if ( chan[i] >= emin && chan[i] <= emax )
                            H2MapC()[ ip ]++;
                    }
                }
            }

            hits_copy.pop();
            if (hits_copy.empty())
	      break;
        }
    }
    m_fill.release();
}

void VATAmodule_analysis::decode(unsigned int *data)
{
    ConfigParser *cfg = ConfigParser::theConfig();
    GeneralConfig *gnrl = dynamic_cast<GeneralConfig *>(cfg->get_handler("general"));

    //clear hits from previous event
    while (hits.size()) hits.pop();

    // We assume here that Analysis has made all the consistency checks
    _romode = (data[2] & 0xf000) >> 12;
    // Get RO mode
    std::vector<int> list_of_channels;
    bool find_pulse=false;
    switch (_romode)
    {
        case 2:
            DoCommonMode(true);
            list_of_channels.reserve(Ntot());
            decode_serial(data, list_of_channels);
            find_pulse=true;
            break;
        case 4:
        case 8:
            DoCommonMode(false);
            list_of_channels.reserve( Ntot() / 4);
            (this->*decode_sparse)(data, list_of_channels);
            break;
    }
    if (gnrl->get_ped_method() == GeneralConfig::normal)
    {
        update_pedestals(list_of_channels, find_pulse);
    }
    else
    {
        update_pedestals_fast(list_of_channels, find_pulse);
    }
}

//---------------------------------------------------------------------
// Decodes serial data
//---------------------------------------------------------------------
void VATAmodule_analysis::decode_serial(unsigned int *data, std::vector<int> &list_of_channels)
{
    double *out = Event();
    double *chan = Chan();
    // We assume the Analysis::Fill has made all teh checks on
    // the data integrity

    unsigned int dt;
    int sz = data[2] & 0xfff;
    int i, j, ioffs;

    // Clear output array
    for (i = 0; i < Ntot(); i++)
        chan[i] = out[i] = 0.;

    list_of_channels.clear();

    // Decode
    sz -= 3; // ignore the 3 xtra clocks needed by the pipeline
    ioffs = 0;
    for (i = 0, j = 6; i < sz && i < Ntot(); i++, j++)
    {
        dt = data[j];
        out[i] = dt & 0x3ff; // Data analogue down
        list_of_channels.push_back(i);
    }
}

//---------------------------------------------------------------------
// Decose sparse data
//---------------------------------------------------------------------
void VATAmodule_analysis::decode_sparse_gp3(unsigned int *data, std::vector<int> &list_of_channels)
{
    double *out = Event();
    double *chan = Chan();
    double *ped = Ped();
    int last_chip;
    // We assume the Analysis::Fill has made all teh checks on
    // the data integrity

    // Another asusmption: We already have the pedestals

    int romode = (data[2] & 0xf000) >> 12;
    int nadj = romode == 8 ? (data[1] & 0xff000000) >> 24 : 0;
    int i, j, sz, add, aux;
    int dsz = (data[0] & 0xfff)-1;
    
    // Clear output array
    for (i = 0; i < Ntot(); i++)
        chan[i] = out[i] = 0.;

    list_of_channels.clear();

    // Get the size and decode
    unsigned int dt;
    sz = data[2] & 0xfff;
    /**
     * How it works. When receiving the data, the address comes in the
     * first clock cycle, but the ADCs needs 3 extra clock cycles, so
     * the amplitude of the first address is shifted by 3 words.
     *
     * For each channel, there are nadj+2 words, and the addres is
     * always there. For example, nadj=2 with 2 channels is:
     *
     * |         3: 8370fc22 chan. 55  sig down 34 up 63
     * |         4: 83710022 chan. 55  sig down 34 up 64
     * |         5: 8370f823 chan. 55  sig down 35 up 62
     * |         6: 83797647 chan. 55  sig down 583 up 605
     * |         7: 8d014035 chan. 208  sig down 53 up 80
     * |         8: 8d014836 chan. 208  sig down 54 up 82
     * |         9: 8d01143c chan. 208  sig down 60 up 69
     * |        10: 8d0a2e76 chan. 208  sig down 630 up 651
     * |        11: 80015038 chan. 0  sig down 56 up 84
     * |        12: 80013431 chan. 0  sig down 49 up 77
     * |        13: 80015425 chan. 0  sig down 37 up 85
     */

    int nblk = 2 + nadj; // size of a channel block
    int ndat = nblk - 1; // size of sensible data in a channel block
    int nTrig, up, down;
    last_chip = -1;
    add = 0;
    for (j = 6, nTrig=0; j < dsz;nTrig++)
    {
        if (nTrig>1)
            j++;
        
        if ((dsz-j)<ndat)
            break;
        
        for (i = 0; i < ndat && j<dsz; i++, j++)
        {
            dt = (data[j-3] & 0x7ff00000) >> 20; // Decode lines 21-28;
            down = data[j] & 0x3ff;
            up = (data[j] & 0xffc00) >> 10;
            if (i == 0)
            {
                add = dt;
                if (add >= 0&& add < Ntot())
                {
                    out[ add ] = down;
                    chan[add] = out[add] - ped[add];
                    add_hit(add);
                    if (romode == 4)
                    {
                        if (Noise()[add] == 0.)
                            Noise()[add] = 0.2886751; // 1./sqrt(12.)
                    }
                }
            }
            else if (i < ndat)
            {
                aux = add + i;
                if (aux >= 0&& aux < Ntot() )
                {
                    out[aux] = down;
                    list_of_channels.push_back(aux);
                }

                // TODO: we ignore by now the UP channels 
                //      aux = add-i;
                //      if ( aux >=0 && aux<Ntot() ) {
                //         out[aux]=up;
                //         list_of_channels.push_back(aux);
                //      }
            }
        }
    }
}
void VATAmodule_analysis::decode_sparse_gp5(unsigned int *data, std::vector<int> &list_of_channels)
{
    // TODO: We assume GP5 sparse is like GP3
    decode_sparse_gp3(data, list_of_channels);
}

//---------------------------------------------------------------------
// Decose sparse data
//---------------------------------------------------------------------
void VATAmodule_analysis::decode_sparse_gp7(unsigned int *data,
                                            std::vector<int> &list_of_channels)
{
    double *out = Event();
    double *chan = Chan();
    double *ped = Ped();
    int last_chip;
    // We assume the Analysis::Fill has made all teh checks on
    // the data integrity

    // Another asusmption: We already have the pedestals
    int dsz = (data[0] & 0xfff)-1;
    int romode = (data[2] & 0xf000) >> 12;
    int nadj = romode == 8 ? (data[1] & 0xff000000) >> 24 : 0;
    int i, j, sz, add, aux;

    // Clear output array
    for (i = 0; i < Ntot(); i++)
        chan[i] = out[i] = 0.;

    list_of_channels.clear();

    // Get the size and decode
    unsigned int dt;
    sz = data[2] & 0xfff;

    int nblk = 2 + nadj; // size of a channel block
    int ndat = nblk - 1; // size of sensible data in a channel block
    int nTrig, up, down;

    //block size is smaller for pure sparse data
    if (nadj==0)
        nblk=ndat=nadj+1;

    last_chip = -1;
    add = 0;

    for (j = 6, nTrig=0; j < dsz; nTrig++)
    {
        if (nTrig>1)
            j++;

        if ((dsz-j)<ndat)
            break;

        ndat = nblk;
        for (i = 0; i < ndat && j<dsz; i++, j++)
        {
            dt = (data[j-3] & 0x7ff00000) >> 20; // Decode lines 21-28;
            down = data[j] & 0x3ff;
            up = (data[j] & 0xffc00) >> 10;
            if (i == 0)
            {
                add = dt;
                if (add >= 0 && add < Ntot())
                {
                    out[ add ] = down;
                    chan[add] = out[add] - ped[add];
                    add_hit(add);
                    if (romode == 4)
                    {
                        if (Noise()[add] == 0.)
                            Noise()[add] = 0.2886751; // 1./sqrt(12.)
                    }
                    ndat--;
                }
            }
            else if (i < ndat)
            {
                aux = add + i;
                if (aux >= 0&& aux < Ntot() )
                {
                    out[aux] = down;
                    list_of_channels.push_back(aux);
                }

                // TODO: we ignore by now the UP channels 
                //      aux = add-i;
                //      if ( aux >=0 && aux<Ntot() ) {
                //         out[aux]=up;
                //         list_of_channels.push_back(aux);
                //      }
            }
        }
    }
}

