#include <iostream>
#include <string>
#include <vector>
#include <data/DataFormat.h>
#include <daq/VMEmanager.h>
#include "FPGACaenV1495.h"
#include "FPGACaenV1495_analysis.h"

#include <cmath>

FPGACaenV1495::FPGACaenV1495( const DAQpp::DAQid &id, unsigned long address)
    : GenModule(id, address)
    , _hold(false)
    , _waiting(false)
    , _paused(false)
{
    int nchip = 16; // in this case nchip represents possible coincidences
    int nchan = 32; // nchan represents possible triggers
    // reset pedestals and noise to sensible values
    // pedestals and noise do not have meaning in this context
    fill_n(back_inserter( get_par_value< std::vector<double> > ("pedestal") ), nchan*nchip, 0.);
    fill_n(back_inserter( get_par_value< std::vector<double> > ("noise") ), nchan*nchip, 0.);

    // Using DAQ++ parameters can be useful, especially if you want to
    // access them the DAQcorba
    add_par("gatewidth", new DAQpp::Par<unsigned int>(200));
    add_par("glevel", new DAQpp::Par<unsigned int>(0));
    add_par("amask", new DAQpp::Par<unsigned int>(0xFFFFFFFF));
    add_par("bmask", new DAQpp::Par<unsigned int>(0xFFFFFFFF));
    add_par("cmask", new DAQpp::Par<unsigned int>(0xFFFFFFFF));
    // custom fpga design parameters
    add_par("si_gate", new DAQpp::Par<unsigned int>(40));
    add_par("si_reset_delay", new DAQpp::Par<unsigned int>(40));
    add_par("wait_delay", new DAQpp::Par<double>(1));
    add_par("flood_mode", new DAQpp::Par<bool>(false));
    add_par("coinc_mask", new DAQpp::Par<unsigned int>(0xffff));

    //default width/delay pairs for silicon; width=128*5 ns=640 ns, delay=0
    add_par("si_delay0", new DAQpp::Par<unsigned int>(0x8000));
    add_par("si_delay1", new DAQpp::Par<unsigned int>(0x8000));
    add_par("si_delay2", new DAQpp::Par<unsigned int>(0x8000));
    add_par("si_delay3", new DAQpp::Par<unsigned int>(0x8000));

    //delay as rrss where ss is the number of clocks that pass between
    //the trigger and bgo_window start and rr is number of clocks
    //after >the trigger< when bgo_window is reset. So rr should be
    //bigger than ss and actual width is rr-ss. Two windows are
    //generated - the short one is intended for bgo-bgo coincidences
    //and the long one is for bgo-si; e.g for the short one
    //width=52*5-48*5=20 ns, delay 48*5=240 ns
    add_par("bgo_delay_short", new DAQpp::Par<unsigned int>(0x3430));
    add_par("bgo_delay_long", new DAQpp::Par<unsigned int>(0x3410));

    //block delay - block new triggers in module for this many clocks
    add_par("bgo_block_delay", new DAQpp::Par<unsigned int>(0x0100));
    // now we should create the Analysis for the monitoring
    // never forget to do this !!!
    this->ana = new FPGACaenV1495_analysis(nchip, nchan, get_id());

    _coincmap = new CoincMap();
    _triggermap = new TriggerMap();
}

int 
FPGACaenV1495::Trigger()
{
    return 0;
}

bool
FPGACaenV1495::hasData()
{
    if ( _paused ) return false;
    if ( _waiting ) stop_waiting();
    DAQpp::GTimer tim;
    double delay = get_par_value<double>("wait_delay");
    tim.reset();
    tim.start();
    
    unsigned short data;
    do {
      data = VMEmanager::VME()->read16(_add | V1495_CUSTOM_STATUS);
      if ( tim() > delay ) break;
      
    } 
    while ( !data && _hold );

    return (data & 0x0001);
}

char *FPGACaenV1495::getData(int &size)
{
    /*
      Get the data
      The format is the following
      Word
      0     0xCAFEibbb   i= event type bbb= size of data block
      1     0xaaaaaaaa   aaaaaaaaaa event counter
      2     0xiiitnnnn   iii=module id t=data type nnnn=data size
      3     data word 0
      4     data word 1
      .     .
      bbbb    bbbb         size of data block

      In practice, some bits of words 1 and 2 can be used. Note,
      however, that the module id should go where it is.

      the size in word 2 is the number of data words folowing this
      word. 
      
      Also, the event type byte will allow us to identify our data block
      when reading the file. Look into DataFormat.h and choose one
      different from the existing ones:
         NormalData = 0x10000,
         PulseData  = 0x20000,
         ScanData   = 0x30000,
         ADCdata    = 0x40000,
         CSPRINTdata= 0x80000,
         FBTDCdata  = 0x90000
    */

    increase_evt_counter();

    // Allocate space for the data. This has to be done with the
    // allocate method of DAQpp::Module
    //
    // Note that here we assume that we have one word per channel. It
    // may well happen that we get from the module a complicated
    // structure that we do not want ot decode here, so we just dumpt
    // it into the header and the trailer words and leave the analysis
    // program with the burden of decoding.

    int nRead16=71;
    int ntot = (nRead16+1)/2; // read out all data every time (138 x 16 bit reads)
    size     = 4   // the 3 words in the header plus the trailing word
             + ntot;

    unsigned int *data =
        reinterpret_cast<unsigned int *>(allocate(sizeof(unsigned int) * size));


    // Fill the data array
    int iw = 0;
    data[iw++] = DataHeader | DataRecord | (size);
    data[iw++] = get_evt_counter();
    data[iw++] = ((int)get_id() & 0xfff) << 20 | get_data_type() | (ntot);
    data[size - 1] = size;

    // read the data via VME
    std::vector<VMEcommand> stack;
    std::vector<unsigned int> vdata;
    stack.reserve(nRead16);
    vdata.reserve(nRead16);

    unsigned long long_add;
    //read the trigger register; bit (i) indicates whether channel i was triggered in this event
    //long_add = _add + V1495_CUSTOM_ATRIGREG_L; // H is comes just after, so we increase add by 2 to get it
    //for ( int i=0; i<2;   i++ ) stack.push_back(VMEcommand(VMEcommand::read16,long_add + 2*i));

    long_add = _add + V1495_CUSTOM_COUNTERREG_0;
    for ( int i=0; i<4;   i++ ) stack.push_back(VMEcommand(VMEcommand::read16,long_add + 2*i));
    long_add = _add + V1495_CUSTOM_A0TIME_L;
    //128 is for both A and B (could use 64 if desired since B is not used)
    for ( int i=0; i<64; i++ ) stack.push_back(VMEcommand(VMEcommand::read16,long_add + 2*i));
    //from 20100730 we just have 16 bit coincreg
    //L1 <=> A0; L2 <=> A1; R1<=>A16; R2 <=> A17; BGOL=or(A2..A15); BGOR=or(A18..A31)
    //bits from 0 to 3 are silicon-silicon (0 L1 R1, 1 L1 R2, 2 L2 R1 3 L2 R2 )
    //from 4 to 7 are Si-BGO ( 4 L1 BGOR 5 L2 BGOR 6 R1 BGOL 7 R2 BGOL)
    //bit 8 (or bit 0 of COINC_REG_1) is BGO-BGO

    //read the trigger register
    long_add = _add + V1495_CUSTOM_ATRIGREG_L; // H is comes just after, so we increase add by 2 to get it
    for ( int i=0; i<2;   i++ ) stack.push_back(VMEcommand(VMEcommand::read16,long_add + 2*i));

    long_add = _add + V1495_CUSTOM_COINCREG_0;
    for ( int i=0; i<1;   i++ ) stack.push_back(VMEcommand(VMEcommand::read16,long_add + 2*i));



    int check = VMEmanager::VME()->stack_execute(stack,vdata);
    if (check<0 || vdata.size()!=(unsigned int)nRead16)
      for (int i=0;i<ntot;i++) data[iw++] = 0;
    else{
      for (int i=0;i<ntot-1;i++) data[iw++] = (vdata[2*i]&0xffff) | ((vdata[2*i+1]&0xffff)<<16);
      data[iw++]=vdata[nRead16-1]& 0xffff;
    }
    // print the data for debugging
    if (false) {
      std::cout << std::hex;
      std::cout << "\nPrinting FPGACaenV1495 data:\n";
      for (int i=0;i<size;i++) std::cout << data[i] << "\n";
      std::cout << std::dec << std::endl;
    }
    // enable the module to trigger again
    if ( _hold ) wait();
    else coinc_reset();

    size *= sizeof(unsigned int);
    return reinterpret_cast<char *>(data);
}

int
FPGACaenV1495::PrepareForRun ()
{
    std::cout << "\t+-> FPGACaenV1495.PrepareForRun [" << get_id() << "]"
    << std::endl;
    Reset();
    return 0;
}

int
FPGACaenV1495::Start (int)
{
    std::cout << "\t+-> FPGACaenV1495.Start         [" << get_id() << "]"
    << std::endl;
    if ( get_analysis() )
        get_analysis()->init_run();

    stop_waiting();
    return 0;
}

int
FPGACaenV1495::Stop (int)
{
    std::cout << "\t+-> FPGACaenV1495.Stop          [" << get_id() << "]"
    << std::endl;
    wait();

    if ( get_analysis() )
        get_analysis()->end_run();

    return 0;
}

int
FPGACaenV1495::Pause ()
{
    std::cout << "\t+-> FPGACaenV1495.Pause         [" << get_id() << "]"
    << std::endl;
    wait();
    _paused = true;
    return 0;
}

int
FPGACaenV1495::Continue ()
{
    std::cout << "\t+-> FPGACaenV1495.Continue      [" << get_id() << "]"
    << std::endl;
    stop_waiting();
    _paused = false;
    return 0;
}

int
FPGACaenV1495::Reset ()
{
    std::cout << "\t+-> FPGACaenV1495.Reset         [" << get_id() << "]"
    << std::endl;
    VMEmanager::VME()->write16(_add | V1495_RESET, 1);
    init();
    return 0;
}

int
FPGACaenV1495::GetReady ()
{
    std::cout << "\t+-> FPGACaenV1495.GetReady      [" << get_id() << "]"
    << std::endl;
    return 0;
}

void
FPGACaenV1495::init()
{
    config();
    wait();
}

void
FPGACaenV1495::config()
{
    std::vector<VMEcommand> stack;
    std::vector<unsigned int> vdata;
    // apply the A mask
    unsigned long add = _add | V1495_COIN_REF_A_MASK_L;
    unsigned int data = get_par_value<unsigned int>("amask");
    stack.push_back(VMEcommand(VMEcommand::write16,add,data&0xffff));
    stack.push_back(VMEcommand(VMEcommand::write16,add+2,(data>>16)&0xffff));
    // apply the B mask
    add  = _add | V1495_COIN_REF_B_MASK_L;
    data = get_par_value<unsigned int>("bmask");
    stack.push_back(VMEcommand(VMEcommand::write16,add,data&0xffff));
    stack.push_back(VMEcommand(VMEcommand::write16,add+2,(data>>16)&0xffff));
    // apply the C mask
    add  = _add | V1495_COIN_REF_C_MASK_L;
    data = get_par_value<unsigned int>("cmask");
    stack.push_back(VMEcommand(VMEcommand::write16,add,data&0xffff));
    stack.push_back(VMEcommand(VMEcommand::write16,add+2,(data>>16)&0xffff));
    // set the gatewidth
    add  = _add | V1495_COIN_REF_GATEWIDTH;
    data = get_par_value<unsigned int>("gatewidth");
    stack.push_back(VMEcommand(VMEcommand::write16,add,data&0xffff));
    // set the G output level
    add  = _add | V1495_COIN_REF_G_CONTROL;
    data = get_par_value<unsigned int>("glevel");
    stack.push_back(VMEcommand(VMEcommand::write16,add,data&0x1));
    // set the si gatewidth
    add  = _add | V1495_CUSTOM_SIGATEWIDTH;
    data = get_par_value<unsigned int>("si_gate");
    stack.push_back(VMEcommand(VMEcommand::write16,add,data&0xffff));
    // set the si reset delay
    add  = _add | V1495_CUSTOM_SIRESDELAY;
    data = get_par_value<unsigned int>("si_reset_delay");
    stack.push_back(VMEcommand(VMEcommand::write16,add,data&0xffff));

    add  = _add | V1495_CUSTOM_CONTROL;
    data=0;
    if (get_par_value<bool>("flood_mode")) data=0x1;
    stack.push_back(VMEcommand(VMEcommand::write16,add,data));
    std::cout<<"Setting mode to "<<data<<" (normal=0, flood=1)"<<std::endl;
    
    add = _add | V1495_CUSTOM_COINCMASK_0 ;
    data= get_par_value<unsigned int>("coinc_mask");
    stack.push_back(VMEcommand(VMEcommand::write16,add,data&0xffff));
    std::cout<<"Setting coinc_mask to "<<std::hex<<data<<std::dec<<std::endl;

    //set silicon window parameters
    add = _add | V1495_CUSTOM_SIDELAY_0;
    data= get_par_value<unsigned int>("si_delay0");
    stack.push_back(VMEcommand(VMEcommand::write16,add,data&0xffff));
    std::cout<<"Setting silicon delay to "<<std::hex<<data<<std::dec<<std::endl;

    add = _add | V1495_CUSTOM_SIDELAY_1;
    data= get_par_value<unsigned int>("si_delay1");
    stack.push_back(VMEcommand(VMEcommand::write16,add,data&0xffff));
    add = _add | V1495_CUSTOM_SIDELAY_2;
    data= get_par_value<unsigned int>("si_delay2");
    stack.push_back(VMEcommand(VMEcommand::write16,add,data&0xffff));
    add = _add | V1495_CUSTOM_SIDELAY_3;
    data= get_par_value<unsigned int>("si_delay3");
    stack.push_back(VMEcommand(VMEcommand::write16,add,data&0xffff));

    //set bgo window parameters
    add = _add | V1495_CUSTOM_BGODELAY_SHORT;
    data= get_par_value<unsigned int>("bgo_delay_short");
    stack.push_back(VMEcommand(VMEcommand::write16,add,data&0xffff));
    add = _add | V1495_CUSTOM_BGODELAY_LONG;
    data= get_par_value<unsigned int>("bgo_delay_long");
    stack.push_back(VMEcommand(VMEcommand::write16,add,data&0xffff));

    //send bgo block duration
    add = _add | V1495_CUSTOM_BGOBLOCKDELAY;
    data= get_par_value<unsigned int>("bgo_block_delay");
    stack.push_back(VMEcommand(VMEcommand::write16,add,data&0xffff));

    VMEmanager::VME()->stack_execute(stack,vdata);
}

void
FPGACaenV1495::wait()
{
    VMEmanager::VME()->write16(_add | V1495_CUSTOM_COINCRESET, 0);
    _waiting = true;
}
void
FPGACaenV1495::stop_waiting()
{
    VMEmanager::VME()->write16(_add | V1495_CUSTOM_COINCRESET, 1);
    _waiting = false;
}

void
FPGACaenV1495::coinc_reset()
{
  std::vector<VMEcommand> stack;
  std::vector<unsigned int> data;
  unsigned long add = _add | V1495_CUSTOM_COINCRESET;
  stack.push_back(VMEcommand(VMEcommand::write16,add,0));
  stack.push_back(VMEcommand(VMEcommand::write16,add,1));
  VMEmanager::VME()->stack_execute(stack,data);
}

