#ifndef __FPGACaenV1495_h__
#define __FPGACaenV1495_h__

/* This is a module for vmedaq.
 * They all should inherit from GenModule, which adds some extra
 * functionality on top of DAQpp::Module:
 *
 *   - The module works around the concept of something with a number
 *     of chips, each having the same number of channels. So nchip
 *     (number of chips) and nchan (number of channels per chip) is
 *     something we should supply somehow. In paticular for the
 *     Analysis instance of the module. That brings me to the second
 *     point.
 *
 *   - There is an Analysis pointer (you have to create the object)
 *     which will store all the monitoring stuff. This guy needs the
 *     number of chips and channels per chip.
 *
 *   - If you plan to make scans with that module, you will need to
 *     override the set_scan_point method to set your module in a
 *     reasonable state every time you are in a new scan point.
 *
 *   - pedestals and noise are there as DAQ++ parameters
 *      + can be updated from the values in Analysis or copied into
 *        the Analysis
 *
 *      + Pedestals are stored in the data file if you override the
 *      write_pedestals method.
 *
 *
 * See the DAQ++ documentation for further information
 */
#include <map>

#include <daq/GenModule.h>
#include <data/DataFormat.h>


class FPGACaenV1495 : public GenModule
{
    typedef std::pair<int,int> CoincMapEntry;
    typedef std::map<int,CoincMapEntry> CoincMap;
    typedef std::vector<int> TriggerArray;
    typedef std::map<int,TriggerArray> TriggerMap;

    public:

        /* Constructor
         * The constructor needs 2 parameters:
         * id: the Module id
         * address: the address (VME, or whatever protocol) to
         *          communicate with the module
         *
         * Note: although DAQ++ allows for any type of identifier,
         * however VMEDAQ only handles properly integer IDs. The only
         * reason for that is that it needs integer values to save them
         * properly in the data file
         */
        FPGACaenV1495(const DAQpp::DAQid &id, unsigned long address);
        ~FPGACaenV1495()
        {
            delete _coincmap;
            delete _triggermap;
        }

        /* This is the method called by the RunManager to know if there
         * is data ready in the module. If the answer if true, then it
         * will try to read the module. If there is no way to know a
         * priory about the availavility of data, one can defer the
         * decision until getData
         */
        bool hasData();

        /* This is the method called to retrieve the data. If hasData
         * could not figure out the right answer, one can still say here
         * that there is no data at all by returning a NULL pointer and
         * setting size to 0.
         *
         * The memory to store the data should be allocated with the
         * allocate method. See the DAQ++ documentation...
         */
        char *getData(int &size);

        int PrepareForRun();

        int Start(int = 0);

        int Stop(int = 0);

        int Pause();

        int Continue();

        int Reset();

        int GetReady();

        int Trigger();

        void flush()
        {
        }
        
        /* This returns the data type. The existing data types
         * can be found in DataFormat.h It is a number constructed
         * like: (n & 0xF)<<16;
         */
        unsigned long get_data_type() const
        {
           return FPGACaenData;
        }

        bool hold() { return _hold; }
        void hold(bool x) { _hold = x; }

        void coincmap_add(int i, int j, int k)
        {
           _coincmap->insert(std::pair<int,CoincMapEntry>(i,CoincMapEntry(j,k)));
        }

        std::pair<int,int> coincmap_find(int i)
        {
           CoincMap::iterator it = _coincmap->find(i);
           if (it == _coincmap->end()) return std::pair<int,int>(-1,-1);
           else return it->second;
        }

        void triggermap_add(int i, std::vector<int> array)
        {
           _triggermap->insert(std::pair<int,TriggerArray>(i,array));
        }

        std::vector<int> triggermap_find(int i)
        {
           TriggerMap::iterator it = _triggermap->find(i);
           if (it == _triggermap->end()) return std::vector<int>();
           else return it->second;
        }

    private:
        bool _hold,_waiting,_paused;

        // reset coincidence logic after reading out data
        void coinc_reset();
        void wait();
        void stop_waiting();
        void init();
        void config();

        CoincMap *_coincmap;

        TriggerMap *_triggermap;

        enum Registers
        {
            // CAEN model v1495 FPGA
            // VME Register Address Map
            // A24/A32 Addressing

            // USER FPGA Access 0x0000 to 0x7FFC
            V1495_CONTROL_REGISTER  = 0x8000, // D16 R/W
            V1495_STATUS_REGISTER   = 0x8002, // D16 R 
            V1495_INTERRUPT_LEVEL   = 0x8004, // D16 R/W
            V1495_INTERRUPT_STATUS  = 0x8006, // D16 R/W
            V1495_GEO_ADDRESS       = 0x8008, // D16 R
            V1495_RESET             = 0x800A, // D16 W
            V1495_FIRMWARE_REVISION = 0x800C, // D16 R
            V1495_VME_FPGA_SELECT   = 0x800E, // D16 R/W
            V1495_VME_FPGA_MEMORY   = 0x8010, // D16 R/W
            V1495_USER_FPGA_SELECT  = 0x8012, // D16 R/W
            V1495_USER_FPGA_MEMORY  = 0x8014, // D16 R/W
            V1495_USER_FPGA_CONFIG  = 0x8016, // D16 R/W
            V1495_SCRATCH16         = 0x8018, // D16 R/W
            V1495_SCRATCH32         = 0x8020, // D32 R/W

            // Configuration ROM 0x8100 to 0x801FE

            // CAEN model v1495 FPGA
            // COIN Reference register map
            // A24/A32 Addressing

            // USER FPGA Access 0x0000 to 0x7FFC
            // as defined by CAEN supplied COIN reference user code
            V1495_COIN_REF_A_STATUS_L  = 0x0000, // D16 R
            V1495_COIN_REF_A_STATUS_H  = 0x0002, // D16 R
            V1495_COIN_REF_B_STATUS_L  = 0x0004, // D16 R
            V1495_COIN_REF_B_STATUS_H  = 0x0006, // D16 R
            V1495_COIN_REF_C_STATUS_L  = 0x0008, // D16 R
            V1495_COIN_REF_C_STATUS_H  = 0x000A, // D16 R
            V1495_COIN_REF_A_MASK_L    = 0x000C, // D16 W
            V1495_COIN_REF_A_MASK_H    = 0x000E, // D16 W
            V1495_COIN_REF_B_MASK_L    = 0x0010, // D16 W
            V1495_COIN_REF_B_MASK_H    = 0x0012, // D16 W
            V1495_COIN_REF_C_MASK_L    = 0x0014, // D16 W
            V1495_COIN_REF_C_MASK_H    = 0x0016, // D16 W
            V1495_COIN_REF_GATEWIDTH   = 0x0018, // D16 W
	    V1495_COIN_REF_C_CONTROL_L = 0x001A, // D16 W
            V1495_COIN_REF_C_CONTROL_H = 0x001C, // D16 W
            V1495_COIN_REF_MODE        = 0x001E, // D16 W
            V1495_COIN_REF_SCRATCH     = 0x0020, // D16 R/W
            V1495_COIN_REF_G_CONTROL   = 0x0022, // D16 W
            V1495_COIN_REF_D_CONTROL_L = 0x0024, // D16 R/W
            V1495_COIN_REF_D_CONTROL_H = 0x0026, // D16 R/W
            V1495_COIN_REF_D_DATA_L    = 0x0028, // D16 R/W
            V1495_COIN_REF_D_DATA_H    = 0x002A, // D16 R/W
            V1495_COIN_REF_E_CONTROL_L = 0x002C, // D16 R/W
            V1495_COIN_REF_E_CONTROL_H = 0x002E, // D16 R/W
            V1495_COIN_REF_E_DATA_L    = 0x0030, // D16 R/W
            V1495_COIN_REF_E_DATA_H    = 0x0032, // D16 R/W
            V1495_COIN_REF_F_CONTROL_L = 0x0034, // D16 R/W
            V1495_COIN_REF_F_CONTROL_H = 0x0036, // D16 R/W
            V1495_COIN_REF_F_DATA_L    = 0x0038, // D16 R/W
            V1495_COIN_REF_F_DATA_H    = 0x003A, // D16 R/W
            V1495_COIN_REF_REVISION    = 0x003C, // D16 R/W
            V1495_COIN_REF_PDL_CONTROL = 0x003E, // D16 R/W
            V1495_COIN_REF_PDL_DATA    = 0x0040, // D16 R/W
            V1495_COIN_REF_D_IDCODE    = 0x0042, // D16 R
            V1495_COIN_REF_E_IDCODE    = 0x0044, // D16 R
            V1495_COIN_REF_F_IDCODE    = 0x0046, // D16 R

            // CAEN model v1495 FPGA
            // Compton PET custom COIN Reference register map
            // A24/A32 Addressing

            // USER FPGA Access 0x0000 to 0x7FFC
            // as defined by modified COIN reference user code
            V1495_CUSTOM_REVISION     = 0x0100, // D16 R
            V1495_CUSTOM_SCRATCH      = 0x0102, // D16 R/W
            V1495_CUSTOM_CONTROL      = 0x0104, // D16 W
            V1495_CUSTOM_STATUS       = 0x0106, // D16 R
            V1495_CUSTOM_COINCRESET   = 0x0108, // D16 W
            V1495_CUSTOM_SIGATEWIDTH  = 0x010A, // D16 W
            V1495_CUSTOM_SIRESDELAY   = 0x010C, // D16 W
	    
	    //dump of trigger bits
	    V1495_CUSTOM_ATRIGREG_L   = 0x0110,  //D16 R
	    V1495_CUSTOM_ATRIGREG_H   = 0x0112,  //D16 R
	    
            V1495_CUSTOM_COINCREG_0   = 0x0120, // D16 R
//             V1495_CUSTOM_COINCREG_1   = 0x0122, // D16 R
//             V1495_CUSTOM_COINCREG_2   = 0x0124, // D16 R
//             V1495_CUSTOM_COINCREG_3   = 0x0126, // D16 R
//             V1495_CUSTOM_COINCREG_4   = 0x0128, // D16 R
//             V1495_CUSTOM_COINCREG_5   = 0x012A, // D16 R
	    V1495_CUSTOM_COINCMASK_0  = 0x0160, //D16 W
	    //each coincreg should have corresponding coinc mask
            V1495_CUSTOM_COUNTERREG_0 = 0x01F8, // D16 R
            V1495_CUSTOM_COUNTERREG_1 = 0x01FA, // D16 R
            V1495_CUSTOM_COUNTERREG_2 = 0x01FC, // D16 R
            V1495_CUSTOM_COUNTERREG_3 = 0x01FE, // D16 R

	    //delay/width pairs coded as 0xwwdd (ww are 8 bits for width, dd are 8 bits for delay)
	    V1495_CUSTOM_SIDELAY_0    = 0x0300,
	    V1495_CUSTOM_SIDELAY_1    = 0x0302,
	    V1495_CUSTOM_SIDELAY_2    = 0x0304,
	    V1495_CUSTOM_SIDELAY_3    = 0x0306,
	    V1495_CUSTOM_BGODELAY_SHORT= 0x0308,
	    V1495_CUSTOM_BGODELAY_LONG= 0x0310,
	    V1495_CUSTOM_BGOBLOCKDELAY= 0x0312,

            // Time stamp registers
            V1495_CUSTOM_A0TIME_L     = 0x0200, // D16 R
            V1495_CUSTOM_A0TIME_H     = 0x0202, // D16 R
            V1495_CUSTOM_A1TIME_L     = 0x0204, // D16 R
            V1495_CUSTOM_A1TIME_H     = 0x0206, // D16 R
            V1495_CUSTOM_A2TIME_L     = 0x0208, // D16 R
            V1495_CUSTOM_A2TIME_H     = 0x020A, // D16 R
            V1495_CUSTOM_A3TIME_L     = 0x020C, // D16 R
            V1495_CUSTOM_A3TIME_H     = 0x020E, // D16 R
            V1495_CUSTOM_A4TIME_L     = 0x0210, // D16 R
            V1495_CUSTOM_A4TIME_H     = 0x0212, // D16 R
            V1495_CUSTOM_A5TIME_L     = 0x0214, // D16 R
            V1495_CUSTOM_A5TIME_H     = 0x0216, // D16 R
            V1495_CUSTOM_A6TIME_L     = 0x0218, // D16 R
            V1495_CUSTOM_A6TIME_H     = 0x021A, // D16 R
            V1495_CUSTOM_A7TIME_L     = 0x021C, // D16 R
            V1495_CUSTOM_A7TIME_H     = 0x021E, // D16 R
            V1495_CUSTOM_A8TIME_L     = 0x0220, // D16 R
            V1495_CUSTOM_A8TIME_H     = 0x0222, // D16 R
            V1495_CUSTOM_A9TIME_L     = 0x0224, // D16 R
            V1495_CUSTOM_A9TIME_H     = 0x0226, // D16 R
            V1495_CUSTOM_A10TIME_L    = 0x0228, // D16 R
            V1495_CUSTOM_A10TIME_H    = 0x022A, // D16 R
            V1495_CUSTOM_A11TIME_L    = 0x022C, // D16 R
            V1495_CUSTOM_A11TIME_H    = 0x022E, // D16 R
            V1495_CUSTOM_A12TIME_L    = 0x0230, // D16 R
            V1495_CUSTOM_A12TIME_H    = 0x0232, // D16 R
            V1495_CUSTOM_A13TIME_L    = 0x0234, // D16 R
            V1495_CUSTOM_A13TIME_H    = 0x0236, // D16 R
            V1495_CUSTOM_A14TIME_L    = 0x0238, // D16 R
            V1495_CUSTOM_A14TIME_H    = 0x023A, // D16 R
            V1495_CUSTOM_A15TIME_L    = 0x023C, // D16 R
            V1495_CUSTOM_A15TIME_H    = 0x023E, // D16 R
            V1495_CUSTOM_A16TIME_L    = 0x0240, // D16 R
            V1495_CUSTOM_A16TIME_H    = 0x0242, // D16 R
            V1495_CUSTOM_A17TIME_L    = 0x0244, // D16 R
            V1495_CUSTOM_A17TIME_H    = 0x0246, // D16 R
            V1495_CUSTOM_A18TIME_L    = 0x0248, // D16 R
            V1495_CUSTOM_A18TIME_H    = 0x024A, // D16 R
            V1495_CUSTOM_A19TIME_L    = 0x024C, // D16 R
            V1495_CUSTOM_A19TIME_H    = 0x024E, // D16 R
            V1495_CUSTOM_A20TIME_L    = 0x0250, // D16 R
            V1495_CUSTOM_A20TIME_H    = 0x0252, // D16 R
            V1495_CUSTOM_A21TIME_L    = 0x0254, // D16 R
            V1495_CUSTOM_A21TIME_H    = 0x0256, // D16 R
            V1495_CUSTOM_A22TIME_L    = 0x0258, // D16 R
            V1495_CUSTOM_A22TIME_H    = 0x025A, // D16 R
            V1495_CUSTOM_A23TIME_L    = 0x025C, // D16 R
            V1495_CUSTOM_A23TIME_H    = 0x025E, // D16 R
            V1495_CUSTOM_A24TIME_L    = 0x0260, // D16 R
            V1495_CUSTOM_A24TIME_H    = 0x0262, // D16 R
            V1495_CUSTOM_A25TIME_L    = 0x0264, // D16 R
            V1495_CUSTOM_A25TIME_H    = 0x0266, // D16 R
            V1495_CUSTOM_A26TIME_L    = 0x0268, // D16 R
            V1495_CUSTOM_A26TIME_H    = 0x026A, // D16 R
            V1495_CUSTOM_A27TIME_L    = 0x026C, // D16 R
            V1495_CUSTOM_A27TIME_H    = 0x026E, // D16 R
            V1495_CUSTOM_A28TIME_L    = 0x0270, // D16 R
            V1495_CUSTOM_A28TIME_H    = 0x0272, // D16 R
            V1495_CUSTOM_A29TIME_L    = 0x0274, // D16 R
            V1495_CUSTOM_A29TIME_H    = 0x0276, // D16 R
            V1495_CUSTOM_A30TIME_L    = 0x0278, // D16 R
            V1495_CUSTOM_A30TIME_H    = 0x027A, // D16 R
            V1495_CUSTOM_A31TIME_L    = 0x027C, // D16 R
            V1495_CUSTOM_A31TIME_H    = 0x027E, // D16 R

            V1495_CUSTOM_B0TIME_L     = 0x0280, // D16 R
            V1495_CUSTOM_B0TIME_H     = 0x0282, // D16 R
            V1495_CUSTOM_B1TIME_L     = 0x0284, // D16 R
            V1495_CUSTOM_B1TIME_H     = 0x0286, // D16 R
            V1495_CUSTOM_B2TIME_L     = 0x0288, // D16 R
            V1495_CUSTOM_B2TIME_H     = 0x028A, // D16 R
            V1495_CUSTOM_B3TIME_L     = 0x028C, // D16 R
            V1495_CUSTOM_B3TIME_H     = 0x028E, // D16 R
            V1495_CUSTOM_B4TIME_L     = 0x0290, // D16 R
            V1495_CUSTOM_B4TIME_H     = 0x0292, // D16 R
            V1495_CUSTOM_B5TIME_L     = 0x0294, // D16 R
            V1495_CUSTOM_B5TIME_H     = 0x0296, // D16 R
            V1495_CUSTOM_B6TIME_L     = 0x0298, // D16 R
            V1495_CUSTOM_B6TIME_H     = 0x029A, // D16 R
            V1495_CUSTOM_B7TIME_L     = 0x029C, // D16 R
            V1495_CUSTOM_B7TIME_H     = 0x029E, // D16 R
            V1495_CUSTOM_B8TIME_L     = 0x02A0, // D16 R
            V1495_CUSTOM_B8TIME_H     = 0x02A2, // D16 R
            V1495_CUSTOM_B9TIME_L     = 0x02A4, // D16 R
            V1495_CUSTOM_B9TIME_H     = 0x02A6, // D16 R
            V1495_CUSTOM_B10TIME_L    = 0x02A8, // D16 R
            V1495_CUSTOM_B10TIME_H    = 0x02AA, // D16 R
            V1495_CUSTOM_B11TIME_L    = 0x02AC, // D16 R
            V1495_CUSTOM_B11TIME_H    = 0x02AE, // D16 R
            V1495_CUSTOM_B12TIME_L    = 0x02B0, // D16 R
            V1495_CUSTOM_B12TIME_H    = 0x02B2, // D16 R
            V1495_CUSTOM_B13TIME_L    = 0x02B4, // D16 R
            V1495_CUSTOM_B13TIME_H    = 0x02B6, // D16 R
            V1495_CUSTOM_B14TIME_L    = 0x02B8, // D16 R
            V1495_CUSTOM_B14TIME_H    = 0x02BA, // D16 R
            V1495_CUSTOM_B15TIME_L    = 0x02BC, // D16 R
            V1495_CUSTOM_B15TIME_H    = 0x02BE, // D16 R
            V1495_CUSTOM_B16TIME_L    = 0x02C0, // D16 R
            V1495_CUSTOM_B16TIME_H    = 0x02C2, // D16 R
            V1495_CUSTOM_B17TIME_L    = 0x02C4, // D16 R
            V1495_CUSTOM_B17TIME_H    = 0x02C6, // D16 R
            V1495_CUSTOM_B18TIME_L    = 0x02C8, // D16 R
            V1495_CUSTOM_B18TIME_H    = 0x02CA, // D16 R
            V1495_CUSTOM_B19TIME_L    = 0x02CC, // D16 R
            V1495_CUSTOM_B19TIME_H    = 0x02CE, // D16 R
            V1495_CUSTOM_B20TIME_L    = 0x02D0, // D16 R
            V1495_CUSTOM_B20TIME_H    = 0x02D2, // D16 R
            V1495_CUSTOM_B21TIME_L    = 0x02D4, // D16 R
            V1495_CUSTOM_B21TIME_H    = 0x02D6, // D16 R
            V1495_CUSTOM_B22TIME_L    = 0x02D8, // D16 R
            V1495_CUSTOM_B22TIME_H    = 0x02DA, // D16 R
            V1495_CUSTOM_B23TIME_L    = 0x02DC, // D16 R
            V1495_CUSTOM_B23TIME_H    = 0x02DE, // D16 R
            V1495_CUSTOM_B24TIME_L    = 0x02E0, // D16 R
            V1495_CUSTOM_B24TIME_H    = 0x02E2, // D16 R
            V1495_CUSTOM_B25TIME_L    = 0x02E4, // D16 R
            V1495_CUSTOM_B25TIME_H    = 0x02E6, // D16 R
            V1495_CUSTOM_B26TIME_L    = 0x02E8, // D16 R
            V1495_CUSTOM_B26TIME_H    = 0x02EA, // D16 R
            V1495_CUSTOM_B27TIME_L    = 0x02EC, // D16 R
            V1495_CUSTOM_B27TIME_H    = 0x02EE, // D16 R
            V1495_CUSTOM_B28TIME_L    = 0x02F0, // D16 R
            V1495_CUSTOM_B28TIME_H    = 0x02F2, // D16 R
            V1495_CUSTOM_B29TIME_L    = 0x02F4, // D16 R
            V1495_CUSTOM_B29TIME_H    = 0x02F6, // D16 R
            V1495_CUSTOM_B30TIME_L    = 0x02F8, // D16 R
            V1495_CUSTOM_B30TIME_H    = 0x02FA, // D16 R
            V1495_CUSTOM_B31TIME_L    = 0x02FC, // D16 R
            V1495_CUSTOM_B31TIME_H    = 0x02FE, // D16 R
        };
};

#endif
