#include <iostream>
#include <iomanip>
#include <config/ConfigParser.h>
#include <config/ModuleParser.h>
#include <config/GeneralConfig.h>
#include <DAQ++/DAQmanager.h>

#include "FPGACaenV1495_analysis.h"
#include "FPGACaenV1495.h"

bool
FPGACaenV1495_analysis::data_valid(unsigned int *data) const
{
    unsigned int dsz = data[0] & 0xfff;
    unsigned int sz  = data[2] & 0xfff;

    // Standard check of data integrity
    if ( dsz != data[dsz-1] ||  dsz - sz != 4 ) return false;

    // We do not want to process twice the same event
    if ( trigger == data[1] ) return false;

    // Check that FPGACaenV1495 expected data structure appears intact
    if ( sz != 36 ) return false;

    return true;
}

void
FPGACaenV1495_analysis::Analyze()
{
  DAQpp::Module *m = DAQpp::DAQmanager::theDAQmanager()->find_module(_id);
  FPGACaenV1495 *module = reinterpret_cast<FPGACaenV1495 *>(m);
  if (!module) return;

  m_fill.acquire();
  double bgo_trig[]={1e30,1e30};
  for (int i=0;i<Ntot();i++){
    //std::cout<<i<<' '<<Chan()[i]<<" > "<<bgo_trig[i/16]<<std::endl;
    if (((i%16) >1) && (Chan()[i]<bgo_trig[i/16])){
      bgo_trig[i/16]=Chan()[i];
      //std::cout<<"Setting bgo_trig["<<i/16<<"] to "<<bgo_trig[i/16]<<std::endl;
    }
  }
  //std::cout<<"BGO: "<<bgo_trig[0]<<' '<<bgo_trig[1]<<std::endl;

  for (int i=0; i<Nchip(); i++) {
    if (_coincidence[i/32]&(1<<(i%32))) {
      std::pair<int,int> pair = module->coincmap_find(i);
      if (pair.first>=0 && pair.second>=0) {
        HitMap()[i]++;
        //double val = Chan()[pair.first]-Chan()[pair.second];
	double v0 = (pair.first == 15 ? bgo_trig[0] : Chan()[pair.first] );
	double v1 = (pair.second == 31 ? bgo_trig[1] : Chan()[pair.second] );
        double val = v1-v0;
        //std::cout << "Coincidence: " << i << " " << pair.first << " " << pair.second << " " << val << " = " << Chan()[pair.first] << " - " << Chan()[pair.second] << std::endl;
        int ib = get_bin(val);
        if ( ib >= 0 && ib < get_nbins() ) {
          Signal(i)[ib]++;
          Signal(-1)[ib]++;
          Signal2D(i)[ib]++;
          if (get_map()) {
            int ix, iy, ip;
            get_map()->pad(i, ix, iy);
            ip = ix + get_map()->nx() * iy;
            H2Map()[ ip ]++;
            if ( val >= emin && val <= emax ) H2MapC()[ ip ]++;
          }
        }
      }
    }
  }
  m_fill.release();
}

void
FPGACaenV1495_analysis::decode(unsigned int *data)
{
    // Get some important handlers of the configuration
    //ConfigParser *cfg = ConfigParser::theConfig();
    //GeneralConfig *gnrl = dynamic_cast<GeneralConfig *>(cfg->get_handler("general"));

    // In order to make a fast analysis of the data, we should provide
    // the pedestal calculation routines the list of channels to be
    // considered.
    //std::vector<int> list_of_channels;

    // We assume the Analysis::Fill has made all the checks on
    // the data integrity
    unsigned int dsz = data[0] & 0xfff;
    _coincidence[0] = data[dsz-2];
    //_coincidence[1] = data[4];
    //_coincidence[2] = data[5];
    long long int hld = data[4];
    _time_counter = data[3] | (hld << 32);

    if (false) {
      std::cout << "Printing raw data...";
      for (int i=0; i<(int)(data[0]&0xfff); i++) std::cout << "\n  " << std::hex << data[i] << std::dec;
      std::cout << std::endl;
    }
    //std::cout << "Dumping channels..." << std::endl;
    for (int i=5,j=0; j<_ntrigger; i++,j++) {
      if (data[i]==0) Chan()[j] = 0;
      else {
	double sig= double(data[i]) - double(_time_counter&0xffffffff);
	if (sig<-256 || sig>10) Chan()[j]=0;
	else Chan()[j] = sig;
      }
      Event()[j]  = Chan()[j];
      //std::cout << "   " << j << " " << Chan()[j] << std::endl;
    }

    // If we are not interested in pedestals and in finding the
    // channels with signal, ignore this...
/*  // ... We are not interested in pedestals ...
    bool find_pulse = true; // or false
    if (gnrl->get_ped_method() == GeneralConfig::normal)
    {
        // pedestals are computed in the usual way and, eventualy
        // updated. Analysis will find the channels with signal by
        // applying a simple cut on S/N
        update_pedestals(list_of_channels, find_pulse);
    }
    else
    {
        // pedestals are computed by an adaptative algorithm. One has to
        // specified if Analyis has to find the channels with signal
        update_pedestals_fast(list_of_channels, find_pulse);
    }
*/
}

void
FPGACaenV1495_analysis::init_run()
{
    std::cout << "** FPGACaenV1495_analysis::init_run\n" << std::endl;
}

void
FPGACaenV1495_analysis::end_run()
{
    std::cout << "\n**FPGACaenV1495_analysis::end_run" << std::endl;
}

