#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <iostream>
#include <cstdio>
#include <DAQ++/DAQmanager.h>
#include <DAQ++/Thread.h>
#include <daq/VMEmanager.h>
#include "corbo.h"
#include <cstdio>


static void __attribute__ ((constructor)) __init__lib__()
{
    printf("### Loading corbo_trigger...\n");
    const unsigned int ADD = 0x50000;
    VMEmanager *vme = VMEmanager::VME ();
    vme->write16 (ADD, 0xec);
    vme->write16 (ADD + 0x58, 0);
}

static void __attribute__ ((destructor)) __close__lib__()
{
    printf("### Unloading corbo_trigger...\n");
    const unsigned int ADD = 0x50000;
    VMEmanager *vme = VMEmanager::VME ();
    vme->write16 (ADD, 0xec);
    vme->write16 (ADD + 0x58, 0);
}

static void TriggerHandler(int sig);


//=====================================================================
// Trigger interrupt stuff
//=====================================================================
/*
  This is the trigger handler that DAQmanager needs.
  It should return true when a trigger is available and
  false otherwise. As it is implemented it will block
  until there is data
*/
class VMETrigger : public DAQpp::DAQTrigger
{
private:
    VMEmanager *vme;
    ROcntl corbo;
    double _timeout;
    int    _offset;
    int    evIRQ;
    int    tmIRQ;
    bool   have_to_stop;

    class Condition : public DAQpp::Condition
    {
    private:
        bool ready;
    public:
        Condition() : ready(false)
        {}
        bool condition()
        {
            return ready;
        }

        void set_ready(bool x)
        {
            ready = x;
            if (ready)
                notify_all();
        }
    };
    Condition sem;
public:
    VMETrigger();
    ~VMETrigger();
    void enableInterrupts();
    void disableInterrupts();
    bool operator()();

    unsigned long event_number()
    {
        return corbo.evtnum(0).Get();
    }
    void reset_timer()
    {
        corbo.dtim(0).Set( 0 );
    }
    void reset()
    {
        corbo.clr(0).Set(0);
    }
    void start();
    void stop();
    bool busy()
    {
        return corbo.csr(0).LclBusy();
    }
    void new_trigger()
    {
        sem.set_ready(true);
    }
    void enable_irq(int level)
    {
        corbo.evcr(0).IRQ(level).IRE(1).IRAC(1).FAC(1);
    }
    int tm_irq() const
    {
        return tmIRQ;
    }
    int evt_irq() const
    {
        return evIRQ;
    }
};

VMETrigger::VMETrigger()
    : vme(VMEmanager::VME()), _timeout(6000.), _offset( int(65536. - _timeout / 0.1)),
      evIRQ(3), tmIRQ(0), have_to_stop(false)
{
    const unsigned long ADD = 0x50000;
    corbo.SetAddress(ADD);
    // Corbo initialization
    corbo.csr(0)
        .ChnEna(0)   // Enable channel (=0 enables, =1 disables)
        .InpSel(3)   // Select internal test pulse
        .BusyMode(0) // Busy is set by the input
        .BusyOut(0)  // Ouput is local busy
        .CntSel(1)   // Count the number of busy's
        .FastClr(1)  // Allow fast clean
        .PushEna(1)  // push button resets the IRQ
        ;

    // Write test1 to assert BUSY and disable external triggers
    corbo.test(0).Set(1);

    // Write vector number of the event and timeout
    if (evIRQ > 0)
        corbo.evvr(0).Set(1 << evIRQ);

    if (tmIRQ > 0)
        corbo.tmvr(0).Set(1 << tmIRQ);

    // Set the IRQ level
    int irq;
    for (irq = 0;irq < 8;irq++)
        corbo.evcr(0).IRQ(irq).IRE(0);


    // Clear Event number
    corbo.evtnum(0).Set(0L);

    // Set the timeout delay
    corbo.dtim(0).Set( 0 );
}

VMETrigger::~VMETrigger()
{
    disableInterrupts();
}


void VMETrigger::start()
{
    // Set Channel 0 to accept triggers
    corbo.csr(0).InpSel(0);

    have_to_stop = false;

    // Clear the BUSY
    reset();
}

void VMETrigger::stop()
{
    have_to_stop = true;
    corbo.csr(0).InpSel(3);
    new_trigger();
}

void VMETrigger::enableInterrupts()
{
    int   irqmask = 0;
    if (evIRQ > 0)
    {
        irqmask |= 1 << evIRQ;
        corbo.evcr(0).IRQ(evIRQ).IRE(1).IRAC(1).FAC(1);
    }
    if (tmIRQ > 0)
    {
        irqmask |= 1 << tmIRQ;
        corbo.tmcr(0).IRQ(tmIRQ).IRE(1).IRAC(1).FAC(1);
    }
    // Clear Event number
    corbo.evtnum(0).Set(0L);


    vme->enable_interrupts(TriggerHandler, irqmask);

}
void VMETrigger::disableInterrupts()
{
    corbo.csr(0)
        .ChnEna(1)   // Enable channel (=0 enables, =1 disables)
        .BusyMode(0) // Busy is set by the input
        .InpSel(3)   // Select internal test pulse
        .BusyOut(0)  // Ouput is local busy
        .CntSel(1)   // Count the number of busy's
        .FastClr(0)  // Allow fast clean
        .PushEna(1)  // push button resets the IRQ
        ;
    if (evIRQ > 0)
    {
        corbo.evcr(0).IRQ(evIRQ).IRE(0);
    }
    if (tmIRQ > 0)
    {
        corbo.tmcr(0).IRQ(tmIRQ).IRE(0);
    }

    // Make sure we are not leaving the BUSY
    reset();
    reset();
    vme->disable_interrupts();
}

bool VMETrigger::operator()()
{
    bool rc=false;
    if (have_to_stop)
        return false;

    do
    {
        rc = false;
        DAQpp::GTimer tm;
        tm.start();
        if (!sem.wait(1500000))
        {
            std::cout << "Time: " << tm() << " stop? " << have_to_stop << std::endl;
            // There has been a time out... Check the BUSY line
            if (!have_to_stop && busy())
            {
                reset();
            }
            if (have_to_stop)
                return false;
        }
        else
        {
            /* The acknowledge have to be done here, outside of the
             * interrupt code
             */
            VMEmanager::IRQinfo irq;
            vme->acquire();
            vme->acknowledge(&irq);

            if (irq.level != evt_irq())
            {
                if (irq.level == tm_irq())
                {
                    std::cout << "*** We got a time out. Reset BUSY" << std::endl;
                    reset();
                    reset_timer();
                }
            }
            else
            {
                rc = true;
            }
            enable_irq(irq.level);
            vme->release();
        }
    }
    while (!rc);

    sem.set_ready(false);
    return true;
}

static VMETrigger *trigger;
void TriggerHandler(int sig)
{
    trigger->new_trigger();
}


extern "C"
{
    /*
      This is the routine called to enable trigger
      interrupts. If things went smooth it should return
      0, otherwise -1.
    */
    int enable_interrupts()
    {
        std::cout << "*** VMEintr: Enabling Interrupts ***" << std::endl;
        if (!trigger)
            trigger = new VMETrigger();

        trigger->enableInterrupts();
        // Make DAQmanager aware of this
        DAQpp::DAQmanager::theDAQmanager()->set_trigger(trigger);

        return 0;
    }

    /*
      This is the routine called to disable trigger interrupts
    */
    void disable_interrupts()
    {
        if (trigger)
        {
            std::cout << "*** VMEintr: Disabling Interrupts ***" << std::endl;
            DAQpp::DAQmanager::theDAQmanager()->set_trigger(0);
            trigger->disableInterrupts();
            delete trigger;
            trigger = 0;
        }
    }

    void reset_interrupts()
    {
        if (trigger)
        {
            trigger->reset();
        }
    }

    void start_interrupts()
    {
        std::cout << "Start CORBO interrupts" << std::endl;
        if (trigger)
            trigger->start();

    }
    void stop_interrupts()
    {
        std::cout << "Stop CORBO interrupts" << std::endl;
        if (trigger)
        {
            trigger->stop();
        }
    }

}
