#include <iostream>
#include <iomanip>
#include <daq/VMEmanager.h>
#include "tdc488.h"


tdc488::tdc488(unsigned int base) {
   setAddresses(base);
}

void tdc488::ID() {
   VMEmanager *vme = VMEmanager::VME();
   unsigned short mn = vme->read16(_manuf);
   unsigned short vrs= vme->read16(_version);
   std::cout << "CAEN 8 channel multievent TDC. Mod. V488 A" << std::endl
	     << "Manufacturer " << std::hex << ((mn&0xfc00)>>10) << std::endl
	     << "Module type  " << std::hex << (mn&0x3ff) << std::endl
	     << "Version      " << std::hex << ((vrs&0xf000)>>12) << std::endl
	     << "serial no.   " << std::hex << (vrs&0xfff) << std::endl
	     << "Code         " << std::hex << vme->read16(_code)
	     << std::dec << std::endl;
}

void tdc488::setAddresses(unsigned int base) {
   _IR      = (unsigned long ) base;
   _version = (unsigned long ) (base + 0xfe);
   _manuf   = (unsigned long ) (base + 0xfc);  
   _code    = (unsigned long ) (base + 0xfa);
   _hfreg   = (unsigned long ) (base + 0x1e);
   _reset   = (unsigned long ) (base + 0x1c);
   _control = (unsigned long ) (base + 0x1a);
   _outb    = (unsigned long ) (base + 0x18);
   _ffreg   = (unsigned long ) (base + 0x16);
   _range   = (unsigned long ) (base + 0x14);
   _thrh    = (unsigned long ) (base + 0x12); 
   _thrl    = (unsigned long ) (base + 0x10);
}
void tdc488::Dump() {
   VMEmanager *vme = VMEmanager::VME();

   ID();
   std::cout << "IR      " << std::hex << vme->read16( _IR ) << std::endl;
   std::cout << "hfreg   " << std::hex << vme->read16( _hfreg  ) << std::endl;
   std::cout << "control " << std::hex << vme->read16( _control ) << std::endl;
   std::cout << "ffreg   " << std::hex << vme->read16( _ffreg ) << std::endl;
   std::cout << "range   " << std::hex << vme->read16( _range ) << std::endl;
   std::cout << "thrh    " << std::hex << vme->read16( _thrh ) << std::endl;
   std::cout << "thrl    " << std::hex << vme->read16( _thrl ) << std::endl;
   std::cout << std::dec;
}

tdc488 *tdc488::reset() {
   VMEmanager *vme = VMEmanager::VME();
   vme->write16(_reset,0);
   return this;
}

unsigned short tdc488::range2dac(float x) const {
   float gain,val;
   unsigned short ss;
   if ( x<=0. ) x=90.;
   if ( x>770.) x=770.;
   gain = 3840./x;
   val  = -224.*(gain-45.)/40.;
   ss =  (unsigned short)(val+0.5);
   return ss;
}

float tdc488::dac2range(unsigned short v) const {
   float gain = (-40./224.)*v+45.;
   float x = 3840./gain;
   return x;
}
tdc488 *tdc488::setRange(float x) {
   VMEmanager *vme = VMEmanager::VME();
   unsigned short ss = range2dac(x);
   vme->write16(_range, (ss & 0xff) );
   std::cout << "Range = " << std::hex << ss << std::dec << std::endl;
   return this;
}

tdc488 *tdc488::setMode( modeType tp ) {
   VMEmanager *vme = VMEmanager::VME();
   unsigned short cnf = vme->read16(_control);
   switch ( tp ) {
      case start:
	 cnf &= ~(1<<15);
	 break;
      case stop:
	 cnf |= 1<<15;
	 break;
      default:
	 std::cout << "tdc488::setMode - Unknown operation mode" << std::endl;
	 break;
   }
   vme->write16(_control, cnf);
   return this;
}

tdc488 *tdc488::enableChan( unsigned short val) {
   VMEmanager *vme = VMEmanager::VME();
   unsigned short cnf = vme->read16( _control );
   cnf |= (val & 0xff);
   vme->write16( _control, cnf);
   return this;
}

tdc488 *tdc488::disableChan( unsigned short val) {
   VMEmanager *vme = VMEmanager::VME();
   unsigned short cnf = vme->read16(_control);
   cnf &= ~(val & 0xff);
   vme->write16(_control, cnf);
   return this;
}

bool tdc488::empty() {
   VMEmanager *vme = VMEmanager::VME();
   unsigned short cnf = vme->read16(_control);
   return ((cnf & 0x4000)==0) ;
}

bool tdc488::full() {
   VMEmanager *vme = VMEmanager::VME();
   unsigned short r = vme->read16(_range);
   unsigned short cnf = vme->read16(_control);
   if ( r & 0x1000 )  // Full FIFO
      return (cnf & 0x2000);
   else
      return ( cnf & 0x1000);
}

tdc488 *tdc488::HighThreshold(int val) {
   VMEmanager *vme = VMEmanager::VME();
   vme->write16(_thrh, (val&0xff));
   return this;
}
tdc488 *tdc488::LowThreshold(int val) {
   VMEmanager *vme = VMEmanager::VME();
   vme->write16(_thrl,  (val&0xff));
   return this;
}
tdc488 *tdc488::HFmode() {
   VMEmanager *vme = VMEmanager::VME();
   vme->write16(_hfreg,0);
   return this;
}
tdc488 *tdc488::FFmode() {
   VMEmanager *vme = VMEmanager::VME();
   vme->write16(_ffreg,0);
   return this;
}

unsigned short tdc488::data() {
   return VMEmanager::VME()->read16(_outb);
}
