#ifndef __VelmexModule_h__
#define __VelmexModule_h__
/**
 * This is a module for vmedaq.
 * They all should inherit from GenModule, which adds some extra
 * functionality on top of DAQpp::Module:
 *
 *   - The module works around the concept of something with a number
 *     of chips, each having the same number of channels. So nchip
 *     (number of chips) and nchan (number of channels per chip) is
 *     something we should supply somehow.
 *
 *     For the SIS3300 module we replace nchip by the number of channels (8)
 *     and nchan by the sample size (== page size)
 *
 *     In paticular for the
 *     Analysis instance of the module. That brings me to the second
 *     point.
 *
 *   - There is an Analysis pointer (you have to create the object)
 *     which will store all the monitoring stuff. This guy needs the
 *     number of chips and channels per chip.
 *
 *   - If you plan to make scans with that module, you will need to
 *     override the set_scan_point method to set your module in a
 *     reasonable state every time you are in a new scan point.
 *
 *   - pedestals and noise are there as DAQ++ parameters
 *      + can be updated from the values in Analysis or copied into
 *        the Analysis
 *
 *      + Pedestals are stored in the data file if you override the
 *      write_pedestals method.
 *
 *
 * See the DAQ++ documentation for further information
 *
 **/
#include <string>
#include <sstream>
#include <daq/GenModule.h>
#include "driver/velmexDriver.h"

class VelmexModule : public GenModule
{
public:
    /**
     * Constructor
     * The constructor needs 2 parameters:
     * id: the Module id
     * address: the address (VME, or whatever protocol) to
     *          communicate with the module
     *
     * Note: although DAQ++ allows for any type of identifier,
     * however VMEDAQ only handles properly integer IDs. The only
     * reason for that is that it needs integer values to save them
     * properly in the data file
     **/
    VelmexModule( const DAQpp::DAQid &id, std::string address );

    ~VelmexModule();

    /**
     * This is the method called by the RunManager to know if there
     * is data ready in the module. If the answer if true, then it
     * will try to read the module. If there is no way to know a
     * priory about the availavility of data, one can defer the
     * decision until getData
     **/
    bool hasData();

    /**
     * This is the method called to retrieve the data. If hasData
     * could not figure out the right answer, one can still say here
     * that there is no data at all by returning a NULL pointer and
     * setting size to 0.
     *
     * The memory to store the data should be allocated with the
     * allocate method. See the DAQ++ documentation...
     **/
    char *getData( int &size );

    int PrepareForRun ();

    int Start ( int = 0 );

    int Stop ( int = 0 );

    int Pause ();

    int Continue ();

    int Reset ();

    int GetReady ();

    int Trigger ()
    {
        return 0;
    };
    void flush ()
    {};

    void set_scan_point( ScanPoint & );
    void set_header_description( unsigned int &data );
    unsigned long get_data_type() const;
    // *** module specific methods
    // method for creating/recreating Analysis
    void move_to( int mtr, float pos );

    void create_analysis( int nchip = 1, int nchan = 1 );

    // * get and set methods
    // module parameters
    // parameters for the motors
    std::string get_motor_s_ScanType ( int mtr )
    {
        std::ostringstream oss;
        oss << "motor" << mtr << "s_scanType";
        return DAQpp::get_par_value<std::string>( get_par( oss.str().c_str() ) );
    }
    void set_motor_s_ScanType ( int mtr, std::string x )
    {
        std::ostringstream oss;
        oss << "motor" << mtr << "s_scanType";
        set_par<std::string>( oss.str().c_str(), x );
    }

    int get_motorScanType ( int mtr )
    {
        std::ostringstream oss;
        oss << "motor" << mtr << "scanType";
        return DAQpp::get_par_value<int>( get_par( oss.str().c_str() ) );
    }
    void set_motorScanType ( int mtr, int x )
    {
        std::ostringstream oss;
        oss << "motor" << mtr << "scanType";
        set_par<int>( oss.str().c_str(), x );
    }

    bool get_motorUseMap ( int mtr )
    {
        std::ostringstream oss;
        oss << "motor" << mtr << "useMap";
        return DAQpp::get_par_value<bool>( get_par( oss.str().c_str() ) );
    }
    void set_motorUseMap ( int mtr, bool x )
    {
        std::ostringstream oss;
        oss << "motor" << mtr << "useMap";
        set_par<bool>( oss.str().c_str(), x );
    }

    float get_motorScale ( int mtr )
    {
        std::ostringstream oss;
        oss << "motor" << mtr << "scale";
        return DAQpp::get_par_value<float>( get_par( oss.str().c_str() ) );
    }
    void set_motorScale ( int mtr, float x )
    {
        std::ostringstream oss;
        oss << "motor" << mtr << "scale";
        set_par<float>( oss.str().c_str(), x );
    }

    int get_motorOffset ( int mtr )
    {
        std::ostringstream oss;
        oss << "motor" << mtr << "offset";
        return DAQpp::get_par_value<int>( get_par( oss.str().c_str() ) );
    }
    void set_motorOffset ( int mtr, int x )
    {
        std::ostringstream oss;
        oss << "motor" << mtr << "offset";
        set_par<int>( oss.str().c_str(), x );
    }

    int get_motorSpeed ( int mtr )
    {
        std::ostringstream oss;
        oss << "motor" << mtr << "speed";
        return DAQpp::get_par_value<int>( get_par( oss.str().c_str() ) );
    }
    void set_motorSpeed ( int mtr, int x )
    {
        std::ostringstream oss;
        oss << "motor" << mtr << "speed";
        set_par<int>( oss.str().c_str(), x );
    }

    void get_motorMap ( int mtr, std::vector<float> &v )
    {
        v.clear();
        std::ostringstream oss;
        oss << "motor" << mtr << "map";
        v = DAQpp::get_par_value< std::vector<float> > ( get_par( oss.str().c_str() ) );
    }
    void set_motorMap ( int mtr, std::vector<float> &v )
    {
        std::ostringstream oss;
        oss << "motor" << mtr << "map";
        set_par< std::vector<float> > ( oss.str().c_str(), v );
    }

private:
    VelmexDriver* velmexDriver;
};

#endif
