/**
   This class is responsible to parse the corresponding section in the
   configuration file to configure the module. It is also responsible
   for create the module instance according to the paremeters retrived
   from the configuration file.

   It derives from ModuleParser (see ModuleParser.h in the config
   directory) which, in turn, derives from SectionParser (again, have
   a look at the header file).

   It has, then, to implement a number of methods from SectionParser

      - parse_node:  this method will parse the corresponding section
                     and store locally the configuration parameters
       defined in there.
      - save_config: this method will be called if the configuration
                     is to be saved to an external file. The idea is
       that the class will check which parameters have
       changed since the file was loaded and update them
       in the document before saving.

   and a few more from ModuleParser:

      - config : called by the gui to change configuration parameters
      - monitor: called by the gui to create the monitor display
      - decode : called to decode the data
      - range_min, range_max: to retrieve the range of the monitor
                              histograms

   One can also define here the factory function specified in the
   configuration file.

   FOr this module we will assume that we only have 2 parameters to
   appear in the configuration file:
      - the amplitude of the pulse that we simulate and,
      - the width of the peak we want to simulate.

   So in the modules section we add

   <modules>
   .
   .
   .
   <test_module id="100" address="deadbeef">
      <!-- monitor is a section common to all the modules -->
      <monitor>
      ...
      </monitor>

      <!-- padmap is also common to all the modules -->
      <padmap>
      .....
      </padmap>

      <!-- and here come our parameters -->
      <mean>100</mean>
      <width>5</width>

   </test_module>
   .
   .
   .
   </modules>
*/
#include <string>
#include <sstream>
#include <DAQ++/DAQmanager.h>
#include <config/ConfigParser.h>
#include <config/ModuleParser.h>
#include <config/GeneralConfig.h>
#include <gtk/gtkwidget.h>
#include <gui/GladeApp.h>
#include <gui/WidgetUpdate.h>
#include <gui/MonitorWindow.h>
#include <data/Analysis.h>
#include <data/DataFormat.h>
#include "VelmexModule.h"
#include "VelmexModuleGuiConf.h"
#include "VelmexModuleGuiMon.h"

class VelmexModuleConf : public ModuleParser
{
private:
    std::string _glade_file;

public:
    VelmexModuleConf()
    {}

    ~VelmexModuleConf()
    {}

    /**
     * This is the method responsible for parsing the configuration
     * file. 
     **/
    void parse_node( xmlNodePtr node )
    throw ( ConfigException );

    /**
     * The method that saves the current configuration. It should
     * update all the xml nodes below the module node.
     **/
    void save_config()  throw (ConfigException);

    /**
     * This is called by VMEDAQ to create a window that will allow
     * to change the configuration of the module while running the
     * program. Probably the easiest is to use glade-2 to create the
     * layout of the window, define the signals (callbacks that are
     * associated to given actions) and then something derived from
     * GladeApp (see the gui folder) to implement the stuff and
     * update the module state accordingly.
     **/
    GtkWidget *config( const std::string &id );

    /**
     * This is called by VMEDAQ to create a window to display the
     * monitor data. The input parameters are:
     *    - id: the module id (that should help finding the module)
     *    - glade_file: The path of the glade file. This is, in
     *      principle the path of the standard VMEDAQ glade
     *      file. One can certainly ignore this value.
     **/
    MonitorWindow *monitor( const std::string &id, const std::string &glade_file );


    // This is the method responsible for parsing the configuration paramters
    // for up to 3 motors attached to the velmex module
    void parse_motors( const std::string &element, const std::string &id );
};

/**
 * This is the hook specified in the config file to return a pointer
 * to the VelmexModuleConf instance that will interact with VMEDAQ.
 *
 * Note that it should be declared as extern "C" so that no demangling
 * mechanism is applied to its name that prevent finding easily the
 * symbol in the library
 *
 **/
extern "C"
{
    VelmexModuleConf *create_VelmexModuleConf()
    {
        VelmexModuleConf *cnf = new VelmexModuleConf();
        return cnf;
    }
}


/**
 * Now come the implementations of the class methods
 */
void
VelmexModuleConf::parse_node( xmlNodePtr _node )
throw ( ConfigException )
{
    unsigned long address;
    std::string saddress;

    // The configuration manager handler
    ConfigParser *cfg = ConfigParser::theConfig();

    // store the node pointer
    node = _node;

    // get the module id
    id = cfg->get_attribute<std::string>( node, "name" );

    // get the address
    saddress = cfg->get_attribute<std::string>( node, "address" );

    // Check if the module is already there
    DAQpp::Module *dm;
    VelmexModule *module = 0;

    if ( ( dm = DAQpp::DAQmanager::theDAQmanager()->find_module( id ) ) )
        module = dynamic_cast<VelmexModule *>( dm );

    if ( !module )
    {
        // The module is not there... create
        module = new VelmexModule( id, saddress );
    }

    // Now we have the module. From now own one should parse the
    // section from the configuration file
    try
    {
        _glade_file = cfg->get_value<std::string>( node, "guiFile" );
    }
    catch ( ConfigException &e )
    {
        // IT was not there... leave the default or current value
    }

    // Now parse the motors
    parse_motors( "//VelmexModule", id );

    // Now parse the things which are common to all the modules, that
    // is the monitor and padmap elements
    parse_common( "//module", id );
}

void
VelmexModuleConf::parse_motors( const std::string &element, const std::string &id )
{
    std::ostringstream rstr;
    std::vector<xmlNodePtr> vr, vval;
    int nr, nval;
    xmlNodePtr nd;
    ConfigParser *cfg = ConfigParser::theConfig();

    std::string s_scanType;
    int   scanType = -1;
    bool  useMap;
    float scale;
    int   offset;
    int   speed;
    std::vector<float> mmap;
    std::ostringstream oss_s_scanType, oss_scanType, oss_useMap, oss_scale, oss_offset, oss_speed, oss_map;

    int n_names = ScanVar::get_n_names();

    rstr << element << "[@name=\"" << id << "\"]/motor";
    nr = cfg->get_node( rstr.str(), vr );
    if ( nr )
    {
        VelmexModule *module = 0;
        DAQpp::Module *dm;
        if ( !( dm = DAQpp::DAQmanager::theDAQmanager()->find_module( id ) ) )
            return;
        if ( !( module = dynamic_cast<VelmexModule *>( dm ) ) )
            return;

        if ( nr > 3 )
        {
            std::cout << "VelmexModuleConf: number of motors = " << nr
            << " is too large, only the first 3 will be used" << std::endl;
            nr = 3;
        }

        for ( int nditer = 0; nditer < nr; nditer++ )
        {
            nd = vr[nditer];
            oss_s_scanType.str( "" );
            oss_scanType.str( "" );
            oss_useMap.str( "" );
            oss_scale.str( "" );
            oss_offset.str( "" );
            oss_speed.str( "" );
            oss_map.str( "" );

            try
            {
                s_scanType = cfg->get_attribute<std::string>( nd, "scanType" );
                for ( int nmiter = 0; nmiter < n_names; nmiter++ )
                {
                    std::cout << "ScanVar::get_name " << nmiter << " = "
                    << ScanVar::get_name( nmiter ) << std::endl;
                    if ( s_scanType == ScanVar::get_name( nmiter ) )
                    {
                        scanType = nmiter;
                        break;
                    }
                }
                if ( scanType < 0 )
                {
                    std::cout << "motor " << nditer << " in VelmexModule " << id
                    << " does not have a recognized ScanType: "  << s_scanType
                    << std::endl;
                    throw;
                }

                useMap = cfg->get_attribute<bool>( nd, "useMap" );

                scale  = cfg->get_value<float>( nd, "scale" );

                offset = cfg->get_value<int>( nd, "offset" );

                speed  = cfg->get_value<int>( nd, "speed" );

                if ( useMap )
                {
                    nval = cfg->get_node( "map", nd, vval );
                    cfg->get_array( vval[0], mmap );
                }
            }
            catch ( ConfigException e )
            {
                std::cout << "motor " << nditer << " in VelmexModule " << id
                << " does not have expected attributes" << std::endl;
                throw;
            }

            oss_s_scanType << "motor" << nditer << "s_scanType";
            module->add_par( oss_s_scanType.str().c_str(), new DAQpp::Par<std::string>( s_scanType ) );
            oss_scanType << "motor" << nditer << "scanType";
            module->add_par( oss_scanType.str().c_str(), new DAQpp::Par<int>( scanType ) );
            oss_useMap << "motor" << nditer << "useMap";
            module->add_par( oss_useMap.str().c_str(), new DAQpp::Par<bool>( useMap ) );
            oss_scale << "motor" << nditer << "scale";
            module->add_par( oss_scale.str().c_str(), new DAQpp::Par<float>( scale ) );
            oss_offset << "motor" << nditer << "offset";
            module->add_par( oss_offset.str().c_str(), new DAQpp::Par<int>( offset ) );
            oss_speed << "motor" << nditer << "speed";
            module->add_par( oss_speed.str().c_str(), new DAQpp::Par<int>( speed ) );
            if ( useMap )
            {
                oss_map << "motor" << nditer << "map";
                module->add_par( oss_map.str().c_str(), new DAQpp::Par< std::vector<float> > ( mmap ) );
            }

            // unnecessary printout to check parameters
            if ( true )
            {
                std::cout << "VelmexModuleConf::parse_motors: motor      = " << nditer << std::endl;
                std::cout << "VelmexModuleConf::parse_motors: s_scanType = "
                << DAQpp::get_par_value<std::string>( module->get_par( oss_s_scanType.str().c_str() ) )
                << std::endl;
                std::cout << "VelmexModuleConf::parse_motors: scanType   = "
                << DAQpp::get_par_value<int>( module->get_par( oss_scanType.str().c_str() ) )
                << std::endl;
                std::cout << "VelmexModuleConf::parse_motors: useMap     = "
                << DAQpp::get_par_value<bool>( module->get_par( oss_useMap.str().c_str() ) ) << std::endl;
                std::cout << "VelmexModuleConf::parse_motors: scale      = "
                << DAQpp::get_par_value<float>( module->get_par( oss_scale.str().c_str() ) ) << std::endl;
                std::cout << "VelmexModuleConf::parse_motors: offset     = "
                << DAQpp::get_par_value<int>( module->get_par( oss_offset.str().c_str() ) ) << std::endl;
                std::cout << "VelmexModuleConf::parse_motors: speed      = "
                << DAQpp::get_par_value<int>( module->get_par( oss_speed.str().c_str() ) ) << std::endl;

                if ( useMap )
                {
                    int i;
                    mmap = DAQpp::get_par_value< std::vector<float> > ( module->get_par( oss_map.str().c_str() ) );
                    std::vector<float>::iterator viter;
                    std::cout << "VelmexModuleConf::parse_motors: map         = ";
                    for ( i = 0, viter = mmap.begin();viter != mmap.end();++viter, i++ )
                    {
                        if ( i % 8 == 0 ) std::cout << std::endl << "    ";
                        std::cout << "  " << *viter;
                    }
                    std::cout << std::endl;
                }
            }
        }
    }
    else if ( nr == 0 )
    {
        std::cout << "VelmexModuleConf: no motors found" << std::endl;
    }
}

void
VelmexModuleConf::save_config()  throw (ConfigException)
{
    VelmexModule *m = 0;
    DAQpp::Module *dm;

    // find the module
    if ( !( dm = DAQpp::DAQmanager::theDAQmanager()->find_module( id ) ) )
        return;

    if ( !( m = dynamic_cast<VelmexModule *>( dm ) ) )
        return;

    // The configuration manager handler
    ConfigParser *cfg = ConfigParser::theConfig();

    // Save common stuff
    save_common();

    // Save the current values in the configuration
}

//=====================================================================
// GUI
//=====================================================================
GtkWidget *
VelmexModuleConf::config( const std::string &id )
{
    VelmexModule *m = 0;
    DAQpp::Module *dm;

    if ( !( dm = DAQpp::DAQmanager::theDAQmanager()->find_module( id ) ) )
        return 0;

    if ( !( m = dynamic_cast<VelmexModule *>( dm ) ) )
    {
        std::cout << "VelmexModuleConf: module " << id
        << " is not of type VelmexModule"
        << std::endl;
        return 0;
    }
    /*
    VelmexModuleGuiConf *conf = new VelmexModuleGuiConf(m->get_id());
    conf->init(_glade_file.c_str(),"SIS3300-Config");

    return conf->get_main_window();
    */
    std::cout << "VelmexModuleConf::config "
    << " config gui not implemented"
    << std::endl;
    return 0;
}

MonitorWindow *
VelmexModuleConf::monitor( const std::string &id, const std::string &glade_file )
{
    VelmexModule *m = 0;
    DAQpp::Module *dm;

    if ( !( dm = DAQpp::DAQmanager::theDAQmanager()->find_module( id ) ) )
        return 0;

    if ( !( m = dynamic_cast<VelmexModule *>( dm ) ) )
    {
        std::cout << "VelmexModuleConf: module " << id
        << " is not of type VelmexModule"
        << std::endl;
        return 0;
    }

    /*
    VelmexModuleGuiMon *mon = new VelmexModuleGuiMon(m->get_id());
    mon->init(_glade_file.c_str(),"SIS3300-Monitor");

    MonitorWindow *mw = new MonitorWindow(mon->get_main_window());
    return mw;
    */
    std::cout << "VelmexModuleConf::monitor "
    << " monitor gui not implemented"
    << std::endl;
    return 0;
}
