#!/usr/bin/env python
"""
    Implementation of an xmlrpc server in the process where the DAQ 
    is running
"""
import xmlrpclib

import sys
import SimpleXMLRPCServer
import daqpp
import base64

__verbose__ = False

class RunObserver(object):
    """ This class receives notifications when a RunManages changes
        its internal state
    """
    def __init__(self):
        """ Initialization of the instance
        """
        pass

    def update(self,arg):
        """ Method called by the Observable
        """
        print "Observer: State is ",
        if arg==daqpp.Ready:
            print "*** Ready"
        elif arg==daqpp.NotReady:
            print "*** NotReady"
        elif arg==daqpp.Starting:
            print "*** Starting"
        elif arg==daqpp.Paused:
            print "*** Paused"
        elif arg==daqpp.Running:
            print "*** Running"
        elif arg==daqpp.Stopped:
            print "*** Running"
        elif arg==daqpp.Stopping:
            print "*** Stopping"
        elif arg==daqpp.TimeOut:
            print "*** TimeOut"
        else:
            print "Unknown [",arg,"]"


class ParamObserver(object):
    """ Receives notifications when a Parameter changes
        its value
    """
    def __init__(self):
        """ Initialization of the instance
        """
        pass
    
    def update(self, arg):
        """ Called by the Observable 
        """
        print "Parameter:", arg 


class MsgManager(object):
    """ The message manager.
        Receives the main notifications of the DAQ process and 
        distributes them to the registered subscribers
    """
    def __init__(self):
        """ Initialization.
            Here we subscribe to DAQmanager
        """
        daqpp.add_observer(self)
        self.run_observer = RunObserver()
        self.param_observer = ParamObserver()
        
    def update(self, arg):
        """ We receive notifications here
        """
        
        print "MsgManager.update - ", arg
        try:
            msgtype = arg[0]
            if msgtype == daqpp.startDAQ:
                print "*** Starting DAQ ****"
            
            elif msgtype == daqpp.stopDAQ:
                print "*** Stopping DAQ ****"

            elif msgtype == daqpp.nwModule:
                print "new Module", arg[1]
                md = daqpp.find_module(arg[1])
                if md:
                    md.addObserver(self.param_observer)

            elif msgtype == daqpp.rmModule:
                print "remove Module", arg[1]
        
            elif msgtype == daqpp.nwRunmanager:
                print "new RunManager",arg[1]
                runm = daqpp.find_runmanager(arg[1])
                if runm:
                    runm.addObserver(param=self.param_observer, state=self.run_observer)

            elif msgtype == daqpp.rmRunmanager:
                print "remove RunManager",arg[1]
                
            else:
                print "*** Unknown msg type ***"
                
        except TypeError:
            pass
        
class DAQXMLrpcServer(SimpleXMLRPCServer.SimpleXMLRPCServer):
    """ This class extends the basic server 
    """
    msg_manager = MsgManager()
        
    def _dispatch(self, method, params):
        """ The methods made public by the server have the 
            prefix export_
        """
        try:
            # We are forcing the 'export_' prefix on methods that are
            # callable through XML-RPC to prevent potential security
            # problems
            func = getattr(self, 'export_' + method)
        except AttributeError:
            raise Exception('method "%s" is not supported' % method)
        else:
            return func(*params)        

    #
    # General interface
    #

    def export_get_module_list(self):
        rc = daqpp.get_module_list()
        if __verbose__:
            print "get_module_list:", rc
            sys.stdout.flush()
        
        return rc
    
    def export_get_runmanager_list(self):
        rc = daqpp.get_runmanager_list()
        if __verbose__:
            print "get_runmanager_list:",rc
            sys.stdout.flush()
            
        return rc
        
    #
    # DAQObject interface
    #
    def get_obj(self, daqid):
        obj = None
        if daqid in daqpp.get_module_list():
            obj = daqpp.find_module(daqid)
            
        elif daqid in daqpp.get_runmanager_list():
            obj = daqpp.find_runmanager(daqid)

        if __verbose__:
            print "get_obj / DAQobject ", daqid," -> ", obj
            sys.stdout.flush()
            
        return obj
        
    def export_get_par_names(self, daqid):
        """ Returns the list of parameter names
        """
        obj = self.get_obj(daqid)  
        if __verbose__:
            print 'get_par_names / DAQobject:',daqid,'->',obj
            sys.stdout.flush()
            
        out = ()
        if obj:
            out = obj.par_names()
        
        elif __verbose__:
            print "DAQobject", daqid, "does not exist"
            sys.stdout.flush()
    
        if __verbose__:
            print "get_par_names(", daqid,"): ", out
            sys.stdout.flush()
            
        return out

    def export_set_par(self, daqid, name, val, notify):
        obj = self.get_obj(daqid)  
        if obj:
            obj.set_par(name, val, notify)
            return 0
        
        else:
            return 1 

    def export_get_par(self, daqid, name):
        obj = self.get_obj(daqid)
        out = None  
        if obj:
            out = obj.get_par(name)

        if __verbose__:
            print "get_par(%s, %s) = %s" % (daqid, name, out)
            sys.stdout.flush()
            
        return out

    def export_add_par(self, daqid, name, value):
        obj = self.get_obj(daqid)
        out = None  
        if obj:
            out = obj.add_par(name, value)

        return True

    def export_reset_mon_data(self, daqid):
        obj = self.get_obj(daqid)
        out = None
        if obj and hasattr(obj, "reset_mon_data"):
            out = obj.reset_mon_data()
        
        return out
    
    def export_get_mon_data(self, daqid):
        obj = self.get_obj(daqid)
        out = None
        if obj and hasattr(obj, "get_mon_data"):
            try:
                data = obj.get_mon_data()
                if data:
                    out = base64.b64encode( data.tostring() )
                    
            except Exception, e:
                print "server.get_mon_data: ", e

        return out

    def export_get_status(self, daqid):
        obj = self.get_obj(daqid)
        if __verbose__:
            print 'get_status / DAQobject:',daqid,'->',obj
            sys.stdout.flush()
            
        out = None
        if obj:
            status = obj.GetStatus()
            if status.isRunning():
                st = "running"
            elif status.isReady():
                st = "ready"
            elif status.isPaused():
                st = "paused"
            elif status.isTimeout():
                st = "timeout"
            else:
                st="notready"
                
            out = {"name":status.get_name(),
                   "status":st,
                   "ntrig":status.Ntrig(),
                   "nwriten":status.Nwriten(),
                   "eff":status.eff(),
                   "rate":status.rate(False),
                   "time":status.time(),
                   "throughput":status.throughput()
                   }

        return out
    
    def get_runm(self, daqid):
        runm = None
        if daqid in daqpp.get_runmanager_list():    
            runm = daqpp.find_runmanager(daqid)
            
        return runm
    def export_runm_is_active(self, daqid):
        runm = self.get_runm(daqid)
        if runm:
            return runm.isActive()
        else:
            return None
        
    def export_runm_activate(self, daqid):
        runm = self.get_runm(daqid)
        if runm:
            return runm.Activate()
        else:
            return None
        
    def export_start_runm(self, daqid):
        runm = self.get_runm(daqid)
        if runm:
            if not runm.isActive():
                return runm.Activate()
            
            daqpp.startRunManager(runm)
        else:
            return None
        
    def export_stop_runm(self, daqid):
        runm = self.get_runm(daqid)
        if runm:
            daqpp.stopRunManager(runm)
        else:
            return None
        
    def export_getready_runm(self, daqid):
        runm = self.get_runm(daqid)
        if runm:
            runm.GetReady()
        else:
            return None

    def export_pause_runm(self, daqid):
        runm = self.get_runm(daqid)
        if runm:
            runm.Pause()
        else:
            return None

    def export_resume_runm(self, daqid):
        runm = self.get_runm(daqid)
        if runm:
            runm.Continue()
        else:
            return None
        
the_server=None
def start_server(port):
    """ Starts the server
    """
    import socket
    global the_server
    print "*** Starting the xmlrpc server ***"
    the_server = DAQXMLrpcServer( (socket.gethostname(), port), logRequests=False, allow_none=True)
    return the_server

