#include <DAQ++/python/pydaqpp_buffer.h>
#include <config/Config.h>
#include "pyanalysis.h"

pyanalysis::pyanalysis(int nchip, int nchan) :
    Analysis(nchip, nchan)
{

}

pyanalysis::~pyanalysis()
{
}

void pyanalysis::Analyze()
{
    bool has_method;

    daqpp_observer_start();
    has_method = PyObject_HasAttrString(self(),"Analyze");
    daqpp_observer_stop();
    if (has_method)
        call_void_method("Analyze",NULL);
    else
        Analysis::Analyze();

}
bool pyanalysis::data_valid(unsigned int *data) const
{
    daqpp_observer_start();
    unsigned int dsz = data[0] & 0xfff;
    daqpp_buffer *bf = (daqpp_buffer *)PyDAQpp_Buffer_New('I',dsz*sizeof(unsigned int), (char *)data);
    daqpp_observer_stop();

    int rc=const_cast<pyanalysis *>(this)->call_int_method("data_valid","O",(PyObject *)bf);
    daqpp_observer_start();
    Py_DECREF(bf);
    daqpp_observer_stop();

    return rc;
}
void pyanalysis::decode(unsigned int *data)
{
    bool error = false;
    daqpp_observer_start();
    unsigned int dsz = data[0] & 0xfff;
    PyObject *bf = PyDAQpp_Buffer_New('I',dsz*sizeof(unsigned int), (char *)data);
    if (PyErr_Occurred())
    {
        error = true;
        daqpp_print_error();
        PyErr_Print();
    }
    daqpp_observer_stop();
    if (!error)
    {
        call_void_method("decode","O",bf);
        daqpp_observer_start();
        Py_DECREF(bf);
        daqpp_observer_stop();
    }
    return;
}
void pyanalysis::init_run()
{
    call_void_method("init_run",NULL);
}
void pyanalysis::end_run()
{
    call_void_method("end_run", NULL);
}

static PyObject *
vmedaq_analysis_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
    vmedaq_analysis *self;
    self =  PyObject_New(vmedaq_analysis,type);
    if (self!=NULL) {
        self->ana  = 0;
        self->dict = PyDict_New();
    }
    return (PyObject *)self;
}

static void
vmedaq_analysis_dealloc(vmedaq_analysis* self)
{
    delete self->ana;
    Py_DECREF(self->dict);
    self->ana = 0;
    self->ob_type->tp_free((PyObject*)self);
}

static int
vmedaq_analysis_init(vmedaq_analysis *self, PyObject *args, PyObject *kwargs)
{
    int _nchip, _nchan;
    if (!PyArg_ParseTuple(args, "ii:Analysis.init", &_nchip, &_nchan))
        return -1;

    self->ana = new pyanalysis(_nchip, _nchan);
    self->ana->set_object((PyObject *)self);
    return 0;
}

static PyObject *
vmedaq_analysis_decode(vmedaq_analysis *self, PyObject *args, PyObject *kwargs)
{
    PyObject *buff=Py_None;
    if (self->ana->external_call())
    {
        Py_INCREF(Py_None);
        return Py_None;
    }
    if (!PyArg_ParseTuple(args, "O:Analysis.decode", &buff))
        return NULL;

    Py_ssize_t size;
    const char *cbuff;
    if (buff!=Py_None)
    {
        if (PyObject_AsReadBuffer(buff, (const void **)&cbuff, &size) )
        {
            if (PyErr_Occurred())
            {
                daqpp_print_error();
                PyErr_Print();
            }

            PyErr_SetString(DAQpp_Error, "Analysis.decode should receive a buffer object");
            Py_DECREF(buff);
            return NULL;
        }
        self->ana->set_python_call();
        self->ana->decode((unsigned int *)cbuff);
        self->ana->clear_python_call();
    }
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
vmedaq_analysis_data_valid(vmedaq_analysis *self, PyObject *args, PyObject *kwargs)
{
    PyObject *buff=Py_None;
    if (self->ana->external_call())
    {
        return PyInt_FromLong(1);
    }
    if (!PyArg_ParseTuple(args, "O:Analysis.data_valid", &buff))
        return NULL;

    Py_ssize_t size;
    const char *cbuff;
    bool valid = false;
    if (buff!=Py_None)
    {
        if (PyObject_AsReadBuffer(buff, (const void **)&cbuff, &size) )
        {
            if (PyErr_Occurred())
            {
                daqpp_print_error();
                PyErr_Print();
            }

            PyErr_SetString(DAQpp_Error, "Analysis.decode should receive a buffer object");
            Py_DECREF(buff);
            return NULL;
        }
        self->ana->set_python_call();
        valid = self->ana->data_valid((unsigned int *)cbuff);
        self->ana->clear_python_call();
    }
    return PyInt_FromLong( valid ? 1 : 0);
}

static PyObject *
vmedaq_analysis_analyze(vmedaq_analysis *self, PyObject *args, PyObject *kwargs)
{
    Py_INCREF(Py_None);
    return Py_None;
}



static PyObject *
vmedaq_analysis_init_run(vmedaq_analysis *self, PyObject *args, PyObject *kwargs)
{
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
vmedaq_analysis_end_run(vmedaq_analysis *self, PyObject *args, PyObject *kwargs)
{
    Py_INCREF(Py_None);
    return Py_None;
}

PyDoc_STRVAR(resize_doc,"Change the number of chips and channels.\n\
");
static PyObject *
vmedaq_analysis_resize(vmedaq_analysis *self, PyObject *args, PyObject *kwargs)
{
    int _nchip, _nchan;
    if (!PyArg_ParseTuple(args, "ii:Analysis.resize", &_nchip, &_nchan))
        return NULL;

    self->ana->Resize(_nchip, _nchan);

    Py_INCREF(Py_None);
        return Py_None;
}

static PyObject *
vmedaq_analysis_addhit(vmedaq_analysis *self, PyObject *args, PyObject *kwargs)
{
    int chan;
    if (!PyArg_ParseTuple(args, "i:Analysis.add_hit", &chan))
        return NULL;

    self->ana->add_hit(chan);
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
vmedaq_analysis_nhits(vmedaq_analysis *self, PyObject *args, PyObject *kwargs)
{
    return PyInt_FromLong( self->ana->nhits() );
}

static PyObject *
vmedaq_analysis_pophit(vmedaq_analysis *self, PyObject *args, PyObject *kwargs)
{
    int chan = self->ana->pop_hit();
    return PyInt_FromLong(chan);
}

static PyObject *
vmedaq_analysis_pedestals(vmedaq_analysis *self, PyObject *args, PyObject *kwargs)
{
    Py_ssize_t i, len;
    PyObject *chanlist;
    PyObject *item=0;
    int ipulse=1;
    if (!PyArg_ParseTuple(args, "O|i:Analysis.compute_pedestals", &chanlist, &ipulse))
        return NULL;

    if (!PySequence_Check(chanlist))
    {
        PyErr_SetString(PyExc_TypeError, "Argument of compute_pedestals should be a sequence");
        return NULL;
    }
    std::vector<int> clist;
    len = PySequence_Length(chanlist);
    for (i=0;i<len;i++)
    {
        item = PySequence_GetItem(chanlist,i);
        clist.push_back( PyInt_AsLong(item));
        Py_DECREF(item);
    }

    bool find_pulse = (ipulse!=0);
    ConfigParser *cfg = ConfigParser::theConfig();
    GeneralConfig *gnrl = dynamic_cast<GeneralConfig *>(cfg->get_handler("general"));
    if (gnrl->get_ped_method() == GeneralConfig::normal)
    {
        self->ana->update_pedestals(clist, find_pulse);
    }
    else
    {
        self->ana->update_pedestals_fast(clist, find_pulse);
    }

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
pyanalysis_get_nchip(vmedaq_analysis *self, void *closure)
{
    return PyInt_FromLong(self->ana->Nchip());
}

static PyObject *
pyanalysis_get_nchan(vmedaq_analysis *self, void *closure)
{
    return PyInt_FromLong(self->ana->Nchan());
}

static PyObject *
pyanalysis_get_ntot(vmedaq_analysis *self, void *closure)
{
    return PyInt_FromLong(self->ana->Ntot());
}

static PyObject *
pyanalysis_get_range_min(vmedaq_analysis *self, void *closure)
{
    return PyFloat_FromDouble(self->ana->get_range_min());
}
static PyObject *
pyanalysis_get_range_max(vmedaq_analysis *self, void *closure)
{
    return PyFloat_FromDouble(self->ana->get_range_max());
}
static PyObject *
pyanalysis_get_nbins(vmedaq_analysis *self, void *closure)
{
    return PyInt_FromLong(self->ana->get_nbins());
}


static PyObject *
pyanalysis_get_ped(vmedaq_analysis *self, void *closure)
{
    return PyDAQpp_Buffer_New('d',self->ana->Ntot()*sizeof(double),(char *)self->ana->Ped());
}

static PyObject *
pyanalysis_get_noise(vmedaq_analysis *self, void *closure)
{
    return PyDAQpp_Buffer_New('d',self->ana->Ntot()*sizeof(double),(char *)self->ana->Noise());
}

static PyObject *
pyanalysis_get_hitmap(vmedaq_analysis *self, void *closure)
{
    return PyDAQpp_Buffer_New('d',self->ana->Ntot()*sizeof(double),(char *)self->ana->HitMap());
}
static PyObject *
pyanalysis_get_signal(vmedaq_analysis *self, void *closure)
{
    return PyDAQpp_Buffer_New('d',self->ana->get_nbins()*sizeof(double),(char *)self->ana->Signal());
}
static PyObject *
pyanalysis_get_event(vmedaq_analysis *self, void *closure)
{
    return PyDAQpp_Buffer_New('d',self->ana->Ntot()*sizeof(double),(char *)self->ana->Event());
}

static PyObject *
pyanalysis_get_chan(vmedaq_analysis *self, void *closure)
{
    return PyDAQpp_Buffer_New('d',self->ana->Ntot()*sizeof(double),(char *)self->ana->Chan());
}




static PyGetSetDef pyanalysis_getseters[] =
{
 {"nchip",(getter)pyanalysis_get_nchip, (setter)0, "Number of chips",NULL},
 {"nchan",(getter)pyanalysis_get_nchan, (setter)0, "Number of channels per chip",NULL},
 {"ntot",(getter)pyanalysis_get_ntot, (setter)0, "Total number of channels",NULL},
 {"ped",(getter)pyanalysis_get_ped, (setter)0,"",NULL},
 {"noise",(getter)pyanalysis_get_noise, (setter)0,"",NULL},
 {"hipmap",(getter)pyanalysis_get_hitmap, (setter)0,"",NULL},
 {"signal",(getter)pyanalysis_get_signal, (setter)0,"",NULL},
 {"event",(getter)pyanalysis_get_event, (setter)0,"",NULL},
 {"channels",(getter)pyanalysis_get_chan, (setter)0,"",NULL},
 {"range_min",(getter)pyanalysis_get_range_min, (setter)0,"",NULL},
 {"range_max",(getter)pyanalysis_get_range_max, (setter)0,"",NULL},
 {"nbins",(getter)pyanalysis_get_nbins, (setter)0,"",NULL},
 {NULL}  /* Sentinel */
};

static PyMethodDef vmedaq_analysis_methods[] =
{
 {"decode",(PyCFunction)vmedaq_analysis_decode,METH_VARARGS,""},
 {"data_valid",(PyCFunction)vmedaq_analysis_data_valid,METH_VARARGS,""},
 {"init_run",(PyCFunction)vmedaq_analysis_init_run,METH_NOARGS,""},
 {"end_run",(PyCFunction)vmedaq_analysis_end_run,METH_NOARGS,""},
 {"resize",(PyCFunction)vmedaq_analysis_resize,METH_VARARGS,resize_doc},
 //{"Analyze", (PyCFunction)vmedaq_analysis_analyze,METH_VARARGS,""},
 {"add_hit",(PyCFunction)vmedaq_analysis_addhit,METH_VARARGS,""},
 {"nhits",(PyCFunction)vmedaq_analysis_nhits,METH_NOARGS,""},
 {"pop_hit",(PyCFunction)vmedaq_analysis_pophit,METH_NOARGS,""},
 {"compute_pedestals",(PyCFunction)vmedaq_analysis_pedestals,METH_VARARGS,""},
 {NULL}  /* Sentinel */
};


static PyTypeObject vmedaq_analysis_Type =
{
     PyObject_HEAD_INIT(NULL)
     0,                         /*ob_size*/
     "pyvmedaq.Analysis",             /*tp_name*/
     sizeof(vmedaq_analysis),       /*tp_basicsize*/
     0,                         /*tp_itemsize*/
     (destructor)vmedaq_analysis_dealloc,                         /*tp_dealloc*/
     0,                         /*tp_print*/
     0,                         /*tp_getattr*/
     0,                         /*tp_setattr*/
     0,                         /*tp_compare*/
     (reprfunc)0,                         /*tp_repr*/
     0,                         /*tp_as_number*/
     0,                         /*tp_as_sequence*/
     0,                         /*tp_as_mapping*/
     0,                         /*tp_hash */
     (ternaryfunc)0,                         /*tp_call*/
     (reprfunc)0,                 /*tp_str*/
     PyObject_GenericGetAttr,                         /*tp_getattro*/
     PyObject_GenericSetAttr,                         /*tp_setattro*/
     0,                         /*tp_as_buffer*/
     Py_TPFLAGS_DEFAULT| Py_TPFLAGS_BASETYPE,        /*tp_flags*/
     "Analysis class",           /* tp_doc */
     0,                      /* tp_traverse */
     0,                     /* tp_clear */
     0,                     /* tp_richcompare */
     0,                     /* tp_weaklistoffset */
     0,                     /* tp_iter */
     0,                     /* tp_iternext */
     vmedaq_analysis_methods,             /* tp_methods */
     0,             /* tp_members */
     pyanalysis_getseters,                         /* tp_getset */
     0,                         /* tp_base */
     0,                         /* tp_dict */
     0,                         /* tp_descr_get */
     0,                         /* tp_descr_set */
     offsetof(vmedaq_analysis, dict),                         /* tp_dictoffset */
     (initproc)vmedaq_analysis_init,      /* tp_init */
     0,                         /* tp_alloc */
     vmedaq_analysis_new,                 /* tp_new */
    };

void register_analysis(PyObject *dict)
{
    daqpp_register_class(dict,"Analysis",
                          &vmedaq_analysis_Type,
                          0);
}
int PyVMEdaq_Analysis_Check(PyObject *O)
{
    return PyType_IsSubtype(O->ob_type, &vmedaq_analysis_Type);
}
PyObject *PyVMEdaq_Analysis_New(int nchip, int nchan)
{
    vmedaq_analysis *ret = (vmedaq_analysis *)PyObject_New(vmedaq_analysis, &vmedaq_analysis_Type);
    ret->ana = new pyanalysis(nchip, nchan);
    ret->ana->set_object((PyObject *)ret);
    return (PyObject *)ret;
}

